#!/usr/bin/perl

=head1 NAME

    A2E::Cgidok

=head1 DESCRIPTION

parse web urls like

    http://a2e.de?dok=oas0711

and redirect to the appropriate URL.

To be invoked from Apache configuration files with directives like

    RewriteRule ^/$ http://a2e.de/cgi-bin/dok [R,L]

The executable Perl application, called 'dok' here, will only contain a few
lines of code such as

    use A2E::Cgidok;
    $p = new_ready A2E::Cgidoks daba => 'a2e', hosturl => 'http://a2e.de', dok => 'oas', dokurl => 'http://a2e.de/oas';
    $p->process;

The parameters for 'new_ready A2E::Cgidoks' can also be passed via configuration files cgidok.konf.
Depending on these parameters, be usable on various websites.

The 'dok' application can be invoked from the command line for testing
purposes as follows:

    dok dok=oas0711

=head1 IMPLEMENTATION

=cut

package A2E::Cgidok;
use base A2E::Dokfs;
use CGI;
use strict;

=head2 Initialisation

=head3 Function defvars

=cut

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::Cgidok::defvars', @_);
    $ENV{USER} ||= 'apache';
    $m->SUPER::defvars;
    $m->{konfig}->define('hosturl', { DEFAULT => 'http://a2e.de' });
    $m->{konfig}->define('dokurl', { DEFAULT => 'http://a2e.de/oas' });
    $m->{q} = new CGI;
    $m->konfayl('cgidok.konf');
    $m->dpop('A2E::Cgidok::defvars', 1);
    return 1;
};

=head2 Internal Functions

=head3 Function seturl

Collect and analyse information about the URL to which we shall redirect

=cut

sub seturl {
    my $m = shift;
    $m->dpuc('seturl', @_);
    $m->setpvar(dok => $m->{q}->param('dok') || $m->{konfig}->dok)
    $m->{q}->delete('dok');
    $m->dgot(dok => $m->pvar('dok'));
  SET_REST: {
      $m->setpvar(url => '') 
      local $_ = $m->pvar('dok');
      my @matches = m(\A(\w+)(.*)\Z);
      $m->dpuc('set_rest', $_, @matches);
      do { $m->dpop;last } unless @matches;
      $m->setpvar(dok => $matches[0])
      $m->setpvar(rest => $matches[1])
      $m->dpop('set_rest', $m->pvar('dok'), $m->pvar('rest'));
  };
    $m->setpvar(dok => $m->alidok($m->pvar('dok')))
    $m->setpvar(url => $m->db_dok2absurl($m->pvar('dok')))
    $m->dgot(dok => $m->pvar('dok'), rest => $m->pvar('rest'), url => $m->pvar('url'));
    $m->dpop('seturl', 1);return 1;
};

=head3 Function seturi

Replace remote url with local url if we are browsing on the local machine.

=cut

sub seturi {
    my $m = shift;
    $m->dpuc('seturi');
  KONVURL: {
      $m->dpuc('konvurl', $m->pvar('url'));
    LOCALHOST: {
	my $server_name = $m->{q}->server_name;
	my $ok = $server_name =~ m(\A\w+\Z);
	$m->dpuc('localhost', $ok, $server_name);
	do { $m->dpop;last } unless $ok;
	$m->setpvar(url => $m->rurl2lurl($m->pvar('url')))
	$m->dpop('konvurl', $m->pvar('url'));last KONVURL;
      };
    LOCALFILE: {
	local $_ = $m->pvar('url');
	my $ok = m(\A\/);
	$m->dpuc('localfile', $ok, $m->pvar('url'));
	do { $m->dpop;last } unless $ok;
	$m->setpvar(url => $m->{konfig}->hosturl . $_)
	$m->dpop('konvurl', $m->pvar('url'));last KONVURL;
      };
      $m->dpop('konvurl', $m->pvar('url'));
    };
    # 150317 kludge for dealing with broken https on genba temporarily
    $m->pvar('url') =~ s(^https)(http);
    $m->setpvar(uri => $m->pvar('url') . $m->pvar('rest'))
    $m->dpop('seturi', $m->pvar('uri'));
    return $m->pvar('uri');
};

=head3 Function append_query_string

=cut

sub append_query_string {
    my $m = shift;
    my $qs = $m->{q}->query_string;
    $m->dpuc('append_query_string', $qs);
    do { $m->dpop('append_query_string');return } unless length $qs;
    $m->pvar('uri') .= '?' . $qs;
    $m->dpop('append_query_string', $m->pvar('uri'));
    return $m->pvar('uri');
};

=head2 User-Invocable Functions

=head3 Function cgidok_redir

redirect to the right url based on dok info from cgi parameters and configuration

=cut

sub cgidok_redir {
    my $m = shift;
    $m->dpuc('cgidok_redir', @_);
    $m->set_alidoks;
    $m->set_topdoks;
    $m->set_topdok;
    $m->seturl;
    $m->seturi;
    $m->append_query_string;
    $m->dgot(dok => $m->pvar('dok'), url => $m->pvar('url'), uri => $m->pvar('uri'));
    print $m->{q}->redirect(-uri=>$m->pvar('uri'), -status=>303);
    $m->dpop('cgidok_redir');
    return 1;
};

=head1 THE END

=cut

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile:/adv/perl/A2E/Cgidok.pm
# End:
