# -*- coding: utf-8; srcfile:/adv/perl/A2E/Daba.pm; mode: perl -*-

=head1 NAME

A2E::Daba

=cut

package A2E::Daba;
use strict;

=head1 DESCRIPTION

High level access to the database.

Encapsulate database access and provide some generic routines for it.

This library provides
 - a mechanism for automated (passwordless) and yet fairly secure access to the system database  
   - connect to local database, hide authentication info
   - use remote db if available, local otherwise
 - translation between Perl and SQL structures, assuming that the database obeys certain rules
 - interface for most common db uses such as 'get' and 'put', the latter being a combination of update and insert
 - treat all input and output as utf-8 

The following functions are stable and for general use

  - put
  - delete
  - get_record
  - get_records

See usage examples below.

The functions of the A2E::Prog library such as

  - vecho
  - decho
  - exek

are also available in the object.

A2E::Daba is intended to be the successor of A2E::db.

=head1 SYNOPSIS

use A2E::Daba;
   
my $db = new_ready A2E::Daba;
my ($var1, $var2, ...) = $db->get_record(qq{select var1, var2, .... from ...}) ;
$db->put('my_table', { key1=>'keyval1', key2=>'keyval2' }, val1=>'valval1', ... ;

=head1 BUGS, LIMITATIONS

The squoted function may have to be adapted to non-PostgreSQL databases.
The documentation effort has just begun.
The configuration mechanism needs clarification.

Further rationalisation could be achieved by introducing a database schema configuration mechanism similar to what has been done in some other applications such as mlht and php_users.

=head1 Related Systems

=head2 PREREQUISITES

    A2E::OpenPGP(3)
    DBI(3)

=cut

use base 'A2E::Prog';
use DBI;
use Encode qw(_utf8_on);
use File::Spec::Functions;
our $novitas = $ENV{novitas};

sub daba_defvars {
    my ($m,$prog) = shift->qpuc('daba_defvars',@_) or return;

=head1 OPTIONS

=over

=item --sqlecho | --sql | -s

echo any executed SQL commands and database feedback

=cut
    $m->konfig_define(['sqlecho', 'sql', 's'], '!');


=item --daba|--db a2e

name of database to be used by default

=cut

    $m->cache_define('daba');

=item --dabas|--dbs a2e=1 

make a database available and activate/deactivate it by assigning it a 1/0 value

=cut

    $m->konfig_define(['dabas','dbs'], '%');

=item --maxrows 1

maximum number of rows that may be affected by an 'update' operation.
If more than this is affected we die.

=cut

    $m->cache_define(maxrows => i => 1);

=item --daba_auto

Automatic/batch mode, use anonymous access, don't even try prompting for a password

=cut 

    $m->{konfig}->define('daba_auto!');

=item --daba_anon_user|--dau a2e=apache

User to use as long as none is specified.  This user should have at least minimal access rights to the database.

=cut

    $m->cache_define([ 'daba_anon_user', 'dau' ], 's%');

=item --daba_pghost|--pgh /var/lib/pgsql

=cut

    $m->cache_define([ 'daba_pghost', 'dph' ], 's', 'localhost');

=item --daba_user|--du a2e=phm

User specific to database.   This is usually specified in personal configuration files.  By default it is the system user.  If no daba_pass is specified, we fall back to daba_anon_user

=cut

    $m->cache_define([ 'daba_user', 'du' ], 's%');

=item --daba_anon_pass|--dap I<topsikret> 

Cleartext passwords for the given databases belonging to the user specified by --daba_anon_user.
These should not be directly stored in any configuration file but could be useful as a commandline options.
They can be supplied by environment variables such as C<A2E_ANON_PASS>, assuming that daba is I<a2e>, or via the --daba_anon_pass_fayl file.
They can be cached in cleartext files whose names are derived using the value of the --daba_anon_pass_fayl option.
The authoritative source information should reside in the --daba_anon_pass_pgp option.  Edit only that option (in a config file) and then delete the derived files so they will be regenerated. 

=cut

    $m->cache_define([ 'daba_anon_pass', 'dap' ], 's%');

=item --daba_pass|--dp a2e=topsikret

Cleartext passwords of the current user for the given databases.
They should never be stored in a configuration file.
They can be specified via environment variables like C<A2E_PASS>, where C<A2E> is the concerned daba (internal database name) in upper case.
They can be cached in cleartext in files whose names are derived using the sprintf format specified through the --daba_pass_fayl option with the daba (database identifier) as its only argument.
The authoritative source information should reside in the --daba_pass_pgp option.  Edit only that option (in a config file) and then delete the derived files so they will be regenerated. 

=cut

    $m->{konfig}->define('daba_pass=s%');

=item --daba_anon_pass_pgp|--dapp a2e=C<pgp_message>

THE FOLLOWING PGP-related options are not usable in this version!

Encrypted passwords for the anonymous user of each database.  Store in sitewide config file using heredoc syntax.
This should usually be the master source from which cached passwords, if any, are generated.a

=cut

    $m->konfig_define('daba_anon_pass_pgp','%');

=item --daba_pass_pgp|--dpp

Encrypted password for the individual user of each database.  Store in individual config file using heredoc syntax.
This should usually be the master source from which cached passwords, if any, are generated.

=cut

    $m->cache_define('daba_pass_pgp', '%');

=item --daba_anon_pass_fayl|--dapf a2e=/var/tmp/.a2e_pass

group-readable file where to cache the anonymous user's cleartext password so that it needn't be read interactively or from memory next time; as long as this is not defined no cleartext storing is done.
This file is created by the program if not present already.  For this reason it should be in a directory where the program has write access.
Enabling this option is the way to go if you want your applications to modify remote databases while working in the background, e.g. invoked by cronjobs, without any human initialisation (and without provoking your system's users to store even more precious cleartext passwords elsewhere).

=cut

    $m->konfig_define('daba_anon_pass_fayl','%');

=item --daba_pass_fayl|--dpf a2e=~/.a2e_pass

user-readable file where to store the individual user's cleartext password so that it needn't be read interactively or from memory next time; as long as this is not defined no cleartext storing is done.
This file is created by the program if not present already.  For this reason it should be in a directory where the program has write access.
By setting this option you take an unnecessary security risk for the sake of making it easier for more automatic processes to use your database applications.  The preferred way to meet such needs is by using the C<daba_anon_pass_fayl> option.

=cut

    $m->konfig_define('daba_pass_fayl','%');

=item --daba_hsys a2e=a2e

machine (hardware system) where the database is found.  Each hsys has its own arbitrary symbol.

=cut

    $m->konfig_define('daba_hsys','%', { a2e => 'a2e', ffii => 'ffii', amend => 'ffii' }); 

=item --daba_name a2e=a2e

the actual name of the database that is used when connecting to it.  This may differ from the arbitrary daba symbol that we use here.

By default it is the same.

=cut 

    $m->konfig_define('daba_name','%', { a2e => 'a2e', ffii => 'ffii' });
 
=item --hsys_host a2e=a2e.de

Domain name of the hsys machine

=cut

    $m->konfig_define('hsys_host','%', { a2e => 'a2e.de', ffii => 'ffii.org' });

=item --hsys_mach a2e=bartu.bos.a2e.de

Fully qualified domain name of the hsys machine

=cut

    $m->konfig_define('hsys_mach','%', { a2e => 'bartu.bos.a2e.de', ffii => 'vic.ffii.org' });

=item --daba_lokal

Whether to limit queries to localhost.

Same effect was previously obtained by setting --host to 'localhost'.

=cut

    $m->konfig_define(daba_lokal => b => $ENV{DABA_LOKAL} ? 1 : 0);

=item --fldtyp tadatum=d

field type of a database field/column.

This is part of a quick & dirty sql-perl mapping syntax which is subject to change

=cut

    $m->cache_define('fldtyp', 's%');

=item --optyps

operator definitions.  a means of extension that is still in development and not ready for public use yet.

=cut

    $m->konfig_define(optyps => '%', { null => ['u', '%s is null'], notnull => ['u', '% is not null' ], true => ['u', '%s'], false => ['u', 'not %s'], '~' => 's', like => 's', '==' => 'n', '=' => 'nd', '<' => 'nd', '>' => 'nd', '<=' => 'nd', '>=' => 'nd', '<=..<' => 'r', '<..=<' => 'r', '<..<' => 'r', '<=..=<' => 'r', 'in' => 'l' });

=item --sqllog /var/opt/a2e/log/daba_%s%d%s.sql

filename template of file where to log sql commands that were used, so that these may be applied to another database later.
template argument 1 (%s) is the name of the database, argument 2 (%d) is 1 for commands on the local database (daba_lokal true), 0 otherwise, 3 (%d) is the time of invocation (progtime) of the program. 
Normally this should be set in a configuration file, so that the $user variable or tilde expansion ~ can be used.
TODO: possibly we should send this stuff to a list of concerned addresses where mail processors listen and react by updating their respective local databases.  This could be done by a special call at the end of program execution which sends out the accumulated files.

=cut

    $m->konfig_define('sqllog', '$', '/var/opt/a2e/log/daba/%s%d%s.sql');

=item --dpllog /var/opt/a2e/log/daba_%s%d%d.dpl

filename template of file where to log high-level deplate database modification instructions that were used, so that these may be applied to another database later.
structure like --sqllog.

=cut

    $m->konfig_define('dpllog', '$', '/var/opt/a2e/log/daba%s%d%s.dpl');

=back

=head1 VERSION

This describes version B<0.1.1>

=cut

    $m->{progver} ||= '0.1.1';
    $m->{dbi} = {};
    $m->{daba_dsn} = {};

=head1 FILES

=head2 Configuration

    /etc/opt/a2e/daba.konf
    ~/.daba_konf
    ./@daba

=cut

    $m->konfayl('daba.konf');
    $m->qpop1($prog,1);
};

=head2 Applications

    A2E::Dokfs(3) -- built on A2E::Daba
    A2E::Mondeb::Daba(3)
  http://a2e.de/adv/perl/A2E

=head2 Related Tools

    daba_pass(1) -- generate statements to be source into a shell environment so as to cache passwords
    env_cache(1) -- invoke daba_pass in bash
    mktdir_bashrc(1) -- contains definition of shell function cdok that changes a directory to a document (identified by its symbol rather than its directory location) and, if needed, creates the password cache
    a2e_pgp(1) -- use public key encryption and optionally read password from environment
									       
=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

    Copyright (c) 2007-8 Hartmut Pilch

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

=head2 Construct the Object

=head3 Function defvars

=cut

sub defvars {
    my ($m,$prog) = shift->qpuc('A2E::Daba::defvars',@_) or return;
    $m->SUPER::defvars;
    $m->daba_defvars;
    $m->qpop1($prog,1);
}; 
sub get_daba {
    my ($m,$prog) = shift->ppuc('get_daba',@_);
    my $daba = $m->pvar('daba');
  GET: {
      last if $daba;
      $daba = $m->{cache}->{daba} || $m->{konfig}->daba;
      $m->pstop('cant get daba') if !$daba;
      $m->setpvar(daba => $daba);
    };
    $m->dpop1($prog,$daba);
};
=head3 Function daba_postkonfig

Some setup work that needs to be performed after all configuration options have been read and calculated: find the database and connect to it.

Normally done via setup, but some applications may need to access the functionality directly so as to build their own setup routine.

=cut

sub daba_postkonfig {
    my ($m,$prog) = shift->qpuc('daba_postkonfig',@_) or return;
    $m->{konfig}->dabas->{$m->{konfig}->daba} = 1;foreach my $daba (keys %{$m->{konfig}->dabas}) { next unless $m->{konfig}->dabas->{$daba};$m->daba_konekt($daba); };
    my $daba = $m->get_daba;
    my $fnom = $m->{konfig}->sqllog;
    my $lokal = $m->{konfig}->daba_lokal;
    $m->set_timestamp;
    if ($fnom) {
	my $fo = sprintf $fnom, $daba, $lokal, $m->{timestamp};
	$m->dgot(fnom => $fnom, lokal => $lokal, daba => $daba, timestamp => $m->{timestamp}, fo => $fo, user => $m->{konfig}->user, expand => $m->{konfig}->{STATE}->_expand('sqllog')); # dump => $m->{konfig}->{STATE}->_dump_var('sqllog')
	$m->open_fayl($fo, 'sqllog', -w $fo ? 'a' : 'w', lock => 1, mkdir => 1);
	$m->print_tu_fayl('sqllog', sprintf q{select '%s' as comment ;}."\n", join(' ', $m->{timestamp}, $0, @ARGV));
    };
    $fnom = $m->{konfig}->dpllog;
    if ($fnom) {
	my $fo = sprintf $fnom, $daba, $lokal, $m->{timestamp};
	$m->open_fayl($fo, 'dpllog', -w $fo ? 'a' : 'w', lock => 1, mkdir => 1);
	$m->print_tu_fayl('dpllog', join(' ', '%', $m->{timestamp}, $0, @ARGV) . "\n");
    };
    $m->qpop1($prog,1);
};

=head3 Function postkonfig

daba_postkonfig for overloading

=cut

sub postkonfig {
    my ($m,$prog) = shift->qpuc('A2E::Daba::postkonfig',@_) or return;
    $m->SUPER::postkonfig;
    $m->daba_postkonfig;
    $m->qpop1($prog,1);
};

=head3 Function daba_konfig

Parse database dependent configuration file.  A convenience subroutine used in lastkonfig overloads e.g. in A2E::Privat::Db2auth(3)

Deprecated: to support simultaneous use of multiple databases, use hash variables instead.

=cut

sub daba_konfig {
    my ($m,$prog,$prognom,$konf,@dirs) = shift->ppuc('daba_konfig',@_);
    warn 'still using deprecated daba_konfig mechanism';
    $prognom ||= $m->{konfig}->prognom;
    $konf ||= 'konf';
    my $daba = $m->{konfig}->daba;
    $m->pstop('daba not found') if !length $daba;
    my $fi = join '_', $prognom, $daba;
    $fi = join '.', $fi, $konf;
    $m->konfayl(@dirs, $fi);
    # $m->pars_konfaylz;
    $m->dpop1('daba_konfig', 1);
};

=head3 Function dbi_konekt

Subroutine of daba_konekt

=cut

sub dbi_konekt {
    my ($m,$prog,$daba) = shift->ppuc('dbi_konekt',@_);
    $daba ||= $m->pvar(daba => 1);
    $m->{dbi}->{$daba} = DBI->connect($m->{daba_dsn}->{$daba}, $m->{daba_user}->{$daba}, $m->{daba_pass}->{$daba});
    my $ok = $m->daba_aktiv($daba);
    $m->dpop1($prog,$ok);
};

=head3 Function daba_aktiv

=cut

sub daba_aktiv {
    my ($m,$prog,$daba) = shift->ppuc('daba_aktiv',@_);
    my $aktiv = ref $m->{dbi}->{$daba} && $m->{dbi}->{$daba}->{Active};
    $m->dpop1($prog,$aktiv);
}

=head3 Function daba_konekt

Connect to the database using the collected identification+authentication info.

Quickly return true if connection is already established.

Connect to the central server if possible, to a local copy otherwise.

=cut

sub daba_konekt {
    my ($m,$prog,$daba) = shift->ppuc('daba_konekt',@_);
    if ($daba) {
	$m->setpvar(daba => $daba);
    } else {
	$daba = $m->get_daba;
    };
    return $m->dpop($prog,1) if $m->daba_aktiv($daba);
    my $hsys = $m->{konfig}->daba_hsys->{$daba} || $daba;
    my $name = $m->{konfig}->daba_name->{$daba} || $daba;
    my $mach = $m->{konfig}->hsys_mach->{$hsys} || die sprintf 'no idea which mach corresponds to hsys %s', $hsys;
    my ($mach0) = split m(\.), $mach;
    my $host = $m->{konfig}->hsys_host->{$hsys} || die sprintf 'no idea which host corresponds to hsys %s', $hsys;
  REMOT: {
      $m->dpuc('remot');
    LOKAL: {
	my $ok = $m->{konfig}->daba_lokal;
	my $host = $m->pvar('host');
	$ok ||= $host eq 'localhost';
	$m->dpuc('lokal', $ok, $host);
	do { $m->dpop;last } unless $ok;
	$m->dpop('remot');last REMOT;
    };
    ON_HOST: {
	my $ok = $mach0 eq $host;
	$m->dpuc('on_host', $ok, $host, $mach0, $mach);
	do { $m->dpop;last } unless $ok;
	$m->dpop('remot');last REMOT;
    };
    PING: {
	last if $novitas;
	my $ok = $m->netp($host);
	$m->dpuc('ping', $ok, $host);
	do { $m->dpop;last } if $ok;
	$m->dpop('remot');last REMOT;
    };
    USER_PASS: {
	$m->dpuc('user_pass', $daba);
	my $daba_user_var = join '_', uc($daba), 'USER';
	$m->{daba_user}->{$daba} = $ENV{$daba_user_var} if exists $ENV{$daba_user_var};
	my $daba_pass_var = join '_', uc($daba), 'PASS';
	my $fayl = $m->{konfig}->daba_pass_fayl->{$daba};
	my $faylstr = '';
	$m->{daba_user}->{$daba} ||= $m->{konfig}->user || $ENV{USER} || do { $m->dpop;last };
	$m->{daba_pass}->{$daba} ||= $ENV{$daba_pass_var} || do { $faylstr = $fayl && -r $fayl && $m->fayl_get_str($fayl,1) };
	$ENV{$daba_pass_var} = $m->{daba_pass}->{$daba};
      FAYL_PUT: {
	  $m->dpuc('fayl_put', $fayl, $faylstr);
	  do { $m->dpop;last } unless $fayl;
	  do { $m->dpop;last } if length $faylstr;
	  my $mode = '0600';
	  $m->vecho('write pass to file %f in mode %m', f => $fayl, m => $mode);
	  $m->fayl_put_str($fayl, $m->{daba_pass}->{$daba});
	  chmod oct($mode), $fayl or warn sprintf 'cant chmod file %s to mode %s', $fayl, $mode;
	  $m->dpop;
	};
	$m->dpop('user_pass', $m->{daba_user}->{$daba}, $m->{daba_pass}->{$daba}, $fayl);
      };
    ANON_USER_PASS: {
	$m->dpuc('anon_user_pass');
	do { $m->dpop;last } if $m->{daba_user}->{$daba} and $m->{daba_pass}->{$daba};
	my $daba_user_var = join '_', uc($daba), 'ANON', 'USER';
	if (length $ENV{$daba_user_var}) {
	    $m->{konfig}->daba_anon_user->{$daba} = $ENV{$daba_user_var};
	};
	my $daba_pass_var = join '_', uc($daba), 'ANON', 'PASS';
	my $fayl = $m->{konfig}->daba_anon_pass_fayl->{$daba};
	my $faylstr = '';
	$m->{konfig}->daba_anon_pass->{$daba} ||= $ENV{$daba_pass_var} || do { $faylstr = $fayl && -r $fayl && $m->fayl_get_str($fayl,1) }; 
	die sprintf 'can not find any username whatsoever for daba %s', $daba unless exists $m->{konfig}->daba_anon_user->{$daba};
	die sprintf 'no anonymous password found for daba %s', $daba unless length $m->{daba_anon_pass}->{$daba};
	$ENV{$daba_pass_var} = $m->{daba_anon_pass}->{$daba};
      ANON_FAYL_PUT: {
	  $m->dpuc('fayl_put', $fayl, $faylstr);
	  do { $m->dpop;last } unless length $fayl;
	  do { $m->dpop;last } if length $faylstr;
	  my $mode = '0660';
	  $m->vecho('write anon pass to file %f in mode %m', f => $fayl, m => $mode);
	  $m->fayl_put_str($fayl, $m->{daba_anon_pass}->{$daba});
	  chmod oct($mode), $fayl or warn sprintf 'cant chmod file %s to mode %s', $fayl, $mode;
	  $m->dpop;
	};
	$m->{daba_user}->{$daba} = $m->{konfig}->daba_anon_user->{$daba};
	$m->{daba_pass}->{$daba} = $m->{konfig}->daba_anon_pass->{$daba};
	$m->dpop('anon_user_pass', $m->{daba_user}->{$daba}, $m->{daba_pass}->{$daba}, $fayl);
      };
    REMOT_OK: {
        $m->dpuc('remot_ok');
	$m->{daba_dsn}->{$daba} = sprintf 'dbi:Pg:dbname=%s;host=%s', $name, $host;
	my $ok = $m->dbi_konekt($daba);
	do { $m->dpop;last } unless $ok;
	$m->dpop('daba_konekt', 1);
	return 1;
      };
      $m->dpop('remot');
    };
  LOKAL: {
      $m->uwarn('failed to ping localhost') unless $m->netp('localhost');
      $host = $ENV{PGHOST} || $m->{konfig}->daba_pghost || 'localhost';
      $m->{daba_dsn}->{$daba} = sprintf 'dbi:Pg:dbname=%s;host=%s', $name, $host;
      $m->{daba_dsn}->{$daba} = 'dbi:Pg:dbname='.$name;
      $m->{daba_dsn}->{$daba} .= ';host='.$host if $host and 'localhost' ne $host;
      $m->{daba_user}->{$daba} = $ENV{USER};$m->{daba_pass}->{$daba} = undef;
      $m->dbi_konekt($daba) or $m->ustop('user %u failed to connect anonymously to database %d', u => $ENV{USER}, d => $daba);
    };
    $m->dpop1($prog,1);
};

=head2 Subroutines for translating Perl to SQL

=head3 Function squoted

enclose a string with single quotes for SQL, escape existing single quotes by reduplication (as required by PostgreSQL, other databases may differ)

=cut

sub squoted {
    my ($m,$s) = @_;local $_ = $s;
    # return $_ if m(\A[\d\.]*\d\Z);
    s(\')('')g;
    "'".$_."'";
};


=head3 Function squoted_or_null

Same as squoted but for update/insert statements where empty strings are replace with NULL.

=cut

sub squoted_or_null {
    my ($m,$s) = @_;local $_ = $s;
    return 'NULL' unless length $s;
    s(\')('')g;
    "'".$_."'";
};

=head3 Function deplate_quoted

enclose a string with double quotes, escape existing double quotes by backslash, as needed by deplate et al

=cut

sub deplate_quoted {
    my ($m,$prog,$s) = shift->ppuc('deplate_quoted',@_);local $_ = $s;
    do { s(\")(\\")g;$_ = '"'.$_.'"' } unless m(\A\w+\Z);
    $m->dpop1($prog,$_);
};

=head3 Function fldtyp

type of a given field, as far as known from available database schema information.

=cut

sub fldtyp {
    my ($m,$prog,$fld,$tab) = shift->ppuc('fldtyp',@_);
    my %fldtyp = $m->phac(fldtyp => 1);
    my $typ = $fldtyp{$fld};
    return $m->dpop1($prog,$typ) if $typ;
    return $m->dpop($prog) if !$tab;
    $fld = join '_', $tab, $fld;
    $typ = $fldtyp{$fld};
    $m->dpop1($prog,$typ);
};

=head3 Function sql_query

Send an SQL command and collect number of affected rows.

A subroutine for higher-level commands below.

The end-user also needs this for special cases that can not be expressed well with higher-level commands.

=cut

sub sql_query {	
    my ($m,$prog,$sql) = shift->ppuc('sql_query',@_);
    my $daba = $m->get_daba;
    $m->daba_konekt($daba);
    my ($rows) = $m->{dbi}->{$daba}->do($sql);
    $m->echo($m->{konfig}->sqlecho, $sql);
    die join "\n", $daba, $sql, $m->{dbi}->{$daba}->errstr if $m->{dbi}->{$daba}->err;
    $rows = $rows =~ m(\A0E0\Z) ? 0 : $rows;
    $m->print_tu_fayl('sqllog', $sql . ";\n") if $rows and $m->{konfig}->sqllog;
    $m->dpop1($prog,$rows);
};
    
=head3 Function sql_set_pair

A key-value pair in a 'set' clause

Input
	privat => 1, 'dokprop'
Output
    	privat = true

=cut

sub sql_set_pair {
    my ($m,$prog,$nom,$val,$tab) = shift->ppuc('sql_set_pair',@_);
    my $typ = $m->fldtyp($nom, $tab) || 's';
    my $res = '';
  BOOLEAN: {
      last unless $typ eq 'b';
      $val = $val ? 'true' : 'false';
      $res = sprintf join ' = ', $nom, $val;
      return $m->dpop1($prog,$res);
    };
  NUMER: {
      last unless $typ eq 'n' or $typ eq 'i';
      $val ||= '0';
      $res = sprintf join ' = ', $nom, $val;
      return $m->dpop1($prog,$res);
    }
    $res = sprintf join ' = ', $nom, $m->squoted_or_null($val);
    $m->dpop1($prog,$res);
}

=head3 Function sql_get_pair

like C<sql_set_pair> but for use in 'where' clause':

Input
	privat => 0, 'dokprop'
Output
	not privat    


Input
        verdatum => ''
Output
	verdatum is null	


=head4 ($frm,$val) = $m->sql_get_frm_val($typ,$val): return sql-form for condidional statement $frm that varies according to value $val, along with actual $val used in it

    args := typ, val
    rets := frm, val
    typ := 'b' || 's' || 'd' || 'n' || 'i'
    typ :: field type as returned by function fldtyp 
    frm => '%s = %s'
    frm :: sprintf format expression to which key and value are fed; ignored by sql_get_val when we are dealing with a listref value 
    val :: value formated to be fed to frm for use in sql where expression

=cut

sub sql_get_frm_val {
    my ($m,$prog,$typ,$val) = shift->ppuc('sql_get_frm_val',@_);$typ ||= 's';
    my $frm = '%s = %s';
  BOOLE: {
      last unless $typ eq 'b';
      $frm = $val ? '%s' : 'not %s';
      return $m->dpop($prog,$frm);
    };
  NUMER: {
      last unless $typ eq 'n' or $typ eq 'i';
      $val ||= '0';
      return $m->dpop($prog,$frm,$val);
    };
    return $m->dpop($prog,'%s is %s','null') if !$val;
    $val = $m->squoted($val);
    $m->dpop($prog,$frm,$val);    
}

sub sql_get_pair {
    my ($m,$prog,$nom,$val,$tab) = shift->ppuc('sql_get_pair',@_);
    my $typ = $m->fldtyp($nom, $nom);
    my $frm = '%s in ( %s )';
    my $res = '';
    if (ref $val) {
	my @vals = $m->ref2list($val);
	$m->pstop('In sql_get_pair I needed a reference to a list but got something empty of type %t', t => ref($val)) unless @vals;
	$res = sprintf $frm, $nom, join ', ', map { my ($f,$v) = $m->sql_get_frm_val($typ, $_);$m->pstop('In sql_get_pair got an empty list element') unless defined $v;"$v" } @vals;
    } else {
	($frm, $val) = $m->sql_get_frm_val($typ,$val);
	$res = sprintf $frm, $nom, $val;
    };
    $m->dpop1($prog,$res);
};

=head2 Edit database fields (update/insert, delete)

=head3 Function put

example: 

    $db->put('users', { user=>'johndoe' }, mail=>'john@doe.com', sex=>'m' )

    $db->put('diary', { user=>'johndoe', day=>'2007-10-07' }, entry=>'great job done today') 

arg 1 is the table name, arg 2 a reference to an array that identifies the range of searched records (normally consisting of key-value pairs of the indexed fields which together constitute a unique identifier), args 3 and onward specify values to write into the found records.

B<put> tries to update an existing record, and, if that fails, inserts a new record.
If the 'update' would result in change of more than one record, nothing is done.

This behaviour can be modified by setting a maximum number of rows that may be
updated by one 'put' command through $m->pvar('maxrows').   If maxrows is 0, there
is no limit and you are on your own.

=cut

sub put {
    my ($m,$prog,$tab,$keysref,$valsref,@args) = shift->ppuc('put',@_);
    my (%vals, %opts) = ();
    if ('HASH' eq ref $valsref) {
	%vals = %$valsref;
	%opts = @args;
    } else {
	unshift @args, $valsref;
	%vals = @args;
    };
    my $sqlecho = $opts{sqlecho} || $m->{debug} || $m->{konfig}->sqlecho;
    for (keys %vals) { if (defined $vals{$_}) { _utf8_on $vals{$_} } else { delete $vals{$_} } };
    my $daba = $m->pvar(daba => 1);
    my %keys = ();
    return $m->dpop($prog) unless %vals;
    $m->daba_konekt($daba);
    local $m->{dbi}->{$daba}->{AutoCommit} = 0;
  KREF: {
      do { $m->pwarn('bad keysref %k', k => $keysref);last } if 'HASH' ne ref $keysref;
      %keys = %$keysref; 
    };
  DPLLOG: {
      last unless $m->{konfig}->dpllog;
      my $keys = join " ", keys %keys;
      my @pairs_keys = map { join '=', $_, $keysref->{$_} } keys %keys;
      my @pairs_vals = map { join '=', $_, $m->deplate_quoted($vals{$_}) } keys %vals;
      my $pairs = join " ", @pairs_keys, @pairs_vals;
      my $dpllog = " #. {dabaput tab=$tab keys=\"$keys\" $pairs}\n";
      $m->print_tu_fayl('dpllog', $dpllog);
    };
    my $sqlset = join ', ', grep { $_ } map { $m->sql_set_pair($_, $vals{$_}, $tab) } keys %vals;
    my $sqlwhere = join ' and ', map { $m->sql_get_pair($_, $keys{$_}, $tab) } keys %keys;
    my $sql = "update $tab set $sqlset";
    $m->pstop(q{refusing %k without where condition. stop}, k => $sql) unless $sqlwhere;
    $sql = join ' ', $sql,  'where',  $sqlwhere;
    my $rows = $m->sql_query($sql);
    $m->echo($sqlecho, 'UPDATE %n', n => $rows);
  ROWS: {
      #150321 local $_ = $rows;
      $m->dpuc('rows', $rows);
    INSERT: {
	$m->dpuc('insert', $rows, %opts);
	do { $m->dpop;last } if $rows;
	do { $m->dpop;last } if $opts{update_only};
	# sqlkeys don't need squote, sqlvals do
	my %keyvals = %keys;
	$keyvals{$_} = $vals{$_} for keys %vals;
	my $sqlkeys = join ', ', keys %keyvals;
	my $sqlvals = join ', ', map { $m->squoted_or_null($_) } values %keyvals;
	($rows) = $m->sql_query(qq{insert into $tab ( $sqlkeys ) values ( $sqlvals )}, 'edit');
	die $m->{dbi}->{$daba}->errstr if $m->{dbi}->{$daba}->err;
	$m->echo($sqlecho, q{INSERT %n}, n => $rows);
	$m->{dbi}->{$daba}->commit;
	$m->dpop('rows', $rows);
	last ROWS;
      };
      my $maxrows = $opts{maxrows} || $m->{konfig}->maxrows; 
      if (!$maxrows or $rows <= $maxrows) { $m->{dbi}->{$daba}->commit;last };
      $m->uwarn('NO WAY. You are not allowed to change more than %n rows at a time.', n => $maxrows) ;
      $m->{dbi}->{$daba}->rollback;
      $rows = 0;
      $m->dpop('rows',$rows);
    };
    $m->dpop1($prog,$rows);
};
sub upd {
    my ($m,$prog,$tab,$keysref,%vals) = shift->ppuc('upd',@_);
    my $rows = $m->put($tab, $keysref, { %vals }, update_only => 1);
    $m->dpop1('upd',$rows);
}

=head3 Function edit

An alias for 'put', deprecated, may be removed soon.

=cut

sub edit { my $m = shift;warn 'using deprecated alias edit for put';$m->put(@_); };

=head3 Function delete

Similar to 'put', but delete the found records.

Example

    $db->delete('users', { user=>'johndoe' });

    $db->delete('diary', { user=>'johndoe', day=>'2007-10-07' });


=cut

sub delete {
    my $m = shift;
    $m->dpuc('delete', @_);
    my $tab = shift;
    my $keysref = shift;
    my %opts = @_;
    my $daba = $m->pvar(daba => 1);
    my $sqlecho = $opts{sqlecho} || $m->{debug} || $m->{konfig}->sqlecho;
    $m->daba_konekt($daba);
  DPLLOG: {
      last unless $m->{konfig}->dpllog;
      my $keys = join " ", keys %$keysref;
      my @pairs_keys = map { join '=', $_, $keysref->{$_} } keys %$keysref;
      my $pairs = join " ", @pairs_keys;
      my $dpllog = " #. {dabadel tab=$tab $pairs}\n";
      $m->print_tu_fayl('dpllog', $dpllog);
    };
    my $sql_kond = $m->sql_kond($keysref);
    $m->ustop(q{attempted delete without where condition}) unless $sql_kond;
    local $m->{dbi}->{$daba}->{AutoCommit} = 0;
    my $rows = $m->sql_query(qq{delete from $tab where $sql_kond});
    die $m->{dbi}->{$daba}->errstr if $m->{dbi}->{$daba}->err;
    $m->echo($sqlecho, q{DELETE %n}, n => $rows);
  KOMROL: {
      $m->dpuc('komrol', $rows);
      do { $m->dpop;last } unless $rows;
      my $maxrows = $opts{maxrows};
      $maxrows = $m->{konfig}->maxrows unless defined $maxrows;  
      do { $m->{dbi}->{$daba}->commit;last } if !$maxrows or $rows <= $maxrows; 
      $m->{dbi}->{$daba}->rollback;
      $m->pstop('NO WAY. You are not allowed to change more than %n rows at a time in table %t', n => $maxrows, t => $tab) ;
      $m->dpop('komrol', $rows);
    };
    $m->dpop('delete', $rows);
    return $rows;
};

=head2 Build Select Queries

=head3 Function fldtyp_squoted

=cut

sub fldtyp_squoted {
    my $m = shift;
    my $val = shift;
    if (ref $val) { return join '.', @$val };
    my $fld = shift;
    my $tab = shift;
    my $typ = $m->fldtyp($fld, $tab);
  TYP: {
      last if $typ;
      local $_ = $val;
      return $m->squoted($_) if m(\A\d{4}\-\d{2}\-\d{2}\Z);
      return $_ if m(\A[\d\.]*\d\Z);
    };
    return $m->squoted($val) unless $typ eq 'n';
    return $val;
}

=head3 Function sql_kol_val

Build SQL column value comparison expression from a Perl key-value pair.

Takes a key and a more or less complex value.

    app => 'phpbb' ---> app = 'phpbb'
    app => ['=', 'phpbb'] --->  app = 'phpbb'
    app => ['like', '%phpbb'] ---> app like '%phpbb'
    app => ['in', 'a', 'b', 'c'] ---> app in ( 'a', 'b', 'c')
    app => ['~', '^phpbb'] ---> app ~ '^phpbb' 
    app => ['>', 'a'] --> app > 'a'
    app => ['<', 'z'] --> app < 'z'
    app => ['<=', 'z'] --> app <= 'z'
    app => 1 --> app = true  # if fldtyp('tab.app') eq 'boolean'

=cut

sub sql_kol_val {
    my $m = shift;
    $m->dpuc('sql_kol_val', @_);
    my $fld = shift; 
    my $val = shift;
    my $tab =  shift;
    do { $m->dpop;return } unless defined $fld and defined $val;
    my $sql = '';
  SIMPLEX: {
      my $ok = not ref $val;
      $m->dpuc('simplex');
      do { $m->dpop;last } unless $ok;
      $val = $m->squoted($val);
      $sql = join ' = ', $fld, $val;
      $m->dpop('sql_kol_val', $sql);return $sql;
    };
    my @vals = @$val;
    my $op = shift @vals;
    my $op_typ = $m->{konfig}->optyps->{$op};
    my $op_fmt = '%s';
  NO_OP_TYP: {
      my $ok = not $op_typ;
      $m->dpuc('no_op_typ', $ok, $op, $op_typ);
      do { $m->dpop;last } unless $ok;
      die sprintf 'invalid field value comparison operator %s', $op;
    };
    if (ref $op_typ) {
	($op_typ, $op_fmt) = @$op_typ;
    };
  TRY_OP: {
      $m->dpuc('try_op', $op_typ);
    NUM_OP: {
	my $ok = $op_typ eq 'n';
	$m->dpuc('num_op', $ok, $op_typ, 'n');
	do { $m->dpop;last } unless $ok;
	$sql = join ' ', $fld, $op, $vals[0];
	$m->dpop('sql_kol_val', $sql);
	return $sql;
      };
    ND_OP: { # numeric or date
	my $ok = $op_typ eq 'nd';
	$m->dpuc('str_op', $ok, $op_typ, 'nd');
	do { $m->dpop;last } unless $ok;
 	$sql = join ' ', $fld, $op, $m->fldtyp_squoted($vals[0], $fld, $tab); 
	$m->dpop('sql_kol_val', $sql);
	return $sql;
      };
    STR_OP: {
	my $ok = $op_typ eq 's';
	$m->dpuc('str_op', $ok, $op_typ, 's');
	do { $m->dpop;last } unless $ok;
	$sql = join ' ', $fld, $op, $m->fldtyp_squoted($vals[0], $fld, $tab); 
	$m->dpop('sql_kol_val', $sql);
	return $sql;
      };
    UNU_OP: {
	my $ok = $op_typ eq 'u';
	$m->dpuc('unu_op', $ok, $op_typ, 'u');
	do { $m->dpop;last } unless $ok;
	$sql = sprintf $op_fmt, $fld;
	$m->dpop('sql_kol_val', $sql);
	return $sql;
      };
    LST_OP: {
	my $ok = $op_typ eq 'l';
	$m->dpuc('lst_op', $ok, $op_typ, 'l');
	do { $m->dpop;last } unless $ok;
	@vals = map { $m->fldtyp_squoted($_, $fld, $tab) } @vals;
	$val = join ', ', @vals;
	$sql = $fld . ' in ( ' . $val . ' )';
	$m->dpop('sql_kol_val', $sql);return $sql;
      };
    RNG_OP: {
	my $ok = $op_typ eq 'r';
	$m->dpuc('rng_op', $ok, $op_typ, 'r');
	do { $m->dpop;last } unless $ok;
	my @ops = split m(\.\.), $op;
	@vals = map { $m->fldtyp_squoted($_, $fld, $tab) } @vals;
	$sql = join ' ', $vals[0], $ops[0], $fld, 'and', $fld, $ops[1], $vals[1]; 
	$m->dpop('sql_kol_val', $sql);
	return $sql;
      };
      die sprintf 'unknown operator type %s', $op_typ;
    };
    die 'sql_kol_val at end without result';
    return 1;
};

=head3 Function sql_kond

Build an SQL Where expression from a Perl hash as follows

Examples

{ person => ffii, date => ['<=..<', '2007-01-01', '2008-01-01' ] }

	--> person == 'ffii' and '2007-01-01' <= date and date < '2008-01-01'

	{ 'a.taktg' => [ '=',  [ 'b' 'dokktg' ] ] }
 
	-->  a.taktg == b.dokktg

Complex comparison functions can be expressed as list references as explained in sql_kol_val

=cut

sub sql_kond {
    my $m = shift;
    my $keysref = shift;
    my $tab = shift;
    my @sqls = map { $m->sql_kol_val($_, $keysref->{$_}, $tab) } keys %$keysref;
    @sqls = grep {$_} @sqls;
    my $sql = join ' and ', @sqls;
    return $sql;
};

=head3 Function get_select_tables

Accept a simple or complex expression describing a table or joined set
of tables, return the corresponding SQL string and the list of concerned tables.

Examples:

=over

=item 'taktg'

    --> 'taktg', 'taktg'

=item [ 'taktg', 'tadok' ]

    --> 'taktg natural join tadok', 'taktg', 'tadok'

=item [ [ 'taktg', 'k' ], [ [ 'tadok', 'd' ], 'join', 'k.ktg' => [ '=', 'd.dok' ] ] ]

    --> 'taktg as k join tadok as d on ( k.ktg == d.dok )', 'taktg', 'tadok'

=back

=cut

sub get_select_tables {
    my $m = shift;
    $m->dpuc('get_select_tables', @_);
    my $tabtok = shift;
    my ($tab, $tabstr, @tabrek, @tabs) = ();
    if (ref $tabtok) {
	my @tabtoks = @$tabtok;
	my $tabtok = shift @tabtoks;
	if (ref $tabtok) {
	    $tab = $tabtok->[0];
	    $tabstr = join ' as ', @$tabtok;
	} else {
	    $tab = $tabtok;
	    $tabstr = $tabtok;
	};
	push @tabs, $tab;
	my $s = $tabstr;
	while ($tabtok = shift @tabtoks) {
	    if (ref $tabtok) { # [ 'my_next_table', 'natural left outer join' ]
		@tabrek = @$tabtok;
		$tabtok = shift @tabrek;
		if (ref $tabtok) {
		    $tab = $tabtok->[0];
		    $tabstr = join ' as ', @$tabtok;
		} else {
		    $tab = $tabtok;
		    $tabstr = $tabtok;
		};
		my $join = shift @tabrek;
		if ($join eq 'join') { # [ 'next_table', 'join', 'a.taktg' => [ '=',  [ 'b' 'dokktg' ] ] ]
		    $s = join ' ', $s, 'join', $tabstr, 'on', '( ' . $m->sql_kond({ @tabrek }) . ' )';
		} else { # [ 'next_table', 'natural', 'left', 'outer', 'join' ]
		    $join = join ' ', $join, @tabrek;
		    $s = join ' ', $s, $join, $tabstr;
		};
	    } else {
		$tab = $tabtok;
		$tabstr = $tabtok;
		$s = join ' ', $s, 'natural join', $tabstr;
	    };
	    push @tabs, $tab;
	};
	$tabstr = $s;
    } else {
	$tabstr = $tabtok;
    };
    $m->dpop('get_select_tables', $tabstr, @tabs);
    return $tabstr, @tabs;
}

=head3 Function get_select_query

A subroutine of get_record and get_records, builds an SQL select statement from perl variables of the following structure:

 arg 1: tab: a string or complex expression denoting the table(s) to search
 arg 2: keysref: a reference to a hash, specifically a set of key value records where the key is a field and the value either a simple string or a reference to a list expression that specifies a condition to be met by the field, e.g. ['~', '^phpbb'] or ['like', '%phpbb'] or ['in', 'a', 'b', 'c']
 arg 3: opts: further modifiers, such as group, order, having, limit etc 

=cut

our %opt_form = ( group => 'group by %s', order => 'order by %s', where => 'where %s', having => 'having %s', limit => 'limit %d', offset => 'offset %s' );

sub get_select_query {
    my $m = shift;
    $m->dpuc('get_select_query', @_);
    my ($tabstr, @tabs) = $m->get_select_tables(shift);
    $m->dgot(tabstr => $tabstr, tabs => \@tabs);
    my $keysref = shift;
    my $fldsref = shift;
    my (@tabflds, %opts) = ();
    if (ref $fldsref) {
	@tabflds = @$fldsref;
	my $tok = shift || '';
	if (ref $tok) {
	    %opts = %$tok;
	} else {
	    while ($tok) {
		$opts{$tok} = shift; 
		$tok = shift;
	    };
	};
    } else {
	@tabflds = ($fldsref, @_);
    };
    $m->dgot(opts => \%opts, tabflds => \@tabflds);
    my ($query, $key, $val, $form, $tok, @q);
    if (ref $keysref) {
	my $flds = @tabflds ? join(', ', @tabflds) : '*';
	my @q = (qq{select $flds from $tabstr});
	my $where = $m->sql_kond($keysref, @tabs);
	$opts{where} =  $opts{where} ? join(' and ', $where, $opts{where}) : $where;
	foreach $key (qw(where group having order limit offset)) {
	    $val = $opts{$key};
	    next unless $val;
	    $form = $opt_form{$key};
	    die sprintf 'no opt_form for key %s', $key unless $form;
	    $tok = sprintf $form, $val;
	    push @q, $tok;
	};
	$query = join ' ', @q;
    } else { $query = $tabstr };
    $m->dpop('get_select_query', $query);
    return $query;
}

=head1 INTERFACE

=head2 Querying Functions

=head3 Function get_record

Return columns values of one found record.

Example:

    my ($mail, $sex) = $db->get_record('users', { user=>'johndoe' }, 'mail', 'sex');

=cut

sub get_record {
    my $m = shift;
    $m->dpuc('get_record', @_);
    my $query = $m->get_select_query(@_);
    my $daba = $m->pvar(daba => 1);
    $m->daba_konekt($daba);
    my @record = $m->{dbi}->{$daba}->selectrow_array($query); 
    if ($m->{dbi}->{$daba}->err) { $m->decho(join ': ', 'get_record', $query, $m->{dbi}->{$daba}->errstr) };
    _utf8_on($_) for @record;
    $m->dpop('get_record', @record);
    return @record;
};

=head3 Function get_record_tuples

Like get_record, but give back a hashref.
Not tested yet.

=cut

sub get_record_tuples {
    my ($m,$prog,@args) = shift->ppuc('get_record_tuples',@_);
    my $query = $m->get_select_query(@args);
    my $daba = $m->get_daba;
    $m->daba_konekt($daba);
    my $tups = $m->{dbi}->{$daba}->selectrow_hashref($query); 
    if ($m->{dbi}->{$daba}->err) { $m->decho(join ': ', 'get_record', $query, $m->{dbi}->{$daba}->errstr) };
    _utf8_on $_ for values %$tups;
    $m->dpop1($prog,$tups);
};

=head3 Function get_records

Return a list of references to retrieved records.

Example:

    @daily_entries = $db->get_records('diary', { user=>'johndoe' }, 'day', 'entry');

    foreach $daily_entry (@daily_entries) {
	($day, $entry) = @$daily_entry;
	print "entry of johndoe on day $day: $entry\n";
    };

=cut

sub get_records {
    my $m = shift;
    $m->dpuc('get_records', @_);
    my $query = $m->get_select_query(@_);
    my $sqlecho = $m->{konfig}->sqlecho;
    $m->echo($sqlecho, $query);
    print STDERR "query: $query\n";
    my $select;
    my $daba = $m->pvar(daba => 1);
    $m->daba_konekt($daba);
    $select = $m->{dbi}->{$daba}->prepare($query) or die sprintf 'prepare %s: %s', $query, $m->{dbi}->{$daba}->errstr;
    $select->execute or die sprintf 'execute %s: %s', $query, $m->{dbi}->{$daba}->errstr;
    my (@row, @rows) = ();
    while (@row = $select->fetchrow_array) {
	_utf8_on $_ for @row;
	push @rows, [ @row ];
    };
    $m->dpop('get_records', scalar @rows);
    return @rows;
};

sub get_tuple_records {
    my $m = shift;
    $m->dpuc('get_records', @_);
    my $query = $m->get_select_query(@_);
    my $select;
    my $daba = $m->pvar(daba => 1);
    $m->daba_konekt($daba);
    $select = $m->{dbi}->{$daba}->prepare($query) or die sprintf 'prepare %s: %s', $query, $m->{dbi}->{$daba}->errstr;
    $select->execute or die sprintf 'execute %s: %s', $query, $m->{dbi}->{$daba}->errstr;
    my ($row, @rows) = ();
    while ($row = $select->fetchrow_hashref) {
	_utf8_on $_ for values %$row;
	push @rows, $row;
    };
    $m->dpop('get_records', scalar @rows);
    return @rows;
};

=head2 Produce some frequently needed text output

=head3 Function sqltab

generate a sql table in HTML --- useful in an txt2html context but solved more elegantly by deplate.

=cut

sub sqltab {
    my ($m, $sql) = @_;
    $m->dpuc('sqltab', @_);
    my @records = $m->get_records($sql) ;
    my ($row, $fld);
    my $res = "<table>\n" . join("\n", map { "<tr>" . join('', map { "<td>$_</td>" } @$_) . "</tr>"; } @records) . "\n</table>";
    $m->dpop('sqltab', scalar @records);
    return $res;
};

=head3 Function sqltab_text

same for plain text, e.g. wiki source

=cut

sub sqltab_text {
    my $m = shift;
    $m->dpuc('sqltab_text', @_);
    my $sql = shift;
    my @records = $m->get_records($sql) ;
    my $res = join "\n", map { join " | ", @$_ } @records;
    $m->dpop('sqltab_text', scalar @records);
    return $res;
};

=head1 TODO

=cut

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile:/adv/perl/A2E/Daba.pm.tmpl
# End:
