=head1 NAME

       Hanzi::Konv

=cut

package A2E::Hanzi::Konv;

=head1 SYNOPSIS

	my $p = new_ready A2E::Hanzi::Konv;
	$p->set_konv('x2tw');				# set direction
	my $fanti = $p->konv('简体');			# convert string in direction
	$p->konv_fayl('jianti.txt', 'fanti.txt');	# convert file in direction

=head1 DESCRIPTION

rewrite string or file of simplified-character Chinese text in traditional characters, using publicly available conversion tables and obeying to all kinds of bells and whistles.

=head1 PREREQUISITES

	A2E::Prog(3) as base class with options inherited
	SDBM_File(3)
	DBM_Filter(3)
	Fcntl(3)
	conversion table data as in RPM package a2e-hanzi maintained at http://a2e.de/adv/pak/hanzi
        /sig/adv/hzfj

=cut

use base 'A2E::Prog', 'A2E::Tmplkonf', 'A2E::Tmplvars';
use Fcntl;
use SDBM_File;
use DBM_Filter;
use Encode qw(_utf8_on);
sub hanzi_konv_defvars {
    my ($m,$prog) = shift->qpuc('hanzi_konv_defvars',@_) or return;

=head1 OPTIONS

=over

=item --konv|-k C<x2cn>

conversion direction/table to use.  Choices are x2cn, x2tw, x2hk, x2sg, default is x2cn.

=cut

    $m->konfig_define(['konv','k'],'$', 'x2cn');
=item --tml|-t C<html>

file:/adv/perl/A2E/Tmplkonf.pm.tmpl::tmplang

=item --konv_tab_fmt|--tab C</etc/opt/a2e/hanzi/%s_tab.txt>

format by which to derive file name from a conversion parameter like I<x2tw>

Corresponding tables and SDBM files (see SDBM_File(3)) should be available at that location.
The file name format contains a conversion direction symbol such as x2tw as its only parameter.
    
=cut

$m->konfig_define(['konv_tab_fmt','tab'], '$', '/etc/opt/a2e/hanzi/%s_tab.txt');

=item --maxlen C<8>

maximal length of item to be looked up, counted in number of multibyte characters. default is 8.

=cut

$m->konfig_define('maxlen','i', 8);

=back

=cut

$m->{konv_lex} = {};

=head1 FILES

	/etc/opt/a2e/hanzi_konv.konf
	~/.hanzi_konv_konf
	./@hanzi_konv

=cut
    $m->konfayl('hanzi_konv.konf');
    $m->qpop1($prog,1);
};

=head1 BUGS

Test suite A2E::Hanzi::KonvTest helps to keep them out

=head1 IMPLEMENTATION

=head2 Constructor Functions

=head3 Function defvars

=cut

sub defvars {
    my ($m,$prog) = shift->qpuc('A2E::Hanzi::Konv::defvars',@_) or return;
    $m->SUPER::defvars;
    $m->tmplkonf_defvars;
    $m->hanzi_konv_defvars;
    $m->qpop1($prog,1);
};

=head3 Function set_konv

Tie DBM file to the %{$m->{konv_lex}} hash.

=cut

our $tk2fun = {
    latex => {
	x2cn => sub {
	    local $_ = shift;
	    s(\\begin\{(CJK\*?)\}\{UTF8\}\{bsmi\})(\\begin{$1}{UTF8}{gbsn});
	    s(\\CJKcaption\{zh\-Hant\})(\\CJKcaption{zh-Hans});
	    s(\A\\input\{(\w+)([\_\.])(zh\_tw|zh)\})(\\input{$1$2zh_cn});
	    return $_;
	},
	x2tw => sub {
	    local $_ = shift;
	    s(\{UTF8\}\{gbsn\})({UTF8}{bsmi});
	    s(\\CJKcaption\{zh\-Hans\})(\\CJKcaption{zh-Hant});
	    s(\A\\input\{(\w+)\_(zh\_cn|zh)\})(\\input{$1_zh_tw});
	    return $_;
	},
    },
};

our $tkxfun = sub { return shift };

sub set_konv {
    my ($m,$prog,$konv) = shift->ppuc('set_konv',@_);
    return $m->dpop($prog) if defined $m->{konv_db};
    my $fi = sprintf $m->{konfig}->konv_tab_fmt, $konv;
    $m->ustop('vocabulary file %f not found', f => $fi) unless -r $fi;
    $m->{konv_db} = tie %{$m->{konv_lex}}, 'SDBM_File', $fi, O_RDONLY, 0666 or die sprintf 'can not tie file %s to hash: %s', $fi, $!;
    my $tml = $m->pvar(tmplang => 1);
    my $fun = $tk2fun->{$tml}->{$konv};
    $tkxfun = $fun if $fun;
    $m->{konv_db}->Filter_Push('DBM_Filter::utf8');
    $m->{konfig}->konv($konv);
    $m->dpop1($prog,$konv);
}

=head3 Function postkonfig

=cut

sub postkonfig {
    my ($m,$prog) = shift->qpuc('A2E::Hanzi::Konv::postkonfig',@_) or return;
    $m->SUPER::postkonfig;
    $m->set_konv($m->{konfig}->konv);
    $m->qpop1($prog,1);
};

=head2 Executor Functions

=head3 Function konv

Convert one string

=cut

sub konv {
    my ($m,$prog,$s) = shift->ppuc('konv',@_);
    my $maxlen = $m->{konfig}->maxlen;
    my @chars0 = unpack 'U*', $s;
    my @chars1 = ();
    my $i = 0;
  LEX: while ($i <= $#chars0) {
      if ($chars0[$i] == 32 and $chars0[$i-1] > 8192 and $chars0[$i+1] > 8192) { $i++;next LEX }; # remove spaces between hanzi
      if ($chars0[$i] < 8192) { # non-hanzi
	  push @chars1, $chars0[$i];
	  $i++;next LEX;
      };
    MAXLEN: { # look up item of maximal length maxlen, giving preference to two-char words over one-char words
	my $len = $#chars0 - $i + 1;
	$len = $maxlen if $len > $maxlen;
	while ($len > 0) {
	    my $s0 = pack 'U*', @chars0[$i..$i-1+$len];
	    my $s1 = $m->{konv_lex}->{$s0} || '';
	    if (length $s1) {
		$m->decho("+$s0+$s1");
		my @s1_chars = unpack 'U*', $s1;
		push @chars1, @s1_chars;
		$i += $len;
		next LEX;
	    };
	} continue { $len-- };
      };
      push @chars1, $chars0[$i];
      $i++;
  };
    $s = pack 'U*', @chars1;
    $s = &$tkxfun($s);
    $m->dpop1($prog,$s);
};

=head3 Function konv_fayl

Read file arg1 and write to file arg2

=cut

sub konv_faylz {
    my ($m,$prog,@args) = shift->ppuc('konv_faylz',@_);
    my $n = 0;
    # binmode STDOUT, ':utf8';
    for (@args) {
	_utf8_on $_;
	unless (-r $_) { print $m->konv($_), "\n";next };
	$m->open_fayl($_, 'i', 'r');
	$n++;
	while ($m->getlini('i')) {
	    my $s = $m->konv($_);
	    print $s;
	};
	$m->kloz_fayl('i');
    };
    $m->sav2out;
    $m->dpop1($prog,$n);
};

=head1 The END

=cut

1;

# Local Variables:
# coding: utf-8
# srcfile:/adv/perl/A2E/Hanzi/Konv.pm
# mode: perl
# End:

