# Time-stamp: <2022-07-02 09:51:40 phm>

=head1 NAME

A2E::Lokvars

=head1 VERSION

This describes version B<0.0.1>

=cut

$VERSION = '0.0.1';

=head1 DESCRIPTION

Writing local variables and other meta information that can be read by editing
programs like Emacs.

This class needs to be mixed into either A2E::Prog or one of its descendants.

This method is of a mix-in type, i.e. it contains no fields and its methods have names which try not to duplicate any names of methods used in any other classes.  Other mix-in libraries include 

    A2E::Osver
    A2E::Sqltab

Since it contains only methods, no fields, and none of these is intended to
overwrite any other method of another class, invocation with the base pragma
in the sens of

	package MyApp;
	use base 'A2E::Daba', 'A2E::Lokvars';
        sub defvars {
	    my $m = shift;
	    $m->SUPER::defvars;
	    $m->lokvars_defvars();
            $m->{konfig}->define('my_app_var|a=s');
	    [...]
	    $m->konfayl('my_app', 'my_app.konf');
	    return 1;		
	}
	sub process {
	    my $m = shift;	
	    [...]
	}
	package Main;
	my $app = new_ready MyApp;
	$app->process;

will do the job.

=head1 BUGS

Seems to work fine.

=head1 PREREQUISITES

       A2E::Prog(3)

=head1 APPLICATIONS

        db2auth(1)
 	mlht(1)
	A2E::Mktdir::Mktdir(3)

=head1 SEE ALSO

    mktdir5(1)
    db2auth(1)
    perl(1)

=head2 Alternatives on multiple inheritance

    mixin(3)
    Class::C3(3)

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

    Copyright (c) 2007 Hartmut Pilch (phm)

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

=head2 Header

=cut

package A2E::Lokvars;
use base 'A2E::Prog';
use File::Spec::Functions;
use strict;

=head2 Construction and extension of underlying object

We inherit the constructor 'new' from the A2E::Prog package, but supply 'lokvars_defvars' to be used within the 'setup' routine of applications. 

=head3 Function lokvars_defvars

=cut

sub lokvars_defvars {
    my $m = shift;
    $m->dpuc('lokvars_defvars', @_);
    do { $m->dpop;return } if $m->{ready}->{lokvars};
    $m->konfig_define('emacs_local_variables', 's%');
    $m->konfayl('lokvars.konf');
    $m->{ready}->{lokvars} = 1;
    $m->dpop('lokvars_defvars', 1);
    return 1;
};

=head2 Elements of Emacs text files

Nice text chunks for the header and footer of text files in formats that are recognised by Emacs and described in the GNU Emacs info manual.

This should be implemented as a mix-in.

=head3 Function emacs_timestamp

Return a string like

    Time-stamp: <2008-01-14 09:54:46 phm>

that can be filled in by the C<time-stamp> function of Emacs Lisp, if it is
placed at the beginning of the text file in a comment syntax.

=cut

sub emacs_timestamp {
    my $m = shift;
    $m->dpuc('emacs_timestamp', @_);
    my $s = join(': ', 'Time-stamp', $m->enclose('<', '>', join(' ', $m->timestamp('%Y-%m-%d %H:%M:%S'), $m->{user})));
    $m->dpop('emacs_timestamp', $s);
    return $s;
};

=head3 Function emacs_init_local_variables

Common part of emacs_top_local_variables and emacs_bottom_local_variables.

Needs same arguments as emacs_top_local_variables: a reference to a variable hash and an optional list of variable names.

=cut

sub emacs_init_local_variables ($$@) {
    my $m = shift;$m->dpuc(emacs_init_local_variables => \@_);
    my ($elvs,@lvnoms) = @_;
    my %khac = ();
    if (@lvnoms) {
	foreach my $k (@lvnoms) { next unless $elvs->{$k};$khac{$k} = $elvs->{$k} };
    } else { 
	%khac = %$elvs ;
    };
    my $kelvs = $m->{konfig}->emacs_local_variables;
    for (keys %$kelvs) { $elvs->{$_} ||= $kelvs->{$_};$khac{$_} = $elvs->{$_} };
    @lvnoms = keys %khac;
    my ($pre, $post) = ('','');
    for ($elvs->{mode}) {
	do { $pre= '['.'%'.'#';$post = '%'.']';last } if m(\Att\Z); # 小心模板轉換 file:/phm/18/12/06/sig/_lng.phm_pub_sig181206.txt::ttlvars
	do { $pre= '%';last } if m(\A(tex|latex|deplate)\Z);
	do { $pre= ';';last } if m(\A(lisp|)\Z);
	do { $pre= '<!--';$post = '-->';last } if m(\A(html|xml|sgml)\Z);
	$pre = '#';
    };
    $m->dpop(emacs_init_local_variables => [$pre, $post, $elvs, @lvnoms]);return $pre, $post, $elvs, @lvnoms;
};

=head3 Function emacs_top_local_variables

=over

=item arg1: variable hash 

reference to a variable hash, e.g.

	{ dokl => 'ProgPerlEn', srcfile => '/adv/perl/A2E/Prog.pm', junk => 'yippee!' }

=item arg2ff: keys for selection from variable hash

optional list of variable names, e.g.

	qw(dokl srcfile)

If nothing is provided then the whole variable hash is used.

=item configuration parameters

The emacs_local_variables hash contains some defaults, e.g. values for mode and coding.

=item result

A string to be written at the top of Emacs files, specifying file variables like

    -*- dokl: ProgPerlEn; coding: utf-8; mode: fundamental; srcfile: /adv/perl/A2E/Prog.pm -*-

=back

=cut

sub emacs_top_local_variables ($$;@) {
    my $prog = 'emacs_top_local_variables';my $m = shift;$m->dpuc($prog, @_);
    my ($elvs,@lvnoms) = @_;
    (my $pre, my $post, $elvs, @lvnoms) = $m->emacs_init_local_variables($elvs, @lvnoms);
    my $s = join ' ', grep {$_} $pre, $m->enclose('-*- ', ' -*-', join '; ', grep {$_} map { $_ && $elvs->{$_} && join ': ', $_, $elvs->{$_} } @lvnoms), $post;
    $m->dpop($prog, $s);return $s;
};


=head3 Function emacs_bottom_lokvars_fun

Function generating closure

=cut

sub make_emacs_bottom_lokvars_fun {
    my $m = shift;$m->dpuc(make_emacs_bottom_lokvars_fun => \@_);
    my ($pre,$post) = @_;
    if ($post) {
	return sub { my $s = shift;return join ' ', grep {$_} $pre, $s, $post };
    } else {
	return sub { my $s = shift;return join ' ', grep {$_} $pre, $s; };
    };
    $m->dpop(make_emacs_bottom_lokvars_fun => 1);return 1;
};

=head3 Function emacs_bottom_local_variables

Example:

    $m->emacs_bottom_local_variables({ dokl => 'ProgPerlEn', junk => 'yippee!', mode => 'fundamental' }, qw(coding srcfile));

should return something like

    <!-- Local Variables: -->
    <!-- dokl: ProgPerlEn -->
    <!-- mode: fundamental -->
    <!-- coding: utf-8 -->

assuming that the emacs_local_variables configuration variable has the I<mode> and I<coding> keys preset to I<fundamental> and I<utf-8> respectively.

Arguments are similar to emacs_top_local_variables, except that as arg 1 a format specification is needed.

We are misspelling "Lokal Variables" on purpose, in order not to escape from interpretation of this by Emacs.
 
=over

=item arg1

reference to a hash containing local variables

=item arg2

optional list of names of variables to be included

=back

=cut

sub emacs_bottom_local_variables {
    my $m = shift;$m->dpuc(emacs_bottom_local_variables => \@_);
    my ($pre, $post, $elvs, @lvnoms) = $m->emacs_init_local_variables(@_);
    my $fun = $m->make_emacs_bottom_lokvars_fun($pre, $post);
    my $s = join "\n", map { &$fun($_) } ('Local Variables:', (grep {$_} map { $_ && $elvs->{$_} && sprintf '%s: %s', $_, $elvs->{$_} } @lvnoms), 'End:');
    $m->dpop(emacs_bottom_local_variables => $s);return $s;
};

=head3 Function set_emacs_lokvars

=cut

our @lvnoms = ('mode','coding');    
sub emacs_lokvars {    
    my ($m,%opts) = @_;
    my $elvs = $m->{konfig}->emacs_local_variables;
    foreach my $opt (keys %opts) { $elvs->{$opt} = $opts{$opt} };
    my $lokvars = {
	emacs_top_local_variables => $m->emacs_top_local_variables($elvs, @lvnoms),
	emacs_bottom_local_variables => $m->emacs_bottom_local_variables($elvs, @lvnoms),
	emacs_timestamp => $m->emacs_timestamp,
	mode => $elvs->{mode},
    };
    $lokvars;
};
sub set_emacs_lokvars {
    my ($m,$prog,$lihs,%opts) = shift->lpuc('set_emacs_lokvars',@_);
    my $lokvars = $m->emacs_lokvars(%opts);
    while (my ($k,$v) = each %$lokvars) { $lihs->putstr($k => $v) };
    $lokvars;
};

=head1 END

=cut

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/Lokvars.pm.tmpl"
# End:
