
package A2E::Mktdir;
use utf8;

=head1 NAME

A2E::Mktdir

=head1 DESCRIPTION

Routines for creating hypertext nodes.

An intermediate layer between mktdir5(1) and A2E::Dokfs, 
does the work of mktdir5, mktdir4 and mktdir0.

=head1 PREREQUISITES

    A2E::Subdoks(3)
    A2E::Mbox2dok(3)
    Template(3)

=cut

use base 'A2E::MLDK', 'A2E::Subdoks', 'A2E::Mbox2dok';
use strict;
use Data::Dumper;
use utf8;
use Encode qw(_utf8_on decode encode);
use English -no_match_vars;
use POSIX qw(strftime setgid getgrnam);
use File::Spec::Functions qw(splitdir splitpath catdir catfile);
$Template::BINMODE = ':utf8';use Template;
use Term::ReadLine;
use Cwd qw(getcwd);
sub defvars {
    my $m = shift;
    $m->dpuc('A2E::Mktdir::defvars', @_);
    
=head1 VERSION

This describes version B<0.0.3> of A2E::Mktdir.

=cut

    $m->{progver} ||= '0.0.3';
    $m->SUPER::defvars;
    $m->subdoks_defvars;
    # $m->tmplfil_defvars; included in A2E::MLHT
    $m->mbox2dok_defvars;
    $m->{$_} = {} for qw(lngtags);
    $m->{$_} = [] for qw(oldlangs novlangs suplangs);
    $m->{$_} = '' for qw(ftmp);
    $m->{timestamp} ||= $m->timestamp;

=head1 OPTIONS

=over

=item --rekurs_mktdir|-R

recursively create subdirectories that may be defined using the sems2subtmpl and sems2subsems chunks.

=cut

    $m->konfig_define(['rekurs_mktdir', 'R'], '!');

=item --basurl I<url>

=cut

    $m->{konfig}->define('basurl=s');
=item --http I<http>

=cut

    $m->{konfig}->define('http=s');

=item --langnoms de=Deutsch

=cut

    $m->cache_define('langnoms', 's%');

=item --lang0 de

=cut

    $m->{konfig}->define('lang0=s');

=item --minimal

=cut

    $m->{konfig}->define('minimal|minim!');

=item --topsty tra

Document style.  By setting this you imply that the document is a top node.  Otherwise the style is inherited from the topmost preceding node (topdok).

=cut

    $m->konfig_define(['topsty', 'top'], 's');

=item --mailuser adv

=cut

    $m->konfig_define('mailuser', 's');

=item --conv deplate

=cut

    $m->konfig_define('conv', 's', 'deplate');

=item --menuitem_template

Obsolete, maybe still used for legacy directories at a2e?

=cut

    $m->{konfig}->define('menuitem_template=s%');

=item --mailhost a2e.de

=cut

    $m->{konfig}->define('mailhost=s');

=item --mailadr adv@a2e.de

=cut

    $m->{konfig}->define('mailadr=s');

=item --makefile_tmpl makefile_tmpl.txt

=cut

    $m->{konfig}->define('makefile_tmpl=s', { DEFAULT => 'makefile_tmpl.txt' });

=item --skrib_antefaylz

write language-specific parameter files, named as specified in antefnom

=cut

    $m->{konfig}->define('skrib_antefaylz!');

=item --tmplvars|--tv|-V mm=01 -V aa=09 -V thema=Italia,de:Italien,en:Italy,fr:Italie

Extra template variables to be set at the beginning and written into the local --antefnom0 (@pre.mydok.txt) or --antefnom (pre.mydok.en.txt) file. 
These variables are also written into the language-neutral --antefnom0 file, so that they precede any vocabulary definitions at compile-time.
The values of these variables can, even though this is not recommended, be multilingual, using the syntax indicated above.  In this case, by implication, skrib_antefaylz is true.
We recommend monolingual --tmplvars values because these variables are, in contrast to the langlexvars, intended to be more closely related to process control than to the textual content of the documents.

=cut

    $m->{konfig}->define('tmplvars|tv|V=s%'); 

=item --langlexvars|llv can=";canned definition;abgepackte Definition" --llv can2='|another ${can}|noch ne ${can}'

extra multilingual textchunk groups to be written into vocubulary files postfayl such as _lng.mydok.en.txt.

These are evaluated to become 

	$m->{langlexvars}->{can}->{en} = 'canned definition'
	$m->{langlexvars}->{can}->{de} = 'abgepackte Definition'

, provided that the languages of the document to be created are en and de.

The value is retrieved&cached using the function C<langlexvar>.

As of 2009-12-12, it seems that we may never want to use this, given that these variables are read late and belong to the text level rather than the program operation level.  For the latter purpose, we already have --tmplvars|-V.

=cut
    
    $m->{konfig}->define('langlexvars|llv=s%');

=item --lexvars|--lv aa=09 --lv mm=01

Extra variables to be written into language-neutral vocabulary file (unless that already exists).
This option is not very useful because the files written to the antefnom (parameter element) file is more important.

=cut

    $m->{konfig}->define('lexvars|lv=s%');

=item --tmplang_outfmts multemp="html pdf" mlwebdb="html"

=cut

    $m->{konfig}->define('tmplang_outfmts|outfmts=s%');

=item --tratoks_extra 'kws'

=cut

    $m->{konfig}->define('tratoks_extra=s@');

=item --subopts_vars getkanals

Variables that can appear as elements in the subopts hash, i.e. those that have special values in the descendant (creatable) document which need to be specified in that descendant's local configuration file (e.g. _dokfs), exist in this hash as keys with a corresponding true value.  

This is used by &A2E::Privat::Lnsig::do_mktdir to determine which variables belong to the descendant level.

=cut

    $m->{konfig}->define('subopts_vars=s%');
=item --topdok_kopi_extra

Populate a topdok directory with lots of templates that normally are not needed.  Do this if you want to extensively customise your section.

# =cut
#     $m->{konfig}->define('topdok_kopi_extra!');
# =item --orig_annus 1987
#
# origo year for a 1-9,A-Z year counting system used at a2e.de
# 
# =cut
#    $m->{konfig}->define('orig_annus=i', { DEFAULT => 1987 });

=item --fol testM4 --dok testM4

leaf of directory to be created.
normally we only give a sems (semantic features) argument and let the system derive the leaf (--fol) and the document (--dok) based on rules specified in the sems2fol and sems2dok template variables, but we can also override these rules and explicitely name these items.

=cut
    $m->{konfig}->define('fol=s');
    $m->konfig_define('subdok_regexp', 's', '\A[a-z][a-z,0-9,A-Z,_]+\Z');
    $m->konfig_define('subfol_regexp', 's', '\A\w+\Z');
    $m->konfig_define('logourl', 's', '/a2e.jpg');

=back

=head2 Quasi-Options realised as Textchunk Language Variables

=over

=item sems2subtmpl

We used to have an option --tmpl_subtmpls oas_akt = tra --tmpl_subtmpls oas_aa_ont = oas_aamm --tmpl_subtmpls pub_aa = ont:pub_ont_aa,int:pub_int_aa,tem:pub_tem_aa.

This was used to specify the default subtemplates (--subtmpl plus --subtmpls, see A2E::Tmplkonf(3)) that belong to each template.
The key is a --tmpl, the argument consisted of a --subtmpl (obligatory) and an optional list of keyvals in which each key is a subdirectory name (usable as mktdir argument no 1) and each val is the subtmpl that is to be used for such a subdirectory.

Now we use a textchunk variable instead, e.g.

	in tmpl/a2e/oas_akt/lang.txt:

	sems2subtmpl = $(case ${1},tra,ont,tra_ont)

=item sems2subsems

Subdirectories to be obligatorily created for each directory of a given document type.
These are leaf names which should have corresponding document types assigned by sems2subtmpl
Formerly the variables tmpl_subtmpls or subtmpls were used for this.

Now we use the a textchunk variable instead, e.g.

      in a2e/pub_aa/lang.txt:

	sems2subsems = int ont tem

      in a2e/pub_int_aa/lang.txt

        sems2subsems = $(fromto 01,12)

      in a2e/pub_int_aamm/lang.txt

        sems2subsems = $(mondags ${aaaa},${mm})

TODO (done already?):
/adv/tmpl/a2e/pub_ont_aamm/lang_pre.txt
SArb/Make.pm monwoks
support constructions such as what is to be returned by render_monwoks
  
    |W26,01|w27,02,03,04,05,06,07,08|w28,09,10,11,12,13,14,15|w29,16,17,18,19,20,21,22|w30,23,24,25,26,27,28,29|w31,30,31|

or

    |W26 01 01|w27 02 08|w28 09 15|w29 16 22|w30 23 29|w31 30 31|

where each primary argument consists of a list of sems.

OBSOLETE: We used to have an option --tmpl_subsems pub_aa = int,ont,tem --tmpl_subsems pub_int_aa = 01,02,03,04,05,06,07,08,09,10,11,12.

=item getkanals

	getkanals = oas_akt 
	getkanals = eupat_ont,eupat_int

This ensures that the webpage generated from the current document has the named rss channels in its news column.

Moreover, if a textchunk of this name has been defined (e.g. in the document type definition or in superordinate document vocabularies), it is written into the _dokfs file (by the function C<vars_skrib_konf>).

OBSOLETE?:
All documents that use a certain template may use certain rss channel(s).
Specify this as a rule, expressed by a hash, in the same way as for individual channels it is done by the getkanals variable.

=item putkanals

	putkanals oas_tra = oas_akt
	putkanals = oas_adv,adv

If a textchunk of this name has been defined (e.g. in the document type definition or in superordinate document vocabularies), it is written into the _dokfs file (by the function C<vars_skrib_konf>).

OBSOLETE?:
All documents that use a certain template may be suitable for insertion into certain news/rss channels.
Specify this as a rule, expressed by a hash, in the same way as for individual channels it is done by the --putkanals option.

=item grpdir

	grpdir = 1

Documents of this type have their descendants in peer directories rather than subdirectories.

Deprecated, better specify C<sems2subflat> in the document type lang_pre.txt file.

=back

=head1 FILES

    mktdir.konf

=cut

    $m->konfayl('mktdir.konf');
    $m->{subsems_funs} = {};
    $m->dpop('A2E::Mktdir::defvars', 1);
    $m->{term} = new Term::ReadLine $m->{prognom};
    return 1;
};

=head1 BUGS

None open at the moment

=head1 LIMITATIONS

=over

=item one document template set for everything

=back

=head1 SEE ALSO

    mktdir0(1)
    mktdir5(1)
    A2E::Dokfs(3)

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

    Copyright (c) 2008-9 Hartmut Pilch (phm)

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

Names in this program are based on English with International Latin spelling, e.g. 'kopi' for 'copy' and 'meikfayl' for 'makefile', as (to be) explained in

  http://a2e.de/alfabet/interlatin

with with occasional Esperanto (such as replacing words such as 'write'/'rait', that become unwieldy in interlatin spelling, with 'skrib')

  http://a2e.de/lang/eo

for which the same spelling is used.

=cut

=head2 Global switches and variables for internal configuration

=head3 @*_tmplvars et al

Template variables that are not reset or treated specially when template variables are read in from files.

=cut

our @tratoks = ('fol', 'lab', 'tit', 'sut', 'des');

# variables that are not read from textchunk files during any run -- probably obsolete, use __user __hodie for current internal state instead
our %basis_non_tmplvars = ();$basis_non_tmplvars{$_} = 1 for qw(mktdir_time mktdir_user);

# variables that must be null before a further document is read.
# these are typically specified in lang_pre.txt or @pre.*.txt files with '?=' assignment statements and will therefore retain their obsolete values if not wiped out before reading

# variables that are no longer read after the first generation-run of one document (e.g. in a second language)
# OBSOLETE? 2010-03-10 發現有害無益，但尚待觀察
our @extra_non_tmplvars = qw(aa aaaa mm dd ddmax KA KAM KAMD);

# variables that are deleted before any further run (in another language), so that they are freshly read once even if specified with ?=
our @extra_del_tmplvars = (@tratoks, 'thema', 'AA', 'MM', 'DD', 'lab_valids', 'tit_valids', 'sut_valids', 'des_valids');

# variables that need to be set in any document type definition /tmpl/a2e/*/lang_pre.txt: if any of these is null the document type is incomplete
our @nonnull_tmplvars = qw(sems2dok sems2fol);

# persistent variables that are not deleted when proceding to a new subdocument in recursive mode
our %rekurs_sav_tmplvars = ();$rekurs_sav_tmplvars{$_} = 1 for qw(user hodie mktdir_time mktdir_user outfmts sems outfmts);
# further candidates, needed when rekurs_del_tmplvars was not followed by rereading: no_subsems privat dokstem topdok topdir supdir supdok hodie sems tmpl

# ?variables that are copied from the object $m into the template variables store $m->{tmplvars}?
our @kopi_tmplvars = qw(dok dir fol tmpl topdir topdok supdir supdok hodie flatsupdok);

# ?variables that are copied from the configuration store $m->{konfig} into the template variables store $m->{tmplvars}
# this is done in postkonfig at the beginning only, so as not to interfere with subdirectory recursion
our @konfig_tmplvars = qw(topsty logourl tmpl);

=head2 General Basic Routines

=head3 Function copyfile

Copy a file from $1 to $2.
A wrapper around Unix cp.

=cut

sub kopi_fayl {
    my $m = shift;
    my @komand = grep {$_} ('cp', $m->{verbose} && '-v', '-f');
    return $m->exek(@komand, @_);
};

=head3 Function tcek_mailadr_varz

=cut

sub tcek_mailadr_varz {
    my $m = shift;
    $m->dpuc('tcek_mailadr_varz', @_);
    do { $m->dpop('tcek_mailadr_varz');return 1 } if $m->{konfig}->mailuser and $m->{konfig}->mailhost;
  MAILADR_OK: {
      $m->dpuc('mailadr_ok', $m->{konfig}->mailadr);
      do { $m->dpop;last } unless $m->{konfig}->mailadr;
      my ($user, $host) = split '@', $m->{konfig}->mailadr;
      $m->ustop('invalid mail address %s', $m->{konfig}->mailadr) unless $host;
      $m->{konfig}->mailuser($user);
      $m->{konfig}->mailhost($host);
      $m->dpop('mailadr_ok', $user, $host);
    };
  USER_HOST_KO: {
      $m->dpuc('user_host_ko', $m->{konfig}->mailuser, $m->{konfig}->mailhost);
      do { $m->dpop;last } if defined $m->{konfig}->mailuser and defined $m->{konfig}->mailhost;
      $m->konfinit('mailhost', $m->{konfig}->basurl);
      $m->konfinit('mailuser', 'info');
      $m->dpop('user_host_ko', $m->{konfig}->basurl, 'info') ;
    };
    $m->{konfig}->mailadr(join '@', $m->{konfig}->mailuser, $m->{konfig}->mailhost);
    $m->dgot(mailadr => $m->{konfig}->mailadr, mailuser => $m->{konfig}->mailuser, mailhost => $m->{konfig}->mailhost);
    $m->dpop('tcek_mailadr_varz', 1);
    return 1;
};

=head3 Function gdp_set_pwd

If the current node is a group node (pwd_get_subflat) and we are not using a dot argument, go to the parent directory.

    arg1 : subflat
    arg2 : dotdir_p

This must be done after reading the parent doctype info and before reading the descendant doctype info.

=cut

sub gdp_set_pwd {
    my $m = shift;
    $m->dpuc('gdp_set_pwd', @_);
    $m->set_pwd;
    my $subflat = shift || $m->{subflat};
    do { $m->dpop;return } unless $subflat;
    my $dotdir_p = shift || $m->{dotdir_p};
    my ($vol, $path, $fol) = splitpath getcwd;
    $path = $m->dir2dir($path);
    $m->dgot(vol => $vol, path => $path, fol => $fol);
  NOFOL: {
      $m->dpuc('nofol');
      do { $m->dpop;last } if $fol;
      $m->ustop("cant parse working directory %s.", $m->{pwd});
    };
    $m->{supdir} = $m->{pwd};
  CHWD: {
      last if $dotdir_p;
      $m->{pwd} = $path;
      $m->chwd($m->{pwd});
    };
    $m->{flatsupbas} = $m->{pwd};
    $m->{flatsupfol} = $fol;
  NOFOLDIR: {
      my $ver = ($dotdir_p || $m->{flatsupfol}) ? 0 : 1;
      $m->dpuc('nofoldir', $ver, $dotdir_p, $m->{flatsupfol});
      do { $m->dpop;last } unless $ver;
      $m->ustop('cant find a leaf in menuwise parent %s', $m->{supdir});
    };
    $m->{flatsupdok} = $m->dirdok($m->{flatsupbas});
    $m->dpop('gdp_set_pwd', $m->{flatsupbas}, $m->{flatsupfol});
    return $m->{flatsupbas}, $m->{flatsupfol};
};

=head3 Function fol2dir

Turn a directory value as specified on the commandline, e.g. '.' or a leaf or an absolute dir, into an absolute directory

Optional second argument $pwd is cleaned (as returned by dir2dir) present working directory 

=cut

sub fol2dir {
    my $m = shift;
    $m->dpuc('fol2dir', @_);
    my $fol = shift;
    my $dir = $fol;
    my $pwd = shift || $m->{pwd};
  DOT: {
      $m->dpuc('dot', $fol);
      do { $m->dpop;last } unless $fol eq '.';
      $dir = $pwd;
      $m->dpop('fol2dir', $dir);
      return $dir;
    };
  REL: {
      $m->dpuc('rel', $fol);
      do { $m->dpop;last } if $fol =~ m(\A$m->{dirsep_re});
      $dir = catdir $pwd, $fol;
      $m->dpop('fol2dir', $dir);
      return $dir;
    };
    $m->dpop('fol2dir', $dir);
    return $dir;
};

=head3 Function sems2dok

A typical use might be

    $m->sems2dok('victoria', '31')  ==> victoria090731

where two semantic features, given as arguments, are translated into a
document symbol according to a rule specified by the user via template
variables like sems2dok or other mechanisms to be implemented here.

=cut

sub sems2dok {
    my $m = shift;
    $m->dpuc('sems2dok', @_);
    my @sems = @_;
    $m->uwarn('no sems arguments given to sems2dok') unless @sems;
    my $val = $m->{tmplvars}->{sems2dok};
    $m->dgot(val => $val);
    $m->ustop('no sems2dok variable known in document type %s', $m->{tmpl}) unless $val;
    my $dok = $m->{sarb}->val_args_call($val, undef, @sems);
    do { $m->ustop(q{can't obtain dok by 'call'ing sems2dok on %s}, $m->koniug_et(@sems));$m->dpop;return } unless $dok;
    $m->dpop('sems2dok', $dok);
    return $dok;
}

=head3 Function sems2fol

Like sems2dok, but generate current directory leaf name rather than document symbol.

=cut

sub sems2fol {
    my $m = shift;
    $m->dpuc('sems2fol', @_);
    my @sems = @_;
    do { $m->dpop;return } unless @sems;
    my $val = $m->{tmplvars}->{sems2fol};
    my $fol = $m->{sarb}->val_args_call($val, undef, @sems);
    do { $m->ustop(q{cannot obtain fol by 'call'ing sems2fol %s on %s}, $val, $m->koniug_et(@sems));$m->dpop;return } unless $fol;
    $m->dpop('sems2fol', $fol);
    return $fol;
}

=head3 Function rekurs_del_tmplvars

Prevent reuse of variables from previous document except special cases defined in %rekurs_sav_tmplvars.

Takes an options hash as argument rest.

This does not delete the group properties $m->{nom2lits}; that is done by lang_set_tmplvars because needed each time.

OBSOLETE?:
If option C<extra> is true, shield certain already-defined variables that are not template-specific and do not change in a subdocument, such as those in C<@basis_non_tmplvars> and C<@extra_non_tmplvars>, from deletion.

If option C<minim> is true, delete only very few tmplvars, namely those that definitely can not be inherited from another document, e.g. because they are document-type-specific.

Mark all variables as not-yet-read-from-file, i.e. make %{$m->{finivars}} empty. 

=cut

sub rekurs_del_tmplvars {
    my $m = shift;
    my %optz = @_;
    %{$m->{finivars}} = ();
    $m->dpuc('rekurs_del_tmplvars', @_);
  TMPLVAR: foreach my $nom (keys %{$m->{tmplvars}}) { 
      $rekurs_sav_tmplvars{$nom} = 1 if $nom =~ '\A\_';
      next if $rekurs_sav_tmplvars{$nom} or $m->{non_tmplvars}->{$nom};
    NON_RO_MINIM: {
	# when minim =>1, leave all variables except the read-one ones such as 'sems2subtmpl' & co alone
	last if $m->{tmplvars_ro}->{$nom};
	# version until 2010-03-10: next if $optz{extra} and $m->{non_tmplvars}->{$nom};
	# use of optz extra should be obsolete since we no longer support @extra_non_tmplvars
	# it was wrong in the first place to ever delete any non_tmplvars
	next TMPLVAR if $optz{minim};
      };
      delete $m->{tmplvars}->{$nom};
  };
    $m->dpop('rekurs_del_tmplvars', 1);
    return 1;
}

=head3 Function set_dir_varz

Set the subdirectory leaf argument $fol nr 1 either directly or based on a semantic element/structure $sem nr 2, determine the tmpl/doctype of the document to be created and, based on that, the values of related variables that refer to related documents and their locations, e.g. 

    $m->{supdok}:	parent node
    $m->{supdir}:	absolute path of parent node directory
    $m->{dir}:		absolute path of directory to be created
    $m->{subflat}:      whether $m->{dir} is directory-wise-parallel to $m->{supdir}
    $m->{flatsupfol}:  if $m->{subflat}, leaf name of $m->{supdir}
    $m->{flatsupbas}:  if $m->{subflat}, base name of $m->{supdir}
    $m->{dirstem}:	absolute path of directory under which $m->{dir} will be created -- this differs from $m->{supdir} if $m->{subflat}
    $m->{fol}:	leafname of directory to be created

Go to $m->{dirstem}.

Return $m->{fol} and $m->{dirstem}.

Allow a dot for a default value of $sem, expand that.

This should come before 'mkchfol', which depends on the supdir-related behaviour and uses the dotdir info.

It should come before 'set_dok_varz'; it should not write anything to the database.

Do not confuse with the supdir and supdok functions of A2E::Dokfs; here we take a different approach; we create the data that the supdir/supdok functions can later use.

The $m->{tmpl} and $m->{subflat} information is gathered from the parent doctype, most other info is then gained from the descendant doctype.

The variable $m->{flatsupfol} is reset in gdp_set_pwd only if needed.  The boolean flag $m->{subflat} is always set to tell us whether we are working from a group node.  If so the flatsupfol must have been set.

=cut

sub set_dir_varz {
    my $m = shift;
    $m->dpuc('set_dir_varz', @_);
    @{$m->{suplangs}} = $m->get_langs;
    my $fol = shift;
    my $sems = shift;
    $m->{dotdir_p} = $sems eq '.';
    @{$m->{sems}} = $m->set_sems_tmplvars($sems);
    $m->{supdok} = $m->get_dok;
    $m->{tmplvars}->{__supdok} = $m->{supdok};
  SUPDBM: {
      my $topsty = $m->{konfig}->topsty;
      $m->dpuc('supdbm', $topsty);
      do { $m->dpop;last } if $topsty;
      my $ok = $m->mldk('dbms');
      # $m->{cache}->{sdbm_ok} = $ok;
      do { $m->dpop;last } if $ok;
      $m->ustop('Failed to make parent dbms at %d. You must either work from a proper parent directory or create a new root directory with the --topsty option', d => &getcwd);
      $m->dpop('supdbm', $ok);
    };
    $m->dgot(m_lastdok => $m->{lastdok}, m_supdok => $m->{supdok});
  DOKTMPLVARS: { 
      last if $m->{supdok} eq $m->{lastdok};
      $m->rekurs_del_tmplvars(minim => 1);
      $m->dok_set_tmplvars($m->{supdok}, $m->{suplangs}->[0]);
    } 
    $m->{subflat} = $m->subflat(@{$m->{sems}});
    $m->{dokstyp} = $m->{subflat} ? 'grpdoks' : 'subdoks';
    $m->gdp_set_pwd($m->{subflat}, $m->{dotdir_p});
    my $subtmpl = $m->subtmpl(@{$m->{sems}});
    my $subvars = $m->subvars(@{$m->{sems}}); # { dokstem => 'phm_pub_moral1805' }
  LANGTMPLVARS: {
      last if $subtmpl eq $m->{tmpl};
      $m->{tmpl} = $subtmpl;
      $m->set_templates;
      $m->rekurs_del_tmplvars(minim => 1);
      # TODO: reduce to reading of tmpl-related stuff e.g. file:/adv/tmpl/a2e/oas_akt_aammdd/lang_pre.txt
      $m->lang_set_tmplvars($m->{suplangs}->[0], $m->{supdok});
    };
    $m->{tmplvars}->{__tmpl} = $m->{tmpl} || $m->pstop('__tmpl unknown');
    foreach my $nom (keys %$subvars) { $m->{tmplvars}->{$nom} = $subvars->{$nom} };
    $fol ||= $m->{dotdir_p} ? '.' : $m->sems2fol(@{$m->{sems}});
    $m->{subfol_regexp_str} = $m->{tmplvars}->{subfol_regexp} || $m->{konfig}->subfol_regexp;
    $m->{subfol_regexp} = qr($m->{subfol_regexp_str});
    $m->{dir} = $m->fol2dir($fol, $m->{pwd});
    my @dirlist = splitdir $m->{dir};
    $m->{fol} = pop @dirlist;
    $m->ustop('invalid leaf name %s, failed to match subfol_regexp %s', $fol, $m->{subfol_regexp_str}) unless $m->{fol} =~ m($m->{subfol_regexp});
    $m->{tmplvars}->{fol} = $m->{fol};
    $m->{dirstem} = catdir @dirlist;
    $m->{supdir} = $m->{dirstem};
  SUPDIR: {
      $m->dpuc('supdir', $m->{subflat}, $m->{fol}, $m->{dirstem}, $m->{dotdir_p});
      do { $m->dpop;last } unless $m->{subflat};
      die sprintf 'invalid flatsupfol %s', $m->{flatsupfol} unless $m->{flatsupfol} =~ m(\A\w+\Z);
      $m->{supdir} = catdir $m->{supdir}, $m->{flatsupfol};
      die sprintf 'invalid supdir %s', $m->{supdir} unless -d $m->{supdir};
      $m->dpop('supdir', $m->{supdir});
    };
    $m->{supdok} = $m->dirdok($m->{supdir});
    $m->set_topdoks;
    $m->set_topdir($m->{supdok});
    foreach my $nom (@kopi_tmplvars) { my $val = $m->{$nom};next unless $val;$m->{tmplvars}->{$nom} = $val };
    $m->pstop('template variable topdok still not defined at end of set_dir_varz') unless $m->{tmplvars}->{topdok};
    $m->dgot(m_dir => $m->{dir}, m_fol => $m->{fol}, m_dirstem => $m->{dirstem}, m_subflat => $m->{subflat}, m_flatsupfol => $m->{flatsupfol}, m_supdir => $m->{supdir}, m_supdok => $m->{supdok}, m_dotdir_p => $m->{dotdir_p}, tmpl => $m->{tmpl});
    $m->dpop('set_dir_varz', $m->{fol}, $m->{dirstem});
    return $m->{fol}, $m->{dirstem};
};

=head2 Function lang_set_tmplvars

Read variables from the vocabulary files of the document type and write them into the template variables hash; omit those that belong to a group of forbidden variables (non_tmplvars).
Choose the default language if none was given.
Preconditions: suplangs (languages of superior document node) must be known.

Arguments:
    $lng : language; if not given, read language-neutral textchunks only;
    $dok : document; if not given, read doctype/tmpl related textchunks only
    %opts : options
        OBSOLETE: extra: treat @extra_non_tmplvars (such as 'aa') as special variables, ignore any further assignment statements of these variables as far as they have already been read
        no_par: (obsolete? passthru) do not read both language-specific and language-neutral file at every level

New 2013-06-22: SUPDBM: overwrite template variables with any prepared subdocument values from the parent
    
=cut

sub lang_set_tmplvars {
    my $m = shift;
    $m->dpuc('lang_set_tmplvars', @_);
    my $lng = shift or $m->pwarn('lang_set_tmplvars called without lang argument');
    my $dok = shift;
    my %opts = @_;
    my $extra = $opts{extra};
    my $par = !$opts{no_par};
    my $tmpl = $opts{tmpl} || $m->{tmpl};
    my @faylz = $m->lng_faylz($lng, $dok, DabaP => 1, NFaylP => 1, par => $par, nosrclang => 1, readable => 1, tmpl => $tmpl);
    $m->pstop('lang_set_tmplvars: no textchnunk files to read for lng %s, dok %s, tmpl %s, par %s', $lng, $dok, $tmpl, $par) unless @faylz;
    $dok ||= $m->get_dok;
    lng: {
	last if $lng;
	($lng) = $m->get_langs($dok);
	last if $lng;
	$lng = $m->{novlangs}->[0];
	last if $lng;
	$m->pwarn('lang_set_tmplvars unable to set lang by itself ');
    };
    my $nom = '';
    %{$m->{nom2lits}} = ();
    if ($extra) { 
	foreach $nom (@extra_del_tmplvars) { delete $m->{tmplvars}->{$nom};delete $m->{finivars}->{$nom} };
    };
    my $user = $m->{tmplvars}->{user}; # defined in A2E::Prog by tmplvars_define 
    $m->dgot(_tmpllsts_antevars_ => $m->{tmpllsts}->{_antevars_});
    $m->riid_tmplvars_faylz({ lang => $lng, dok => $dok, tmpl => $tmpl }, { msglini => 1}, @faylz);
    $m->dgot(_tmpllsts_antevars_ => $m->{tmpllsts}->{_antevars_});
  SUPDBM: { # overwrite template variables with any prepared subdocument values from the parent
      my $topsty = $m->{konfig}->topsty;
      $m->dpuc('supdbm', $m->{supdok}, $m->{dok}, $topsty);
      do { $m->dpop;last } if $topsty;
      do { $m->dpop;last } if $m->{supdok} eq $m->{dok};
      my $dbmvars = {};
      my $sems = join '_', @{$m->{sems}};
      my $kiifilter = sub { local $_ = shift;my ($r) = m(\A(?:$m->{dok}|$sems)\_([a-z]\w+)\Z);$m->dgot(_ => $_, r => $r);return $r };
      my $dbm_ok = $m->dok2tmplvars($m->{supdok}, $dbmvars, $lng, kiifilter => $kiifilter);
      $m->dgot(dbm_ok => $dbm_ok, dbmvars => $dbmvars, sems => $sems);
      local $_;for (keys %$dbmvars) { my ($k) = m(\A(?:$m->{dok}|$sems)\_(\w+)\Z);next unless $k;my $v = $dbmvars->{$_};next unless $v;$m->{tmplvars}->{$k} = $v;$m->dgot(k => $k, _ => $_, v => $v); };
      $m->dpop('supdbm', $dbm_ok);
    };
  LNGTAGS: { # overwrite with any values from the commandline in pars_args
      foreach my $nom (keys %{$m->{lngtags}->{$lng}}) { $m->{tmplvars}->{$nom} = $m->{lngtags}->{$lng}->{$nom} };
    }
    $m->dpop('lang_set_tmplvars', 1);
    return 1;
}

=head2 set_dok_varz

Find out what kind of a location our leaf is (whether we can write there, whether it is a tree trunk), store some basic configuration values into memory, write others to the database.

The aliases and topdoks must be read into memory in any case.

The database must be edited only if we are creating a new document, i.e. not SAMDOK.

=cut

sub set_dok_varz {
    my $m = shift; 
    $m->dpuc('set_dok_varz', @_);
    # M-x -op-p file:/phm/18/09/21/sig/_lng.phm_pub_sig180921.txt::kdok
    my $dok = $m->sems2dok(@{$m->{sems}}) || '';
    #150211 $m->pstop('document %s has wrong name', $dok) if $ENV{DEBUG_SEMS2DOK} eq '150211' and not $dok eq 'phm_pub_adv_sig15';
    $m->ustop('document %s looks like a language symbol, you probably misspecified your arguments.  If you are sure you know what you are doing, override with --force', $dok) if $m->{cache}->{langnoms}->{$dok} and not $m->{konfig}->force; 
    $m->{dok} = $dok eq '.' ?  $m->{fol} : $dok;
    my $subdok_regexp = $m->{tmplvars}->{subdok_regexp} || $m->{konfig}->subdok_regexp;
    $m->{subdok_regexp} = qr($subdok_regexp);
    $m->ustop('invalid document name %s, failed to match regexp %s', $m->{dok}, $subdok_regexp) unless $m->{dok} =~ m($m->{subdok_regexp});
    $m->{tmplvars}->{'__'.$_} = $m->{$_} for qw(dok fol dir);
    $m->{tmplvars}->{$_} = $m->{$_} for qw(dok);
    $m->set_alidoks;
    $m->set_topdoks;
    my $db_absurl = $m->db_dok2absurl($m->{dok});
  OLDURL: {
      $m->dpuc('oldurl', $db_absurl);
      do { $m->dpop;last } unless $db_absurl;
      do { $m->dpop;last } if $m->{url} eq $db_absurl;
      do { $m->rmdok($m->{dok});last } if $m->{force};
      $m->ustop('Das Dokument %s existiert bereits mit URL %s statt %s.'."\n".'Rufe mich mit --force auf, wenn du mit der Brechstange vorgehen willst.', $m->{dok}, $db_absurl, $m->{url});
    };
    $m->put('mlhtdok', { dok=>$m->{dok} }, dokurl => $m->{url}, doktmp=>'now()');
    $m->put('dokprop', { dok=>$m->{dok} }, doknom => $m->{fol}, dokvers => 1, dokdatum => $m->{hodie}, dokuid => $m->{tmplvars}->{user}, doktyp => $m->{tmpl}, lastmod => $m->{hodie});
    if ($m->{konfig}->topsty) {
	$m->put('dokflags', { dok => $m->{dok} }, varnom => 'topdok', varval => 'true'); # obsolete 2009-04-19
	$m->put('topdok', { topdok => $m->{dok} }, lastmod => $m->{hodie}, inddok=> $m->{cache}->{inddok});
    };
    $m->dpop('set_dok_varz', $m->{dok});
    return $m->{dok};
};

=head2 Set further variables

Once we have identified the directory, document and position in the hierarchy,
we are ready to set some more variables and perform some checks.

=head3 Function dir_set_url

Derive the url from the directory which we have chosen for our document.

This does not read anything from an existing directory, can therefore come relatively early.

Since the sequence of checking directories against urls is critical, we can not use a hash variable for our dirurls config option!

It must come between set_dir_varz and set_dok_varz.

=cut

sub dir_set_url {
    my $m = shift;
    $m->dpuc('dir_set_url', @_);
    $m->{url} = $m->dir2url($m->{dir});
    $m->dpop('dir_set_url', $m->{url});
    return $m->{url};
};

=head3 Function num2alpha35

generate triantapentesimal alphanumeric digit, e.g. for 'K4F' style date 

=cut

sub num2alpha35 ($$) {
    my $m = shift;
    $m->dpuc('num2alpha35', @_);
    my $n = shift;
    $n %= 35;
  NUM: {
      $m->dpuc('num', $n);
      do { $m->dpop;last } unless $n < 10;
      $m->dpop('num2alpha35', $n);
      return $n;
    };
    my $k = chr $n - 9 + 64;
    $m->dpop('num2alpha35', $k);
    return $k;
};

=head2 Install the target directory with some basic files

=head3 Function mkchfol

Safely create a subdirectory after moving blocking data out of the way, change to that directory.
Refrain from creating if the data indicates that the document node shouldn't be here.
Make sure we are svn-wise up to date.
Retrieve-store info about languages of any pre-existing dok (oldlangs).
This comes before pars_args.

=cut

sub mkchfol ($$) {
    my $m = shift;
    $m->dpuc('mkchfol', @_);
    my $fol = shift or $m->ustop('mkchfol needs leaf argument');
    my $dirstem = shift;
    $m->chwd($dirstem) if $dirstem;
  SVN_UPDATE: {
      last if $m->{svn_repos}->{offline};
      eval { $m->svn_update };
      last unless $@;
      $m->topdir_svn_cleanup;
      $m->svn_update;
    };
  DOTDIR_P: {
      $m->dpuc('dotdir_p', $m->{dotdir_p});
      do { $m->dpop;last } unless $m->{dotdir_p};
      # Almost nothing needs to be done if we are just adding a language to an existing document
      $m->set_oldlangs;
      $m->dpop('mkchfol', 1);
      return 1;
    };
    $m->add_subdoks($m->{dok}, $m->{supdir}, $m->{dokstyp}) or $m->ustop('kann %s nicht zu subdoks hinzufügen', $m->{dok});
  MKDIR: {
      $m->dpuc('mkdir', $fol);
    DIR_P: {
	$m->dpuc('dir_p', $fol);
	do { $m->dpop;last } unless -d $fol;
	$m->vecho('Verzeichnis %s existiert bereits.  Gut so.', $fol);
	$m->chwd($fol);
	$m->dpop('mkdir');
	last MKDIR;
      }
    EXISTS: {
	$m->dpuc('exists', $fol);
	do { $m->dpop;last } unless -f $fol;
	my $ftmp = `mktemp`;chomp($ftmp);
	rename $fol, $ftmp or die sprintf "kann %s nicht zu %s umbenennen", $fol, $ftmp;
	$m->dpop;
      }; 
      $m->svn_mkdir($fol);
      $m->chwd($fol);
    };
    $m->set_oldlangs;
    $m->dpop('mkchfol', 1);
    return 1;
};

=head3 Function transkopi

Copy a template from its source to its destination and transform it, using the Template toolkit.

=cut

sub transkopi { 
    my $m = shift;
    $m->dpuc('transkopi', @_);
    my $fi = shift;
    my $fo = shift;
    do { $m->dpop('transkopi', 'no_file');return } unless -s $fi;
    $m->open_fayl($fi, 'i', 'r');
    $m->open_fayl($fo, 'o', 'w');
    $m->{star_template}->process($m->{f}->{i}->{h}, $m->{tmplvars}, $m->{f}->{o}->{h});
    $m->kloz_fayl($_) for qw(i o);
    $m->dpop('transkopi', 1);return 1;
};

=head3 Function set_oldlangs

Find out for which languages the present node already contains body documents and store these in @{$m->{oldlangs}}.

This is relevant only for existing documents, i.e. cases where dokdir_p is true or where we are entering into an existing directory in mkchfol.
It must be there before we have parsed the commandline arguments i.e. C<pars_args>.

=cut

sub set_oldlangs ($) {
    my $m = shift;
    $m->dpuc('set_oldlangs', @_);
    @{$m->{oldlangs}} = $m->get_langs;
    $m->dgot('m_oldlangs' => $m->{oldlangs});
    do { $m->dpop;return } unless @{$m->{oldlangs}};
    $m->{konfig}->lang0($m->{oldlangs}->[0]);
    $m->dpop('set_oldlangs', @{$m->{oldlangs}});
    return @{$m->{oldlangs}};
};
 
=head3 Function pars_args

Parse language-version-related arguments.

Take @ARGV as arguments.

Check all arguments and store them into a hash $m->{lngtags}, from which they are transferred to $m->{tmplvars} shortly before use, e.g. at the end of lang_set_tmplvars.

This way we can detect syntax errors early in the process.

This must come after mkchfol, at end of setup.

=cut

sub pars_args {
    my $m = shift;
    $m->dpuc('pars_args', @_);
    my @args = @_;
    @{$m->{novlangs}} = ();
    my $lng = '';
    my $val = '';
    my $tok = '';
    my $fi = '';
    $m->pstop('failed to pass a tratoks listref to mktdir object') unless 'ARRAY' eq ref $m->{tratoks};
    my @tratoks = @{$m->{tratoks}}; # pass via new_ready
    $m->dgot(tratoks => \@tratoks);
  LNG: while (@args) {
      $lng = shift @args;
      $m->dpuc('lng', $lng, \@tratoks, \@args);
      $m->ustop('unbekannte Sprache %s', $lng) unless $m->{cache}->{langnoms}->{$lng};
      $m->ustop('Sprache %s mehrfach angegeben.', $lng) if grep { $_ eq $lng } @{$m->{novlangs}};
      $m->{lang} = $lng;
      $m->{lngtags}->{$lng} = {};
      foreach $tok (@tratoks) {
	  $m->dpuc('tok', $tok);
	  $val = shift @args;
	  $m->ustop('unstimmige Argumentezahl: kein %s für Sprache %s', $tok, $lng) unless $val;
	  do { $m->dpop;next } if $val eq '.';
	  _utf8_on $val;
	  $m->dgot(lng => $lng, tok => $tok, val => $val, args => \@args);
	  $m->{lngtags}->{$lng}->{$tok} = $val;
	  $m->dpop;
      };
    LNGTAGS_DEFAULT: { 
	$m->dpuc('lngtags_default', $m->{tmpl}); 

=pod

set default values for title, subtitle etc unless we know our template (tmpl).

=cut

	  do { $m->dpop;last } if $m->{tmpl};
	  $m->{lngtags}->{$lng}->{lab} ||= $m->{lngtags}->{$lng}->{fol};
	  $m->{lngtags}->{$lng}->{tit} ||= $m->{lngtags}->{$lng}->{lab};
	  $m->{lngtags}->{$lng}->{sut} ||= $m->{lngtags}->{$lng}->{tit};
	  $m->{lngtags}->{$lng}->{des} ||= $m->{lngtags}->{$lng}->{sut};
	  $m->dpop('lngtags_default', $m->{lngtags}->{$lng});
      }
    OLD_LANG: {
	my $old = grep { $_ eq $lng } @{$m->{oldlangs}}, @{$m->{novlangs}};
	$m->dpuc('old_lang', $old, $lng, $m->{oldlangs}, $m->{novlangs});
	do { $m->dpop;last } unless $old;
	$m->uwarn('omitting duplicate language %s', $lng);
	$m->dpop('lng');next LNG;
      };
      push @{$m->{novlangs}}, $lng;
      $m->dpop('lng', $lng);
  };
  SUPLANGS: {
      $m->dpuc('suplangs', __dok__ => $m->{dok}, __nov__ => $m->{novlangs}, __old__ => $m->{oldlangs}, __sup__ => $m->{suplangs});
    LANGSF: {
	my $fi = $m->hsprintf($m->{konfig}->langsfnom, p => $m->{srcfpfx});
	last if -e $fi;
	$m->uwarn('no langs file %s for current directory %s', $fi, &getcwd);
      }
      do { $m->dpop;last } if @{$m->{novlangs}};
      $m->ustop('in same directory and not adding any new language') if $m->{dotdir_p};
      if ($m->{tmplvars}->{novlangs}) { @{$m->{novlangs}} = split m(\W+), $m->{tmplvars}->{novlangs} } else { foreach my $lang (@{$m->{suplangs}}) { next if grep { $_ eq $lang } @{$m->{oldlangs}};push @{$m->{novlangs}}, $lang } };
      $m->{tmplvars}->{novlangs} = join ' ', @{$m->{novlangs}};
      $m->mktdir_input_lit('novlangs');
      @{$m->{novlangs}} = split m(\W+), $m->{tmplvars}->{novlangs};
      $m->dpop('suplangs', $m->{novlangs});
    };
    @{$m->{novlangs}} = grep { m(\A\w{2}) } @{$m->{novlangs}};
    $m->dgot(m_lngtags => $m->{lngtags}, m_novlangs => $m->{novlangs});
    $m->ustop('nonlingual document %s?', $m->{dok}) unless @{$m->{novlangs}} or @{$m->{oldlangs}};
    $m->dpop('pars_args', @{$m->{novlangs}});
    return @{$m->{novlangs}};
};

=head2 Write the files

Write stubs for the main document (@dok_*) and its auxiliary files.

First evaluate all the commandline parameters, start writing only if they are OK.

=head3 Function tmplsrcfayl

Return full pathname for basename arg1, obtained by searching through the template directories.
The returned pathname has been found to exist.
arg2ff is a list of options containing 
    C<insist> ends with error if nothing is found
    C<tmpldir>  limits search to template directory tree, requires that something must be specifically defined for this template (this option is becoming obsolete, was used for 'dok_tmpl.txt' until 2015-11-09)

=cut

sub tmplsrcfayl {
    my $m = shift;
    $m->dpuc('tmplsrcfayl', @_);
    my $ffol = shift;
    my %opts = @_;
  TMPLDIR: { # Nearly OBSOLETE since 2015-11-09; no need to define anything per template, INCLUDE_PATH should be the way to go everywhere; previously this was used for 'dok_tmpl.txt' but generic '/adv/tmpl/a2e/multemp/dok_tmpl.txt' is just fine for many templates.  Defining templates by paramters in lang_pre.txt only is quite desirable.
      last unless $opts{tmpldir};
      $m->dpuc('tmpldir', \%opts);
      my $dir = $m->{tmpldir};
      $m->ustop('required template directory tmpldir is not known') unless $dir;
      do { $m->dpop;last } unless $dir;
      my $fayl = catfile $dir, $m->{tmplang}, $ffol;
      $m->ustop('required template file source %s does not exist', $fayl) unless -e $fayl;
      $m->dpop('tmplsrcfayl', $fayl);
      return $fayl;
    };
    foreach my $dir (@{$m->{cache}->{tmplopts}->{INCLUDE_PATH}}) {
	$m->dpuc('dir', $dir);
	my $fayl = catfile $dir, $ffol;
	do { $m->dpop;$m->decho('template %s not found, falling back', $fayl);next } unless -e $fayl;
	$m->dpop('tmplsrcfayl', $fayl);return $fayl;
    };
    $m->ustop_or_warn(!$opts{insist}, 'unable to find a template source file for %f', f => $ffol);
    $m->dpop('tmplsrcfayl');return;
};

=head3 Function fnom_tmplsrcfayl

Return name of template generation source file, such as e.g.

    /adv/tmpl/a2e/oas_adv_aamm/lng_tmpl.txt
    /adv/tmpl/a2e/oas_adv_aamm/lng_tmpl.de.txt

if such is found based on an expression $fnom e.g. posttmplfnom0 or posttmplfnom such as

    %d/lng_tmpl.txt
    %d/lng_tmpl.%l.txt

or others such as

    %ppre.%d.%l.txt

etc.  Return nil if nothing is found

=cut

sub fnom_tmplsrcfayl {
    my $m = shift;
    $m->dpuc('fnom_tmplsrcfayl', @_);
    my $fnom = shift;
    my %tups = @_;
    my $tmpl = $m->{tmpl} || $m->pstop('tmpl not found in fnom_tmplsrcfayl'); 
    $tups{d} ||= $tmpl;
    $tups{p} ||= $m->{konfig}->srcfpfx; 
    my $fib = $m->hsprintf($fnom, %tups);
    my $fi = catfile $m->{tmpldir}, $m->{tmplang}, $fib;
    do { $m->dpop('fnom_tmplsrcfayl', $fi);return $fi } if -r $fi;
    $fi = catfile $m->{tmpldir}, $fib;
    do { $m->dpop('fnom_tmplsrcfayl', $fi);return $fi } if -r $fi;
    $m->pwarn('template %s not found in tmpldir %s with tmplang %s, using internal default', $fib, $m->{tmpldir}, $m->{tmplang});
    $m->dpop('fnom_tmplsrcfayl');
    return;
};

=head3 Function topdok_kopi

Copy templates that are needed for top nodes of microsites, i.e. topdoks.

Arg1 is the filename suffix.  The rest of the filename is determined by other factors: the tmplsupdir and topsty config variables.

If template variables are present in arg 2 then use transkopi for the transformations, using the Template toolkit with the star style.

Try with more or less specific directory and base names.  Die if no suitable default file is found.

=cut

sub topdok_kopi ($$;) {
    my $m = shift;
    $m->dpuc('topdok_kopi', @_);
    my $sfx = shift;
    my $tmplvars = shift;
    my %opts = @_;
    my $sep = $sfx =~ m(\.) ? '_' : '.';
    my $fo = join $sep, $m->{dok}, $sfx;
  FO_EKZISTAS: {
      $m->dpuc('fo_ekzistas', $fo);
      do { $m->dpop;last } unless -e $fo;
      $m->dpop('topdok_kopi');return;
    };
    my $topsty = $m->{konfig}->topsty or $m->pstop('no topsty given');
    my $fib = join $sep, $topsty, $sfx;
    my $fi = $m->tmplsrcfayl($fib, %opts);
  FI_NACK: {
      last if $fi and -e $fi;
      $m->uwarn('can not find template source %s anywhere', $fib);
      $m->dpop('topdok_kopi');return;
    };
    $m->vecho('copying microsite-wide default from %s to %s', $fi, $fo);
    if ($tmplvars) {
	$m->transkopi($fi, $fo, $tmplvars);
    } else {
	$m->kopi_fayl($fi, $fo);
    };
    $m->svn_add($fo) if -s $fo;
    $m->dpop('topdok_kopi', 1);
    return 1;
};

=head3 Function topdok_install

Install all the topdok templates using the Template toolkit with star style.

NO_TOPSTY: stop unless we are in a top level document.

This does not yet involve any templates that depend on the vocabulary database.

=cut

sub topdok_install {
    my $m = shift;
    $m->dpuc('topdok_install', @_);
    my $topsty = $m->{konfig}->topsty;
  NO_TOPSTY: { 
      my $ok = !$topsty;
      $m->dpuc('no_topsty', $ok, $topsty);
      do { $m->dpop;last } unless $ok;
      $m->dpop('topdok_install', $ok);
      return $ok;
    };
    my $tmplvars = { topdir => $m->{dir}, topdok => $m->{dok}, topurl => $m->{url}, tmplang => $m->{tmplang}, tmpl => $m->{tmpl}, news_limit => '20', news_level => '0' };
    # $m->topdok_kopi($_, $tmplvars) for qw(css xhtml);
    # $m->topdok_kopi($_) for qw(png tex dba);
    my $fnom = $m->{konfig}->konf_tmpl_fnom;
  EXTRA: {
      last unless $m->{konfig}->topdok_kopi_extra; 
      $m->topdok_kopi($_, $tmplvars) for map { sprintf $fnom, $_ } split m(\W), $m->{konfig}->tmplapp_tmpls->{$m->{tmplapp}};
      $m->topdok_kopi($_) for 'nav_html_tmpl.txt', 'doksrow_html_tmpl.txt', 'news_html_tmpl.txt', 'news_rss_tmpl.txt';

    };
    $m->dpop('topdok_install', 1);
    return 1;
};

=head3 Function riid_frazfayl

If a text building block file already exists, store any variable assignements found therein in their raw original form as whole lines in a hash %vars, but identify the variable and use it as hash index.  Return %vars.

This is used by skrib_antefayl0, skrib_antefayl, skrib_postfayl and skrib_postfayl0

=cut

sub riid_frazfayl {
    my $m = shift;
    $m->dpuc('riid_frazfayl');
    my $fi = shift;
    my %vars = ();
    do { $m->dpop('riid_frazfayl', %vars);return %vars } unless -r $fi;
    $m->open_fayl($fi, 'i', 'r');
    my $nom = '';
    while ($m->getlini('i')) {
	($nom) = m(\A(?:export|unexport)\s+(\w+)\s+\??\:?\=\s+\S);
	do { $vars{$nom} = $_;next } if $nom;
	($nom) = m(\A(\w+)\s+\??\:?\=\s+\S);
	do { $vars{$nom} = $_;next } if $nom;
    };
    $m->kloz_fayl('i');
    $m->dpop('riid_frazfayl', %vars);
    return %vars;
}

=item Function lexvars2vals

Take hash argument containing keys (e.g. 'lab') and values (e.g. 'lab := Bar').

Return a list of entries (e.g. 'lab := Bar') in the order in which they should appear in a text chunk file, with obligatory document metadata tokens lab, tit, sut, des coming first.

=cut

sub lexvars2vals {
    my $m = shift;
    $m->dpuc('lexvars2vals', @_);
    my %vars = @_;
    my @vals = ();
    foreach my $nom (@tratoks, @{$m->{konfig}->tratoks_extra}) {
	my $val = $vars{$nom};
	$m->dpuc('nom', $nom, $val);
	push @vals, $val if $val;
	delete $vars{$nom};
	$m->dpop('nom');
    };
    push @vals, values %vars;
    $m->dpop('lexvars2vals', @vals);
    return @vals;
}

=item Function frazfayl_vals: obtain values from existing vocabulary file

If a vocabulary file exists, store any variable assignments found therein in their raw original form as whole lines in the hash %vars, but identify the variable and use it as hash index. then overwrite these variables with any newer values that we might have and produce an ordered list of whole lines without the hash index.

 arg1 : vocabulary file, e.g. 'pre.abc.de.txt'
 arg2 : vocabulary definition format, e.g. '%s := %s' 
 arg3 : hashref of values that have already been set and take precedence over anything found in the file
 arg4ff : keys of hashref that we actually will use, all if untrue

=cut

sub frazfayl_vals {
    my $m = shift;
    $m->dpuc('frazfayl_vals', @_);
    my $fayl = shift;
    my $lexdef_fmt = shift;
    my $tmplvars = shift;
    my @tmplvars_noms = @_;
    @tmplvars_noms = keys %$tmplvars unless @tmplvars_noms;
    my %vars = $m->riid_frazfayl($fayl);
  TMPLVARS_HODIE: { # kludge 2017-07-01: set reasonable date that due to other bugs was not there
      $tmplvars->{hodie} = $m->{hodie};
      $tmplvars->{datum} = $m->{hodie};
    };
  VARS: {
      $m->dpuc('vars', $tmplvars);
      do { $m->dpop;last } unless %$tmplvars;
      foreach my $nom (@tmplvars_noms) {
	  my $val = $tmplvars->{$nom};
	  next unless defined $val;
	  $vars{$nom} = sprintf $lexdef_fmt, $nom, $val;
	  my $prompt = join '_', $nom, 'prompt';
	  next unless $tmplvars->{$prompt};
	  $vars{$nom} = '## '.$tmplvars->{$prompt}."\n".$vars{$nom};
      };
    };
    $m->dgot(%vars);
    my @vals = ();
    foreach my $nom (@tmplvars_noms) { my $val = $vars{$nom};next unless $val;push @vals, $val };
    $m->dpop('frazfayl_vals', @vals);
    return @vals;
};

=head3 Function mktdir_input_lit

Read a lingual text from the command line and validate it.
Return input value and whether it is old, i.e. the input just confirmed what was already in the template.

=cut

sub mktdir_input_lit {
    my $m = shift;
    $m->dpuc('mktdir_input_lit', @_);
    my $nom = shift;
    my $dialog = shift || !$m->{cache}->{batch};
    my ($val, $old) = $m->litvalid($nom, $dialog);
    $m->dpop('mktdir_input_lit', $val, $old);
    return $val, $old;
};

=head3 Function langlexvar

Cache/retrieve textchunks coming from the langlexvars configuration variable.

arguments:

    $nom : variable name
    $lang : language

return values:

    $val : value derived from $m->{konfig}->langlexvars

intermediate variable cache:

   $m->{langlexvars}

store values of present variable in the cache if not already there and return the one of them that matches $lang

=cut

sub langlexvar {
    my $m = shift;
    $m->dpuc('langlexvar', @_);
    my $nom = shift;
    my $lang = shift || $m->{lang};
    my $val = $m->{langlexvars}->{$nom}->{$lang};
    do { $m->dpop('langlexvar', $val);return $val } if $val;
    do { $m->dpop;return } if $m->{langlexvars}->{$nom};
    my $kval = $m->{konfig}->langlexvars->{$nom};
    do { $m->dpop;return } unless $kval;
    my @vals = $m->{sarb}->textsep2listref($kval);
    my @langs = @{$m->{novlangs}};
    while (@vals and @langs) {
	$val = shift @vals;
	my $ll = shift @langs;
	$m->{langlexvars}->{$nom}->{$ll} = $val;
    };
    $val = $m->{langlexvars}->{$nom}->{$lang};
    $m->dpop('langlexvar', $val);
    return $val;
}


=head3 Function konf_skrib_postfayl

Write a postfayl 'lng.$dok.$lang.txt' from scratch based on configuration variables and preset variables.

    $filepath
    $lang
    $asis : hashref : { lab => 'lab := abc' }
    $lits : hashref : { lab => 'abc' }
    $olds : hashref : whether text chunks are old

=cut

sub konf_skrib_postfayl {
    my $m = shift;
    $m->dpuc('konf_skrib_postfayl', @_);
    my $fo = shift;
    my $lang = shift;
    my $Lang = ucfirst $lang;
    my $asis = shift;
    my $lits = shift;
    my $olds = shift;
    my %vars = ();
    $m->open_fayl($fo, 'o', 'w');
    my ($nom, $val, $lexdef) = ('', '', '');
    my $lexdef_fmt = $m->{konfig}->post_lexdef_fmt;
    my $tmpl = $m->{tmpl};
    $m->dgot(fayl => $fo, asis => $asis, lits => $lits, tmpl => $tmpl, olds => $olds);
    $m->print_tu_fayl('o', $m->emacs_top_local_variables({ mode => 'fundamental', coding => 'utf-8', dokl => $m->{Dok}.'Lng'.$Lang })."\n\n");# 150209 suppressed srcfile => $fo
  TRATOK: foreach my $nom (@{$m->{langlits}}) {
      $m->dpuc('tratok', $nom);
      $val = $lits->{$nom} || $m->{lngtags}->{$lang}->{$nom} || $m->{tmplvars}->{$nom} || $m->langlexvar($nom, $lang);
      do { $m->dpop;next } unless $val;
      $vars{$nom} = sprintf $lexdef_fmt, $nom, $val;
      delete $asis->{$nom};
      # $vars{$nom} = '# ' . $vars{$nom} if $olds->{$nom};
      # 13.11.22 no longer adequate since (SUPDBM we deliberately pre-store the intended values in the parent document) 
      my $prompt = join '_', $nom, 'prompt';
      my $prompt_val = $m->{tmplvars}->{$prompt};
      do { $m->dpop;next } unless $prompt_val;
      $vars{$nom} = '## '.$prompt_val."\n".$vars{$nom};
      $m->dpop;
  };
    my @vals = $m->lexvars2vals(%vars, %$asis);
    $m->print_tu_fayl('o', $_ . "\n") for @vals;
    $m->print_tu_fayl('o', '#EOF');
    $m->kloz_fayl('o');
    $m->dpop('konf_skrib_postfayl', 1);
    return 1;
};

=head3 Function skrib_postfayl

Write language dependent tags from object variable lngtags into a lng_fayl and add that to the repository.
If a file exists, reuse variables found therein.
If a template exists, use that, otherwise write from memory based on configuration and preexisting variables.

=cut

sub skrib_postfayl {
    my $m = shift;
    $m->dpuc('skrib_postfayl', @_);
    my $lang = shift;
    my $Lang = ucfirst $lang;
    my $fnom = $m->{konfig}->lexfnom;
    $m->stop('postfayl name format lexfnom not defined') unless $fnom;
    my $fob = $m->hsprintf($fnom, d => $m->{dok}, l => $lang, p => $m->{konfig}->srcfpfx);
    my %vars = $m->riid_frazfayl($fob);
    my %lits = ();
    my %olds = ();
    my $fo = catfile $m->{dir}, $fob;
  TRY: {
      $m->dpuc('try', $fo);
      my %input_lits = ();
      ($lits{$_}, $olds{$_}) = $m->mktdir_input_lit($_) for @{$m->{langlits}};
    TMPL: {
	$m->dpuc('tmpl', %vars);
	my $fnom = $m->{konfig}->posttmplfnom or $m->pstop('posttmplfnom not defined');
	my $fi = $m->fnom_tmplsrcfayl($fnom, l => $lang);
	last TMPL unless $fi;
	my $oldvars = join "\n", values %vars; 
	$m->tmpl_skrib_fayl($fi, $fo, mode => 'fundamental', oldvars => $oldvars, %lits); # oldvars TODO: unify interface for templates
	$m->dpop('try', $fo);
	last TRY;
      };
      $m->vecho('writing my own file %s', $fo);
      $m->konf_skrib_postfayl($fo, $lang, \%vars, \%lits, \%olds);
      $m->dpop('try', $fo);
    };
    $m->svn_add($fob);
    $m->dpop('skrib_postfayl', 1);
    return 1;
};    

=head3 Function skrib_antefayl0

Write out language neutral definitions file unless it already exists.

=cut

sub emacs_fayl_put_str {
    my $m = shift;
    $m->dpuc('emacs_fayl_put_str', @_);
    my $fo = shift;
    my $emacs_mode = shift;
    my $srcfile = shift;
    my @vals = @_;
    my $res = $m->fayl_put_str($fo, join "\n", $m->emacs_top_local_variables({ mode => $emacs_mode, coding => 'utf-8', srcfile => $srcfile }), '', @vals, '#EOF');
    $m->dpop('emacs_fayl_put_str', $res);
    return $res;
};

sub skrib_antefayl0 {
    my $m = shift;
    $m->dpuc('skrib_antefayl0', @_);
    my $fnom = $m->{konfig}->antefnom0;
    do { $m->pwarn('no fnom for antefayl0');$m->dpop('skrib_antefayl0');return } unless $fnom;
    my $fo = $m->hsprintf($fnom, p => $m->{konfig}->srcfpfx, d => $m->{dok});
    do { $m->uwarn('antefayl0 %f already exists', f =>  $fo);$m->dpop('skrib_antefayl0');return } if -e $fo;
    $fnom = $m->{konfig}->antetmplfnom0;
    my $fi = $m->fnom_tmplsrcfayl($fnom);
    my $lexdef_fmt = $m->{konfig}->ante_lexdef_fmt;
    # do { $m->dpop('skrib_antefayl0');return } if -s $fo;
    my $srcfile = '';# 150209 suppressed catfile $m->{dir}, $fo; 
    my @dabalits = ();
    foreach my $lit (@{$m->{dabalits}}) { my ($val, $old) = $m->mktdir_input_lit($lit);$m->{tmplvars}->{$lit} = $val;next unless $val;push @dabalits, $lit };
    $m->{tmplvars}->{dokprogver} = $m->{konfig}->progver;
    $m->dgot(dabalits => \@dabalits, tmpllsts_antevars_ => $m->{tmpllsts}->{_antevars_}, tmplvars_antevars_ => $m->{tmplvars}->{_antevars_}, tmplvars_sems => $m->{tmplvars}->{sems}, tmplvars_sem1 => $m->{tmplvars}->{sem1});
    $m->{tmplvars}->{url} ||= $m->{url} || $m->dok2absurl;
  TMPL: {
      last if $m->{tmplvars}->{tmpl} eq $m->{tmpl};
      $m->uwarn('Has the template variable tmpl has been rigidly set to %v in a superior directory using = instead of ?= ?  Setting to %w anyway.', v => $m->{tmplvars}->{tmpl}, w => $m->{tmpl});
      $m->{tmplvars}->{tmpl} = $m->{tmpl}; # phm_pub_sig180114.oas_pub18
    };
    my @vals = $m->frazfayl_vals($fo, $lexdef_fmt, $m->{tmplvars}, 'dokprogver', 'url', @dabalits, @{$m->nom2xlst('_antevars_')});
    my $emacs_mode = 'makefile';
    if ($fi) { $m->tmpl_skrib_fayl($fi, $fo, mode => $emacs_mode, vals => \@vals) } else { $m->emacs_fayl_put_str($fo, $emacs_mode, $srcfile, @vals) }; 
    $m->svn_add($fo);
    $m->dpop('skrib_antefayl0', 1);
    return 1;
};

=head3 Function skrib_antefayl 

Write language-specific parameter file.

=cut

sub skrib_antefayl {
    my $m = shift;
    $m->dpuc('skrib_antefayl', @_);
    my $fnom = $m->{konfig}->antefnom;
    do { $m->dpop('skrib_antefayl');return } unless $fnom;
    my $lang = shift;
    my $Lang = ucfirst $lang;
    my $lexdef_fmt = $m->{konfig}->ante_lexdef_fmt;
    my $fo = $m->hsprintf($fnom, d => $m->{dok}, l => $lang);
    $fnom = $m->{konfig}->antetmplfnom;
    my $fi = $m->fnom_tmplsrcfayl($fnom, l => $lang);
    my $srcfile = '';# 150209 suppressed catfile $m->{dir}, $fo;
    my $old_fayl_p = -e $fo;
    my @vals = $m->frazfayl_vals($fo, $lexdef_fmt, $m->{lang_tmplvars}->{$lang});
    my $emacs_top_local_variables = $m->emacs_top_local_variables({ mode => 'makefile', coding => 'utf-8'});# 150209 suppressed srcfile => $srcfile
    if ($fi) { $m->tmpl_skrib_fayl($fi, $fo, mode => 'fundamental', vals => \@vals) } else { $m->fayl_put_str($fo, join "\n", $emacs_top_local_variables, '', @vals, '#EOF') };
    $m->svn_add($fo) unless $old_fayl_p;
    $m->dpop('skrib_antefayl', 1);
    return 1;
}

=head3 Function skrib_postfayl0

Write out language neutral vocabulary file if it does not already exist.

=cut

sub skrib_postfayl0 {
    my $m = shift;
    $m->dpuc('skrib_postfayl0', @_);
    my $fnom = $m->{konfig}->lexfnom0 || $m->pstop('no lexfnom0 found');
    my $fo = $m->hsprintf($fnom, p => $m->{konfig}->srcfpfx, d => $m->{dok});
    do { $m->dpop;return } if -s $fo;
    my $tmpl = $m->{tmpl};
    my $lexdef_fmt = $m->{konfig}->post_lexdef_fmt;
    my %vars = $m->riid_frazfayl($fo);
    my $srcfile = '';# 150209 suppressed catfile $m->{dir}, $fo;
    my $emacs_top_local_variables = $m->emacs_top_local_variables({ mode => 'fundamental', coding => 'utf-8' });# 150209 suppressed srcfile => $srcfile
  LEXVARS: {
      $m->dpuc('lexvars');
      my %lexvars = %{$m->{konfig}->lexvars};
      foreach my $nom (keys %lexvars) {
	  $m->dpuc($nom);
	  do { $m->dpop;next } if $vars{$nom};
	  my $val = $lexvars{$nom};
	  $vars{$nom} = sprintf $lexdef_fmt, $nom, $val;
	  $m->dpop($nom);
      };
      $m->dpop('lexvars', 1);
    };
    my @vals = $m->lexvars2vals(%vars);
    $fnom = $m->{konfig}->posttmplfnom0 or $m->pstop('posttmplfnom0 not defined');
    my $fi = $m->fnom_tmplsrcfayl($fnom);
    if ($fi) { 
	$m->tmpl_skrib_fayl($fi, $fo, mode => 'fundamental', vals => \@vals)
    } else {
	$m->open_fayl($fo, 'lng', 'w');
	$m->print_tu_fayl('lng', $emacs_top_local_variables."\n\n");
	$m->print_tu_fayl('lng', $_ . "\n") for @vals;
	$m->print_tu_fayl('lng', '#EOF');
	$m->kloz_fayl('lng'); };
    $m->svn_add($fo);
    $m->dpop('skrib_postfayl0', 1);
    return 1;
};

=head3 Function tmplsrcobjfayls

Find the appropriate template file, as specified by fnom2ffol from A2E::Dokfs(3), by seeking it in a series of directories specified in A2E::Tmplkonf(3), from the lowest (most specific) to the highest (most generic) possible template.
Even this may be no longer needed because we will just invoke all members of fnom2ffol to generate the object files from the templates in one run.

=cut

sub tmplsrcobjfayls {
    my $m = shift;
    $m->dpuc('tmplsrcobjfayls', @_);
    my $fnom = shift;
    my ($ffol, $fobj, %opts) = $m->fnom2ffol($fnom);
    $ffol = $m->hsprintf($ffol);
    my $fsrc = $m->tmplsrcfayl($ffol, %opts);
    $fobj = $m->hsprintf($fobj);
    # $m->{tmplsrcfayls}->{$fnom} = $fsrc;
    # $m->{tmplobjfayls}->{$fnom} = $fobj;
    $m->dpop('tmplsrcobjfayls', $fsrc, $fobj, %opts);
    return $fsrc, $fobj, %opts;
};

=head3 Function mldok_set_tmplvars
    
Set some template variables that are needed both in skrib_mldok and skrib_antefayl0.

=cut

sub mldok_set_tmplvars {
    my $m = shift;
    $m->dpuc('mldok_set_tmplvars', @_);
    $m->{Dok} = $m->CamelCase($m->{dok});
    # $m->{tmplvars}->{$_} ||= $m->{$_} for qw(Dok tmplang);
    $m->{tmplvars}->{outfmts} = $m->{konfig}->tmplang_outfmts->{$m->{tmplang}};
    $m->{tmplvars}->{mktdir_user} = $m->{tmplvars}->{__user};
    $m->{tmplvars}->{mktdir_time} = $m->timestamp;
    # $m->{tmplvars}->{$_} = $m->{konfig}->tmplvars->{$_} for keys %{$m->{konfig}->tmplvars};
    # moved to sub mktdir, done only once in interactive invocation, not in recursion
    $m->dgot(m_tmplvars => $m->{tmplvars}, m_konfig_tmplvars => $m->{konfig}->tmplvars);
    $m->dpop('mldok_set_tmplvars', 1);
    return 1;
}


=head3 Function tmpl2langlits

Subroutine to mlht_skrib_faylz

=cut

sub tmpl2langlits ($$) {
    my $m = shift;
    $m->dpuc('tmpl2langlits', @_);
    my $tmpl = shift;
    my @lits = (@tratoks, @{$m->{konfig}->tratoks_extra});
    if ($m->{cache}->{tmpltoks_ok}) {
	my @tmpltoks = @{$m->nom2xlst('_tmpltoks_')};
	push @lits, @tmpltoks;
    };
    $m->dpop('tmpl2langlits', \@lits);
    return \@lits;
};

=head3 Function mlht_skrib_faylz

Set up the text elements (read vocabulary dbm files), write out the main document bodies, first language-dependent parts with different dbm files as basis, then language-independent parts with original of topdok as basis.

=cut

sub mlht_skrib_faylz ($) {
    my $m = shift;
    $m->dpuc('mlht_skrib_faylz', @_);
    $m->mldok_set_tmplvars;
    my $tmpl = $m->{tmpl} or $m->pstop('unknown document type');
    my @langs = @{$m->{novlangs}};
    do { $m->pwarn('no languages to add, nothing to do');$m->dpop('mlht_skrib_faylz') } unless @langs;
    my $lang0 = shift @langs;
    my $lang = $lang0;
    $m->pstop('empty first language') unless $lang0;
    $m->{langlits} = $m->tmpl2langlits($tmpl);
    # $m->{dabalits} = $m->tmpl2dabalits($tmpl, @{$m->{langlits}}, (map { $m->{konfig}->supdok_non_tmplvars->{$_} && $_ } keys %{$m->{konfig}->supdok_non_tmplvars}));
    $m->{dabalits} = $m->tmpl2dabalits($tmpl, @{$m->{langlits}});
    $m->lang_set_tmplvars($lang, $m->{dok}, extra => 1);
    $m->dgot(langs => \@langs, lang0 => $lang0, novlangs => $m->{novlangs}, langlits => $m->{langlits}, dabalits => $m->{dabalits});
    LANG: {
	$m->skrib_postfayl($lang);
	$m->skrib_antefayl($lang) if $m->{konfig}->skrib_antefaylz;
	$lang = shift @langs;
	last unless $lang;
	$m->lang_set_tmplvars($lang, $m->{dok}, extra => 1);
	redo LANG;
    };
    $m->lang_set_tmplvars($lang0, $m->{dok}, extra => 1);
    $m->skrib_antefayl0;
    $m->skrib_postfayl0;
    # $m->skrib_mldok; 
    # $m->skrib_mlnav;
    $m->dpop('mlht_skrib_faylz', 1);
    return 1;
};

=head2 Function skrib_langs_fayl

Write out the @langs file.

This comes after writing the main document(s) but after set_oldlangs

=cut

sub skrib_langs_fayl ($) {
    my $m = shift;
    $m->dpuc('skrib_langs_fayl', @_);
    my @novlangs = @{$m->{novlangs}};
    $m->uwarn("no new language version to add, novlangs is null") unless @novlangs;
    my $fnom = $m->{konfig}->langsfnom || $m->ustop('langsfnom not defined');
    my $pfx = $m->{konfig}->srcfpfx;
    $m->ustop('source file name prefix srcfpfx not defined for @langs or similar in skrib_langs_fayl') unless $pfx;
    my $fayl = $m->hsprintf($fnom, p => $pfx);
    my @langs = grep {$_} (@{$m->{oldlangs}}, @novlangs);
    my $str = join ' ', @langs;
    $m->fayl_put_str($fayl, $str);
    $m->svn_add($fayl);
    $m->dpop('skrib_langs_fayl', 1);return 1;
};

=head3 Function skrib_fnoms

Generate objects from templates as defined in fnom2ffol hash variable.
TODO 2015-11-09: This should render skrib_meikfayl, skrib_konf and other special routines obsolete soon. DONE 2017-07-11 for skrib_meikfayl

=cut

sub skrib_fnoms {
    my $m = shift;
    my @keys = @_;
    @keys = keys %{$m->{konfig}->fnom2ffol} unless @keys;
    foreach my $fnom (@keys) { $m->skrib_fnom($fnom) };
};

=head3 Function subtmpl

Identify the subtmpl (subordinate document type) belonging to the current tmpl (document type), based on what is found in the values of the dynamic textchunk variable C<sems2subtmpl>.
Optionally @sems (semantic units passed to 'mktdir', often only the directory leaf) arguments are given.  These are passed and often required by the C<sems2subtmpl> variable.
If either @sems or the C<sems2subtmpl> variable are missing, we terminate (raise an exception).

=cut

sub subtmpl {
    my $m = shift;
    $m->dpuc('subtmpl', @_);
    my @sems = @_;
    @sems = @{$m->{sems}} unless @sems;
    my $val = $m->{tmplvars}->{sems2subtmpl};
    $m->ustop('no sems2subtmpl found in document type %s', $m->{tmpl}) unless $val;
    $m->dgot(val => $val);
    my $subtmpl = $m->{sarb}->val_args_call($val, undef, @sems);
    $m->ustop('no subtmpl found for %s with %s', $m->{tmpl}, $m->koniug_et('sems2subtmpl', @sems)) unless $subtmpl;
    $m->dpop('subtmpl', $subtmpl);
    return $subtmpl;
};

=head3 Function subflat

Resolve the boolean C<subflat> (whether to create an adjacent directory rather than a subdirectory) value belonging to the current tmpl (document type), based on what is found in the values of the dynamic textchunk variable C<sems2subflat> and the C<sems> (semantic units supplied to mktdir as arguments, often consisting of the directory leaf only).
The sems are passed to the C<sems2subflat> variable as ${1}, ${2} etc.
If nothing is given we assume 0, i.e. opt for the default procedure of creating subdirectory.

=cut

sub subflat {
    my $m = shift;
    $m->dpuc('subflat', @_);
    my @sems = @_;
    @sems = @{$m->{sems}} unless @sems;
    my $val = $m->{tmplvars}->{sems2subflat};
    $m->dgot(sems2subflat_val => $val);
    do { $val = $m->{tmplvars}->{grpdir};$m->dgot(grpdir_val => $val) } unless defined $val;
    my $subflat = $val ? $m->{sarb}->val_args_call($val, undef, @sems) : 0;
    $m->dpop('subflat', $subflat);
    return $subflat;
};

=head3 Function subvars

Set variables for subtemplate as specified in sems2subvars and return them.
E.g.

    sems2subvars ?= $(case ${1},,sig,|privat:1|no_subsems:1|)

leads to

    { privat => 1, no_subsems => 1 }

in case $sems[0] eq 'sig'.

2017-06-24 note: the switch no_subsems seems to mean that subdirectories are not made automatically.

=cut

sub subvars {
    my $m = shift;
    $m->dpuc('subvars', @_);
    my @sems = @_;
    @sems = @{$m->{sems}} unless @sems;
    my $val = $m->{tmplvars}->{sems2subvars};
    do { $m->dpop;return } unless $val;
    $val = $m->{sarb}->val_args_call($val, undef, @sems);
    my @vars = $m->{sarb}->textsep2list($val);
    $m->dgot(val => $val, vars => \@vars, dokstem => $m->{tmplvars}->{dokstem}, __sem1 => $m->{tmplvars}->{__sem1}); 
    my $subvars = {};
    my $nom = '';
    foreach my $var (@vars) {
	($nom, $val) = $var =~ m(\A(\w+)\:(.+)\Z);
	next unless $nom;
	$subvars->{$nom} = $val;
    };
    $m->dpop('subvars', $subvars);
    return $subvars;
};


=head3 Function subsems

Return a list of argument list strings each of which is to be passed to one subdocument generation ('submktdir') call.

In the frequent and simplest case this argument list string consists of one item that is then immediately used as subdirectory leaf name.

This function optionally takes @sems (semantic units passable to C<mktdir>, usually directory leaf) arguments.  It evaluates the textchunk variable C<sems2subsems> of the current template/doctype, which can but in practise rarely does use @sems as arguments.

If the C<sems2subsems> variable does not exist, we are at the end of directory tree and return nil without warning.

A complex argument like

    sems2subsems ?= $(monwoks ${aaaa},${mm},${ddmin})

produces something like

    'W26,01,01', 'w27,02,08', 'w28,09,15', 'w29,16,22', 'w30,23,29', 'w31,30,31'

as a result.

=cut

sub subsems {
    my $m = shift;
    $m->dpuc('subsems', @_);
    my @sems = @_;
    my $sems2subsems = $m->{tmplvars}->{sems2subsems};
    $m->dgot(sems2subsems => $sems2subsems);
    do { $m->dpop;return } unless $sems2subsems;
    my $subsems_str = $m->{sarb}->val_args_call($sems2subsems, undef, @sems);
    $m->dgot(subsems_str => $subsems_str); 
    my ($dum, @subsems) = $m->{sarb}->textsep2seplist($subsems_str);
    $m->dpop('subsems', @subsems);
    return @subsems;
};

=head2 Finishup and postprocessing

Add accumulated files and some further ones to Subversion repository.
Write metadata to the database. 
Set up a link to an mbox if found.

=head3 Function mktdir_finic

Final steps of the mktdir5 procedure, excluding svn_commit and directory changes: everything that needs to be done before any subdocuments are created.

=cut

sub mktdir_finic ($) {
    my $m = shift;
    $m->dpuc('mktdir_finic');
    eval { $m->mbox2dok };warn sprintf 'mbox2dok: %s', $@ if $@;
  NOCACHE: { # like system 'make', 'rmcache'
      my $varsfayl = $m->hsprintf($m->{konfig}->varsfnom, p => $m->{konfig}->srcfpfx, d => $m->{dok});
      foreach my $fi ($varsfayl, glob('*.cache')) { unlink $fi if -r $fi };
    };
  MAKE_MAIN: {
      last if $m->{konfig}->minimal;
      $m->mldk or $m->pstop('failed to make main in directory %s', $m->{cwd});
    };
    $m->dpop('mktdir_finic', 1);
    return 1;
};

=head2 Top level functions

=head3 Function lastkonfig

=cut

sub lastkonfig {
    my $m = shift;
    $m->dpuc('A2E::Mktdir::lastkonfig', @_);
    $m->SUPER::lastkonfig;

=over

=item topsty

always write language-specific parameter files if we are at top document level.

=cut

    $m->konfsetif('skrib_antefaylz', $m->{konfig}->topsty); 
  LANG_TMPLVARS: while (my ($nom, $val) = each %{$m->{konfig}->tmplvars}) {
      $m->dpuc('lang_tmplvars', $nom, $val);

=item LANG_TMPLVARS

transfer multilingual values of variable settings such as

    -V thema=Italia,de:Italien,en:Italy,fr:Italie

from $m->{konfig}->tmplvars into $m->{lang_tmplvars} 

=cut

      my ($val0, @tups) = split ',', $val;
      do { $m->dpop;next } unless @tups;
      $m->dgot(val0 => $val0, tups => \@tups);
      $m->{konfig}->set('skrib_antefaylz', 1);
      $m->{lang_tmplvars} ||= {};
      # $m->{konfig}->tmplvars->{$nom} = $val0;
      $m->{tmplvars}->{$nom} = $val0;
      foreach my $tup (@tups) {
	  $m->dpuc('tup', $tup);
	  my ($ll, $vall) = $tup =~ m(\A(\w+)\:(.+)\Z);
	  do { $m->dpop;next } unless $vall;
	  $m->dgot(ll => $ll, vall => $vall); 
	  $m->{lang_tmplvars}->{$ll} ||= {};
	  $m->{lang_tmplvars}->{$ll}->{$nom} = $vall;
	  $m->dpop('tup');
      };
      $m->dpop('lang_tmplvars');
  };

=over

=cut

    $m->dpop('A2E::Mktdir::lastkonfig', 1);
    return 1;
};

=head3 Function postkonfig

Overload the postkonfig part of the new_ready method from A2E::Prog(3).

=cut

sub postkonfig {
    my $m = shift;
    $m->dpuc('A2E::Mktdir::postkonfig', @_);
    $m->SUPER::postkonfig;
    $m->mldk_postkonfig;
    $m->tcek_mailadr_varz;
    $m->{emacs_timestamp} ||= $m->emacs_timestamp;
    $m->put_rekonfig(subtmpl => \&rekonfig_subtmpl);
    $m->put_rekonfig(tmplvars => \&rekonfig_tmplvars);
    $m->put_rekonfig(subopts => \&rekonfig_subopts);
    %{$m->{non_tmplvars}} = %basis_non_tmplvars;
    $m->{sems} ||= [];
    foreach my $nom (@konfig_tmplvars) { my $val = $m->{konfig}->get($nom);next unless $val;$m->{tmplvars}->{$nom} = $val };
    $m->ustop('source file name prefix srcfpfx not set during configuration') unless $m->{konfig}->srcfpfx;
    $m->dpop('A2E::Mktdir::postkonfig', 1);
    return 1;
};

=head3 Functions rekonfig_subtmpl, rekonfig_tmplvars, rekonfig_subopts

Plugins for &A2E::Prog::old_ready to make sure that top level functions like mktdir* can be invoked repeatedly on a pre-existing object (without having to reconstruct it with new_ready).

Currently this possibility is used by A2E::Privat::Lnsig and in mktdirs().

=cut

sub rekonfig_subtmpl {
    my $m = shift;
    $m->dpuc('rekonfig_subtmpl', @_);
    my $subtmpl = shift;
    $m->{lastdok} = $m->{dok};
    @{$m->{oldlangs}} = ();
    do { $m->dpop;return } unless $subtmpl;
    $m->dpop('rekonfig_subtmpl', 1);
    return 1;
};

sub rekonfig_tmplvars { 
    my $m = shift;
    $m->dpuc('rekonfig_tmplvars', @_);
    my $vars = shift;
    # %{$m->{konfig}->{STATE}->{VARIABLE}->{tmplvars}} = %$vars if $vars;
    %{$m->{tmplvars}} = %$vars;
    $m->dgot(tmplvars_antevars_ => $m->{tmplvars}->{_antevars_});
    $m->dpop('rekonfig_tmplvars', 1);
    return 1;
};

sub rekonfig_subopts { 
    my $m = shift;
    $m->dpuc('rekonfig_subopts', @_);
    my $vars = shift;
    do { $m->dpop;return } unless $vars;
    %{$m->{konfig}->{STATE}->{VARIABLE}->{subopts}} = %$vars;
    $m->dpop('rekonfig_subopts', 1);
    return 1;
};

=head3 Function mktdir

The bulk of the mktdir5/mktdir4/mktdir0 program, starting right after parsing of commandline options.

=cut

sub submktdir {
    my $m = shift;
    $m->dpuc('submktdir', @_);
    $m->{lastdok} = shift;
    my $sems = shift;
    my $fol = shift;
    $m->{dok} = '';
    my $cwd = $m->dir2dir;
    $m->{cwd} = $cwd;
    my $dirstem = '';
    ($fol, $dirstem) = $m->set_dir_varz($fol, $sems);
    $m->dir_set_url;
    my $dok = eval { $m->dirdok($m->{dir}) };
    if ($dok) {
	$m->chwd($m->{dir});
	$m->{dok} = $dok;
    } else {
	$dok = $m->set_dok_varz;
	$m->mkchfol($fol, $dirstem);
    };
    $m->pars_args(@_);
    $m->topdok_install;
    if (@{$m->{novlangs}}) {
	$m->{hsprintf_repl}->{t} = $m->{tmpl};
	$m->{hsprintf_repl}->{d} = $m->{dok};
	$m->mlht_skrib_faylz;
	$m->skrib_langs_fayl;
	my $tmpl_fnoms = $m->{tmplvars}->{tmpl_fnoms} || $m->{konfig}->tmpl_fnoms; 
	my @fnoms = $m->{sarb}->textsep2list($tmpl_fnoms);
	$m->skrib_fnoms(@fnoms);
	my @tmls = $m->{sarb}->textsep2list($m->{konfig}->tmls);
	foreach my $tml (@tmls) {
	    $m->{hsprintf_repl}->{f} = $tml;
	    $m->skrib_fnom('tml');
	}
	# $m->skrib_konf;
    };
    $m->topdir_svn_cleanup;
    $m->mktdir_finic; # 'make main', needs different setting of local directory, dok_kre_news_kanal needed before mktdirs
    $m->submktdirs($m->{dok}, @{$m->{sems}});
    $m->svn_commit($dirstem);
    $m->chwd($cwd); # 150713 added for extra safety, should be redundant but sometimes apparently isnt
    $m->dpop('submktdir', $dok, $fol);
    return $dok, $fol;
};

sub mktdir {
    my $m = shift;
    $m->dpuc('mktdir', @_);
    $m->fayl_predoks($m->get_dok, kmdfun => 'putdok'); # repair database because we rely on it during creation process
    my $fol = $m->{konfig}->fol;
    my $sems = shift or $m->ustop('need at semantic tree, consisting of at least one element, as argument in mktdir');
    $m->{tmplvars}->{sems} = $sems;
    my $dok = '';
    $m->{lastdok} = $dok;
    ($dok, $fol) = $m->submktdir($dok, $sems, $fol, @_);
    $m->dpop('mktdir', $dok, $fol);
    return $dok, $fol;
};

=head3 Function submktdirs

Like mktdirs but for sublevel invocation

=cut

sub submktdirs {
    my $m = shift;
    $m->dpuc('submktdirs', @_);
    do { $m->dpop;return } if $m->{tmplvars}->{no_subsems};
    my $dok = shift;
    my $tmpl = $m->{tmpl};
    my @sems = @_;
    my %tmplvars = %{$m->{tmplvars}};
    my %opts = (tmplvars => \%tmplvars, subopts => { dokdatum => $m->{hodie} });
    my @subsems = $m->subsems(@sems);
    foreach my $sem (@subsems) {
	$m->old_ready(%opts);
	$m->{lastdok} = $dok;
	$m->{tmpl} = $tmpl;
	$m->submktdir($dok, $sem);
    };
    $m->dpop('submktdirs', 1);
    return 1;
}

=head1 Transferred from elsewhere

Functions which we recently transferred because they seem to belong here because related to functionality confined to this level.

=head2 from Dokfs

=head1 TODO

=over

=item never erase existing lang-neutral file lng.$dok.txt

2010-08-25 用 { cdok tra_feasN8;mktdir rep } 増添已有語言時 C<@lng.feasN8_rep.txt> 被移除

=back

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/Mktdir.pm.tmpl"
# End:
