#!/usr/bin/perl

package A2E::Mvtdir;
use strict;

=head1 NAME

    A2E::Mvtdir - move/rename a hypertext document node

=head1 VERSION

    This describes version B<0.0.3> of A2E::Mvtdir

=cut

    our $VERSION = '0.0.3';

=head1 SYNOPSIS

    my $p = new_ready A2E::Mvtdir;
    $p->chwd($p->dokdir('spez08'));
    my $new_dokname = 'ffii_spez08';
    $p->mvtdir($new_dokname);
    $p->mvtdir('ffii_spez0801', 'january', '01');

=head1 DESCRIPTION

Rename and rewrite the files of the current document node from what they are now to argument 1.

The one-argument mode is used in cases where the directory path stays unchanged and only the document identifier is changed.  

In the three-argument mode, the document to be renamed is found in the subdirectory indicated by the old_leaf_name argument.

In the one-argument mode, the document is found in the current directory.

=head1 PREREQUISITES

=over
=item parent library A2E::Dokfs(3)
=back

=cut

use base 'A2E::Dokfs';
use POSIX;
use Encode qw(_utf8_on);

sub mvtdir_defvars {
    my $m = shift;

=head1 OPTIONS

All options of the parent library and its ancestors are inherited.

=over

=item --setalias|--alias|-a

When renaming the document identifier, leave the old identifier as an alias to the new one.
On by default.

=cut 

    $m->{konfig}->define('setalias|alias|a!', { DEFAULT => 1 });

=back

=head1 CONFIGURATION FILES

	/etc/opt/a2e/mvtdir.konf
	~/mvtdir.konf
	@mvtdir

=cut

    $m->konfayl('mvtdir.konf');
    return 1;
}

=head1 BUGS, LIMITATIONS

So far we can move only the identifier of the current directory or leaves in this directory.

Moving and copying document nodes is more complex than doing the same with files as in the Unix mv and cp commands, and the ramifications have yet to be worked out and specified.

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

Copyright (c) 2007-8 Hartmut Pilch (phm)

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=cut

sub defvars {
    my $m = shift;
    $m->{progver} ||= '0.0.3';
    $m->SUPER::defvars;
    $m->mvtdir_defvars;
    return 1;
}

sub open_faylz ($$) {
    my $m = shift;
    $m->dpuc('open_faylz', @_);
    my $fi = shift;
    $m->open_fayl($fi, 'i', 'r');
    my $fo = join '.', $fi, 'nov';
    $m->open_fayl($fo, 'o', 'w');
    $m->dpop('open_faylz', 1);return 1;
};

sub deplate_rewriter ($$$) {
    my ($m, $var, $val) = @_;
    $m->dpuc('deplate_rewriter', @_);
    my ($line, $dok, @doks, @alias) = ();
    for ($var) {
	/\A(descr|title)\Z/ && do { $line = "#SET id=$var: $val";last };
	/\Adok\Z/ && do { ($dok, @alias) = split /\s+/, $val;@alias = ( $m->pvar('dok0'), grep(m(\A.+\Z), map { !m(\A(self->{dok0}|$m->pvar('dok1'))\Z) && $_ } @alias));$line = "#VAR: dok=$dok\n#VAR: alias=@alias";last }; 
	/\A\alias\Z/ && dok { (@alias) = (split(/\s+/, $val), @{$m->{alias}});@alias = ( $m->pvar('dok0'), grep(/^.+$/, map { !m(\A(self->{dok0}|$m->pvar('dok1'))\Z) && $_ } @alias));$line = "#VAR: alias=@alias";last };
	$line = "#VAR: $var=$val";
    };
    $m->dpop('deplate_rewriter', $line);return $line;
};

sub txt2html_rewriter ($$$) {
    my $m = shift;
    $m->dpuc('txt2html_rewriter', @_);
    my $var = shift;
    my $val = shift;
    my ($line, @doks, @alias) = (); 
    for ($var) {
	/\A\dok\Z/ && dok { (@doks) = split /\s+/, $val;@doks = ( $m->pvar('dok1'), $m->pvar('dok0'), grep(/^.+$/, map { !m(\A(($m->pvar('dok1')|$m->pvar('dok0')))\Z) && $_ } @doks) );$line = "# dok: @doks";last };
	/\A\alias\Z/ && dok { my (@alias) = split /\s+/, $val;@alias = ( $m->pvar('dok0'), grep(/^.+$/, map { !m(\A(self->{dok0}|$m->pvar('dok1'))\Z) && $_ } @alias) );$line = "# alias = @alias";last };
	$line = "# $var: $val";
    };
    $m->dpop('txt2html_rewriter', $line);return $line;
};

sub rewrite_dokdata {
    my $m = shift;
    my ($found, $line) = ();
    $m->put('mlhtdok', { dok=>$m->pvar('dok1') }, doktmp=>'now()');
    foreach my $file (glob('@top-*.txt')) {
	$m->open_faylz($file);
	$line = $m->parse_dokdata($m->fh('i'), $m->fh('o'));
	$m->print_tu_fayl('o', $line . "\n");
	$m->print_tu_fayl('o', $_) while $m->getlini('i');
	$m->kloz_faylz('i', 'o');
    };
    $m->put('dokprop', { dok=>$m->pvar('dok1') }, doknom=>$m->{nom1});
    return 1;
};

sub rename_files {
    my $m = shift;
    $m->dpuc('rename_files', @_);
    do { $m->dpop;return } if $m->pvar('dok1') eq $m->pvar('dok0');
    my ($file0, $file1);
    my $err;
    foreach $file0 (glob("*$m->pvar('dok0')*")) {
	$file1 = $file0;
	$file1 =~ s(\b$m->pvar('dok0')\b)($m->pvar('dok1'));
	$m->dgot(file0 => $file0 , file1 => $file1);
	next if $file1 eq $file0;
	eval { $m->svn_move($file0, $file1) };
	next unless $@;
	$m->uwarn('failed to svn-move file %s to %s: %s', $file0, $file1, $@) if $@;
	eval { rename $file0, $file1 };
	$m->uwarn('failed to perl-rename file %s to %s: %s', $file0, $file1, $@) if $@;
    };
    $m->dpop('rename_files', 1);return 1;
};

sub rewrite_files {
    my $m = shift;
    $m->dpuc('rewrite_files', @_);
    my $Dok0 = $m->CamelCase($m->pvar('dok0'));
    my $Dok1 = $m->CamelCase($m->pvar('dok1'));
    foreach my $file ('Makefile', glob('*.txt'), glob('*.mk')) {
	$m->open_faylz($file);
	while ($m->getlini('i')) {
	    chomp;
	    s(\b$m->pvar('dok0')\b)($m->pvar('dok1'))g;
	    s(\b$Dok0)($Dok1)g;
	    s(\b$m->{nom0}\b)($m->{nom1})g if defined $m->{nom0};
	    $m->print_tu_fayl('o', $_."\n");
	};
	$m->kloz_faylz('i', 'o');
    };
    $m->dpop('rewrite_files', 1);return 1;
};

sub mvtdir {
    my $m = shift;
    $m->dpuc('mvtdir', @_);
    my $dok1 = shift;
    $m->setpvar(dok1 => $dok1);
    if (scalar @_ == 2) {
	$m->{nom0} = shift;
	$m->{nom1} = shift;
    } elsif (@_) {
	die "wrong number of arguments: must be either 1 or 3";
    };
    my $supdir = '';
    if ($m->{nom0}) {
	$supdir = `pwd`;chomp $supdir;
	chdir $m->{nom0};
    } else { $supdir = '..' };
    $m->set_dok;
    my $doksfile = $m->doksfile;
    $m->setpvar(dok0 => $m->pvar('dok'));
    $m->dgot(m_dok0 => $m->pvar('dok0'), m_dok1 => $m->pvar('dok1'), m_nom0 => $m->{nom0}, m_nom1 => $m->{nom1});
    unlink $_ for $m->svn_faylz({ nsf => 1 });
    $m->svn_update;
    $m->rename_files;
    $m->rewrite_files;
    $m->rewrite_dokdata;
    $m->setpvar(dok => $m->pvar('dok1'))
  ALIAS: {
      last unless $m->{konfig}->setalias;
      last unless $m->pvar('dok0') and $m->pvar('dok1') and not $m->pvar('dok0') eq $m->pvar('dok1');
      $m->rmdok($m->pvar('dok0'));
      $m->set_alias($m->pvar('dok0'), $m->pvar('dok1'));
    };
    $m->svn_commit;
  MVNOM: {
      my $ok = $m->{nom0} ? 1 : 0;
      $m->dpuc('mvnom', $ok, $m->{nom0});
      do { $m->dpop;last } unless $ok;
      $m->chwd($supdir);
      $m->svn_move($m->{nom0}, $m->{nom1}, 1);
      $m->svn_commit;
      $m->dpop;
    };
    $m->rewrite_doksfile($doksfile, 'mv', $m->pvar('dok0'), $m->pvar('dok1'));
    $m->dpop('mvtdir', 1);return 1;
};

=head1 The End

=cut

return 1;

# Local Variables: ***
# coding: utf-8 ***
# mode: perl ***
# End: ***
