# Time-stamp: <2018-10-30 11:58:18 phm>
package A2E::News2html;
use strict;
use base A2E::Dokfs;

=head1 NAME

A2E::News2html - generate a menu from newsticker info found in a text file,

=head1 DESCRIPTION

analogous to nav2html, yet to be implemented in more detail, also in
combination with an RSS mechanism.

=head1 VERSION

This describes version B<0.0.2> of A2E::News2html.

=cut

our $VERSION = '0.0.2';

=head1 SYNOPSIS

    new $p = new_ready A2E::News2html;
    $p->news2html(@ARGV)

=head1 OPTIONS

=over

=item --help | -h

Display short help and exit.

=item --manpage | -m

Display full documentation and exit.

=back

General options:

=over

=item --debug

Enable copious script debugging output (don't bother, this is for the
developer)

=item --verbose | -v

verbose reporting for the user

=item --urllab | -u

In the input file the URL comes first, the associated title second
(traditionally it has been the other way around)

=item --dok

document identifier, e.g. 'news2html_pod0701'

Preferrably found in the name of the input file.

=item --lang <lang>

two-letter ISO language code, e.g. 'en'. 

Preferrably found in the name of the input file.

=item --infiles <infiles> | -i <infiles> | <infiles>

Files used for input.

Should preferrably be one file with a name like

    news2html_pod0701.nav.en.html

where 'news2html_pod0701' is the document identifier (dok) and 'en' the language (lang).

But it's also possible to specify several files.

Moreover, the dok and lang elements can also be specified by separate commandline options, see above.

=head1 BUGS

Tell adv at a2e de about them.

=head1 PREREQUISITES

    A2E::Dokfs(3)

=head1 SCRIPT CATEGORIES

dokfs

=head1 ENVIRONMENT

=over

=item INFILE FORMAT

The infile could typically read as follows:

	This project

	 href:news2html.pl
	 href:sqltab.pl		SQLTAB

	@Environment

	Participants

	 dok:phm	Hartmut Pilch
	 dok:adv	Data Processing
	# dok:i2p	I2P


This creates a menu, in which the indented items are links.
The href macro's argument is a relative link 
The dok macro generates a relative link based on a document identifier.
The second column may contain a label.
If it isn't pressent, a default value is used for the label.
This default is either a label found in the database or the link itself.

To compile the above form, the '--urllab' option needs to be specified.
Without this option, the sequence of url and label is inversed. 
In a later version, the 'urllab' sequence will probably become the default.

Unindented lines represent titles.

If an unindented line starts with a '@', it is followed by a menu, as far as
this can be generated from the document hierarchy that is registered in the
database.

Lines that begin with hash marks (#) are treated as comments, i.e. ignored.


=item Database Format

As required by the 'A2E::Dokfs' module.


=back

=head1 SEE ALSO

news2html(1)
perl(1)

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

Copyright (c) 2006-8 Hartmut Pilch

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=cut

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::News2html::defvars', @_);
    $m->SUPER::defvars;
    $m->define('outfile|out|o=s');
    $m->define('infiles|in|i=s@');
    $m->define('urllab|u!');
    $m->dpop('A2E::News2html::defvars', 1);return 1;
};

sub open_sidebar {
    my $m = shift;
    ++$m->{sidebar};
    return "$m->{topline}\n<div id=\"sidebar\">\n";
};
sub close_sidebar {
    my $m = shift;
    --$m->{sidebar};
    return "</div>\n";
}
sub open_sidepanel {
    my $m = shift;
    ++$m->{sidepanel};
    return "<div class=\"sidepanel\">\n<p class=\"sideBarTitle\">$_[0]</p>\n";
}
sub close_sidepanel {
    my $m = shift;
    --$m->{sidepanel};
    return "</div>\n";
}
sub open_ulist {
    my $m = shift;
    ++$m->{ulist};
    return "<ul>\n";
};
sub close_ulist {
    my $m = shift;
    --$m->{ulist};
    return "</ul>\n";
};

sub hreftextitem {
    my $m = shift; 
    my ($href, $text) = @_;
    return sprintf "<li>%s</li>\n", $m->hreftext($href, $text);
};

sub news2html_sub {
    my $m = shift;
    $m->dpuc('news2html_sub', @_);
    my $fi = shift;
    $m->open_fayl($fi, 'i', 'r');
    my ($url, $lab, $tit) = ();
    my $dok = $m->pvar('dok');
    while ($m->getlini('i')) {
	($url, $lab, $tit) = ();
	next if /\A\#/;
	chomp;
	s(\s+\Z)();
	/\A(\S.*)\Z/ && do { 
	    $tit = $1;
	    print $m->close_ulist while $m->{ulist};
	    print $m->close_sidepanel while $m->{sidepanel};
	    if (m(\A\@(\w+)\Z)) { 
		$tit = $1;
		print $m->open_sidepanel($m->htmlesc($tit));
		print $m->open_ulist;
		print join "\n", $m->dokmenu_entries($dok);
	    } else {
		print $m->open_sidepanel($m->htmlesc($tit));
		print $m->open_ulist;
	    };
	    next; 
	};
	if ($m->{konfig}->urllab) {
	    /\A\s+href\:(\S+)\s\t\s*(\S.*\S)\Z/ && do { print $m->ahref_item($1, $2), "\n";next };
	    /\A\s+dok\:(\S+)\s*\t\s*(\S.*\S)\Z/ && do { print $m->ahref_item($1, $2), "\n";next; };
	    /\A\s+(\S+)\s*\t\s*(\S.*\S)\Z/ && do { print $m->ahref_item($1, $2), "\n";next; };
	} else {
	    /\A\s+(\S.*\S)\s*\t\s*href\:(\S+)\Z/ && do { print $m->ahref_item($2, $1), "\n";next };
	    /\A\s+(\S.*\S)\s*\t\s*dok\:(\S+)\Z/ && do { print $m->ahref_item($2, $1), "\n";next; };
	    /\A\s+(\S.*\S)\s*\t\s*(\S+)\Z/ && do { print $m->ahref_item($2, $1), "\n";next; };
	};
	/\A\s+dok\:([\w\_]+)\Z/ && do { print $m->ahref_item($1), "\n";next; };
	/\A\s+href\:(\S+)\Z/ && do { print $m->ahref_item($1), "\n";next; };
	/\A\s+([\w\_]+)\Z/ && do { print $m->ahref_item($1), "\n";next; };
	/\A\s+([^\t\n]+)\Z/ && do { print $m->ahref_item( -r $1 && $1, $1), "\n";next; };
    };
    $m->kloz_fayl('i');
};

sub news2html {
    my $m = shift;
    $m->dpuc('news2html', @_);
    $m->init_navdoks;
    $m->{sidebar} = 0;
    $m->{sidepanel} = 0;
    $m->{ulist} = 0;
    my @infiles = @{$m->{konfig}->{infiles}};
    push @infiles, @_;
    $m->{topline} = "<!-- -*- coding: utf-8; mode: fundamental; srcfile: $fi; -*- -->";
    my $fo = $m->{konfig}->{outfile};
    if ($fo) { open SAVOUT, '>&STDOUT';die sprintf "cant redirect STDOUT to %s", $fo unless open STDOUT, '>', $fo };
    binmode STDOUT, ":utf8";
    # print $m->open_sidebar;
    foreach my $fi (@infiles) { $m->news2html_sub($fi) };
    print $m->close_ulist while $m->{ulist};
    print $m->close_sidepanel while $m->{sidepanel};
    # print $m->close_sidebar;
    if ($fo) { close STDOUT;open STDOUT, '>&SAVOUT' };
    $m->dpop('news2html', 1);return 1;
};

=head1 The End

=cut

return 1;

# Local Variables:
# coding: utf-8
# srcfile: /adv/perl/A2E/News2html.pm.tmpl
# mode: perl
# End:
