#!/usr/bin/env perl

package A2E::OpenPGP;

=head1 NAME

    A2E::OpenPGP(3) -- conveniently encrypt, decrypt, sign according to OpenPGP using Crypt::OpenPGP

=head1 DESCRIPTION

Enter the passphrase with reduced interaction if you dare to do without the nanny services of the gpg-agent(1)

=head1 SYNOPSIS

    my $p = new_ready A2E::OpenPGP progver => '0.0.1';
    $p->sign('my_file.txt');       

=head1 PREREQUISITES

    A2E::Prog(3) -- base library from where options are inherited
    Crypt::OpenPGP(3) -- the workhorse behind our interface

=cut

use base A2E::Prog;
use Crypt::OpenPGP;
use File::Basename;

sub defvars {
    my $m = shift;
    $m->SUPER::defvars;

=head1 VERSION

    0.0.1

=cut

    $m->{progver} = '0.0.1';

=head1 OPTIONS

see prerequisites, plus the following

=cut

=over

=item --pgp_key I<3FD329EE>

pgp key, can also be supplied via the environment variable OpenPGP_KEY.

=cut

    $m->{konfig}->define('pgp_key=s');

=item --pgp_pass I<........>

pgp passphrase, can also be supplied via the environment variable OpenPGP_PASS.

=cut

    $m->{konfig}->define('pgp_pass=s');

=item --detach

produce a detached signature, to be stored in a file whose name is derived by adding .asc (in case of --armour) or .gpg to the source file name

=cut

    $m->{konfig}->define('detach!');

=item --clearsign|clear

sign at end of text

=cut

    $m->{konfig}->define('clearsign|clear!');

=item --armour|--armor

write 7-bit ascii for safe mail transport

=cut

    $m->{konfig}->define('armour|armor!');

=item --sign

produce a signature

=cut

    $m->{konfig}->define('sign!');

=item --decrypt

decrypt and write either to outfayl (supplied via -o) or to a file derived by removing .asc or .gpg suffix, or, if all that is not available, to stdout.

=cut

    $m->{konfig}->define('decrypt!');

=item --binmode I<utf8>

When decrypting binary files this should be set to C<bytes>

=cut

    $m->{konfig}->define('binmode=s');

=back

=cut

    $m->{pgp} = new Crypt::OpenPGP;

=head1 FILES

=head2 Configuration

	/etc/opt/a2e/openpgp.konf
	~/.openpgp_konf
	./@openpgp

=cut

    $m->konfayl('openpgp.konf');
    $m->konfayl('a2e_pgp.konf');
    return 1;
}

=head1 IMPLEMENTATION

=head2 Execution Functions

=head3 Function pgp_args

Conveniently read basic arguments into the hash that Crypt::OpenPGP uses and return that hash.
The input can be a reference to a ready hash or a list consisting of
 - arg1: filename or string
 - arg2: passphrase
If no passphrase is given, it is read from the pgp_pass variable, the PGP_PASS environment variable or the terminal.

This function is used by C<sign> as well as C<decrypt>

=cut

sub pgp_args {
    my $m = shift;
    $m->dpuc('pgp_args', @_);
    my $arg1 = shift;
    my %args = eval { %$arg1 };
    if (%args) {
	$m->dgot(args => \%args);
    } elsif (ref $arg1) {
	$args{Data} = eval { $$arg1 };
    } elsif (-e $arg1) {
	$args{Filename} = $arg1;
    } elsif (1) {
	$args{Data} = $arg1;
    };
    $args{Passphrase} ||= shift || $m->{konfig}->pgp_pass || $ENV{PGP_PASS} || '';
    unless (length $args{Passphrase}) {
	delete $args{Passphrase};
	$args{PassphraseCallback} = sub { $ENV{PGP_PASS} = Crypt::OpenPGP::_default_passphrase_cb @_;return $ENV{PGP_PASS} };
    };
    $args{KeyID} ||= $m->{konfig}->pgp_key || $ENV{PGP_KEY};
    $m->dpop('pgp_args', %args);
    return %args;
}

=head3 Function sign

=cut

sub sign {
    my $m = shift;
    my (%args) = $m->pgp_args(@_);
  SAVOUT: {
      last unless $m->{konfig}->detach;
      my $fo = join '.', $fi, $m->{konfig}->armour ? 'asc' : 'gpg';
      $m->out2sav($fo);
    };
    die 'need pgp key id' unless $args{KeyID};
    my $res = $m->{pgp}->sign(%args, Armour => $m->{konfig}->armour, Clearsign => $m->{konfig}->clearsign, Detach => $m->{konfig}->detach) or die $m->{pgp}->errstr;
    print $res if $m->{savout};
    return $res;
};

=head3 Function passphrase_cb

=cut

sub passphrase_cb {
    my $cert = shift;
    print STDERR $cert ? sprintf('Enter passphrase for secret key %s: ', $cert->key_id_hex) : 'Enter your OpenPGP passphrase: ';
};

=head3 Function decrypt

=cut

sub decrypt {
    my $m = shift;
    $m->dpuc('decrypt', @_);
    my (%args) = $m->pgp_args(@_);
  SAVOUT: {
      $m->dpuc('savout', $args{Data}, $args{Filename}, $args{Outfile}, $m->{savout_fayl});
      do { $m->dpop;last } if $args{Data};
      do { $m->dpop;last } if $m->{savout};
      $m->ustop('need either a string or a name of a file to decrypt') unless exists $args{Filename};
      $args{Outfile} ||= basename $args{Filename}, '.gpg', '.asc';
      die sprintf 'cant figure out how to derive an output file name from filename %s', $args{Filename} if $args{Outfile} eq $args{Filename};
      $m->out2sav($args{Outfile});
      $m->vecho('decrypting file %s to file %s', $args{Filename}, $args{Outfile});
      $m->dpop('savout');
    };
    my $binmode = $args{binmode} || $m->{konfig}->binmode || '';
    binmode STDOUT, ':' . $binmode if length $binmode;
    my $res = $m->{pgp}->decrypt(%args) or $m->ustop('failed typ decrypt with args %s: %s', \%args, $m->{pgp}->errstr);
    print $res if $args{Outfile};
    $m->dpop('decrypt', $res);
    return $res;
}

=head3 Function pgp

process a commandline of a program similar to gpg(1)

=cut

sub pgp {
    my $m = shift;
    $m->dpuc('pgp', @_);
    if ($m->{konfig}->sign) {
	$m->sign(@_);
    } elsif ($m->{konfig}->clearsign) {
	$m->sign(@_);
    } else {
	$m->decrypt(@_);
    };
    $m->sav2out;
    $m->dpop('pgp', 1);return 1;
};

=head1 The END

=cut

return 1;

# Local Variables:
# coding: utf-8
# url: https://a2e.de/adv/perl/A2E/OpenPGP.pm
# srcfile: /adv/perl/A2E/OpenPGP.pm
# mode: perl
# End:
