# -*- coding: utf-8; srcfile:/adv/perl/A2E/Prog.pm; mode: perl -*-
=head1 NAME

A2E::Prog

=head1 VERSION

This describes version B<0.4.4>

=head1 DESCRIPTION

Basic program elements useful for almost any program, including 

  vecho: verbose message
  decho: debugging message
  necho: non-quiet message
  techo: output this message when in testing or debugging mode    
  exek: execute a program and return non-null if successful
  setpvar, pvar: storing and retrieving variables that belong to the program rather than userspace

=head1 SYNOPSIS

use A2E::Prog;
$p = new_ready A2E::Prog verbose => 1, dev => '/dev/mapper/perldisk';
$p->vecho('OK') if $p->exek('cryptsetup', 'isLuks', $p->{dev}); 

=head1 DESCRIPTION

Elements common to any program of more than minimal complexity, including
output and debugging messages, commandline parsing, configuration, invocation
of external programs and capturing of their output.

This is a fairly thin wrapper around a series of standard CPAN modules.

This does not comprise database access.  To add that, invoke A2E::Daba.
A2E::Prog is a base class that most classes at A2E will inherit and most
programs will directly or indirectly use.

=head1 BUGS, CAVEATS

The documentation effort has just begun.

=head1 PREREQUISITES

    AppConfig(3)
    Getopt::Long(3)
    Pod::Usage(3)
    Encode(3)
    Carp(3)
    POSIX(3)
    perl(1)
									       
=head1 SEE ALSO

  http://a2e.de/adv/perl/A2E
    A2E::Daba(3)
    perldoc(1)

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

    Copyright (c) 2007 Hartmut Pilch (phm)

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

=head2 Header: Load Libraries

=cut 

our $novitas = $ENV{NOVITAS} || 0;
package A2E::Prog;
use strict;use warnings 'all';no warnings 'uninitialized';
use utf8;
require 5.008;
use Encode qw(_utf8_on decode);
# use Carp qw(carp cluck croak confess);
use POSIX qw(strftime locale_h);
use Getopt::Long;Getopt::Long::Configure('gnu_getopt');
use Pod::Usage;
use AppConfig qw(:expand);
use IO::File;
use Data::Dumper;
$Data::Dumper::Maxdepth = 2;
use File::Spec::Functions qw(catfile catdir splitpath splitdir);
use File::Basename;
use File::Path;
use Fcntl ':flock';
# use SVN::Client;
use Net::Ping;
use Cwd;
use English -no_match_vars;
use Sys::Hostname;
use open ':utf8';
use Scalar::Util qw(reftype blessed);
binmode STDIN, ':utf8';
binmode STDERR, ':utf8';
binmode STDOUT, ':utf8';

=head2 Translatable Messages
=cut

our %MSG = ();
our %READY = ();
our %MSGS = (
    OK => 'OK',
    funnovar => 'at %(PL) no value given for variable %v',
    fnotups => 'no tups to put in file %f',
    badmsgst => 'Bad message syntax: any hashrefs used as message arguments should be for hsprintf only nowadays; %a should no longer stand at the beginning of the arguments of the message %k whose text is %t and whose args are %A',
    binmode => 'cant set handle of file %f to utf8',
    klozfayl => 'cant close handle of file %f',
    openfayl => 'cant open file %f for reading in dir %d',
    dirgflz => 'wildcard %g in dir %d should be matched by just one file rather than %f',
    msgu2v => 'attempting to reset message %k from q %u eoq to q %w eoq',
    nomisref => 'nom %n is a reference here',
    badnom => 'ill-formed variable name %n',
    conoferm => 'in exek_open3 already closed child output',
    cinoferm => 'in exek_open3 failed to close child input',
    snedef => '%s undefined',
    open3perr => 'error in exek_open3: failed to open child',
    pidexit => 'child exited with value %e',
    diedwcd => 'child of command %k died with signal %s and coredump',
    diedwocd => 'child of command %k died with signal %s without a coredump',
    open3_nokmd => 'no command to exec for open3',
    badmsgk => 'invalid message key %k',
    chwdnoarg => 'chwd without argument in dir %d',
    badexek => 'failed to execute command %k: %s',
    pvardef => 'program variable %n should not be undefined here',
    nolangvar => 'Landessprache unbekannt; es könnte helfen, die Umgebungsvariable LANG zu setzen',
    noping => 'failed to construct a ping with protocol %p and timeout %t',
    setlocale => 'konnte Sprachparameter nicht auf %l setzen',
    localenesed => 'locale set to %l instead of %m',
    nolangnolocale => 'lang unknown and no locale found',
    nolanglocale => 'lang unknown and nonlingual locale %l found',
    localenolang => 'konnte locale %l keine Sprache entnehmen',
    fputs => 'cant get write handle on file %f: %m'
 );

=head2 Echo: Message routines

=head3 Function msg

    Subroutine of echo & friends.
    Depending on the nature of the second argument, interpolate the first either with sprintf or with hsprintf.
    The hsprintf mode is activated when the second argument is a hashref with at least one key and with all keys consisting of only one letter, e.g.

    $m->msg('my file %f has %l lines', { f => 'file.txt', l => 15 }); 

    The sprintf mode is

    $m->msg('my file %s has %d lines', 'file.txt', 15); 

    Both modes can be mixed:    

    $m->msg('my file %f has %d lines', { f => 'file.txt' }, 15); 

    Additionally, an unmixable non-hashref hsprintf mode can be used:

    $m->msg('my file %f has %d lines', f => 'file.txt', d => 15)

    In both modes the arguments can be all kinds of variables, including trees built from nested references.

    Using hsprintf is preferrable because translatable and more robust with regard to unaffected percent characters.

    sprintf mode messages like

    $m->msg('symbol %s refers to file %s', 'f', 'file.txt');

    fail because the argument list looks like f => 'file.txt'.

    This happens only when the number of arguments is even and the first and every other consist of one letter only.

    Say

    $m->msg('symbol %s refers to file %s', {}, 'f', 'file.txt');

    instead or switch to

    $m->msg('symbol %s refers to file %f', s => 'f', f => 'file.txt');

    right away.

=cut

sub defmsg {
    my ($m,$k,$v) = @_;
    $m->pstop(badmsgk => k => $k) unless $k =~ m(\A\w+\Z);
    my $u = $m->{msg}->{$k};
    $m->pwarn(msgu2v => k => $k, u => $u, w => $v) if $u;
    $m->{msg}->{$k} = $v;
    return $v;
};
sub msgdefs {
    my ($m,%msgs) = @_;
    while (my($k,$v) = each %msgs) {
	$m->defmsg($k => $v);
    };
};
sub prog_msgdefs {
    my $m = shift;
    my ($nom,%msgs) = @_;
    my $prog = join '_', $nom, 'msgdefs';
    do { $m->pwarn('messages %n already defined', n => $nom);return } if $m->ready($prog);
    while (my($k,$v) = each %msgs) {
	$m->defmsg($k => $v);
    };
    1;
}
sub prog_minidefs { # use in minimally-inheriting files like A2E::Lihs
    my $m = shift;
    $m->hsprintf_qre2fun_defs;
    $m->prog_msgdefs('prog',%MSGS);
};
sub msg {
    my $m = shift;
    my $k = shift or return;
    my $s = '';
  MSGNURSYM: {
      do { $s = $k;last } unless $k =~ m(\A[a-z,_]+\Z);
      $s = $m->{msg}->{$k} || $MSGS{$k};
      last if $s;
      warn qq{using unknown message symbol $k};
      return $k unless @_;
      return Dumper $k, @_;
    };
    _utf8_on $s;
    my @rest = ();
  ARGS: {
      last unless @_;
      my %argh = ();
      my @rest = ();
      if ('HASH' eq ref $_[0]) { my $arg0 = shift;%argh = %$arg0;@rest = @_ } elsif ($#_ % 2) { %argh = @_ } else {
	  my $lok = $m->plok;die qq{uneven number of attribute tuple arguments @_ for message $k with body $s at $lok} 
      };
      last unless %argh;
      for (keys %argh) { 
	  next if '' eq ref $_ and m(\A[a-z,A-Z]{1,3}\Z);
	  warn sprintf qq{bad message argument key %s: any hashrefs used as message arguments should be for hsprintf only nowadays; the message $k with body $s has argh %s and rest %s}, Dumper($_), Dumper(\%argh), Dumper(\@rest);
	  last ARGS;
      };
      $s = $m->hsprintf($s, %argh);
    };
    return $s unless @rest;
    @rest = map { $m->s8dval($_) } @rest;
    $s = sprintf $s, @rest;
    $s;
};

=head3 Function echo

    Arg 1 is a flag which must be true for anything to happen.
    Arg 2 is the message.
    The remaining arguments are, if present, incorporated into the message by sprintf/hsprintf.

=cut

sub echo {
    my $m = shift;
    my $flag = shift;
    return unless $flag;
    my $msg = $m->msg(@_);
    print STDERR $msg . "\n";
    $msg;
};

=head3 Function vecho

    verbose message

=cut

sub vecho {
    my $m = shift;
    my $verbose = $m->{progvars}->{verbose};
    $m->echo($verbose, @_);
};

=head3 Function debug

=cut

sub debug {
    my ($m,$val) = @_;
    return $m->{debug} unless defined $val;
    $m->{debug} = $val;
    $val;
};

=head3 Function decho

    debugging message

=cut

sub decho {
    my $m = shift;
    $m->echo($m->{progvars}->{debug}, @_);
};
sub techo {
    my $m = shift;
    return $m->echo($m->{progvars}->{test_modus} || $m->{progvars}->{debug}, @_);
};

=head2 Function dval

represent perl expressions

=cut

our %DVALREFS = ();

sub dlst { # represent one listref
    my ($m,$val) = @_;
    return '[ '. join(', ', map { $m->dval($_) } @$val) .' ]';
};

sub dhac { # represent one hashref
    my ($m,$val) = @_;
    return join(', ', map { join ' => ', $_, $m->dval($val->{$_}) } keys %$val);
};

sub dval {
    my ($m,$val) = @_;
    my $ref = ref $val;
    my $max = 128; # $m->{konfig}->dval_array_maxlen;
  STR: {
      last if $ref;
      _utf8_on $val;
      return q{'}.$val.q{'};
    };
    INIFIN: {
	last unless $DVALREFS{$val};
	warn qq{dval avoiding infinite loop in reference $ref};
	return '(BADREF:'. Dumper($val) . ')';
    };
    return $ref if $ref eq 'CODE';
    $DVALREFS{$val} = 1;
    my $setdel = sub {
	my $s = shift;
	delete $DVALREFS{$val};
	return $s;
    };
  ARRAY: {
      last unless $ref eq 'ARRAY';
      my $len = scalar @$val;
      return &$setdel('@' . $len) if $max and $len > $max;
      return &$setdel($m->dlst($val));
    };
  HASH: {
      last unless $ref eq 'HASH';
      my $len = scalar keys %$val;
      return &$setdel('%' . $len) if $max and $len > $max;
      return &$setdel('{ '. $m->dhac($val) .' }');
    };
    delete $DVALREFS{$val};
    return "'". $ref . "'" if blessed $val;
    Dumper $val;
};
sub s8dval { # return utf8 string or ref representation
    my ($m,$o) = @_;
    return $m->dval($o) if ref $o;
    _utf8_on $o;
    return $o;
};
    

=head3 Function dmsg

Message about current function start, end or state.
$1 : $msg : 'start' || 'end' || 'inmid' :: initial word
$2 : $popz_ref : [[],[]] : symbols to be popped off in case of end
$3@ : @args : () : resulting values to be displayed

=cut

sub dmsg {
    my ($m,$msg,$popz_ref,@args) = @_;
    my $debug = $m->{progvars}->{debug};
    return if !$debug;
    my $funs = join '.', map { $m->dval($_) } @{$m->{funz}};
    $msg = join ' ad ', $msg, $funs || 'orig' ;
    my $from = join '.', map { $m->dval($_) } @$popz_ref;
    $msg = join ' de ', grep { $_ } $msg, $from;
  SET_VALZ: {
      my $s = @args ? $m->koniug_et(map { $m->dval($_) } @args) : 'nil';
      $msg = join ' cum ', $msg, $s;
    };
    print STDERR $msg. "\n";
    $msg;
};

=head3 Function dpuc

    Lightweight debugging messages when entering and exiting from a function.

    Maintain a funz stack in the object and push the new function name onto it when starting.

    Use dpop to pop it off when leaving.

=cut

sub dpuc {
    my ($m,$prog,@args) = @_;
    push @{$m->{funz}}, [$prog,@args];
    $m->dmsg('start', [], @_);
    ($m,$prog,@args);
};

=head3 Funktion dpop

    Leave a group that was begun by C<dpuc>

=cut

sub dpop {
    my ($m,$prog,@args) = @_;
    my ($pop,@popz) = ();
    if ($prog) {
      POPZ: {
	  my $len = $#{$m->{funz}};
	  my $i = $len;
	  my $j = $i;
	  while ($i >= 0) {
	      $pop = $m->{funz}->[$i]->[0];
	      $i--;
	      do { my @funz = @{$m->{funz}}[0..$i];@popz = @{$m->{funz}}[$j..$len];@{$m->{funz}} = @funz;last POPZ } if $pop eq $prog;
	      $j--;
	  };
	  my @verbs = map { $_->[0] } @{$m->{funz}};
	  $m->pwarn('dpop could not find function %f in chain %C', f => $prog, C => \@verbs);
	};
    } else {
	$pop = pop @{$m->{funz}};
	@popz = ($pop); 
    };
    my $nodebug = $m->{nodebug};
    $m->dmsg('reven', \@popz, @args);
  DPOP_UNDEF_IF_EMPTY: return unless @args; # return undefined rather than just an empty list
    @args;
};
sub dpop1 {
    my ($m,$prog,@args) = @_;
    @args = $m->dpop($prog,@args);
    die sprintf "dpop1 $prog must have exactly one argument, not %s", Dumper(\@args) unless $#args == 0;
    $args[0];
};
sub ddestrukt {
    my ($m,$prog,$destrukts) = shift->ppuc('ddestrukt',@_);
    $m->pstop('bad destructor list %D', D => $destrukts) if 'ARRAY' ne ref $destrukts;
    while (my $destrukt = shift @$destrukts) {
	my $ref = ref $destrukt;
	$m->pstop('invalid destructor %r', r => $ref) unless 'CODE' eq $ref;
	&$destrukt;
    };
    1;
};
sub ddpop {
    my ($m,$prog,$destrukts,@args) = @_;
    $m->ddestrukt($destrukts);
    $m->dpop($prog,@args);
};
sub ddpop1 {
    my ($m,$prog,$destrukts,$arg) = @_;
    $m->ddestrukt($destrukts);
    $m->dpop1($prog,$arg);
};

=head4 wpop, wpop1, Wpop0, Wpop1, Wpop9 leave a block with a warning message and optional argument(s)

Allow rich return from functions by combining dpop/dpop1 and message functions pwarn,uwarn,vecho,techo,decho,echo into one return argument.

     $m->wpopx($prog,{ fun => 'p', msg => 'funnovar', v => 'dok' });
     $m->wpopx($prog,[qw(p funnovar v dok)]);
     $m->wpopx($prog,['u', 'dirnodok', 'd', $dir]);
     $m->wpopx($prog,"directory $dir does not exist");
     $m->wpopx($prog,"directory $dir does not exist");
     $m->wpopx($prog,$m->pwarn(funnovar => v => 'dok'));
     $m->wpopx($prog,do { &my_stuff;$m->pwarn(funnovar => v => 'dok') });
     $m->wpopx($prog,sub { $m->pwarn(funnovar => v => 'dok') });
     $m->wpopx($prog,$my_stuff);

If, in the hash notation, the fun attribute is omitted, 'p' i.e. \&pwarn is used:

     $m->wpopx($prog,{ msg => 'dirnodok', d => $dir });

The available funs are laid down in the %warfuns registry.

The 'e' \&echo function which has a different argument list can still be used in the list but not in the hash notation:

     $m->wpopx($prog,['e',$verbose,'dirnodok','d',$dir]);	

For even more complex variants, the code type argument or perhaps a string resulting from a do { ... } construct can be passed.
Such direct passing of arguments is most efficient, but the list notation seems most elegant.

Return arguments @ret can be passed either as list or as singleton, dependent on $unu switch

     $dir = $m->wpopx($prog,[qw(p funnovar v dok)],1,$dir)
     ($dir) = $m->wpopx($prog,[v => 'dirnodok', d => $dir],0,$dir)

The list version is usually rewritten as wpop, the singleton one as wpop1.

Examples from file:/adv/perl/A2E/Dokfs.pm.tmpl:
 
    return $m->wpop($prog,[u => 'no absurl found for dok %D', D => $dok]) unless $absurl;
    return $m->wpop1($prog,[d => 'Verzeichnis %d nicht vorhanden', d => $dir], $opts{force}) unless -d $fol;

As indicated in the last example, the initial 'd', pointing to &decho, is not in the same namespace as the later ones that point to hsprintf arguments.

These can be rewritten as wpop and wpop1 respectively

The Wpop0, Wpop1, Wpop9 versions are variant of the above arrayref \@warn version but with dereferenced array @warn.
The border between the @warn and @rets arrays is determined by the number of @rets (0 and 1 in Wpop0 and Wpop1 respectively) or a delimiting null value in Wpop9.

    return $m->Wpop0($prog, u => 'no absurl found for dok %D', D => $dok) unless $absurl;
    return $m->Wpop1($prog, d => 'Verzeichnis %d nicht vorhanden', d => $dir, f => $opts{force}) unless -d $fol;
    return $m->Wpop($prog, u => noinfofrom, p => $paf, d => $dir, m => $@) if $@;
    return $m->Wpop9($prog, u => noinfofrom, p => $paf, d => $dir, m => $@, 0, $paf, $dir) if $@;

These return the values of none, the last one or all of the message arguments respectively.
The Wpop9 is probably of no great practical value compared to the less costly 

    return $m->wpop($prog, [u => noinfofrom, p => $paf, d => $dir, m => $@], $paf, $dir) if $@;

or the more elegant

    return $m->Wpop($prog, u => noinfofrom, p => $paf, d => $dir, m => $@) if $@;

In the latter case the extra argument $@ will usually do no harm.

=cut

our  %warnfuns = (p => \&pwarn, P => \&pstop, u => \&uwarn, U => \&ustop, v => \&vecho, d => \&decho, e => \&echo, t => \&techo);
sub wpop0 { 
    my ($m,$prog,$warn,$unu,@ret) = @_;
    for my $ref (ref $warn) {
	if ('CODE' eq $ref) { &$warn;last };
	if ('HASH' eq $ref) { my %args = %$warn;(my $f,%args) = $m->shifthash('fun',%args);(my $msg,%args) = $m->shifthash('msg',%args);die 'no msg key in hash' unless $msg;my $fun = $warnfuns{$f} ||= \&pwarn;&$fun($m,$msg,%args);last };
	if ('ARRAY' eq $ref) { my @args = @$warn;my $f = shift @args;my $fun = $warnfuns{$f} || \&pwarn;&$fun($m,@args);last };
	last if $ref;
	$m->pwarn($warn);
    };
    return $m->dpop($prog) unless @ret;
    return $m->dpop($prog,@ret) unless $unu;
    die "during wpop0 from $prog trying to pass arguments @ret as singleton" unless 0 == $#ret;
    $m->dpop1($prog,$ret[0]);
};
sub wpop {
    my ($m,$prog,$warn,@rets) = @_;
    $m->wpop0($prog,$warn,0,@rets);
};
sub wpop1 {
    my ($m,$prog,$warn,$ret) = @_;
    $m->wpop0($prog,$warn,1,$ret);
}
sub Wpop {
    my ($m,$prog,@warn) = @_;
    my $i = 1;my @rets = ();
    while ($warn[++$i]) {
	push @rets, $warn[++$i];
    };
    $m->wpop($m,$prog,\@warn,@rets);
};
sub Wpop1 {
    my ($m,$prog,@warn) = @_;
    $m->wpop1($prog,\@warn,$warn[$#warn]);
}
sub Wpop0 {
    my ($m,$prog,@warn) = @_;
    $m->wpop($prog,\@warn);
};
sub Wpop9 { # nil as border between @warn and @rets: $m->Wpop0($prog, u => noinfofrom => p => $paf, d => $dir, m => $@, undef, $paf, $dir) 
    my $m = shift;my $prog = shift;my @warn = ();
    while (1) { if ($_[0]) { push @warn, shift, shift } else { shift;last } };
    $m->wpop($prog,\@warn,@_);
};

sub ppuc{
    my $m = shift; 
    my $prog = shift;
    $m->dpuc($prog,@_);
    ($m,$prog,@_);
};

=head3 qpuc, qpop, qpop1: do this function only once 

Invocation of qpuc must involve 'or return' or similar.
This is used for initiation routines like postkonfig, prog_postkonfig. 

=cut    
sub qpuc { # quit if ready
    my ($m,$prog,@args) = @_;
    return if $m->{ready}->{$prog};
    $m->dpuc($prog,@args);
    ($m,$prog,@args);
};
sub qpop { # mark as ready and quit
    my ($m,$prog,@args) = @_;
    $m->{ready}->{$prog} = 1;
    $m->dpop1($prog,@args);
}
sub qpop1 {
    my ($m,$prog,@args) = @_;
    $m->{ready}->{$prog} = 1;
    $m->dpop1($prog,@args);
}

=head3 Funktion dgot

    $m->dgot(fnom => $fnom, name => $name, mode => $mode, binmode => $opts{binmode});

=cut

sub dgot {
    my ($m,%optz) = @_;
    $m->dmsg('inmid', [], \%optz);
};

=head3 Function necho

    normal, necessary message.
    Disabled only if quiet option is set.

=cut

sub necho {
    my ($m,@args) = @_;
    $m->echo(!$m->{quiet}, @args);
};

=head3 Functions ustop, pstop, uwarn, pwarn

    User level vs program level error warning/termination.
    
    Wrappers around die, warn, croak, confess, carp and cluck.

=cut

sub ustop {
    my $m = shift;
    my $msg = $m->msg(@_);
    die $msg;
}

=head4 $m->ustop_or_warn($pres,@_)

Stop if not $pres, otherwise press ahead and just warn.

=cut
sub ustop_or_warn {
    my ($m,$pres,@args) = @_;
    $pres ? $m->uwarn(@args) : $m->ustop(@args);
}

=head3 Funktion pstop

=cut

sub plok {
    my $m = shift;
    join '.', map { $m->dval($_) } @{$m->{funz}};
};
sub pstop {
    my $m = shift;
    my $msg = $m->msg(@_);
    my $lok = $m->plok;
    $msg = sprintf "inmid %s with\n\tprogvars %s\nand grupvars %s:\nstopping because\n%s\n", $lok, Dumper($m->{progvarp}), Dumper($m->{grupvars}), $msg;
  ERR: {
      last unless $!;
      my $err = decode 'UTF-8', $!;
      print $err, "\n" if $err;
    };
    printf STDERR 'internal error at %s:'."\n", $lok;
    die $msg;
}

sub pstop_or_warn {
    my ($m,$pres,@args) = @_;
    $pres ? $m->pwarn(@args) : $m->pstop(@args);
}

=head3 Funktion deprek_warn

=cut

sub deprek_warn {
    my ($m,$old,$nov) = @_;
    $m->pwarn('still invoking function %n under old name %o with args %A', n => $nov, o => $old, A => \@_);
    1;
}

=head3 Funktion uwarn

=cut

sub uwarn {
    my $m = shift;
    my $msg = $m->msg(@_);
    my $nodebug = $m->{nodebug};
    unless ($nodebug) {
	warn $msg;
    } else {
	warn $msg;
    };
}
=head3 Funktion pwarn

=cut

sub pwarn {
    my $m = shift;
    my $msg = sprintf 'inmid %s with progvars %s: %s'."\n", $m->plok, $m->dhac($m->{progvars}), $m->msg(@_);
    warn $msg;
};

=head3 $m->koniug('and', 1, 2, 3) => '1, 2 and 3'

    express a list with a conjuction like 'and', e.g. in debugging messages.

=cut

sub koniug {
    my ($m,$iug,@list) = @_;
  MINI: {
      last if $#list > 0;
      return '' if $#list == -1;
      return $list[0];
    };
    my $kls = join ', ', @list[0..$#list-1]; # '1, 2'
    join ' ', $kls, $iug, $list[$#list]; # '1, 2, and 3'
}

=head3 Funktion koniug_et

=cut

sub koniug_et {
    my $m = shift;
    $m->koniug('and', @_);
};

=head3 Function tcek_grup

    If a desired group (i.e. one that should have the writes on the files we write) is set, check whether we are in this group, die if we are not.

=cut

sub tcek_grup {
    my $m = shift;
    my $grup_nomo = $m->{konfig}->get('group');
    return 1;
    my $grup_numero = getgrnam $grup_nomo;
    $m->ustop("Gruppe %g ist hier nicht bekannt.\nBitte Konfiguration ändern oder mit der Option --group eine andere Gruppe angeben", g => $grup_nomo) unless $grup_numero;
    my @grup_numeroi = split /\s+/, $EFFECTIVE_GROUP_ID;
    return $grup_numero if $grup_numeroi[0] =~ m(\A$grup_numero\Z);
    $m->ustop('Du bist nicht Mitglied der Gruppe %g', g => $grup_nomo) unless grep { $_ eq $grup_numero } @grup_numeroi;
    return $grup_numero if setgid $grup_numero;
    my $erar_mesadj = $!;
    _utf8_on $erar_mesadj;
    $m->ustop(join("\n", $erar_mesadj, "Ich möchte mit der Gruppe $grup_nomo aufgerufen werden.", "Bitte sage 'sg $grup_nomo' und versuche es nochmal!"));
};

=head2 Program Time

=head3 Function timestamp

    Return a timestamp that can be used for naming of temporary files.

    As arg1, a time template as used by POSIX::strftime can be given.

=cut

our $timestamp_fmt = '%Y%m%d%H%M%S';
sub set_timestamp {
    my ($m,$fmtarg) = @_;
    my $timestamp = $m->{progvars}->{timestamp};
    return $timestamp if $timestamp and not $fmtarg;
    my $fmt = $fmtarg || $timestamp_fmt;
    my @progtime = @{$m->{progvars}->{progtime}};
    unless (@progtime) {
	@progtime = localtime time;
	$m->{progvars}->{progtime} = \@progtime;
    };
    $timestamp = strftime $fmt, @progtime;
    $m->{progvars}->{timestamp} = $timestamp unless $fmtarg;
    return $timestamp;
};
sub timestamp { 
    my $m = shift;
    $m->dpuc(timestamp => \@_);
    my $fmtarg = shift;
    my $timestamp = $m->set_timestamp($fmtarg);
    $m->dpop1(timestamp => $timestamp);return $timestamp;
};

=head2 Various common program elements

=head3 Functions ref2list, ref2hash

    Return a list/hash for an object that may or may not be a reference to a list/hash.

=cut

sub reftypp ($$$) {
    my ($m,$val,$typ) = @_;
    return $val if $typ eq ref $val;
};
sub ref2list ($$) {
    my $m = shift;
    my $v = $m->reftypp(shift, 'ARRAY');
    return unless $v;
    return @$v;
};
sub elm2list ($$) {
    my ($m,$v) = @_;
    return $v unless 'ARRAY' eq ref $v;
    return @$v;
};
sub ref2hash ($$) {
    my $m = shift;
    my $v = $m->reftypp(shift, 'HASH');
    return unless $v;
    return %$v;
};

=head3 Function enclose

    Internal subroutine.

    An elegant way of expressing a frequent text structure in programming language: 

    enclose '( ', ' )', 'abc' ---> '( abc )'

=cut

sub enclose {
    my $m = shift;
    my $left = shift;
    my $right = shift;
    my $mid = shift;
    my $res = $left . $mid . $right;
    return $res;
};

=head3 function CamelCase

    synopsis

    $CamelDok = $m->CamelCase($dok);

replace abc_def_en with AbcDefEn.
    A subroutine used in template transformation operations.
    Argument 1 is an object property, e.g. 'dok'

=cut

sub CamelCase ($$) {
    my $m = shift;
    my $v = shift;
    my @s = split m([^a-zA-Z0-9]+), $v;
    @s = map { ucfirst $_ } @s;
    $v = join '', @s;
    return $v;
};

=head2 File IO

    A thin wrapper around IO::File. 

    Open files for reading/writing in utf-8 mode, store associated info in $m->{f}->{$name} which points to a hash with keys 'name', 'h', and 'mode'.

=head3 Function open_fayl

    Open a file fpn arg1 under a shorthand name nom arg2 in some mode arg3 which is passed to IO::File.

    Options arg4ff include binmode (default C<:utf8>) and lock (default C<0>).
    
    return handle that is also stored in $m->{f}->{$nom}->{h};

=cut

# open_onim1
sub open_fayl {
    my ($m,$prog,$fpn,$nom,$mode,%opts) = shift->ppuc('open_fayl',@_);
    $m->pstop("I need an external file name argument") if !$fpn;
    $m->pstop("I need an internal file handle argument") if !$nom;
    $m->pstop('can not reopen file handle %h for file %f, in use for %n', h => $nom, f => $fpn, n => $m->{f}->{$nom}->{n}) if $m->{f}->{$nom}->{h};
    my @fnomz = grep {$_} map { $m->{f}->{$_}->{h} && $m->{f}->{$_}->{n} } keys %{$m->{f}};
    $m->pstop('file %f is among open files %F', f => $fpn, F => \@fnomz) if grep { $_ eq $fpn } @fnomz; 
    $m->pstop('I need a file mode argument such as r or w') if !$mode;
  MKDIR: {
      last unless $opts{mkdir};
      my ($vol, $dir, $fol) = splitpath $fpn;
      eval { mkpath $dir };
      $m->ustop('unable to mkpath %d for %f: %m', d => $dir, f => $fol, m => $@) if $@;
    };
    $opts{binmode} ||= ':utf8';
    my $h = new IO::File $fpn, $mode or $m->ustop('can not open file %f in mode %m dir %d', f => $fpn, m => $mode, d => &getcwd);
    $h->binmode($opts{binmode}) or $m->pstop('can not treat %f as coded in %b', f => $fpn, b => $opts{binmode});
    $m->dgot(fnom => $fpn, name => $nom, mode => $mode, binmode => $opts{binmode});
    $m->{f}->{$nom} = { 'n' => $fpn, 'h' => $h, 'm' => $mode };
    $m->lokk_fayl($nom) if $opts{lock};
    $m->dpop1($prog,$h);
};

=head3 Function lokk_fayl

    lock the file identified by arg 1

=cut

sub lokk_fayl {
    my $m = shift;
    $m->dpuc('lokk_fayl', @_);
    my $nom = shift or $m->pstop('I need a file identifier argument');
    $m->pstop('can not lock unknown file handle %h', h => $nom) unless $m->{f}->{$nom}->{h};
    lock $m->{f}->{$nom}->{h}, LOCK_EX or $m->ustop('failed to lock file handle %h', h => $nom);
    $m->{f}->{$nom}->{l} = 1;
    $m->dpop1('lokk_fayl', 1);
    return 1;
}

=head3 Function anlokk_fayl

    unlock file

=cut

sub anlokk_fayl {
    my $m = shift;
    $m->dpuc('anlokk_fayl', @_);
    my $nom = shift or $m->pstop('I need a file identifier argument');
    do { $m->dpop;return } unless $m->{f}->{$nom}->{l};
    lock $m->{f}->{$nom}->{h}, LOCK_UN or $m->ustop('failed to unlock file handle %h', h => $nom);
    $m->{f}->{$nom}->{l} = 0;
    $m->dpop1('anlokk_fayl', 1);
    return 1;
}

=head3 Function anlokk_faylz

    Unlock all locked files or those in @args.
    This is included in kloz_faylz.

=cut

sub anlokk_faylz {
    my $m = shift;
    my @args = @_ ? @_ : keys %{$m->{f}};
    $m->anlokk_fayl($_) for @args;
    return 1
};

=head3 Function fh

=cut

# fayl_i2h
sub fh {
    my $m = shift;
    my $fayl = shift or $m->pstop("fh needs file name argument");
    return $m->{f}->{$fayl}->{h};
};

=head3 Function fn

=cut 

# fayl_i2n
sub fn {
    my $m = shift;
    my $fayl = shift or $m->pstop("fn needs file name argument");
    return $m->{f}->{$fayl}->{n};
};


=head3 Function getlini

    Read a line from opened file $_[1] to $_, like perl readline, trailing
    "\n" needed by while loops to determine whether we are at the end of
    the file.

=cut

# getlin_i2s
sub getlini ($$) {
    my $m = shift;
    my $fayl = shift;
    $_ = readline $m->{f}->{$fayl}->{h};
    # ($_) = m(\A(.*)[\s\n\r]*\Z);
    return $_;
};

=head3 Function xchomp

    extended chomp: remove any blanks, carriage returns or line feeds from
    the end of $_, return the result so as to allow skipping empty lines
    by statements like

    $m->xchomp or next;

=cut

sub xchomp ($) {
    my $m = shift;
    $m->dpuc('xchomp');
    my $s = shift || '';
    $_ = $s || $_ || '';
    s([\s\n\r]+\Z)();
    $m->dpop1('xchomp', $_);
    return $_;
};

=head3 Function print_tu_fayl

    Print to file using handle from open_fayl and arguments from msg.
    No linefeed is added.

=cut

# print_fs1
sub print_tu_fayl {
    my ($m,$prog,$nom,@msgargs) = shift->ppuc('print_tu_fayl',@_);
    my $msg = $m->msg(@msgargs);
  NO_H: {
      last if $m->{f}->{$nom}->{h};
      $m->pstop('cant print messsage to file %f', f => $m->fn($nom));
    };
    $m->{f}->{$nom}->{h}->print($msg);
    $m->dpop1($prog,1);
};

=head3 $m->fayl_get_lst($fi,$trimp,$rxs,$join,$nmax,$nlmax) => @lst

	fayl_get_str(...) => $str

Read complete contents of a file into a string.
Do nothing and return undef if filename argument is void.
Strip comments if $trimp.
Read only matching lines if $rxs.
Use $join (either coderef or a sep argument for join e.g. '|' or "\t"), to package matched substrings into a list representation
Collect only $nmmax matches.
Search only the first $nlmax lines.  	

=cut

sub fayl_get_lst {
    my ($m,$prog,$fi,$trimp,$rxs,$join,$nmmax,$nlmax) = shift->ppuc('fayl_get_lst',@_);
    return $m->dpop($prog) unless $fi and -r $fi and -f $fi;
    my $rx = $rxs ? qr($rxs) : undef; # optional regex to make it work like the grep shell utility
    my $fun = 'CODE' eq ref $join ? $join : sub { join $join, @_ };
    $nmmax ||= 0; # collect only this many
    $nlmax ||= 0; # read only this many lines 
    my $hi = new IO::File $fi, 'r' or $m->ustop(openfayl => { f => $fi, d => &getcwd() });
    $hi->binmode(':utf8') or $m->ustop(binmode => f => $fi);
    my @lins = ();
    my $nl = 0; # number of lines
    my $nm = 0; # number of matching lines
    local $_;
  LINI: while ($_ = readline $hi) {
	chomp;
	_utf8_on $_;
	$nl++;
	last if $nlmax and $nlmax < $nl;
	if ($trimp) {
	    $_ = $m->rtrim_nokoment($_);
	    next LINI unless $_;
	};
      REGEX: {
	  last unless $rx;
	  local @_ = m($rx);
	  last if @_ and !$fun;
	  next LINI unless @_;
	  $_ = &$fun(@_);
	  next LINI unless $_;
	};
	$nm++;
	last if $nmmax and $nmmax < $nm;
	push @lins, $_;
  };
    $hi->close or $m->ustop(klozfayl => f => $fi);
    $m->dpop($prog,@lins);
};

sub fayl_get_str {
    my ($m,$prog,@args) = shift->ppuc('fayl_get_str',@_);
    my @lst = $m->fayl_get_lst(@args);
    my $str = join "\n", @lst;
    $m->dpop1($prog,$str);
};

=head4 $fayl = $m->fayl_put_str($fayl,$str)

    Write a simple string to a file, only once, no bookkeeping of handles etc is done.
    Do nothing and return undef if filename argument is void.

=cut

sub fayl_put_str {
    my ($m,$prog,$fayl,$str) = shift->ppuc('fayl_put_str',@_);
    return $m->dpop($prog) unless length $fayl;
    $m->uwarn('nothing to write to file %f in %p', f => $fayl, p => $prog) unless length $str;
    my $ho = new IO::File $fayl, 'w' or $m->ustop(fputs => f => $fayl);
    $ho->binmode(':utf8') or $m->ustop(binmode => f => $fayl);
    $ho->print($str);
    $ho->close or $m->ustop(klozfayl => f => $fayl);
    $m->dpop1($prog,$fayl);
};

=head4 $fayl = $m->fayl_put_lst($fayl,@lst): Write a raw list of lines to a file
$fayl :: file
@lst :: list of lines

This can be used to write out either lists or hashes.
It works fine as long as the lines do not contain linebreaks.
For real-life text hashes we need the extended version fayl_put_tups.

=cut

sub fayl_put_lst {
    my ($m,$prog,$fayl,@lst) = shift->ppuc('fayl_put_lst',@_);
    my $str = join "\n", @lst;
    $m->fayl_put_str($fayl,$str);
    $m->dpop1($prog,$fayl);
};

=head4 $fayl = $m->fayl_put_tups($fayl,\%tups,%opts(k2v)): write tuples to a file

Make sure that a hash that we write to a file can be read from it again, avoiding errors that arise from linebreaks in the variable values.
Check sanity of keys and values.
We assume that the variables do not contain any 
 ("\r") and if they do it is OK to convert it into linefeed (^J, "\n").
Allow a function C<$v = $k2v($k,$tupsr?)> to be passed that can change what is read from the tuples, e.g. substituting references (A2E::File objects) with their values as in file:/adv/perl/A2E/Lihs.pm::k2v with k2v => sub { $lihs->getstr($_[0]) }
Support A2E::Lihs::put_tups and A2E::MLDK::dkmk.
Rely on A2E::Prog::fayl_put_lst

=cut

sub fayl_put_tups {
    my ($m,$prog,$fayl,$tupsr,%opts) = shift->ppuc('fayl_put_tups',@_);my %tups = $m->ref2hash($tupsr);$m->pstop(fnotups => f => $fayl) if !%tups;
    my $k2v = $opts{k2v} || sub { my ($k,$hacr) = @_;$hacr ||= $tupsr;$hacr->{$k} };
    die 'k2v option of put_tups must be of type CODE' unless 'CODE' eq ref $k2v;
    foreach my $k (keys %tups) {
	$m->pstop('invalid key %k in tups for file %f', f => $fayl, k => $k) unless $k =~ m(\A\w+\Z);
	$_ = &$k2v($k,$tupsr);
	next if !length $_;
	$m->pstop('nonstring value %v in tups for file %f', f => $fayl, v => $_) if ref $_;
      LNBRK: { # linebreaks need to be encoded in a file where each item is on 1 line; decode in 'get_tups'
	  s(\n)(
)g };
	$tups{$k} = $_ };
    $m->fayl_put_lst($fayl,%tups);
    $m->dpop1($prog,$fayl);
};

=head \%stor = $m->fayl_get_tups($fayl,%opts(stor,v2k,kk)): read tuples from a file

Read tuples written by fayl_put_tups, decoding its safeguard-coding (related to linebreak).
Check sanity of keys and values.
Allow inverse function C<$v2k($k,$v)> to be passed.
Alternatively offer a choice of functions in the form of the index keywords of %kk2put
	kk => 'ad': store only when value to be stored is defined
	kk => 'ae': store only when value to be stored exists i.e. is true
	kk => 'bd': store only when not defined in storage
	kk => 'be': store only when not existant i.e. not true in storage
	kk => 'ff': store no matter what
Default is 'bd' (most cautious option).
%kk2put functions are also used in function hashpush.

The v2k/kk function operates on storage \%stor which can also be passed as $opts{stor}.

Supports A2E::Lihs::get_tups and A2E::MLDK::dk_mlk
Depends on fayl_get_lst.

=cut
our %kk2put = ();
$kk2put{bd} = sub { my ($stor,$k,$v) = @_;$stor->{$k} //= $v };
$kk2put{be} = sub { my ($stor,$k,$v) = @_;$stor->{$k} ||= $v };
$kk2put{ad} = sub { my ($stor,$k,$v) = @_;$stor->{$k} = $v if defined $v };
$kk2put{ae} = sub { my ($stor,$k,$v) = @_;$stor->{$k} = $v if $v };
$kk2put{ff} = sub { my ($stor,$k,$v) = @_;$stor->{$k} = $v };
sub fayl_get_tups {
    my ($m,$prog,$fayl,%opts) = shift->ppuc('fayl_get_tups',@_);
    my %tups = $m->fayl_get_lst($fayl);
    my $v2k = $opts{v2k};
    my $stor = $opts{stor} || {}; # ultimate destination of storage, could be external, e.g. $m->{progvars} or $m->{sarb}->{lihs}->{vrbs}
    my $kk = $opts{kk} || 'bd';
    $v2k ||= $kk2put{$kk}; # storing process
    $m->pstop('bad put function %f', f => $v2k) if 'CODE' ne ref $v2k;
    foreach my $k (keys %tups) {
	$m->pstop('invalid key %k in tups of file %f', f => $fayl, k => $k) unless $k =~ m(\A\w+\Z) ; 
	local $_ = $tups{$k};
      LNBRK: { # undo encoding of C<fayl_put_tups::LNBRK>, memorize variables like LF with proper linebreak; 
	  s(
)(\n)g }; 
	&$v2k($stor,$k,$_) };
    $m->dpop1($prog,$stor);
};

=head3 Function kloz_fayl

    Close file handle, remove any lock that may be on it.
    Undef the handle but retain the rest of the info related to the file, so that it can be used for reporting.

=cut

sub kloz_fayl {
    my ($m,$prog,$nom) = shift->ppuc('kloz_fayl',@_);
    return $m->dpop($prog) unless $m->{f}->{$nom}->{h};
    $m->anlokk_fayl($nom);
    eval { $m->{f}->{$nom}->{h}->close };
    $m->pwarn('could not properly close %n file %f: %m', n => $nom, f => $m->{f}->{$nom}->{n}, m => $@) if $@;
    my $h = delete $m->{f}->{$nom}->{h};
    $m->dpop1($prog,$h);
};

=head3 Function kloz_faylz

    Close all files given on the command line or all that are still open.

=cut

# kloz_i9
sub kloz_faylz {
    my ($m,$prog,@faylz) = shift->ppuc('kloz_faylz',@_);
    @faylz = keys %{$m->{f}} unless @faylz;
    $m->kloz_fayl($_) for @faylz;
    $m->dpop1($prog,1);
};

=head2 Accessing the Configuration 

    wrappers around AppConfig, making us depend less on the underlying module, so that we are freeer to migrate e.g. to something like YAML::AppConfig.

=head3 Functions konfget, konfset

=cut

sub konfget {
    my $m = shift;
    $m->dpuc('konfget', @_);
    my $var = shift;
    my $val = $m->{konfig}->get($var);
    $m->dpop1('konfget', $val);
    return $val;
};

sub konfset {
    my $m = shift;
    $m->dpuc('konfset', @_);
    my $var = shift;
    my $val = shift;
    $m->{konfig}->set($var, $val);
    $m->dpop1('konfset', $val);
    return $val;
};

=head3 Function konfsetif

    Set if proposed value is not null.

=cut

sub konfsetif {
    my $m = shift;
    $m->dpuc('konfsetif',@_);
    my $var = shift;
    my $val = shift;
  NULL: {
      $m->dpuc('null');
      do { $m->dpop;last } if $val;
      $m->dpop('konfsetif');
      return;
    };
    $m->{konfig}->set($var, $val);
    $m->dpop1('konfsetif', $val);
    return $val;
};

=head3 Function konfinit

    Set if nothing there yet

=cut

sub konfinit {
    my $m = shift;
    $m->dpuc('konfinit', @_);
    my $var = shift;
    my $val = shift;
    my $old = $m->{konfig}->get($var);
  OLD: {
      $m->dpuc('old');
      do { $m->dpop;last } unless $old;
      $m->dpop1('konfinit', $old);
      return $old;
    };
    $m->{konfig}->set($var, $val);
    $m->dpop1('konfinit', $val);
    return $val;
};

=head2 Invocation of external programs

=head3 Function exek

    execute a system command with arguments and return 1 if execution was successful, 0 otherwise.
    Store the exit value in $EERR and the error message in $EMSG;

=cut

our $EERR = 0;
our $EMSG = '';
sub exek { 
    my $m = shift;
    my @kmd = @_;
    system @kmd;
    my $err = $?;
    $EERR = $err >> 8;
    $EMSG = $!;
    my $ok = ($EERR == 0) ? 1 : 0;
    return $ok if $ok;
    my $signal = $err & 127;
    my $k = join ' ', @kmd;
    if ($EERR == -1) {
	$m->pwarn(badexek => k => $k, s => $EMSG);
    } elsif ($signal) {
	my $msg = $EERR & 128 ? 'diedwcd' : 'diedwocd';
	$m->pwarn($msg => k => $k, s => $signal);
    }
    else {
	$m->pwarn(pidexit => e => $EERR);
    };
    return $ok;
};

# mv sub exek_open3 al file:/adv/perl/A2E/Open.pm::exek_open3

=head3 Function out2sav

    Redirect stdout to the file supplied via the argument or via -o option, unless it is already redirected.

    Append '.tmp' to the file name, so that the intended file will only
    exist after everything has been done successfuly and stdout has been
    restored with sav2out.

    A newer variant with Scalar::Util is used in file:/sig/adv/mnt/hwcgW7/Daikufile::findfols
    In this case the &Scalar::Util::openhandle function helped to avoid misinterpretation of bareword variables like STDOUT

=cut

sub out2sav {
    my ($m,$prog) = shift->ppuc('out2sav',@_);
    return $m->wpop($prog,[p => 'savout_fayl already exists']) if $m->{savout_fayl};
    return $m->wpop($prog,[p =>  'savout already exists']) if $m->{savout};
    my $fo = shift || $m->{konfig}->outfayl || '';
    return $m->dpop($prog) unless $fo;
    open $m->{savout}, '>&STDOUT' or die 'failed to save stdout';
    binmode $m->{savout}, ':utf8';
    my $fo_tmp = join '.', $fo, 'tmp';
    open STDOUT, '>', $fo_tmp or $m->pstop('failed to redirect STDOUT to %f', f => $fo_tmp);
    binmode STDOUT, ':utf8';
    $m->{savout_fayl} = $fo;
    $m->{savout_ftmp} = $fo_tmp;
    $m->dpop1('out2sav', 1);
};

=head3 Function sav2out

    Restore stdout from copy saved by out2sav.
    Needed only in exceptional cases where stdout must be freed.

=cut

sub sav2out {
    my ($m,$prog) = shift->ppuc('sav2out',@_);
    return $m->dpop($prog) unless $m->{savout};
    close STDOUT;
    open STDOUT, '>&=', $m->{savout} or die 'failed to restore stdout';
    my $dir = &getcwd();
    $m->pstop('output file %f was created but does not exist in %d, did we change directory?', f => $m->{savout_ftmp}, d => $dir) unless -e $m->{savout_ftmp};
    do { $m->pwarn('removing empty output file %f', f => $m->{savout_ftmp});unlink $m->{savout_ftmp} } unless -s $m->{savout_ftmp};
    do { $m->pwarn('no output file to rename');$m->dpop;return } unless -s $m->{savout_ftmp};
    rename $m->{savout_ftmp}, $m->{savout_fayl} or $m->pstop('failed to rename %t to %y', t => $m->{savout_ftmp}, y => $m->{savout_fayl});
    $m->dpop1('sav2out',1);
};

=head3 Funktion chwd

    In absolutes oder relatives Verzeichnis $1 wechseln und dessen absolute Version in $m->pvar('cwd') speichern.
    Mit Fehler beenden falls es nicht klappt.
    Optionales $2 ist gegenwärtiges Verzeichnis (in idealer Benennung).
    rueckgabewert 1 ist neues, 2 frueheres Verzeichnis.

=cut

sub get_cwd {
    my $m = shift;
    my $cwd = shift || $m->{progvars}->{cwd} || &getcwd();
    $m->{progvars}->{cwd} = $cwd;$m->{cwd} = $cwd;
    $cwd;
}
sub absfayl { # m->absfayl('../../../ok.txt','/sig/oas/19/06') => '/sig/ok.txt'
    my ($m,$prog,$fyl,$pwd) = shift->ppuc('absfayl',@_);
    $pwd ||= $m->get_cwd; # replaced in A2E::Dokfs
    my @fols = splitdir $fyl; # leaflist of $fyl
    return $m->dpop1($prog,$fyl) unless $fols[0]; # return already absolute $fyl
    my @pwfs = splitdir $pwd; # leaflist (@fols) of $pwd
    foreach my $fol (@fols) {
	next if '.' eq $fol; # in '././dum.txt', '.' leaves are redundant sugar
	do { $m->ustop('not enough directories to pop for generating absolute file from %f at %d', f => $fyl, d => $pwd) unless @pwfs;pop @pwfs;next } if '..' eq $fol;
	push @pwfs, $fol;
    };
    local $_ = catfile @pwfs;
    $m->dpop1($prog,$_);
};

sub inod { my ($m,$f) = @_;my @stat = stat $f;$stat[1] };
sub chwd {
    my ($m,$prog,$dir,$lwd,%opts) = shift->ppuc('chwd',@_);
    my $cwd = $m->{progvars}->{cwd};
    my $inod0 = $m->inod('.');
    $cwd = &getcwd unless $cwd and $m->inod($cwd) == $inod0;
    $m->pstop('cwd inode differs from dotdir inode') unless $m->inod($cwd) == $inod0;
    $dir ||= $cwd;
    $lwd ||= $opts{lwd} || $cwd;
    $lwd = $cwd unless $m->inod($lwd) == $inod0;
    my $ok = chdir $dir;
    do { $m->msg("failed to chdir from $lwd to $dir: $@");if ($opts{msg}) { $m->pwarn($opts{msg}, %opts) };if ($opts{force}) { return $m->dpop($prog,$lwd) } else { die "KO chdir $dir" } } unless $ok;
    $dir = $m->absfayl($dir, $lwd);
    $m->{progvars}->{cwd} = $dir;
    $m->dpop($prog,$dir,$lwd);
}

=head3 Funktion cdup

    Step into parent directory, return the child, i.e. the leaf relative to the parent stem into which the program could step to to get back where it came from.

=cut

sub cdup {
    my $m = shift;
    $m->dpuc('cdup', @_);
    my $cwd = $m->pvar('cwd') || &getcwd();
    my @dirs = splitdir $cwd;
    do { $m->dpop;return } unless @dirs;
    my $leaf = pop @dirs;
    my $stem = catdir @dirs;
    my ($dir, $lwd) = $m->chwd($stem);
    $m->dpop('cdup', $leaf, $dir, $lwd);
    return $leaf;
}

=head3 Function mv_non_dir

Move an object out of the way.

=cut

sub mv_non_dir {
    my $m = shift;
    $m->dpuc('mv_non_dir', @_);
    my $fpn = shift;
    do { $m->dpop;return } unless -e $fpn;
    $m->set_timestamp;
    my $bak = join '_', $fpn, 'bak' . $m->{timestamp};
    eval { rename $fpn, $bak };
    $m->pstop('kann %f nicht aus dem Weg räumen: %m', f => $fpn, m => $@) if $@;
    $m->dpop1('mv_non_dir', 1);
};

=head2 Constructing the Object and reading the configuration

=head3 Function new

Simple constructor that creates an object but does not set its configuration options.  Deprecated, use new_ready instead.

=cut

sub new {
    die "still using deprecated constructor new for class A2E::Prog, please migrate to new_ready immediately so this code can be removed";
    my $class = shift;
    my $m = {};
    bless $m, $class;
    $m->SUPER::new;
    return $m;
};

=head3 Function set_debug_system

subroutine of new_ready: set up debugging-related switches based on environment variables DEBUG_BLOKS, DEBUG_PUCPOP_BLOKS, NON_DEBUG_BLOKS and DEBUG

=cut

sub set_debug_system {
    my $m = shift;
    # $m->setpvar(nonlangp => 1);
    my $DEBUG = $ENV{DEBUG};
    $m->{progvars}->{DEBUG} = $DEBUG;
    $m->{progvars}->{debug} = $DEBUG;
    $m->{debug} = $DEBUG;
    $m->{DEBUG} = $DEBUG;
    $m->{argz} = [];
    $m->{debug_bloks} = { };
    my $bloks = $ENV{DEBUG_BLOKS} || '';
    $m->{debugger} = 1 if $ENV{PERL_DEBUGGER};
    $m->{debug_bloks}->{$_} = 1 for split m([\s\,]+), $bloks;
    $bloks = $ENV{NON_DEBUG_BLOKS} || '';
    $m->{debug_bloks}->{$_} = 0 for split m([\s\,]+), $bloks;
    $bloks = $ENV{DEBUG_PUCPOP_BLOKS} || '';
    $m->{debug_pucpop_bloks}->{$_} = 1 for split m([\s\,]+), $bloks;
    # $m->setpvar(nodebug => !$m->{progvars}->{debug});
    return unless $m->{progvars}->{nodebug};
    foreach my $blok (keys %{$m->{debug_bloks}}) {
	next unless $m->{debug_bloks}->{$blok};
	$m->{progvars}->{nodebug} = 0;
	last;
    }
    1;
};

=head3 Function new_ready

Make an object ready for use. 

Alternative to new and setup, does all the initiating work, offering various parts, especially C<defvars> and C<lastkonfig>, for overloading.  

Takes pre-configuration options hash as arguments, with the following options supported in this module:

=over

=item debug

Debug right from the start, even before the configuration mechanism has been set up.

=item verbose

Be verbose right from the start

=item progver

version number of the invoking application.  If not provided, the version number is the number of the first prerequisite that has one.

See the --progver and --dokprogver options below.

=item noexec

Do not treat me as an interactively invoked executable, do not parse the commandline.

=back

ATTENTION: you are free to screw the program by setting object properties that you should not meddle with.  No checks are performed, because these properties are set only by the programmer, not the end user.  Examples of such use are

=over

=item A2E::SArb(3)

makes heavy use of this in setting properties $m->{funs} and $m->{vars} that are not supported in this module.

=item A2E::Mktdir(5), mktdir4, mktdir5

uses tratoks to determine the command line.  This is clearly a programming option rather than a user interface option.

=back

=cut

our %srcfpfxs = (_ => '_', '%' => '%', '@' => '@', ':' => ':', '__' => '__');
sub check_srcfpfx {
    my ($m,$prog,$pfx0) = shift->ppuc('check_srcfpfx',@_);
    my $pfx = $srcfpfxs{$pfx0};
    $m->ustop('invalid srcfpfx %p', p => $pfx0) unless $pfx;
    $m->setpvar(srcfpfx => $pfx);
    $m->{hsprintf_repl}->{p} = $pfx;
    $m->dpop1($prog,$pfx);
}
sub set_srcfpfx {
    my ($m,$prog,%opts) = shift->ppuc('set_srcfpfx',@_);
    my $dir = $opts{dir} || &getcwd();
    my $pfx = $m->get_srcfpfx(dir => $dir);
    do { $m->uwarn('no srcfpfx found in dir %D, using default', D => $dir);$pfx = '_' } unless $pfx;
    $m->setpvar(srcfpfx => $pfx);
    $m->dpop1($prog,$pfx);
};

sub new_ready {
    my $prog = 'new_ready';
    my $class = shift;
    my %optvalz = @_;
    my $m = {};
    bless $m, $class;
    $m->set_debug_system;
    my $DEBUG = $m->{progvars}->{DEBUG};
    my $nodebug = $m->{progvars}->{nodebug};
    if ($ENV{DEBUG} || $ENV{DEBUG_BLOKS}) {
	print STDERR join('; ', "DEBUG: $DEBUG", "nodebug: $nodebug",
			  join(' ', 'debug bloks:', grep { $m->{debug_bloks}->{$_} } keys %{$m->{debug_bloks}}),
			  join(' ', 'non debug bloks:', grep { !$m->{debug_bloks}->{$_} } keys %{$m->{debug_bloks}}) ) . "\n";
    };
    $m->dpuc($prog);
    $m->{ready} = {};
    # $m->{progvars} ||= {};
    $m->{f} = {};
    $m->{cache} ||= {};
    $m->{progvars}->{verbose} = 0;
    $m->{qrevars} = {};
    $m->{supobj_vars} = [ 'cwd' ];
    my $set_optvalz = sub {
	foreach my $nom (keys %optvalz) {
	    my $val = $optvalz{$nom};
	    if (ref $val) {
		$m->{$nom} = $val;
	    } else {
		$m->{progvars}->{$nom} = $val;
	    };
	};
    };
    &$set_optvalz;
    $m->defvars;
    $m->set_srcfpfx;
    $m->pars_konfaylz;
    $m->konfig_optvalz(%optvalz);
    $m->konfig_getopt unless $optvalz{noexec};
    $m->lastkonfig;
    $m->postkonfig;
    $m->dpop1($prog,$m);
};

=head3 Function ready

Register in $m->{ready} whether a routine has been loaded already; this is applicable to %MSGS and other info that is stored in the package rather than in each object.
Cf usage example in prog_msgdefs
=cut

sub ready { 
    my $m = shift;
    my ($nom,$val) = @_;
    if ($val) { $m->{ready}->{$nom} = $val } else { $val = $m->{ready}->{$nom} };
    return $val;
};

=head3 Function lastkonfig

    Routines at the end of the init process, to be overwritten by inheriting libraries.

    This comes in after the commandline configuration options have been read but before any commandline arguments have been read and before any options have been transferred into object variables.  Parsing of additional configuration files, to be determined by user input, can be done here.

    The arguments are the commandline arguments that have remained after GetOptions.

    Used e.g. in db2auth(1)

=cut

sub lastkonfig {
    my ($m,$prog) = shift->ppuc('lastkonfig',@_);
    $m->kopi_konfigvars;
    $m->dpop($prog);
};

=head3 Function konfayl

    Add a configuration file.

    An element that every application\'s defvars routine needs.

    Shield applications from the File::Spec package.

=cut

sub konfayl ($@) {
    my ($m,$prog,@fols) = shift->ppuc('konfayl',@_);
    my $konfayl = catfile $m->{konfig}->etcdir, @fols;
    push @{$m->{konfaylz}}, $konfayl;
    $m->dpop1('konfayl', 1);
};

=head3 Function konfig_define

    do {konfig}->define with somewhat simpler parameters, e.g.

    $m->konfig_define([ 'tmplang', 'tml', 'T' ], 's', 'multemp');

would map to

    $m->{konfig}->define('tmplang|tml|T=s', { DEFAULT => 'multemp' });

Return the variable name, i.e. in this case 'tmplang'. 

Arguments:

    $nom:	variable name(s): item or listref
    $typ:	variable type: '!' for boolean, 'i' integer, 's' string, 's@' array, 's%' hash    
    $opts:	options: item or hashref; hashref with DEFAULT key would be same as options argument of konfig->define, anything else would be a value for the DEFAULT key
=cut

sub integerp {
    local $_ = shift;
    return unless m(\A\d+\Z);
    $_;
};
sub boolp {
    local $_ = shift;
    return 1 if m(\A(?:1|true|OK)\Z);
    return 0 if m(\A(?:0|00|false|KO)\Z);
};
sub stringp { 
    local $_ = shift;
    return if ref $_;
    $_;
};
sub hashp { 
    local $_ = shift;
    return unless 'HASH' eq ref $_;
    $_;
};
sub listp {
    local $_ = shift;
    return unless 'ARRAY' eq ref $_;
    $_;
};
sub konfig_define {
    my ($m,$prog,$nom,$typ,$val) = shift->ppuc('konfig_define',@_);
    my $spec = $nom;
    my $opts = $val;
  SPEC: { # [ 'chdir','C' ]
      last unless ref $nom;
      $m->pstop('konfig_define variable name must be either scalar or listref') unless 'ARRAY' eq ref $nom;
      my @noms = @$nom;
      $spec = join '|', @noms;
      $nom = shift @noms;
    };
    my $ttyp = '';
  TYP: {
      do { $typ = '!';$ttyp = '!';$m->{konfig_validp}->{$nom} = \&boolp;$val = $val ? 1 : 0;last } if grep { $typ eq $_ } '!', 'b';
      do { $typ = 's';$ttyp = '$';$m->{konfig_validp}->{$nom} = \&stringp;last } if !$typ or grep { $typ eq $_ } '$', 's';
      do { $typ = 's';$ttyp = '$';$m->{konfig_validp}->{$nom} = \&integerp;last } if grep { $typ eq $_ } 'i', '1';
      do { $typ = 's@';$ttyp = '@';$m->{konfig_validp}->{$nom} = \&listp;last } if grep { $typ eq $_ } '@', 'l', 's@';
      do { $typ = 's%';$ttyp = '%';$m->{konfig_validp}->{$nom} = \&hashp;last } if grep { $typ eq $_ } '%', 'h', 's%';
    };
  OPTS: {
      last unless defined $val;
    REF: {
	last unless 'HASH' eq ref $opts;
      DEFAULT: {
	  last unless $opts->{DEFAULT};
	  $opts->{DEFAULT} = $val;
	  last OPTS;
	};
      };
      $opts = { DEFAULT => $val };
    };
    $spec .= '=' unless $typ eq '!';
    $spec .= $typ;
    my @args = ();
    push @args, $spec;
    push @args, $opts if $opts;
    $m->{konfig}->define(@args);
    $m->dpop($prog,$nom,$ttyp,$val);
};

=head3 Function kopi_define, prog_define, cache_define, qre_define, kopivar_define

    like konfig_define, additionally store the variable to %{$m->{kopivars}}, so that it will be copied into C<%$m->{progvars}> within the the C<new_ready> initialisation routine (immediately before postkonfig).

    kopi_define is no longer different from cache_define, now instead prog_define is.
    Previously kopi_define used the $m object as a variable register, whereas prog_define uses $m->{progvars}.

=head3 Function 

like konfig_define, additionally store the variable in %{$m->{cachevars}}, so that it will be read by the program as potential internal variables, a reservoir from which the program sets its progvars according to its rules.
This happens in the middle of C<new_ready> via <lastkonfig> and C<kopi_konfigvars>.
Similar routines like A2E::Tmplvars::lihs_define can be inserted into the process by registering a registry in %{$m->{kopivars_hac2reg}}, which registry informs whether a givern variable belongs to this category.
Additionally a setting function can be registered in %{$m->{kopivars_hac2fun}}

=cut
sub kopi_define { my $m = shift;$m->pwarn('kopi_define is deprecated, decide here whether to use cache_define or prog_define ASAP!');return $m->cache_define(@_) };
sub kopivar_define {
    my ($m,$prog,$hacnom,$nom,$typ,$val) = shift->ppuc('kopivar_define',@_);
    my $hacreg = $m->{kopivars_hac2reg}->{$hacnom}; # hacreg => (?:cachevars|progvarp|qrevars|tmplvarp)
    $m->pstop('unknown configuration variable type %t', t => $hacnom) unless $hacreg;
    ($nom,$typ) = $m->konfig_define($nom,$typ,$val);
    $m->{$hacnom}->{$nom} = $val;
    $m->{$hacreg}->{$nom} = $typ;
    $m->dpop1($prog,$nom);
};

sub cache_define {
    my ($m,$prog,$nom,$typ,$val) = shift->ppuc('cache_define',@_);
    $nom = $m->kopivar_define('cache',@_);
    $m->dpop1($prog,$nom);
};

sub prog_define {
    my ($m,$prog,$nom,$typ,$val) = shift->ppuc('prog_define',@_);
    $nom = $m->kopivar_define('progvars',$nom,$typ,$val);
    $m->dpop1($prog,$nom);
};

=head3 Function qre_define

    like konfig_define, but additionally store the variable to %{$m->{qre}}, in the same way that C<cache_define> stores it to %{$m->{progvars}}, but in regex-compiled fashion.  The transfer of the values is done in new_ready right before invocation of postkonfig.

=cut

sub qre_define {
    my ($m,$prog,@args) = shift->ppuc('qre_define',@_);
    my $nom = $m->kopivar_define('qre',@args);
    $m->dpop1($prog,$nom);
};

=head3 Function kopi_konfigvars

    Copy a variable from $m->{konfig} to the internal location where it ultimately resides.

    Something similar is done independently on all variables in new_ready


=cut

sub kopi_konfigvars {
    my ($m,$prog,@args) = shift->ppuc('kopi_konfigvars',@_);
    my $fun0 = $m->{kopivars_hac2fun}->{_};
    while (my ($hac,$reg) = each %{$m->{kopivars_hac2reg}}) {
	my $fun = $m->{kopivars_hac2fun}->{$hac} || $fun0;
	foreach my $nom (keys %{$m->{$reg}}) {
	    my $typ = $m->{$reg}->{$nom}; # '$', '@', '%'
	    $m->pwarn('no typ defined for konfvar %n in %r', n => $nom, r => $reg) unless $typ;
	    my $val = $m->{konfig}->get($nom);
	    do { $m->decho('konfigvar %n is undefined, not transfering to %h', h => $hac, n => $nom, r => $reg);next } unless defined $val;
	    $m->vecho('transfering %n with %v to %h', n => $nom, v => $val, h => $hac);
	  VALIDP: {
	      my $validp = $m->{konfig_validp}->{$nom};
	      last unless $validp;
	      my $ref = ref $validp;
	      $m->pstop('validator of configuration variable %n is of invalid type %r', n => $nom, r => $ref) unless 'CODE' eq $ref;
	      my $res = &$validp($val);
	      do { $val = $res;last } if defined $res;
	      $m->ustop('invalid value %v for configuration variable %n of type %t', n => $nom, v => $val, t => $typ);
	    };
	    &$fun($hac,$nom,$val,$typ);
	};
    };
    $m->dpop1($prog,1);
};


=head3 Function define_prog_vars and the defvars mechanism

    At each application layer, a defvars routine should be defined which again invokes the corresponding routine of its ancestor layer at the beginning, see e.g. define_daba_vars in A2E::Daba(3), define_dokfs_vars in A2E::Dokfs(3), define_mktdir_vars in A2E::Mktdir(3) and define_mktdir5_vars in mktdir5(1).

    Thus each program will invoke 'new', 'defvars', 'setup' and 'process' in sequence.  The mktdir5(1) and dokfs_subdirs(1) can serve as models of minimal application programs.

    Each parent library, also in case of multiple inheritance, has its defvars.  Even a mix-in library like A2E::Lokvars(3) has its lokvars_defvars method, but no setup method.  Moreover, the mixin-type defvars routine does not invoke those of any parents.

=cut

sub prog_defvars {
    my ($m,$prog) = shift->qpuc('prog_defvars',@_) or return;
    $m->{kopivars_hac2reg} = { cache => 'cachevars', progvars => 'progvarp', qre => 'qrevars', lihs => 'tmplvarp' };
    $m->{kopivars_hac2fun} = { _ =>  sub { my ($hac,$nom,$val) = @_;$m->{$hac}->{$nom} = $val } }; # cf file:/adv/perl/A2E/Tmplvars.pm.tmpl::kopivars_hac2fun
    $m->{$_} = {} for values %{$m->{kopivars_hac2reg}};
    $m->{konfig} = new AppConfig { CASE => 1, GLOBAL => { EXPAND => EXPAND_ALL | EXPAND_WARN } };
    $m->konfig_define(dval_array_maxlen => i => 32);
    $m->{blanks} = [ 8, 9, 32, 12288 ];
    $m->hsprintf_qre2fun_defs;

=head1 OPTIONS

=over

=item --verbose

=cut

    $m->prog_define(['verbose', 'v'], '!', $ENV{VERBOSE} || 0);
    $m->prog_define(['debug'], '!', $ENV{DEBUG} ? 1 : 0);
#    $m->prog_define('novitas', 'i', $ENV{NOVITAS} || 0);
    $m->cache_define(user => s => $ENV{USER});
    $m->cache_define('sexp', '!', $ENV{SEXP} || 0) unless $novitas;

=item --version

=cut

    $m->konfig_define('version','!');

=item --manpage

=cut

    $m->konfig_define(['manpage','man'],'!');

=item --help

=cut

    $m->konfig_define(['help','usage','h'],'!');

=item --quiet

=cut

    $m->cache_define(['quiet','q'],'!');

=item --force

=cut

    $m->cache_define('force', '!');

=item --noexec

=cut

    $m->cache_define('noexec', '!');

=item --test

=cut

    $m->konfig_define('test','!');
    my $incdir = $ENV{PERL5LIB} || $INC[0];
    my $etcdir = catdir $incdir, 'etc';
    my $test_modus = $incdir =~ m(\A\/adv\/perl\b) and -d $etcdir;
    $m->prog_define(test_modus => b => $test_modus);
    my $cwd = getcwd;
    _utf8_on $_ for @ARGV;
    if ($test_modus) { $m->uwarn('%p test run with %e as etcdir in working directory %w with args %A', p => $0, e => $etcdir, w => $cwd, A => \@ARGV) } else { $etcdir = '/etc/opt/a2e' };
    $m->konfig_define('etcdir', '$', $etcdir);
    $m->konfig_define(['outfayl','outfile','out','o']);
    $m->konfig_define(['infaylz','infiles','in'],'@');
=item --srcfpfx

Prefix of directory-specific source and configuration text files such '_' as found in '_prog', '_dokfs', '_langs' and '_lng.myprog.en.txt'

=cut
    $m->konfig_define('srcfpfx', '$', '');

=item --konfbas dokfs

Name of library that is used for the minimal configuration file that must be present in each directory so that the source text file name prefix srcfpfx '@' or '_' can be derived from it, e.g. 'dokfs'.

=cut
    $m->konfig_define('konfbas', 's', 'dokfs');
=item --pinghost 'a2e.de'
=cut
    $m->konfig_define('pinghost','s', 'a2e.de');
=item --konfsfx konf

Suffix used in configuration file names 
invoked in file name templates by '%s'
fillable with $m->hsprintf('%b.%s', b => 'dokfs', 's' => 'konf')

=cut
    $m->konfig_define('konfsfx','s','konf');
    $m->konfig_define('group','s');
    $m->cache_define('umask', 's', '002');
    $m->cache_define('host', 's', $ENV{HOSTNAME} || &hostname);

=item --locale|-l C<de_DE.utf-8>

Landessprache festlegen.

Die Landessprache muss vom Betriebssystem unterstützt sein und es sollten Übersetzungen für sie vorliegen.

In Frage kommen derzeit de_DE.utf-8, en_US.utf-8, zh_TW.utf-8, ja_JP.utf-8 und C.

Was ein Unix-Betriebssystem unterstützt, kann man mit dem Befehl C<locale -a> herausfinden.

Voreingestellt ist der Inhalt der Umgebungsvariable LANG.

In der internen Variable $m->{lang} wird der erste Teil von --locale (also 'de', 'en', 'zh' bzw 'ja') abgespeichert.

S. auch C<perldoc perllocale>.

=cut
   
    $m->cache_define('locale', '$', $ENV{LANG});

=item --lang zh, -l de 

Limit operation to this language.  Can be passed via environment variable MLANG

Could also be derived from locale: do { local @_ = split m([._]), $ENV{LANG};$_[0] }), but the two are unrelated.

=cut

    $m->prog_define(['lang','mlang','ll','l'], '$', $ENV{MLANG});

=item --hom_konfayl_fnom '.%b_%s'

sprintf template for generating name of user home directory configuration file .prog_konf from elements such as

	b base 'prog'
	s suffix 'konf'
	p prefix '_'

=cut

    $m->konfig_define('hom_konfayl_fnom', 's', '.%b_%s');

=item --lok_konfayl_fnom '%p%b'

sprintf template for generating name of local configuration file prog from elements such as

	b base 'prog'
	s suffix 'konf'
	p prefix '_'

=cut

    $m->konfig_define('lok_konfayl_fnom', 's', '%p%b');
    $m->konfig_define('homdir', 's', $ENV{HOME});
    $m->cache_define('mlht_homedir', 's', '/sig') unless $novitas;
=item --progver 12035

Current version of software, corresponding to 1.2.35 in PAK.mk.
This is used by the creating programs such as mktdir to write dokprogver into the created directory's configuration file aka dokfs.

=item --batch

Never ask the user to supply missing information interactively, always use default values instead.
=cut
    $m->cache_define('batch', '!');
=item --hsprintf_repl p = _ --hsprintf_repl a = a2e
preset some values for single letter variables in templates of the sort %pdok.%d.%f.txt from which filenames are made
=cut
    $m->cache_define('hsprintf_repl', 's%') unless $novitas;
=back    

=cut

    $m->{f} = {};
    $m->{konfaylz} = [];
    $m->{konfaylz_old} = {};
    my @progtime = localtime time;
    my $setpvar = sub { my ($k,$v) = @_;$m->{progvars}->{$k} = $v };
    &$setpvar(progtime => \@progtime);
    my $hodie = strftime '%Y-%m-%d', @progtime;
    &$setpvar(hodie => $hodie);
    my $prognom = $0;
    &$setpvar(prognom => $prognom);
    my $progbas = basename $prognom;
    &$setpvar(progbas => $progbas);    
    my $progver = '0.4.4';
    &$setpvar(progver => $progver);
    my $dirsep = catdir('');
    &$setpvar(dirsep => $dirsep);
    my $dirsep_re = quotemeta $dirsep;
    &$setpvar(dirsep_re => $dirsep_re);
    $m->konfayl('prog.konf');
    $m->qpop1($prog, 1);
};

=head3 Function hsprintf

Fill in a simple template string with values from a hash.

Simple call-by-value replacement for positional 'sprintf'.

Expand something like

    %p%d.%l.txt 

to sth like

    _perl_pub.de.txt

when invoked by

    $m->hsprintf('%p%d.%l.txt', p => '_', d => 'perl_pub', l => 'de');

Letters are read either from the argument line or from the object variable hsprintf_repl which may have some default settings such as e.g. the current source filename prefix for %p, the current document for %d and the current language for %l.

=cut 

sub hsprintf_qre2fun {
    my ($m,$qre,$fun) = @_;
    $m->{hsprintf_qre2fun}->{$qre} = $fun;
};

sub hsprintf_stat {
    my ($m,$repl,$tut,$nom) = @_;
    my $val = $repl->{$nom} // $m->{hsprintf_repl}->{$nom};
    return '' if not length $val;
    if (ref $val) { $val = $m->dval($val) } else { _utf8_on $val };
    $val;
};
sub hsprintf_dyna {
    my ($m,$repl,$tut,$nom) = @_;
    return $m->plok if $nom eq 'PL';
    return $tut;
};
sub hsprintf_para { # file:/phm/19/06/07/sig/_lng.phm_pub_sig190607.txt::hsprintf
    my ($m,$repl,$tut,$nom,$etc) = @_;
    my @noms = split qr{\|}, $etc;
    my $num = $repl->{$nom} // $m->{hsprintf_repl}->{$nom};
  NUM: {
      last if $num =~ m(\A\d+\Z);
      $num = $num ? 1 : 0;
    };
    my $val = $noms[$num];
    $val || $tut;
};

sub hsprintf_qre2fun_defs {
    my $m = shift;
    $m->hsprintf_qre2fun(qr(\%(\w)) => \&hsprintf_stat);
    $m->hsprintf_qre2fun(qr(\%\{(\w+)\}) => \&hsprintf_stat);
    $m->hsprintf_qre2fun(qr(\%[(](\w+)[)]) => \&hsprintf_dyna);
    $m->hsprintf_qre2fun(qr(\%[(](\w+)\|([^()]+)[)]) => \&hsprintf_para);
    1;
};

sub hsprintf {
    my $m = shift;
    local $_ = shift or return;
    _utf8_on $_;
    my %repl = $m->ref2hash($m->{hsprintf_repl});
    while (@_) {
	my $k = shift;
	my $v = shift;
	$repl{$k} = $v;
    };
    while (my ($qre,$fun) = each %{$m->{hsprintf_qre2fun}}) {
	s($qre)(&$fun($m,\%repl,$&,$1,$2))eg;
    };
    $_;
};

=head3 Function defvars

Define configuration variables and specify configuration files (with the konfayl method) here.

=cut

sub defvars {
    my ($m,$prog) = shift->qpuc('A2E::Prog::defvars',@_) or return;
    $m->prog_defvars;
    $m->qpop1($prog,1);
};

=head3 Function ekspand_konfayl

=over

=item arg1

e.g.

	'/etc/opt/a2e/prog.konf'

=item result: array of 3 files

e.g.

    '/etc/opt/a2e/prog.konf', '/home/phm/.prog_konf', './prog'

=back


Subroutine of ekspand_konfaylz

=cut

sub ekspand_konfayl {
    my ($m,$prog,$etcfol) = shift->ppuc('ekspand_konfayl',@_);
    my ($vol, $dir, $fol) = splitpath $etcfol;
    my $homfol_old = catfile $ENV{HOME}, $fol;
    my ($folbas, $sfx) = $fol =~ m(\A(.+)\.(\w+)\Z);
    my $pfx = $m->get_srcfpfx;
    my $homfol = $m->hsprintf($m->{konfig}->hom_konfayl_fnom, b => $folbas, 's' => $sfx, p => $pfx);
    $homfol = catfile $m->{konfig}->homdir, $homfol;
    my $lokfol = $m->hsprintf($m->{konfig}->lok_konfayl_fnom, b => $folbas, 's' => $sfx, p => $pfx);
    my @faylz = ($etcfol, $homfol_old, $homfol, $lokfol, join('_', $lokfol, $sfx));
    $m->dpop($prog,@faylz);
};

=head3 Function ekspand_konfaylz

Subroutine of pars_konfaylz

Expand a list of configuration files like '/etc/opt/a2e/prog.konf' and return list about 3 times as long, consisting of something like  

    ('/etc/opt/a2e/prog.konf', "$ENV{HOME}/.prog", 'prog')

in place of each entry.

=cut

sub ekspand_konfaylz {
    my ($m,$prog) = shift->ppuc('ekspand_konfaylz',@_);
    my %konfs = ( etc => [], hom => [], lok1 => [], lok2 => [] );
    my @konfaylz = ();
    my %oldp = ();
    my $push_if_new = sub {
	my ($nom,$elm) = @_;
	return unless $elm;
	return if $oldp{$elm};
	$oldp{$elm} = 1;
	return unless -r $elm;
	push @{$konfs{$nom}}, $elm;
    };
    foreach my $fayl (@{$m->{konfaylz}}) {
	my ($etc,$hom,$lok1,$lok2) = $m->ekspand_konfayl($fayl);
	&$push_if_new(etc => $etc);
	&$push_if_new(hom => $hom);
	&$push_if_new(lok1 => $lok1);
	&$push_if_new(lok2 => $lok2);
    };
    @konfaylz = (@{$konfs{etc}}, @{$konfs{hom}}, @{$konfs{lok1}}, @{$konfs{lok2}});
    $m->dpop($prog, @konfaylz);
};

=head3 Function pars_konfaylz

Read configuration files as accumulated in my konfaylz field as well as extra ones optionally supplied as arguments.  Void the konfaylz field and instead enter the the names of those files that were found onto the konfaylz_old hash.  Never read any file that is already on this hash.

Look for corresponding configuration files also in the user\'s home directory and in the current directory, e.g. if C</etc/opt/a2e/mktdir/mktdir.konf> is among the configuration files, also read C<~/mktdir.konf> and C<./@mktdir> if found.

=cut

sub pars_konfaylz {
    my ($m,$prog,@konfaylz) = shift->ppuc('pars_konfaylz',@_);
    push @{$m->{konfaylz}}, @konfaylz;
    @konfaylz = $m->ekspand_konfaylz;
    my $fi = '';
    my $n = 0;
    while ($fi = shift @konfaylz) {
	$n++; 
	$m->dpuc('fi'.$n, $fi);
	do { $m->dpop;next } unless -r $fi;
	if ($m->{konfaylz_old}->{$fi}) { $m->pwarn('Hey, I have already read the configuration file %f and will not do it again.', f => $fi);next };
	$m->open_fayl($fi, 'konf', 'r'); 
	# we need to read a handle, simply reading $fi wont do, otherwise BlablaTest fails because blabla.konf was not read with proper :utf8 support
	$m->{konfig}->file($m->fh('konf'));
	$m->kloz_fayl('konf');
	$m->{konfaylz_old}->{$fi} = 1;
	# $m->{konfig}->_debug(1) if $m->{konfig}->get('debug');
	# $m->decho("in pars_konfaylz after $fi got konfig " . Dumper($m->{konfig}));
	$m->dpop;
    };
    $m->dpop1($prog,1);
};

# =head3 Function konfig_faylz
# 
# =cut
# 
# sub konfig_faylz {
#     my $m = shift;
#     $m->pwarn('using deprecated name konfig_faylz for function pars_konfaylz');
#      return $m->pars_konfaylz(@_);
# };

=head3 Function konfig_optvalz

Part of the new_ready constructor.  The argument is a hash with configuration defaults of the current application.
This allows any arguments supplied to new_ready to override any options of the same name as found in configuration files, while still allowing users to override this from the commandline.

WARNING: HACK!

The solution relies on internals of AppConfig & AppConfig::State.

We need to directly access the $m->{konfig}->{STATE}->{VARIABLE} hash.

We cannot use the set/get routines on lists and hashes because they perform unwanted parsing.

=cut

sub konfig_optvalz (%) {
    my ($m,$prog,%optvalz) = shift->ppuc('konfig_optvalz',@_);
    foreach my $var (keys %optvalz) {
	my $ok = $m->{konfig}->_exists($var);
	next unless $ok;
	my $val = $optvalz{$var};
	if (ref $val) { $m->{konfig}->{STATE}->{VARIABLE}->{$var} = $val } else { $m->{konfig}->set($var => $val) };
    };
    $m->dpop1($prog,1);
};

=head3 Function konfig_getopt

read konfig values from commandline.

show help, manpage or version if appropriate.

=cut

sub konfig_getopt {
    my ($m,$prog) = shift->ppuc('konfig_getopt',@_);
    my $ok = $m->{konfig}->getopt();
    pod2usage({ -message => "$0", -exitval => 1, -verbose => 0 }) unless $ok;
    my $prognom = $m->{progvars}->{prognom};
    my $progver = $m->{progvars}->{progver};
    do { $m->necho("$prognom version $progver");exit 0 } if $m->{konfig}->version;
    pod2usage { -message => "$prognom version $progver", -exitval => 0, -verbose => 2 } if $m->{konfig}->manpage;
    pod2usage { -message => "$prognom version $progver", -exitval => 0, -verbose => 1 } if $m->{konfig}->help;
    $m->{konfig}->_debug(1) if $m->{progvars}->{debug};
    $m->dgot(konfig => $m->{konfig});
    $m->dpop1($prog,1);
};

=head3 Function postkonfig

Some final sanity checks and finishing steps to get the object ready for use.
To be overwritten by inheriting objects.

=cut

sub netp {
    my ($m,$prog,$host) = shift->ppuc('netp',@_);
    return $m->{ping}->ping($host) if grep { $_ eq $host } 'localhost', '127.0.0.1', '127.0.0.2';
    return $m->dpop($prog) if $ENV{OFFLINE};
    $host ||= $m->{konfig}->pinghost;
    $m->pstop('ping is not a Net Ping object') unless 'Net::Ping' eq ref $m->{ping};
    printf STDERR 'ping host %s...', $host;
    my $ok = $m->{ping}->ping($host);
    printf STDERR "%s\n", $ok ? 'OK' : 'KO';
    $m->dpop1($prog,$ok);
};
sub prog_postkonfig {
    my ($m,$prog) = shift->qpuc('prog_postkonfig',@_) or return;
    $m->prog_msgdefs('prog',%MSGS);
    $m->set_srcfpfx;
    $m->{ping} = new Net::Ping 'syn', 1 or $m->pstop(noping => p => 'syn', t => 1);
  TCEK_GRUP: {
      last if $novitas;
      $m->tcek_grup;
    };
    my $cwd = getcwd();
    $m->setpvar(cwd => $cwd);
    umask oct($m->{konfig}->umask);
  UTF8BIN: {
      binmode STDIN, ':utf8';
      binmode STDOUT, ':utf8';
      binmode STDERR, ':utf8';
    };
    my $locale = $m->{konfig}->locale || '';
    my $set_locale = sub {
	if ($locale) {
	    setlocale LC_MESSAGES, $locale or $m->ustop(setlocale => l => $locale);
	    my $locale1 = setlocale LC_MESSAGES;
	    $m->uwarn(localenesed => l => $locale1, m => $locale) unless $locale1;
	} else {
	    setlocale LC_MESSAGES, '';
	    $locale = setlocale LC_MESSAGES;
	};
	$m->uwarn('nolangvar') unless $locale;
	$m->setpvar(locale => $locale);
    };
    &$set_locale;
    my $lang = $m->{konfig}->lang || '';
    my $set_lang = sub {
	return $lang if $lang =~ m(\A\w{2,3}\Z);
	$m->ustop('nolangnolocale') unless $locale;
	$m->warn(nolanglocale => l => $locale) if grep { $_ eq $locale } qw(C POSIX);
	local $_ = $locale;
	($lang) = m(\A([a-z]{2,3})[\_\b]) or $m->ustop(localenolang => l => $locale);
	$m->setpvar(lang => $lang);
    };
    $m->out2sav;
    eval { &$set_lang }; # cheat for now, debug later: file:/phm/18/11/04/sig/_log.g.phm_pub_sig181104.txt::11539
    $m->qpop1($prog,1);
};

=head3 Function postkonfig

=cut

sub postkonfig {
    my ($m,$prog) = shift->qpuc('A2E::Prog::postkonfig',@_) or return;
    $m->prog_postkonfig;
    $m->qpop1($prog,1);
};

=head3 Function get_srcfpfx

Return source file name prefix based on what is found in current directory

ARGS
    %opts => { bas =>  'langs', dir => '/sig/oas' }
    bas :: base name of file that must be present, e.g. 'dokfs' for '@dokfs', 'langs' for '__langs'
RET
    $pfx :: directory specific source file name prefix 
    $pfx => '_', '__', '@'

Returns '_' as directory-specific source file name prefix if file '_langs' is found in the directory.

TODO
    use pack and unpack, operate on characters one by one so as to avoid rescanning
    allow '%%'

=cut

sub get_srcfpfx {
    my ($m,$prog,%opts) = shift->ppuc('get_srcfpfx',@_);
    my $bas = $opts{bas} || 'langs';
    my $dir = $opts{dir} || '.';
    local $_;my $pfx = '';my %pfxh = ();
    my $dh;opendir $dh, $dir;
    for (readdir $dh) {
	next unless -f $_;
	($pfx) = m(\A([_%@~])(?:\w{3}\..*\.txt|\w+)\Z);
	next unless $pfx;
	++$pfxh{$pfx};
    };
    closedir $dh;
    return $m->dpop1($prog,$m->{konfig}->srcfpfx) unless %pfxh;
    my @pfxl = keys %pfxh;
    return $m->dpop1($prog,$pfxl[0]) if $#pfxl == 0;
    @pfxl = sort { $pfxh{$b} <=> $pfxh{$a} } @pfxl;
    $m->dpop1($prog,$pfxl[0]);
};

=head4 Function rtrim_nokoment
Remove trailing coments that come after hashmark which is not backslash-escaped and which is followed either by eol or whitespace
Comment lines are suppressed, trainling spaces removed, but before a comment '#' sign only one space is removed.  This way we can define LF TAB SPC

=cut
sub rtrim_nokoment { # remove everything that comes after first '#' and trailing whitespace before it
    my $prog = 'rtrim_nokoment';my $m = shift;$m->dpuc($prog, @_);
    local $_ = shift;
    my $len = length $_;
    my $p = 0;
    my $p0 = 0;
    my $p1 = 0;
    my $s = '';
  FIND: { # TODO: allow escaping '#' with backslash
      $p = index $_, '#', $p0;
      last if $p == -1;
      $p0 = $p + 1;
    BACKSLASH: {
	$p1 = $p - 1;
	$s = substr $_, $p1, 1;
	last unless '\\' eq $s;
	$len--; 
	$s = substr $_, 0, $p1;
	$s .= '#';
	$s .= substr $_, $p0;
	$_ = $s;
	redo FIND;
      }
    FINALHM: { # remove final hashmark
	last if $p0 < $len;
	last FIND 
      }
    NONBLANK: {
	$s = substr $_, $p0, 1;
	last if grep { $_ eq $s } ' ', "\n";
	redo FIND;
      };
    };
    $_ = substr $_, 0, $p if $p >= 0;
    my @chars = unpack 'U*', $_;$p = $#chars;$p-- while $p >= 0 and grep { $_ eq $chars[$p] } @{$m->{blanks}};$p += 1;
    $_ = substr $_, 0, $p;
    $m->dpop1($prog, $_);
};

=head3 Function superuserp

Whether we are a privileged user who can be given unfettered freedoms to invoke Perl or Bash.
It doesn't have to be the root user.
The decision is made on the basis of write access to installed configuration files which are normally only root-writable but could temporarily be made writable to a privileged group such as 'wheel'. 
Cf file:Makefile::wheel

=cut

sub superuserp { return 1 if -w '/etc/opt/a2e/prog.konf' };

=head2 Destructors

=head3 Function prog_destroy

remove file locks

=cut

sub prog_destroy {
    my $prog = 'prog_destroy';my $m = shift;$m->dpuc($prog,@_);
    # $m->anlokk_faylz;
    $m->techo('destroy prog at %L', L => $m->plok);
    $m->kloz_faylz;
    $m->dpop1($prog,1);return 1;
};

=head4 setpvar( $nom => $val ) => 1: make a value unchangeable so that reading out dbmvars files will not alter it
Used for values that belong the the control structure of our process.
setpvar should be used only for variables that are managed by the program regardless of what external files try to impose.


=cut

our $ali2nom = { tml => 'tmplang' };
our $ali2val = { tmplang => { deplate => 'multemp' } };
sub ali2nom {
    my ($m,$ali) = @_;
    my $nom = '';
    while ($ali) { 
	$nom = $ali;
	$ali = $ali2nom->{$ali};
    };
    return $nom;
};
sub ali2val {
    my ($m,$nom,$ali) = @_;
    $m->pstop('ali2val invoked without nom') unless $nom;
    return $ali unless $ali;
    my $nom_a2v = $ali2val->{$nom};
    return $ali unless $nom_a2v;
    my $val;
    while($ali) {
	$val = $ali;
	$m->pstop('invalid ali2val value %v', v => $val) unless $val =~ m(.);
	$ali = $nom_a2v->{$ali};
    };
    $val;
};
sub set_ali2val {
    my ($m,$var,$alias,$real) = @_;
    $ali2val->{$var}->{$alias} = $real;
    $real;
};
sub setpvar {
    my ($m,$alinom,$alival) = @_;
    my $nom = $m->ali2nom($alinom);
    my $val = $m->ali2val($nom,$alival);
    $m->{progvars}->{$nom} = $val;
    $val;
};
sub defpvar { # 如模板語$(q:DEFAULT)
    my ($m,$prog,$nom,$val) = shift->ppuc('defpvar',@_);
    my $oldval = $m->pvar($nom);
    return $m->dpop1($prog,$oldval) if defined $oldval;
    $m->setpvar($nom,$val);
    $m->dpop1($prog,$val);
}
sub setplst {
    my ($m,$prog,$nom,$val,%opts) = shift->ppuc('setplst',@_);
    local $_ = $m->setpvar($nom,$val,%opts);
    return unless 'ARRAY' eq ref $_;
    $m->dpop($prog,@$_);
};
=head $m->setphac(langlands => \%langlands, keys => \@keys, lihs => $lihs)

file:/adv/perl/A2E/MLDK.pm.tmpl::setphac
The C<keys> option sets a list of keys for the hash, stored in key _keys_, so as to put them in sequence, according to the convention further elaborated in A2E::TmplArb.  Which may be a reason to move this functiong there.
The C<lihs> option becomes meaningful only at a higher level where the base function setpvar is replaced by a more complex method in A2E::MLHT.
In analogy to setplst, this function returns the hash rather than hashref; this return value might be useful for functionaly processing where decisions depend on whether it is void.

=cut
sub setphac {
    my ($m,$prog,$nom,$val,%opts) = shift->ppuc('setphac',@_);
    $m->pstop('value given to setphac %v is not a hashref', v => $val) unless 'HASH' eq ref $val;
    my %hac = %$val;
    my $get_keys = sub {
	(my $kisr,%opts) = $m->shifthash('keys',%opts);
	return [] unless $kisr;
	$m->pstop('value given to keys option of setphac %v is not a listref', v => $kisr) unless 'ARRAY' eq ref $kisr;
	$kisr;
    };
    $hac{_keys_} ||= $m->uniqpush(&$get_keys,[keys %hac]);
    $m->setpvar($nom,\%hac,%opts);
    $m->dpop($prog,%hac);
};

sub pvar {
    my ($m,$nom,$insist) = @_;
    $nom = $m->ali2nom($nom);
    my $val = $m->{progvars}->{$nom};
  VAL: {
      last if defined $val;
      $val = $m->{cache}->{$nom};
      last if defined $val;
      $val = $m->{$nom};
    };
  ALI: {
      last if ref $val;
      $val = $m->ali2val($nom,$val);
    };
    $m->pstop(nomisref => n => $nom) if ref $nom;
    $m->pstop(badnom => n => $nom) unless $nom =~ m(\A\w+\Z);
    $m->pstop('pvar %n has no val',  n => $nom) if $insist and not defined $val;
    return $val;
};
sub plst {
    my ($m,$ali,$insist) = @_;
    my $val = $m->pvar($ali,$insist);
    return unless 'ARRAY' eq ref $val;
    @$val;
};
sub ref2hac {
    my ($m,$ref) = @_;
    $m->pwarn('invoking ref2hash under name ref2hac');
    return $m->ref2hash($ref);
};
sub ref2lst {
    my ($m,$ref) = @_;
    $m->pwarn('invoking ref2list under name ref2lst');
    return $m->ref2list($ref);
};
sub uniqpush { # produce a uniq list from two or more listrefs, preserving the order especially at the beginning
    my $m = shift;
    my @ll = (); # list
    my %lp = (); # register
    foreach my $lr (@_) {
	next unless $lr;
	$m->pstop('uniqpush argument %l is not a listref', l => $lr) if 'ARRAY' ne ref $lr;
	do { next if $lp{$_};push @ll, $_;$lp{$_} = 1 } for $m->ref2list($lr); 
    };
    @{$_[0]} = @ll;
    @ll;
};
=head2 \%h = $m->hashpush($kk,$h0,$h1..$h9): merge all hashes $h0..$h9 into $h0

$kk => (?:ad|ae|bd|be|ff)
$kk :: use this index to find the assignment function in %kk2put; this can be left out in which case 'ff' is assumed
$h0,$h1,... :: hashrefs

=cut
sub hashpush { # merge hashes $h0..$h9 into $h0
    my $m = shift;
    my $kk = shift;
    my $put = $kk2put{$kk};
    my $h0;
    if ($put) {
	$h0 = shift;
    } else {
	$h0 = $kk;
	$put = $kk2put{ff};
    };
    $m->pstop('bad put function %f', f => $put) if 'CODE' ne ref $put;
    $h0 ||= {};
    my $i = 0;
    my $ihacp = sub { my ($h) = @_;$m->pstop('arg %i %h is not a hashref', i => $i, h => $h) unless 'HASH' eq ref $h };
    &$ihacp($h0);
    my $kisr0 = $h0->{_keys_};
    my $puthk = $kisr0 ? sub {
	my ($h,$k) = @_;
	push @$kisr0, $k unless $h0->{$k};
	&$put($h0,$k,$h->{$k});
    } : sub {
	my ($h,$k) = @_;
	&$put($h0,$k,$h->{$k});
    };
    foreach my $h (@_) {
	$i++;
	$h ||= {};
	&$ihacp($h);
	my $kisr = ($kisr0 && $h->{_keys_});
	my @keys = $kisr ? @$kisr : keys %$h;
	foreach my $k (@keys) { &$puthk($h,$k) };
    };
    $h0;
};
sub shifthash { # ('b', a => 1, b => 2) ==> ('2', a => 1) file:/adv/perl/A2E/MLHT.pm.tmpl::shifthash
    my $m = shift;
    my $k = shift;
    my %hac = @_;
    my $v = $hac{$k};
    delete $hac{$k} if $v;
    ($v, %hac);
};
sub phac {
    my ($m,$ali,$insist) = @_;
    my $val = $m->pvar($ali,$insist);
    return unless 'HASH' eq ref $val;
    %$val;
};

=head4 $m->listref(qw(a b c)) => ['a','b','c']
Or:
$m->listref(['a','b','c']) => ['a','b','c']

Turn arglist into listref or return single listref as such.
Any argument list that does not consist of a single listref becomes a listref, e.g. a single hashref would become a listref with the original hashref as its sole element: 
$m->listref(1) => [ 1 ]
$m->listref($lihs) => [ $lihs ]
$m->listref([]) => []

=cut
sub listref {
    my $m = shift;
    return $_[0] if !$#_ and 'ARRAY' eq ref $_[0];
    [ @_ ];
}
=head4 $m->hashref(a => 1, b => 2) => { a => 1, b => 2 }

Turn even-numbered arglist into hashref or return single hashref as such.
Report error otherwise.
Under the conditions in the example, scalar(@_) => 4, $#_ => 3, $#_ % 2 => 1

=cut
sub hashref {
    my $m = shift;
    return $_[0] if !$#_ and 'HASH' eq ref $_[0];
    $m->pstop('uneven number of hash arguments %R', R => \@_) unless $#_ % 2;
    { @_ };
}
=head4 $m->uniq_plst('tmls') => @tmls; same as plst but assure that each member occurs only once

=cut
use List::MoreUtils qw(uniq);
sub uniq_plst { return uniq shift->plst(@_) };

=head3 Function DESTROY

=cut

sub DESTROY {
    my $m = shift;
    $m->prog_destroy;
};

=head1 TODO

=cut

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/Prog.pm.tmpl"
# End:
