# -*- coding: utf-8; srcfile:/adv/perl/A2E/SArb.pm.tmpl; mode: perl -*-

package A2E::SArb;

=head1 NAME

A2E::SArb - Parsing a String into a Syntactic Tree, recursively defined as a reference to a list of trees

=head1 DESCRIPTION

Parse and transform lines of text that contain nested parenthesese, function calls
and the like, into semantic tree structures of arbitrarily many levels
of hierarchy.

Usually this is done by writing a subclass to A2E::SArb, see e.g. A2E::SArb::Make and A2E::SArb::MLHT. 

=head1 SYNOPSIS

	use A2E::SArb::Make;

        my $vars = {};
        my $p = new_ready A2E::SArb::Make lihs => $lihs, gvars => $grupvars;
        $vars->{lib} = 'A2E::Prog';
        $p->transform('my library is ${lib}')
            => my library is A2E::Prog

Of course much more complicated nested parsing is possible, see e.g. the A2E::SArb::Make application which imitates and extends the syntax of GNU Make.

=head1 RELATED SOFTWARE & DOCUMENTATION

=head2 Download Source

    http://a2e.de/adv/perl/A2E/SArb.pm

=head2 Prerequisites

    A2E::Prog(3)
    Scalar::Util(3)
    Data::Dumper(3)

=cut

use base 'A2E::Tmplvars';
use strict;
use Data::Dumper;
use Scalar::Util qw(reftype);
use Encode qw(_utf8_on);
our %MSGS = (
    voknonfun => 'trying to call unknown function %n',
    );
=head2 Applications

    A2E::SArb::MLHT(3)
    A2E::SArb::Make(3)
    A2E::SArb::Deplate(3)

=head2 Similar CPAN Modules

=over

=item Parser::Simple(3)

This allows plugins but is line-based and carries a lot of baggage that made it seem less likely to be efficient for the applications envisaged here.
The others are again probably overkill for the purpose of line-to-tree parsing envisaged here.

=item Parser::RecDescent(3)

=item Parse::Yapp, Parse::Eyapp

=item Parse::YYLex, perl-byacc

perl-byacc is not on CPAN as it is a C-based yacc compiler that produces perl programs.

=back

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

    Copyright (c) 2008-20 Hartmut Pilch (phm at a2e de)

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

=head2 Constructor parts

=head3 Function sarb_defvars

=cut

sub get_vars {
    my ($m,$varsr) = @_;
    $varsr ||= $m->{lihs}->{vars};
    $m->pstop('varsr is not a hashref') unless 'HASH' eq ref $varsr;
    $m->hashpush('ff',$varsr, $m->{lihs}->{lvars});
    $m->pstop('variable no 1 has bad value %k', k => $varsr->{1}) if 'HASH' eq ref $varsr->{1};
    %$varsr;
};
sub vpuc {
    my $m = shift;my $varsr = shift;
    $varsr ||= {};
    $m->pstop('varsr is not a hashref') unless 'HASH' eq ref $varsr;
    $m->pstop('variable no 1 has bad value %k', k => $varsr->{1}) if 'HASH' eq ref $varsr->{1};
    ($m,$varsr,@_);
};
sub vrpuc {
    my ($m,$varsr,@args) = shift->vpuc(@_);
    ($m,$varsr,map { $m->render($varsr,$_) } @args); 
};
sub render_concat {
    my ($m,$varsr,@rest) = shift->vpuc(@_);
    @rest = map { $m->render($varsr,$_) } @rest;
    my $rest0 = $rest[0];
    ref $rest0 ? $rest0 : join '', @rest;
};
sub render_join { 
    my ($m,$varsr,$sep,@rest) = shift->vpuc(@_);
    join $sep, map { $m->render($varsr, $_) } @rest;
};
sub render_paren { 
    my ($m,$varsr,$left,$right,@rest) = shift->vpuc(@_);
    @rest = map { $m->render($varsr, $_) } @rest;
    my $body = join '', @rest;
    $left . $body . $right;
}
sub sarb_defvars {
    my ($m,$prog) = shift->qpuc('sarb_defvars',@_) or return;
    # $m->prog_minidefs;
=over
=item --parsers
=cut
    $m->{konfig}->define('parsers=s@');
=back
=cut
    $m->{konfig}->parsers('paren');
    $m->add_parser(paren => sub { my $p = shift;return $m->try_paren($p) });
    # $m->add_parser('keyvals', sub { my $p = shift;return $m->try_keyvals($p) });
    $m->konfayl('sarb.konf');
    # $m->{openers} = [ (40, 91, 123, 65288, 12300) ];
    # $m->{closers} = [ (41, 93, 125, 65289, 12301) ];
    $m->{opener_closers} = { 34 => 34, 40 => 41, 91 => 93, 123 => 125, 12300 => 12301, 65288 => 65289 }; # '('
    $m->{closer_openers} = { 34 => 34, 41 => 40, 93 => 91, 125 => 123, 12301 => 12300, 65289 => 65288 }; # ')'
    $m->pstop('blanks not defined') unless 'ARRAY' eq ref $m->{blanks};
    $m->{blankp}->{$_} = 1 for @{$m->{blanks}};
    $m->{equal_char} = 61;
    my @alphanums = ( ord('0') .. ord('9'), ord('A') .. ord('Z'), ord('_'), ord('a') .. ord('z') );
    $m->{alphanums} = \@alphanums;
    $m->{alphanump}->{$_} = 1 for @alphanums;
    $m->add_renfun(concat => \&render_concat);
    $m->add_renfun(join => \&render_join);
    $m->add_renfun(paren => \&render_paren);
    $m->prog_msgdefs('sarb',%MSGS);
    $m->qpop1($prog,1);
};

=head3 Function defvars

=cut

sub defvars {
    my ($m,$prog) = shift->qpuc('A2E::SArb::defvars',@_) or return;
    $m->SUPER::defvars;
    $m->sarb_defvars;
    $m->qpop1($prog,1);
};

=head3 Function postkonfig

=cut

sub postkonfig {
    my ($m,$prog) = shift->qpuc('A2E::SArb::postkonfig',@_) or return;
    $m->qpop1($prog,1);
};

=head3 Function add_parser

Register a parser, i.e. a function that will try to find a syntactic expression at a given position and return the new position, to which has advanced if successful, and the syntactic expression ($arb) which it found.

Currently 'try_paren' and unused 'try_keyvals' are defined here, and 'make' is defined in A2E::SArb::Make.

=cut

sub add_parser {
    my ($m,$name,$meth) = @_;
    $m->{parsers}->{$name}->{aktiv} = 1;
    $m->{parsers}->{$name}->{meth} = $meth;
    $meth;
};

=head3 Function switch_parser

    $m->switch_parser('paren', 1)
    $m->switch_parser('paren', 0)

activate and deactivate a parser

=cut

sub switch_parser {
    my ($m,$name,$val) = @_;
    $m->{parsers}->{$name}->{aktiv} = $val;
    $val;
};

sub add_renfun {
    my ($m,$nom,$val) = @_;
  NOVVAL: {
      my $oldval = $m->{funs}->{$nom};
      if ($oldval) {
	OLDVAL: {
	    $m->pwarn('renfun %n already registered', n => $nom);
	    last if 'CODE' eq ref $oldval;
	    $m->pwarn('renfun %n was not of type code', n => $nom);
	  };
      } else {
	  $m->vecho('renfun %n registered for the first time', n => $nom);
      };
    };
    $m->pstop('invalid new renfun for %n', n => $nom) unless 'CODE' eq ref $val;
    $m->{funs}->{$nom} = $val;
    $val;
};

=head2 Parser Functions

=head3 Function try_parse

=cut

sub try_parse {
    my ($m,$p,$arb) = @_;
    foreach my $parser (@{$m->{konfig}->parsers}) {
	next unless $m->{parsers}->{$parser}->{aktiv};
	my $meth = $m->{parsers}->{$parser}->{meth};
	next unless $meth;
	($p, $arb) = &$meth($p);
	next unless $arb;
	last;
    };
    ($p,$arb);
};

=head3 Function add_parse

Try to find a syntactic expression at the position p and return with new positions and found tree 

=over 
=item $p0: the position from where the next parsable element should begin
=item $p: the position up to where we parsed (where the sexp ends)
=item arb: the resulting tree
=back

If nothing was found, I<p0> is unchanged, I<p> moves ahead by 1, and I<arb> is undefined.

=cut

sub add_parse {
    my ($m,$p0,$p1,$arb) = @_;
    my ($p,$sub_arb) = $m->try_parse($p1);
  SUBARB: {
      last unless $sub_arb;
    S0: {
	last unless $p0 < $p1;
	my $s0 = $m->substring($p0, $p1 - 1);
	push @$arb, $s0;
      };
      push @$arb, $sub_arb;
      $p0 = $p + 1;
    };
    ($p0,$p,$arb);
};

=head3 Function add_subarb

=cut

sub add_subarb {
    my ($m,$p0,$p,$sub_arb,$arb) = @_;
  S0: {
      last unless $p0 <= $p;
      my $s0 = $m->substring($p0, $p);
      push @$sub_arb, $s0 if $s0;
    };
    push @$arb, $sub_arb if $sub_arb;
    $arb;
};

=head3 Function string

=cut

sub string {
    my $m = shift;
    pack 'U*', @{$m->{chars}};
};

=head3 Function substring

=cut

sub substring {
    my ($m,$p0,$p1) = @_;
    pack 'U*', @{$m->{chars}}[$p0..$p1];
};

=head3 Function try_paren

See whether we have an opening parenthesis under arg1 C<$p> and recursively collect anything found up to the matching closer into tree C<$arb>.

=cut

sub try_paren {
    my ($m,$prog,$p) = shift->ppuc('try_paren',@_);
    my $pa = $p;
    my $ch_a = $m->{chars}->[$p];
    my $ch_z = $m->{opener_closers}->{$ch_a};
    return $m->dpop1($prog, $p) unless $ch_z;
    my $arb = [ 'paren', chr($ch_a), chr($ch_z) ];
    $p++;
    my $p0 = $p;
  CHAR: while ($p < $m->{len}) {
      my $ch = $m->{chars}->[$p];
      my $chr = chr $ch;
    CLOSER: {
	last unless $ch == $ch_z;
	my $s0 = $m->substring($p0, $p-1);
	push @$arb, $s0;
	return $m->dpop($prog,$p,$arb);
      };
      ($p0,$p,$arb) = $m->add_parse($p0,$p,$arb);
  } continue { $p++ };
    $m->dpop1($prog,$pa);
};

=head3 Function skip

=cut

sub skip {
    my ($m,$p,@args) = @_;
    $p++ while $p < $m->{len} and grep { $_ == $m->{chars}->[$p] } @args;
    $p;
};

=head3 Function skip_non

=cut

sub skip_non {
    my ($m,$p,@args) = @_;
    $p++ while $p < $m->{len} and not grep { $_ == $m->{chars}->[$p] } @args;
    $p;
};

=head3 Function parse_from

=over

=item arg1 $p
position in string to start from
=item ret1 $p
position at which we arrived
=item ret2 $arb
resulting syntactic tree 

=back

=cut

sub parse_from {
    my ($m,$p) = @_;
    my $p0 = $p;
    my $arb = ['concat'];
    while ($p < $m->{len}) {
	($p0, $p, $arb) = $m->add_parse($p0, $p, $arb);
    } continue { $p++ };
    my $s0 = $m->substring($p0, $p-1);
    push @$arb, $s0;
    ($p,$arb);
};

=head2 Rendering Subroutines

=head3 Function getval

Return a pre-stored variable value; common subroutine of getfun and getarb.

    args := typ, nom, vars
    rets := val
    typ => 'arbs' or 'vars' or 'funs', default 'vars'
    typ :: name of global symbol table to search in
    nom => 'url'
    nom :: variable name
    vars => { url => 'http://a2e.de' }
    vars :: local symbol table
    val :: string value
    val => 'http://a2e.de'

=cut

sub getval {
    my ($m,$varsr,$typ,$nom) = shift->vpuc(@_);
    my $val = $varsr->{$nom};
    return $val if $val;
    # $val = $m->{gvars}->{$nom};
    # return $val if $val;
    $m->{lihs}->getval($nom);
};

=head3 Function getvar

Return stored variable value
arg1 is key, arg2 is local symbol table (hashref) used in preference to the global one

=cut

sub getvar {
    my ($m,$varsr,$nom) = shift->vpuc(@_);
    $m->getval($varsr,'vars',$nom);
};

=head3 Function getarb

Return the already-parsed arb corresponding to nom or parse+store+return it, along with the string so as to make sure we reparse the string when it changes.

    args := nom, vars?
    rets := arb
    arb :: tree structure corresponding to a dynamic variable (function)

=cut

sub getarb {
    my ($m,$varsr,$nom) = shift->vpuc(@_);
    my $arb = $m->{arbs}->{$nom};
    my $str = $m->getvar($varsr,$nom);
    return $arb if $arb and $str eq $m->{strs}->{$nom};
    $m->ustop(voknonfun => n => $nom) if !$str;
    $arb = $m->parse($str);
    $m->pstop('failed to parse string %t', t => $str) unless $arb;
    $m->{arbs}->{$nom} = $arb;
    $m->{strs}->{$nom} = $str;
    $arb;
};

=head3 Function getfun

arg1 is key, arg2 is local symbol table (hashref)

=cut

sub getfun {
    my ($m,$varsr,$nom) = shift->vpuc(@_);
    $m->getval($varsr,'funs',$nom);
}

=head3 Function maptrafo

Help deal with constructions like

    @tr = |pet|${TRIZ}|ahurl+${url}|

i.e. evaluate variables therein and allow recursion into multilayer structure on the last function argument, which is dealt with as C<lastfun> in C<pred>.

=cut

sub maptrafo { 
    my ($m,$varsr,@args) = @_;
    map { 'ARRAY' eq ref $_ ? [ $m->maptrafo($varsr, @$_) ] : $m->transform($_, $varsr) } @args;
};

=head3 Function getfuns

This takes the C<ahurl> verb in a construction like

   $(ahurl|http://a2e.de|Home)

or, after definition of

   oam2osm := $(fill oam2osmS,$(s|ahurl|../spez))

the locally defined C<s> verb in the value of C<oas2osmS> as first argument and returns a corresponding Perl function with its initial arguments.

    args := verb, vars
    verb #= 'ahurl' or 's' from examples above
    vars := hashref((nom => lst)+)
    rets := funkorp, funarg+
    funkorp := perl_function_code
    funarg := commandline arguments of funkorp that come before any user-supplied command arguments

=cut

sub getfuns {
    my ($m,$varsr,$verb) = shift->vpuc(@_);
    my $funkorp = $m->{funs}->{$verb};
    return $funkorp if 'CODE' eq ref $funkorp;
    my @funrest = ();
  GRUPVARS: { # e.g. set by file:/adv/perl/A2E/Tmplfil.pm.tmpl::sfx2fun
      last unless $varsr->{$verb} and 'ARRAY' eq ref $varsr->{$verb};
      @funrest = @{$varsr->{$verb}};
      last unless @funrest;
      $verb = shift @funrest;
      $verb = $m->{funs}->{$verb} unless ref $verb;
      last unless 'CODE' eq ref $verb;
      @funrest = $m->maptrafo($varsr, @funrest);
      return $verb, @funrest;
    };
  MINIVERB: {
      my $subverb = $m->{lihs}->{vrbs}->{$verb};
      last unless $subverb;
      @funrest = $m->maptrafo($varsr, @funrest);
      unshift @funrest, $subverb;
      return $m->{funs}->{call}, @funrest;
    };
    $funkorp = $m->{funs}->{$verb};
    @funrest = $m->maptrafo($varsr, @funrest);
    return $funkorp, @funrest if 'CODE' eq ref $funkorp;
    ($m->{funs}->{vok},$verb,@funrest);
};

=head2 INTERFACE

=head3 Function parse

analyse a string into a tree/arb

=cut

sub parse {
    my ($m,$s) = @_;
    $m->{chars} = [ unpack 'U*', $s ];
    $m->{len} = scalar @{$m->{chars}};
    my ($p, $arb) = $m->parse_from(0);
    $arb;
};


=head2 Rendering Functions

=head3 Function pred

Recurse into a potentially multilevel last argument as in

    @tr = |pet|${TRIZ}|ahurl+${url}|
    @shout = |underline|+boldface+ahurl/${url}+|

such that the text arguments of this function are applied to the inner-most level.

This is a subroutine of C<pred> which is supported by subroutine C<maptrafo> of C<getfuns>.

=cut

sub pred {
    my ($m,$varsr,$verb,@args) = shift->vpuc(@_);
    my ($fun,@funs) = $m->getfuns($varsr,$verb);
    my $res = '';
    if (ref $funs[$#funs]) {
	my $lastfun = pop @funs;
	$res = &$fun($m, $varsr, @funs, $m->pred($varsr, @$lastfun, @args));
    } else {
	$res = &$fun($m, $varsr, @funs, @args);
    };
    $res;
};

=head3 Function render

Render the syntactic tree arg2 into a string in the target format, using the local definitions found hashref arg1 as well as the global ones found in the namespaces $m->{lihs}->{vars}, $m->{arbs} and $m->{funs} of the object.

This is recursively invoked by many of the command verbs.

$vars stays a ref because needs to be returned.

=cut

sub render {
    my ($m,$varsr,$arb) = shift->vpuc(@_);
    return $arb unless ref $arb;
    my @args = eval { @$arb };
    return unless @args;
    $m->pred($varsr,@args); # render_pred
}

=head3 Function transform

Parse and render: like render, but starts from a string (in the source format) rather than a tree as first argument.
Do this only once.

=cut

sub vars_transform {
    my ($m,$varsr,$str,@args) = shift->vpuc(@_);my %vars = %$varsr;
    my $i = 0;
    foreach my $arg (@args) { 
	$vars{++$i} = $arg;
    };
    $vars{++$i} = undef;
    my $arb = $m->parse($str);
    $m->render(\%vars, $arb);
}
sub transform { 
    my ($m,$str,$gvars,@args) = @_;
    $m->vars_transform($gvars,$str,@args);
};

=head3 Function miniverb

Used in directives like

  itemlist_join = $(aresto|${BRA}list type=ul:|1|join|:call:ul_item|1|${ARB})
  special miniverb ul itemlist_join

to store the special short form 'ul' of verb 'itemlist_join' in a hashref $m->{miniverbs} that is
consulted (by getfuns) when a function is applied to text, allowing users to
put some special verbs into a namespace where they do not interfere with
lingual text chunks (lits).  The local miniverb definitions performed by the
fill function work in the same way and take precedence over the global miniverb
definitions performed here.  However the global miniverbs are pre-compiled into
a tree and thus execute faster.  This also means that in order to change the
verb's meaning later, the miniverb assignement also has to be invoked once
more.

This requires that function $m->{funs_intern}->{arb_args_call} is defined by the implementing library.

=cut

sub miniverb_put {
    my ($m,$prog,$mini,$verb,$lihs) = shift->ppuc('miniverb_put',@_);$lihs ||= $m->{lihs};
    return $m->dpop($prog) unless $mini and $verb;
    my $vrbs = $lihs->{vrbs};
    # $vrbp->{$mini} = 1; # cf file:/phm/19/07/03/sig/_lng.phm_pub_sig190703.txt::miniverb ; compile into listref form only when invoked by getfuns
    $m->pstop('bad miniverb significans %m', m => $mini) unless $mini =~ m(\A\w+\Z);
    $m->pstop('bad miniverb significatum %v', v => $verb) unless $verb =~ m(\A\w+\Z);
    $vrbs->{$mini} = $verb;
    $m->dpop($prog,$mini,$verb);
};
sub miniverb_get {
    my ($m,$prog,$mini,$lihs) = shift->ppuc('miniverb_get',@_);$lihs ||= $m->{lihs};
    my $verb = $lihs->{vrbs}->{$mini};
    $m->dpop1($prog,$verb);
};

=head1 BUGS, LIMITATIONS, TODO

Support for something like the following

		ul = <itemizedlist><title>${1}</title>${2@<listitem><para>@</para></listitem>}</itemizedlist>

will require an extension of &A2E::SArb::Make::pars_make, &A2E::SArb::Make::render_call and possibly &A2E::SArb::render, see A2E::SArb::Make(3) for more.

=head1 THE END

=cut

return 1;

# Local Variables:
# coding: utf-8
# srcfile: "/adv/perl/A2E/SArb.pm.tmpl"
# mode: perl
# End:
