# -*- coding: utf-8; srcfile:/adv/perl/A2E/SArb.pm; mode: perl -*-
# Time-stamp: <2018-01-23 09:03:35 phm>

package A2E::SArb::MLHT;
use strict;

=head1 NAME

A2E::SArb::MLHT - Parsing an MLHT string into a Syntactic Tree, recursively defined as a reference to a list of trees

=head1 SYNOPSIS

	use strict;
	use A2E::SArb::MLHT;
	use Data::Dumper;

	my $p = new_ready A2E::SArb::MLHT;

	my $s = 'an %(ad|%(utf8) %(ae|text|%(af|string))) with {ah dok=adv:bracketed substrings} at the end.';

	my $arb = $p->parse($s);
	print Dumper($arb) . "\n"; 

$arb is a syntactic tree, representing nested syntactic expressions, to be used by a rendering application.  Currently 'paren' (nested parentheses/quotes) and 'mlht' are supported by default.  A plugin for key value pairs 'keyvals' is available.

An example of a rendering application is A2E::Template::Plugin::MLHT(3).

=head2 Download Source

    http://a2e.de/adv/perl/A2E/SArb.pm

=head2 Prerequisites

    A2E::SArb

=cut

    use base qw(A2E::SArb);

=head2 Applications

	mlht(1)
	A2E::Template::Plugin::MLHT(3)

=head2 Syntax Documentation

=over
=item MLHT

    MLHT specification -- a format for minimised translatable hypertext
    http://a2e.de/adv/mlht

=back

=head2 Similar CPAN Modules

=over

=item Parser::Simple(3)

This allows plugins but is line-based and carries a lot of baggage that made it seem less likely to be efficient for the applications envisaged here.

=item A2E::Tratext(3)

These libraries were used for parsing Deplate text files and thus contain code that could easily be turned into a A2E::SArb::MLHT::Deplate plugin.

=back

=head1 BUGS

This has been working fine for a while in the A2E::Template::Plugin::MLHT(3) context, but right now it needs to be subjected to intensive testing in its newly repackaged form.

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

    Copyright (c) 2008 Hartmut Pilch (phm at a2e de)

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

=head2 Internal Constants

=cut

use Data::Dumper;

our $mlht_prefix = 37; # %
our $mlht_opener = 40; # (
our $mlht_closer = 41; # )
our $mlht_argsep = 124; # |

=head2 Overload Functions

=cut

sub sarb_mlht_defvars {
    my $m = shift;
    $m->konfayl('sarb_mlht.konf');
    return 1;
}

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::SArb::MLHT::defvars', @_);
    $m->SUPER::defvars;
    $m->sarb_mlht_defvars;
    $m->dpop('A2E::SArb::MLHT::defvars', 1);
    return 1;
};

sub sarb_mlht_lastkonfig {
    my $m = shift;
    $m->dpuc('sarb_mlht_lastkonfig', @_);
    $m->{konfig}->parsers('mlht');
    $m->dpop('sarb_mlht_lastkonfig', 1);
    return 1;
}

sub lastkonfig {
    my $m = shift;
    $m->dpuc('A2E::SArb::MLHT::lastkonfig', @_);
    $m->SUPER::lastkonfig;
    $m->sarb_mlht_lastkonfig;
    $m->dpop('A2E::SArb::MLHT::lastkonfig', 1);
    return 1;
}

sub sarb_mlht_postkonfig {
    my $m = shift;
    $m->dpuc('sarb_mlht_postkonfig', @_);
    $m->add_parser('mlht', sub { my $p = shift;return $m->pars_mlht($p) });
    $m->dpop('sarb_mlht_postkonfig', 1);
    return 1;
}

sub postkonfig {
    my $m = shift;
    $m->dpuc('A2E::SArb::MLHT::postkonfig', @_);
    $m->SUPER::postkonfig;
    $m->sarb_mlht_postkonfig;
    $m->dpop('A2E::SArb::MLHT::postkonfig', 1);
    return 1;
}


=head2 Plugin Functions

=cut

sub pars_mlht {
    my $m = shift;
    $m->dpuc('pars_mlht', @_);
    my $p = shift;
    my $p0 = $p;
    my $ch = $m->{chars}->[$p];
  NO_PREFIX: {
      $m->dpuc('no_prefix', [], p => $p, ch => $ch, mlht_prefix => $mlht_prefix);
      do { $m->dpop;last } if $ch == $mlht_prefix;
      $m->dpop('pars_mlht', $p);
      return $p; 
  };
    $p++;
    $ch = $m->{chars}->[$p];
  NO_OPENER: {
      $m->dpuc('no_opener', [], p => $p, ch => $ch, mlht_opener => $mlht_opener);
      do { $m->dpop;last } if $ch == $mlht_opener;
      $m->dpop('pars_mlht', $p0);
      return $p0; 
  };
    my $arb = [ 'mlht' ];
    $p++;
    $p0 = $p;
    $ch = 0;
    $m->dpuc('char0', $p, $m->{len});
  CHAR0: while (1) {
      die sprintf "no verb found at $p in\n\t%s\nwith tree\n\t%s", $m->string, Dumper($arb) unless $p < $m->{len};
      $ch = $m->{chars}->[$p];
      $m->dgot(p => $p, ch => $ch, chr => chr($ch));
    ARGSEP0: {
	$m->dpuc('argsep0', $mlht_argsep);
	do { $m->dpop;last } unless $ch == $mlht_argsep;
	my $s0 = $m->substring($p0, $p-1);
	push @$arb, $s0; 
	$p++;
	$p0 = $p;
	$m->dpop('char0', [], s0 => $s0, p => $p, p0 => $p0, arb => $arb );
	last CHAR0;
    };
    CLOSER0: {
	$m->dpuc('closer0', $mlht_closer);
	do { $m->dpop;last } unless $ch == $mlht_closer;
	my $s0 = $m->substring($p0, $p-1);
	push @$arb, $s0; 
	$m->dpop('pars_mlht', $p, $arb);
	return $p, $arb;
    };
  } continue { $p++ };
    my $sub_arb = [ 'concat' ];
    my $s0 = '';
    $m->dpuc('char1', $p, $m->{len});
  CHAR1: while (1) {
      die sprintf "mlht expression not closed at $p in\n\t%s\nwith tree\n\t%s", $m->string, Dumper($arb) unless $p < $m->{len};
      $ch = $m->{chars}->[$p];
      $m->dgot(p => $p, ch => $ch, chr => chr($ch));
    ARGSEP1: { 
	$m->dpuc('argsep1', $mlht_argsep);
	do { $m->dpop;last } unless $ch == $mlht_argsep;
	$arb = $m->add_subarb($p0, $p-1, $sub_arb, $arb);
	$p0 = $p + 1;
	$sub_arb = [ 'concat' ]; 
	$m->dpop('argsep1', $p);
	next CHAR1;
    };
    CLOSER1: {
	$m->dpuc('closer1', $mlht_closer);
	do { $m->dpop;last } unless $ch == $mlht_closer;
	$arb = $m->add_subarb($p0, $p-1, $sub_arb, $arb);
	$m->dpop('pars_mlht', $p, $arb);
	return $p, $arb;
    };
      ($p0, $p, $sub_arb) = $m->add_parse($p0, $p, $sub_arb);
  } continue { $p++ };
    die sprintf "WEIRD: mlht expression\n\t%s\n\twith tree\n\t%s\ndid not end in any of the known ways at $p.", $m->string, Dumper($arb);
};

=head1 THE END

=cut

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: /adv/perl/A2E/SArb/MLHT.pm.tmpl
# End:
