=head1 NAME

A2E::SArb::Make - Parsing an Make string into a Syntactic Tree, recursively defined as a reference to a list of trees

=cut 

package A2E::SArb::Make;

=head1 SYNOPSIS

	use A2E::SArb::Make;
        my $p = new_ready A2E::SArb::Make lihs => $lihs, gvars => $grupvars;
        print "RESULT: ", $p->transform('$(foreach pm,$(wildcard *.pm),$(basename $(pm)).html)'), "\n";

        => RESULT: Deplate.html Make.html MLHT.html


=head1 DESCRIPTION

Analyse variable assignments written in the syntax of makefiles written for GNU Make and produce a syntactic tree with unlimited nesting depth.
This is done in a two-step process, consisting of parsing an original string to produce a tree structure and rendering this tree structure into the resulting string.   The A2E::SArb module is used as a basis for parsing and rendering, so that more parsing and rendering functions can be plugged in.

=head1 RELATED DOCUMENTS

=head2 Download Source

    http://a2e.de/adv/perl/A2E/SArb/Make.pm.tmpl

=head2 Prerequisites

    A2E::SArb(3)

Inherits options from there.

=cut

use strict;
use base 'A2E::SArb', 'A2E::Hannum';
use Data::Dumper;
use Scalar::Util qw(looks_like_number reftype blessed);
use POSIX qw(strftime); # TODO: move everyting that requires this to A2E::Hannum
#( needed by monwoks? maybe strftime is enough
#use Date::Simple;
#)
use utf8;

our @double_width_numbers = ('０', '１', '２', '３', '４', '５', '６', '７', '８', '９');
our %MSGS = (
    chkaaaa => '%p argument %a is not a year number',
    chkmm => '%p argument %m is not a month number',
    chkdd => '%p argument %d is not a day number',
    chkww => '%p argument %w is not a week number');
=head2 Applications

	A2E::Tmplfil(3)

=head2 Syntax Documentation

See the TeXinfo manual of GNU Make.

We aim to implement most of what can appear in assignment statements.

Currently supported functions include 

=over 1

=item call

=item shell

=item wildcard

=item foreach

=item if

=item ifeq

extended to use a perl regexp for comparison.
the regexp is anchored at beginning and end.
the literal string is just the special case of a plain regexp.

=item ifneq

extended like ifeq.

=item value

=item basename

=item notdir

=item subst

=back

We are also adding some functions are not found in GNU Make and used mainly for the purpose of separating the layers of natural language text so as to facilitate maintenance of multilinguality.

=over 1

=item fill

Fill the slots in a dynamic text variable with static and dynamic variables.  Example:

    feast = let's feed the $(pl ${1}), $(pl ${2}) and other $(3 ${adj}) $(pl ${noun}) in ${locus}!
    _my_party = $(fill feast,cat,dog,:call:italic,adj=lovely,pl=:call:plural,noun=pet,locus=my house)
    _your_party = $(fill feast,tiger,snake,:call:hyperref:${wildlife_url},adj=wild,pl=:call:plural,noun=beast,$(locus:your backyard))
    plural = ${1}s
    italic = <em>${1}</em>
    wildlife_url := http://en.wikipedia.org/wiki/Wildlife
    hyperref = <a href="${1}">${2}</a>

For any dynamic variables that begin with '_', corresponding static variables of the same name without initial '_' are created after all template variable files have been read and shortly before the templates are processed.

=item case

Simple array-like statement, see case statement of bash(1), but extended like ifeq to use Perl regular expressions for matching.

    wday = $(case ${1},Noday,0,Sunday,1,Monday,2,Tuesday)
    TUESDAY := $(call wday,2)
    month = $(case ${1},Anymonth,01,January,02,February,03,March, ....)
    MARCH := $(call month,03)
    ddmax = $(case ${1},30,(1|3|5|7|8|10|12),31,2,28)
    LASTMARCHDAY := $(call ddmax,3)

The second argument ('Noday' and 'Anymonth' in the examples above) gives the value that is returned when none of the keys matches.
The uneven-numbered arguments from nr 3 onwards are regular expressions which argument nr 1 must completely match, e.g. in the case of the 'ddmax' (maximum number of days in a given month) function above, the return value (number of days) becomes '31' if the month argument '${1}' matches '\A(1|3|5|7|8|10|12)\Z', '28' if it matches '\A2\Z', and '30' in all other cases.

=item sprintf

like the sprintf function of C, Perl etc, e.g.

    twodigitnum = $(sprintf %02d,${1})

=item or

as in programming languages: 

    _tit = $(or ${oas_int_tit},$(call _oas_int_tit))

The second expression is evaluated only if the first one is null.
Any number of expressions is acceptable.

=item and

as in programming languages: 

    _tit = $(and ${oas_int_tit},$(call _oas_int_tit))

The second expression is evaluated only if the first one is not null.
Any number of expressions is acceptable.

=item rest, restcall

allow processing of an infinite number of remaining arguments, see examples below

=item join

    list := |join|, |a|b|c|

    => a, b, c

Note that in our Make language we are free to use most ascii symbols as argument separators, as long as we
stick to a single one within one expression.

=item substr

substring as in perl

=item minim

choose smallest of the numeral arguments

=item maxim

choose largest of the numeral arguments

=item plus

add args

=item minus

subtract args 2ff from arg1 

=item mult

multiply arguments

=item div

divide arg1 by arg2 [by arg3 ...]

=item mcall

like call, but obtain arguments by matching regex of arg 1 to arg 2.

    keyval := [ ${1} ]: ${2}
    $(mcall keyval,\A(\w+)\:\s(\w+),animal: cat) 

=back

=head3 Adaption to target markup languages

The $m->{tmplangdir} directory defined in A2E::Tmplkonf should also contain vocabulary files that specify certain specific functions that depend on the language-neutral output format (e.g. deplate or xml markup), e.g.

    multemp/lang.txt

		q = {dquote:${1}}

    multemp/lang.zh.txt

                q = 「${1}」


    my_xml/lang.txt

		q = <dquote>${1}</dquote>

so that certain below-sentence-level markup structures can be used within vocabulary building files.

=head3 Processing of remaining arguments with $(rest ...)

Here it becomes necessary to allow a rest parameter, e.g. specifiable as below

    plainlist = <list>$(rest)</list>
    user_groups = ${1}: $(rest|2|comma_join)
    ul = <itemizedlist><title>${1}</title>$(rest|2|lines|:call:listitem)</itemizedlist>
    dl = <variablelist><title>${1}</title>$(rest|2|lines|/call/descitem)|2)</variablelist>

The rest function iterates over the remaining arguments, beginning at position
$arg1 = 2, joins them together with function $arg2 = 'lines', taking $arg4
('1' in unordered list 'ul', 2 in description list 'dl') item-arguments off
the argument list for each item and feeding them as to the item generator
function $arg3 = 'descitem'.

The functions are defined by the user as follows, using the builtin $<list> constructor:

 comma_join := /join/, /
 LF := $(char 10)
 lines := /list/join/${LF}/
 descitem = <varlistentry><term>${1}</term><desc>${2}</desc></varlistentry>

The above definition of 'dl' can be simplified to

    dl = <variablelist><title>${1}</title>$(rest|2|lines|descitem|2)</variablelist>

In other words, C<$(list|call|something)> can be replaced with C<something> in this context.
Since the C<${lines}> function chain does not contain a C<call>, it can not be simplified in this way.

A list type variable like here C<lines> is normally not quoted in scalar context but only
used as an argument in C<rest> constructions and more directly with C<chain>,
see example C<uno_due_tre> below.  

The functions C<char> and C<join> are builtins written in Perl, see render_char and render_join.

All arguments of 'rest' have default values:

    arg1 : 1
    arg2 : space_join
    arg3 : 1
    arg4 : plain_string

Another useful builtin joining function is 'jung', i.e. enumerative conjunction:

    et_list := |jung|, | and |
    uno_due_tre := $(chain|${et_list}|1|2|3)

By invoking ${uno_due_tre} we would get I<1, 2 and 3>.

The C<chain> mechanism gives us a second means of defining a simple function,
beside the $(call ...) mechanism.  This type of function is particularly
useful as an auxiliary function in callable functions that process an infinite
trailing argument list with $(rest ...).

Functions like I<${et_list}> differ from language to language but should be kept
independent of any peculiarities of the markup language.  If one formatter has
a different way of expressing the enumerative comma, then that should be
defined as a variable in the $m->{tmplangdir}/lang*.txt file, whereas definitions
like et_list belong into $m->{doktmpldir}/lang*.txt files or individual template
files.

=head1 OPTIONS 

Inherits from the Prerequisites.    

=head1 BUGS

As of 2008-05-26, intensive unit testing with A2E::SArb::MakeTest has just begun, many bugs are likely to appear.

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

    Copyright (c) 2008 Hartmut Pilch (phm at a2e de)

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 SETUP

=cut

sub sarb_make_defvars {
    my ($m,$prog) = shift->qpuc('sarb_make_defvars',@_) or return;

=head2 Object Variables

=over

=item sddelta: normal difference between double-width equivalents and ascii characters

=cut

    $m->{sddelta} = ord('０') - ord('0');

=item --dwchars

individual mappings from single width to double-width characters (normally mapping is done by addition of sddelta)

=cut

    $m->{konfig}->define('dwchars=s%', { DEFAULT => { 32 => '　' } });

=back

=head2 Configuration Files

        /etc/opt/a2e/sarb_make.konf
	~/.sarb_make_konf
	@sarb_make

=cut

    $m->konfayl('sarb_make.konf');
    $m->prog_msgdefs('sarb_make', %MSGS);
    $m->qpop1($prog,1);
};

sub defvars {
    my ($m,$prog) = shift->qpuc('A2E::SArb::Make::defvars',@_) or return;
    $m->SUPER::defvars;
    $m->sarb_make_defvars;
    $m->hannum_defvars;
    $m->qpop1($prog,1);
}

sub sarb_make_lastkonfig {
    my ($m,$prog) = shift->qpuc('sarb_make_lastkonfig', @_) or return;
    $m->{konfig}->parsers('make');
    $m->qpop($prog,1);
};

sub lastkonfig {
    my ($m,$prog) = shift->qpuc('A2E::SArb::Make::lastkonfig',@_) or return;
    $m->SUPER::lastkonfig;
    $m->sarb_make_lastkonfig;
    $m->dpop1($prog,1);
}

sub sarb_make_postkonfig {
    my ($m,$prog) = shift->qpuc('sarb_make_postkonfig', @_);
    $m->{konfig}->parsers('make');
    $m->{parsers}->{make}->{prefix} = 36; # '$'
    # $m->{parsers}->{make}->{listprefix} = 64; # '@'
    my @openers = (40, 123);
    $m->{parsers}->{make}->{openers} = \@openers;
    $m->{parsers}->{make}->{openerp}->{$_} = 1 for @openers;
    $m->{parsers}->{make}->{argsep0} = 32; 
    $m->{parsers}->{make}->{argsep1} = 44; # ','
    $m->{parsers}->{make}->{inaktiv} = 0;
=pod

argument separators

32, !33, "34, #35, $36, %37, &38, '39, (40, )41, *42, +43, ,44, -45, .46, /47, [ 0..9 ] , :58, ;59, <60, =61, >62, ?63, @64, [ A..Z ], [91, \92, ]93, ^94, _95, `96, [ a..z ], {123, |124, }125, ~126

=cut
    my @argseps = unpack 'U*', ' 	!#%&*+,-./:;=?^|~';
    $m->{parsers}->{make}->{argseps} = \@argseps;
    $m->{parsers}->{make}->{argsepp}->{$_} = 1 for @argseps;
    my @argseps_min = unpack 'U*', '|:/+,;*';
    $m->{parsers}->{make}->{argseps_min} = \@argseps_min;
    $m->{parsers}->{make}->{argsepp_min}->{$_} = 1 for @argseps_min;
    $m->{funs_intern} = {};
    $m->{funs_intern}->{arb_args_call} = \&arb_args_call;
    $m->add_renfun(make => \&render_make);
    $m->add_renfun(shell => \&render_shell);
    $m->add_renfun(f2s => \&render_f2s);
    $m->add_renfun(cat => \&render_f2s);
    $m->add_renfun(wildcard => \&render_wildcard);
    $m->add_renfun(foreach => \&render_foreach);
    $m->add_renfun(or => \&render_or);
    $m->add_renfun(and => \&render_and);
    $m->add_renfun(if => \&render_if);
    $m->add_renfun(ifn => \&render_ifn);
    $m->add_renfun(defp => \&render_defp);
    $m->add_renfun(ifeq => $m->make_ifcmp_fun('eq'));
    $m->add_renfun(ifneq => $m->make_ifcmp_fun('ne'));
    $m->add_renfun(ifgt => $m->make_ifcmp_fun('gt'));
    $m->add_renfun(ifge => $m->make_ifcmp_fun('ge'));
    $m->add_renfun(iflt => $m->make_ifcmp_fun('lt'));
    $m->add_renfun(ifle => $m->make_ifcmp_fun('le'));
    $m->add_renfun(ifmr => $m->make_ifcmp_fun('mr'));
    $m->add_renfun(progn => \&render_progn);
    $m->add_renfun(let => \&render_progn);
    $m->add_renfun(set => \&render_set);
    $m->add_renfun('shift' => \&render_shift);
    $m->add_renfun(catch => \&render_catch);
    $m->add_renfun(throw => \&render_throw);
    $m->add_renfun('last' => \&render_throw);
    $m->add_renfun('next' => \&render_next);
    $m->add_renfun('while' => \&render_while);
    $m->add_renfun('until' => \&render_until);
    $m->add_renfun(value => \&render_value);
    $m->add_renfun(case => \&render_case);
    $m->add_renfun(call => \&render_call);
    $m->add_renfun(subcall => \&render_subcall);
    $m->add_renfun(restcall => \&render_restcall);
    $m->add_renfun(chain => \&render_chain);
    $m->add_renfun(rest => \&render_rest);
    $m->add_renfun(apply => \&render_apply);
    $m->add_renfun(aresto => \&render_aresto);
    $m->add_renfun(fill => \&render_fill);
    $m->add_renfun(basename => \&render_basename);
    $m->add_renfun(sprintf => \&render_sprintf);
    $m->add_renfun(substr => \&render_substr);
    $m->add_renfun(strlen => \&render_strlen);
    $m->add_renfun(lrtrim => \&render_lrtrim);
    $m->add_renfun(ucfirst => \&render_ucfirst);
    $m->add_renfun(uc1 => \&render_ucfirst);
    $m->add_renfun(uc => \&render_uc);
    $m->add_renfun(lc => \&render_lc);
    $m->add_renfun(lcfirst => \&render_lcfirst);
    $m->add_renfun(lc1 => \&render_lcfirst);
    $m->add_renfun(subst => \&render_subst);
    $m->add_renfun(strip => \&render_strip);
    $m->add_renfun(mcall => \&render_mcall);
    $m->add_renfun(revargs => \&render_revargs);
    $m->add_renfun(modnum => \&render_modnum);
    $m->add_renfun(numpad => \&render_numpad);
    $m->add_renfun(dnumpad => \&render_dnumpad);
    $m->add_renfun(dwchars => \&render_dwchars);
    $m->add_renfun(map => \&render_map);
    $m->add_renfun(mapconc => \&render_map);
    $m->add_renfun(jung => \&render_jung);
    $m->add_renfun(char => \&render_char);
    $m->add_renfun(notdir => \&render_notdir);
    $m->add_renfun(fromto => \&render_fromto);
    $m->add_renfun(ddmax => \&render_ddmax);
    $m->add_renfun(mondags => \&render_mondags);
    $m->add_renfun(wokdags => \&render_wokdags);
    $m->add_renfun(monwoks => \&render_monwoks);
    $m->add_renfun(wokmon => \&render_wokmon);
    $m->add_renfun(plus => \&render_plus);
    $m->add_renfun(minus => \&render_minus);
    $m->add_renfun(mult => \&render_mult);
    $m->add_renfun(div => \&render_div);
    $m->add_renfun(intdiv => \&render_intdiv);
    $m->add_renfun(mod => \&render_mod);
    $m->add_renfun(minim => \&render_minim);
    $m->add_renfun(maxim => \&render_maxim);
    $m->add_renfun(strfamd => \&render_strfamd);
    $m->add_renfun(folnum => \&render_folnum);
    $m->add_renfun(hannum => \&render_hannum);
    $m->add_renfun(inaktiv => \&render_inaktiv);
    $m->add_renfun(list => \&render_list);
    $m->add_renfun('@', \&render_list);
    $m->add_renfun(hash => \&render_hash);
    $m->add_renfun('%', \&render_hash);
    $m->add_parser(make => sub { my $p = shift;return $m->pars_make($p) });
    $m->{opfuns} = { 
	lt => sub { $_[0] < $_[1] }, 
	le => sub { $_[0] <= $_[1] },
	gt => sub { $_[0] > $_[1] },
	ge => sub { $_[0] >= $_[1] },
	eq => sub { $_[0] eq $_[1] },
	ne => sub { $_[0] ne $_[1] },
	mr => sub { $_[0] =~ m($_[1]) }, 
	eqn => sub { $_[0] == $_[1] },
	nen => sub { $_[0] != $_[1] },
	'=~' => 'mr',
	'<' => 'lt',
	'<=' => 'le',
	'>' => 'gt',
	'>=' => 'ge',
	'==' => 'eqn',
	'!=' => 'nen' };
    $m->qpop($prog,1);
}

sub postkonfig {
    my ($m,$prog) = shift->qpuc('A2E::SArb::Make::postkonfig', @_);
    $m->SUPER::postkonfig;
    $m->sarb_make_postkonfig;
    $m->qpop1($prog,1);
}

=head2 Parsing Functions

=head3 Function pars_make

Return a new position $p and a resulting tree $arb if a variable/function call such as $(...) ${....} as known from the Makefile language starts here.

=cut

sub pars_make {
    my ($m,$prog,$p) = shift->ppuc('pars_make',@_);
    # die "argseps not defined" unless grep { $_ == 47 } @{$m->{parsers}->{make}->{argseps}}; # '/' must be in argument separators
    my $p0 = $p;
    my $ch = $m->{chars}->[$p];
    my $argsep = $m->{parsers}->{make}->{argsep0};
  NO_PREFIX: {
      my $ok = $m->{parsers}->{make}->{inaktiv} || $ch != $m->{parsers}->{make}->{prefix}; # verbatim mode or no '$' found
      last unless $ok;
      # LIST_PREFIX: { last unless $ch == $m->{parsers}->{make}->{listprefix} };
      return $m->dpop1($prog,$p);
    };
    $p++;
    $ch = $m->{chars}->[$p];
  PREFIX: {
      my $ok = $ch == $m->{parsers}->{make}->{prefix}; # $$
      last unless $ok;
      my $chr = chr $ch;
      # my $arb = [ 'concat', $chr ];
      return $m->dpop($prog, $p, $chr);
    };
  NO_OPENER: {
      my $ok = not $m->{parsers}->{make}->{openerp}->{$ch}; # grep { $ch == $_ } @{$m->{parsers}->{make}->{openers}};
      $m->dpuc('no_opener', $ok, $ch, $p);
      do { $m->dpop;last } unless $ok;
      return $m->dpop1($prog,$p0);
  };
    my $opener = $ch;
    my $closer = $m->{opener_closers}->{$opener};
    my $arb = [ 'make', $opener ];
    $p++;
    $p0 = $p;
    $ch = 0;
  CHAR0: while (1) { # find command verb
      $m->pstop('no verb found at position %p in string\n\t%t\nwith tree\n\t%A', p => $p, t => $m->string, A => $arb) unless $p < $m->{len};
      $ch = $m->{chars}->[$p];
    ARGSEP0: {
	my $ok = $m->{parsers}->{make}->{argsepp}->{$ch};
	last unless $ok;
	$argsep = $ch == $argsep ? $m->{parsers}->{make}->{argsep1} : $ch;
	my $s0 = $m->substring($p0, $p-1);
	push @$arb, $s0; 
	$m->{parsers}->{make}->{inaktiv} = 1 if $s0 eq 'inaktiv';
	$p++;
	$p0 = $p;
	last CHAR0;
      };
    CLOSER0: {
	my $ok = $ch == $closer;
	last unless $ok;
	my $s0 = $m->substring($p0, $p-1);
	push @$arb, $s0; 
	$m->{parsers}->{make}->{inaktiv} = 0;
	return $m->dpop($prog, $p, $arb);
      };
  } continue { $p++ };
    my $sub_arb = [ 'concat' ];
    my $s0 = '';
  CHAR1: while (1) {
      $m->ustop("make expression not closed at $p in\n\t%s\nwith tree\n\t%a", s => $m->string, a => $arb) unless $p < $m->{len};
      $ch = $m->{chars}->[$p];
    ARGSEP1: { 
	my $ok = $ch == $argsep;
	last unless $ok;
	$arb = $m->add_subarb($p0, $p-1, $sub_arb, $arb);
	$p0 = $p + 1;
	$sub_arb = [ 'concat' ]; 
	next CHAR1;
      };
    CLOSER1: {
	my $ok = $ch == $closer;
	last unless $ok;
	$arb = $m->add_subarb($p0, $p-1, $sub_arb, $arb);
	$m->{parsers}->{make}->{inaktiv} = 0;
	return $m->dpop($prog, $p, $arb);
      };
      ($p0, $p, $sub_arb) = $m->add_parse($p0, $p, $sub_arb);
  } continue { $p++ };
    $m->pstop("make expression\n\t%s\n\twith tree\n\t%a\ndid not end in any of the known ways at $p.", s => $m->string, a => $arb);
};

=head2 Rendering Subroutines

=head3 Function fillgetfuns

Information about function and its initial arguments to be stored in local $vars buffer of render routine and used for expansion of dynamic elements.

Consists of function in final coderef form, defaulting to coderef of the 'concat' function, and ensuing arguments in unrendered form.  Rendering is left to the functions.

=cut

sub fillgetfuns {
    my ($m,$varsr,$verb,@args) = shift->vpuc(@_);
    my @funs = $m->getfuns($varsr,$verb);
    push @funs, @args;
    @funs;
};

=head3 Functions fillsetvar_from_string

Parse an argument of the $(fill bodyvar,arg1,arg2,...) function.
Take a position number and a string found at that position, e.g.

    0, 'value'
    1, 'ah=:call:ahurl:myfile.txt'

Return a position number (incremented in case our argument was positional), a key and a value, e.g.

    1, '0', 'value' 
    2, 'ah', [ sub { return call }, 'ahurl', 'myfile.txt' ]

The key is the name of an argument that may appear in bodyvar.
The value is either a string (static) or a reference to a list starting with a perl function from $m->{funs} and optionally continuing with initial arguments to be passed to that function before any arguments from bodyvar are taken.
All arguments are strings, any expansion must have been done before C<fill> does its work.

=cut

sub fillsetvar_from_string {
    my ($m,$varsr,$pos,$s) = shift->vpuc(@_);$pos ||= 0;local $_ = $s;
    # flat syntax 'dog', ':call:italic', 'noun=pet', 'pl=:call:plural', 'pl=:call:mailto:user@host'
    my ($key, $val) = m(\A(\w+)\=(.*)\Z);
  POSARG: { # positional argument
      last if $key;
      $pos++;
      $key = qq{$pos};
      $val = $_;
    };
    my $sep = substr $val, 0, 1;
  STATVAL: { # static argument: dog, pet, etc  
      my $sepch = ord $sep;
      last if $m->{parsers}->{make}->{argsepp}->{$sepch}; # grep { $_ == $sepch } @{$m->{parsers}->{make}->{argseps}};
      return $pos, $key, $val;
    };
    my @toks = $m->textsep2seplist($val, $sep);
    $sep = shift @toks;
    @toks = $m->fillgetfuns($varsr,@toks);
    ($pos, $key, \@toks);
};

=head3 Function fillsetvar_from_arb

Same as fillsetvar_from_string, but, if argument 1 is written in the safe way of

    $(fill text,$(pg|call|ah|${url}))

then it is split before expansion of the contents, so that splitting works
correctly no matter which nefarious characters may be contained in the
expanded value of ${url}.

If argument 1 is static or written in the simple unsafe way

    $(fill text,pg=|call|ah|${url}|)

then it is parsed usint fillsetvar_from_string

=cut

sub fillsetvar_from_arb {
    my ($m,$varsr,$pos,$arb) = shift->vpuc(@_);
    $m->pstop('invalid arb %a', a => $arb) unless 'ARRAY' eq ref $arb; 
    my ($key,$val) = ('','');
    if ($arb->[0] eq 'concat' and 'ARRAY' eq ref $arb->[1] and $arb->[1]->[0] eq 'make' and $arb->[1]->[1] eq 40) {
	($key, $val) = $m->fillsetvar_from_makearb($varsr,$arb->[1]);
    } else {
	($pos, $key, $val) = $m->fillsetvar_from_string($varsr,$pos,$m->render($varsr,$arb));
    };
    ($pos, $key, $val);
};

sub fillsetvar_from_makearb {
    my ($m,$varsr,$arb) = shift->vpuc(@_);
    my @toks = @$arb;
    shift @toks for 0..1; # remove 'make' and 40
    @toks = map { $m->render({}, $_) } @toks;
    my $key = shift @toks; # 'pg'
    my $dyna = $toks[0] ? 1 : 0; # 'call' : ''
    shift @toks unless $dyna;
    @toks = $m->fillgetfuns($varsr,@toks); # [ sub { call }, 'ahurl', 'http://#$%()*#!|||' ]
    my $val;
    if ($dyna) {
	$val = \@toks;
    } else {
	my $verb = shift @toks;
	$val = &$verb($m, {}, @toks);
    };
    ($key, $val);
};


=head2 Top level Rendering Functions

Corresponding to functions used in the A2E GNU Make language that we are parsing.

Named 'render_I<function>',  plugged in in constructor part 'postkonfig'.

=head3 Function render_fill

    _my_party = $(fill feast,cat,dog,$(call|italic),adj=lovely,$(pl|call|plural),noun=pet,$(locus|my house))
    _your_party = $(fill feast,tiger,snake,:call:hyperref:${wildlife_url},adj=wild,pl=:call:plural,noun=beast,locus=your backyard)
    

=cut

sub render_fill { 
    my ($m,$varsr,$body,@args) = shift->vpuc(@_);my %vars = %$varsr;
    my $nom = $m->render({}, $body);
    my $arb = $m->getarb($varsr,$nom);
    my $pos = 0;
    foreach my $arg (@args) { ($pos,my $key,my $val) = $m->fillsetvar_from_arb(\%vars,$pos,$arg);$vars{$key} = $val };
    $m->render(\%vars,$arb);
};

sub render_make { 
    my ($m,$varsr,$opener,$body,@args) = shift->vpuc(@_);
    if ($opener == 40) { # '('
	$body = $m->render($varsr,[ $body, @args ])
    } else { 
	$body = $m->getvar($varsr,$body);
    };
    $body;
}

=head4 $m->arb_args_call($arb,$varsr,@args), $m->vars_arb_args_call($varsr,$arb,@args)

Set arguments to variables '1', '2', '3' etc; 
Call a tree-compiled function textchunk on arguments i.e. on \%vars, not $varsr.
Subroutine of $(call ...) et al.

The old versions without the vars_ prefix are deprecated.

=cut

sub vars_arb_args_call {
    my ($m,$varsr,$arb,@args) = shift->vpuc(@_);my %vars = %$varsr;
    my $i = 0;foreach my $arg (@args) { $vars{++$i} = $arg };$vars{++$i} = undef;
    $m->render(\%vars, $arb);
};
sub arb_args_call { my ($m,$arb,$varsr,@args) = @_;$m->vars_arb_args_call($varsr,$arb,@args) };

=head4 Function arb_args_subcall, vars_arb_args_subcall

Is to arb_args_call what subcall is to call: reuse arguments from superior level in addition to arguments of this level.

=cut

sub vars_arb_args_subcall {
    my ($m,$varsr,$arb,@args) = shift->vpuc(@_);my %vars = %$varsr;
    my $n = 0;while (exists $vars{++$n}) { push @args, $vars{$n} };
    $n = 0;while (@args) { $vars{++$n} = shift @args };
    $m->render(\%vars, $arb);
}
sub arb_args_subcall { my ($m,$arb,$varsr,@args) = @_;$m->vars_arb_args_subcall($varsr,$arb,@args) };
sub vars_nom_args_call {
    my ($m,$varsr,$nom,@args) = shift->vpuc(@_);my %vars = %$varsr;
    my $arb = $m->getarb($varsr,$nom);
    $m->vars_arb_args_call($varsr,$arb,@args);
}
sub nom_args_call { my ($m,$arb,$varsr,@args) = @_;$m->vars_nom_args_call($varsr,$arb,@args) };
sub vars_nom_args_subcall { 
    my ($m,$varsr,$nom,@args) = shift->vpuc(@_);my %vars = %$varsr;
    my $arb = $m->getarb($varsr,$nom);
    $m->vars_arb_args_subcall($varsr, $arb, @args);
}
sub nom_args_subcall { my ($m,$nom,$varsr,@args) = @_;$m->vars_nom_args_subcall($varsr,$nom,@args) };
sub vars_val_args_call { 
    my ($m,$varsr,$val,@args) = shift->vpuc(@_);my %vars = $m->get_vars($varsr);
    my $arb = $m->parse($val);
    $m->vars_arb_args_call($varsr,$arb,@args);
};
sub val_args_call { 
    my ($m,$val,$varsr,@args) = @_;
    $m->vars_val_args_call($varsr,$val,@args) };

=head4 $sarb->render->call(\%vars,$verb,@args) => $body

=cut
sub render_call {
    my ($m,$varsr,$verb,@args) = @_;my %vars = $m->ref2hash($varsr);
    $verb = $m->render($varsr, $verb); # [ 'concat', 'emph' ] => 'emph'
    @args = map { $m->render($varsr, $_) } @args;
    $m->nom_args_call($verb,$varsr,@args);
};

=head4 Function render_subcall

Like render_call but in addition to own arguments take arguments from superior level.

Used for defining a combination of an existing function with specific arguments as a new function:

    MapVel = $(subcall MapJung,${WSP}${jungvel}${WSP})

=cut

sub render_subcall { 
    my ($m,$varsr,$verb,@args) = @_;my %vars = $m->get_vars($varsr);
    $verb = $m->render($varsr, $verb); # 'emph'
    @args = map { $m->render($varsr, $_) } @args;
    $m->nom_args_subcall($verb,$varsr, @args);
}

=head4 Function render_restcall

Make it possible to call a dynamic variable with an infinite argument list, e.g.

    ol = $(rest|1|join+${LF}|call+ol_item)
    lnol = ${1}${LF}$(restcall 3,ol,my first item is ${2})

meaning that dynavar ol should be call'ed with the remaining arguments, starting from 3, processed after the explicitely listed arguments, such as here I<my first item is ${2}> have been processed.
In the future we may discard this and use $(progn ...) instead, see example in Function render_progn.

=cut

sub render_restcall { 
    my ($m,$varsr,$narg,$var,@args) = @_;my %vars = $m->get_vars($varsr);
    $narg = $m->render({}, $narg);
    $var = $m->render({}, $var); # [ 'concat', 'ol' ] => 'ol'
    @args = map { $m->render({}, $_) } @args;
    while (1) {
	last unless exists $vars{$narg};
	push @args, $vars{$narg};
    } continue { $narg++ };
    $m->dgot(args => \@args, narg => $narg);
    my $arb = $m->getarb($varsr,$var);
    my $n = 0;foreach my $arg (@args) { $vars{++$n} = $arg };delete $vars{++$n};
    $m->render(\%vars,$arb);
}

=head4 Function render_jung

for use in enumerations with special joiner for last argument like 'and' or 'or'.

    kai_verb = $(rest|1|:jung:${jungkom}${WSP}:${WSP}${jungkai}${WSP}) 

=cut

sub render_jung {
    my ($m,$varsr,$sep,$jung,@args) = shift->vpuc(@_);$sep = $m->render($varsr,$sep) || ', ';$jung = $m->render($varsr,$jung) || ' & ';
    my $body = join $sep, map { $m->render($varsr,$_) } @args[0..$#args-1];
    join $jung, $body, $m->render($varsr,$args[$#args]);
}

sub render_char {
    my ($m,$varsr,$num) = shift->vpuc(@_);
    $_ = $m->render($varsr,$num);
    chr $_;
};

sub get_map_arg_grups {
    my ($m,$gsz,@args) = @_;
    my $map_arg_grups = [];
    while (@args) {
	die sprintf 'too few arguments, need %d, have only %d left, namely %s', $gsz, @args, join(' ', @args) if @args < $gsz; 
	my $map_args = ();
	foreach my $n (1..$gsz) { push @$map_args, shift @args };
	push @$map_arg_grups, $map_args;
    };
    $map_arg_grups;
}

=head3 Function render_chain

Call a function chain rather than just a function on the arguments.

=cut

sub render_chain {
    my ($m,$varsr,$var) = shift->vpuc(@_);
    my $join_chain_tok = $m->render($varsr,$var);
    $m->pstop('no join chain token found in structure %S based on template variables %V', S => $var, V => $m->{lihs}->{vars}) unless $join_chain_tok;
    my $join_chain_ref = $m->textsep2listref($join_chain_tok);
    my @join_chain = eval { @$join_chain_ref };
    $m->pstop('no join chain found in chain argument %c', c => $join_chain_tok) unless @join_chain;
    my $join_car = shift @join_chain;
    my @args = @_;
    my $join_fun = $m->{funs}->{$join_car} or $m->pstop('no function found for join_car %j', j => $join_car);
    &$join_fun($m,$varsr,@join_chain, @args);
};

=head3 Function textsep2seplist

Take a $text argument like

    'foo, ,'
    'foo,+'
     '|foo|+ |'

with an optional second separator argument,
and return a list like

    ( ',', 'foo', ' ' )
    ( ',', 'foo', '+' )
    ( '|', 'foo', '+ ' )

whose first element is the separator that was actually found, followed by the identified list members.
In no separator was found, return '' and $text.

Empty space after a last separator is ignored.
Initial empty space is ignored, trailing empty space not, unless after the last separator nothing but empty space is found.

If a separator argument is given, other separators are ignored.  Otherwise the first viable separator is chosen and thereafter any others are ignored.

%opts
    separen: expression must start with separator
    noblanksepp: blank char 32 does not count as separator

BUG: file:/phm/18/09/sig/_org.phm_pub_sig1809.txt::textsep

=cut

sub textsep2seplist {
    my ($m,$text,$sep,%opts) = @_;$text ||= '';$sep ||= '';
    @{$m->{chars}} = unpack 'U*', $text;
    my $p = 0;
    my $ch = 0;
    $p++ while $m->{blankp}->{$m->{chars}->[$p]}; # grep { $ch = $m->{chars}->[$p] || 0;$_ == $ch } @{$m->{blanks}};
    my $p0 = $p;
    my $isep = !$sep ? 0 : ord $sep; 
    my %sepp = $isep ? ($isep => 1) : %{$m->{parsers}->{make}->{argsepp}};
    my @list = ();
  SET_SEP: { # if the first character is a separator, set it and advance; enforce requirement that it must be i.e. $opts{separen}
      my $ch = $m->{chars}->[$p] || 0;
    SEPP: { # 
	last if $sepp{$ch};
	do { $m->decho('initial char %c of following text is not a separator: %t', c => $ch, t => $text);return '', $text } if $opts{separen};
	last SET_SEP;
      };
      do { $m->decho('following text starts with wrong separator %c rather than %e: %t', t => $text, c => $ch, e => $sep);return '', $text } if $opts{separen} and $sep and $ch ne $isep;
      $isep ||= $ch;
      $p++;
      $p0++;
    };
    my $eos = sub {
	return unless $p > $#{$m->{chars}};
      PUC: {
	  last unless $p > $p0;
	  my $non_blank = 0;
	  foreach my $ch (@{$m->{chars}}[$p0..$p-1]) {
	      next if $m->{blankp}->{$ch}; # grep { $_ eq $ch } @{$m->{blanks}};
	      $non_blank = $ch;
	      last;
	  };
	  last unless $non_blank;
	  push @list, $m->substring($p0, $p-1);
	  $p0 = $p;
	};
	1;
    };
    my $pucstr = sub {
	push @list, $m->substring($p0, $p-1);
	$p0 = $p + 1;
	1;
    };
  NOSEP: { # finding the separator after the first symbolword
      last if $isep; # is found
      $p++;
      %sepp = %{$m->{parsers}->{make}->{argsepp}};
      $sepp{32} = 0 if $opts{noblanksepp};
      while (1) {
	  last if &$eos;
	  next unless $sepp{$m->{chars}->[$p]}; # grep { $_ == $m->{chars}->[$p] } @{$m->{parsers}->{make}->{argseps_min}};
	  $isep = $m->{chars}->[$p];
	  &$pucstr;
	  $p++;
	  last;
      } continue { $p++ };
    };
    $sep = !$isep ? '' : chr $isep;
  SEP: { # separating with a given separator
      while (1) {
	  my $ch = $m->{chars}->[$p] || 0;
	  my $chr = chr $ch;
	  last if &$eos;
	  next unless $ch == $isep;
	  &$pucstr;
      } continue { $p++ };
    };
    @{$m->{chars}} = ();
    ($sep,@list);
};

=head3 textsep2list,text2list,text2seplist,text2listref,text2seplistref

invocations of textsep2seplist with a subset of parameters and return values

=cut

sub textsep2list {
    my ($m,$text,$sep) = @_;
    ($sep, my @list) = $m->textsep2seplist($text,$sep);
    @list;
};
sub text2list {
    my ($m,$text) = @_;
    my ($sep,@list) = $m->textsep2seplist($text);
    @list;
};
sub text2seplist {
    my ($m,$text) = @_;
    my ($sep,@list) = $m->textsep2seplist($text);
    ($sep,@list);
};
sub text2listref {
    my ($m,$text) = @_;
    my @list = $m->text2list($text);
    \@list;
};
sub text2seplistref {
    my ($m,$text) = @_;
    my ($sep,@list) = $m->text2seplist($text);
    ($sep,\@list);
};
sub textsep2listref {
    my ($m,$text,$sep) = @_;
    my @list = $m->textsep2list($text,$sep);
    \@list;
};
sub chainsplit {
    my $m = shift;
    $m->pwarn('using textsep2listref under old name chainsplit');
    $m->textsep2listref(@_);
}
=head3 $m->render_rest($varsr,$pos,$join,$mapf,$gsz)

Klunky old syntax extension function: allow an infinite number of trailing arguments to be passed to the mother function, starting from $pos.  Group them in groups of size $gsz, map these groups with mapping function $mapf, join the whole together with function $join.  The function chains are scalar variables with an internal separator that is recognised by the textsep2listref subroutine

Given

    ol_item = {item:${1}}
    ol = {list type=ol:$(rest|1|:join:, :|:call:ol_item|1)}

invocation of

    $(call ol,rot,gruen,blau) 

should yield

     {list type=ol:{item:rot}, {item:gruen}, {item:blau}}

=cut

sub render_rest {
    my ($m,$varsr,$pos,$join,$mapf,$gsz,@args) = shift->vrpuc(@_);my %vars = $m->get_vars($varsr);
    $gsz ||= 1;
    $m->ustop('nonnumeric position argument in function rest: %p', p => $pos) unless $pos =~ m(\A\d+\Z);
    # ':jung:,${SP}:${SP}&${SP}' cf file:/phm/19/06/23/sig/_log06adv.phm_pub_sig190623.txt::35492
    my $i = $pos;while ($vars{$i}) { push @args, $vars{$i++} }; # arguments supplied during rendering, starting from $pos;
    my @map_arg_grups = $m->ref2list($m->get_map_arg_grups($gsz,@args));
    my ($sep, @join_chain) = $m->textsep2seplist($join);
    push @join_chain,'join' if !@join_chain;
    my $join_car = shift @join_chain;
    my ($join_fun, @join_extra) = $m->getfuns($varsr,$join_car);
    my @map_chain = ();
    ($sep, @map_chain) = $m->textsep2seplist($mapf);
    push @map_chain, 'concat' if !@map_chain;
    my $map_car = shift @map_chain;
    my ($map_fun, @map_extra) = $m->getfuns($varsr,$map_car);
    push @join_chain, '' unless @join_chain;
    &$join_fun($m, {}, @join_extra, @join_chain, map { my @map_args = @$_;&$map_fun($m, {}, @map_extra, @map_chain, @map_args) } @map_arg_grups);
};
=head4 Function render_apply

Should become something entirely different from the obsolete 'rest' and 'aresto' stuff, namely an invocation of list as a function, such that the predefined list becomes a simple lambda expression which processes any explicit arguments as well as the remaining implicit function arguments following them.
Functions map and mapply likewise eat all unmentioned remaining arguments, and whereas 'map' treats them as strings, 'mapply' treats them as lists, separated with textsep2listsep.

   (
   au_kmd = <a href="${1}">${2}</a>
   @ad_kmd = 1|call|au_kmd|$(url $(value ${1}_dok))|$(or ${2},$(value ${1}_tit))|
   @an_kmd = 1|proc|ahnomval|$(set|nom|$(shift))|$(or|${1}|${nom})|
   CDG =: $(apply|ad_kmd|dg|Charles De Gaulle)
   special = miniverb au au_kmd
   special = miniverb ad |apply|ah_kmd|
   special = miniverb ad @ah_kmd
   @komma_iug = 1|join|, |
   koniug = $(let|last+kai|$(set|last|$(pop)),$(set|kai|$(shift)),$(apply|komma_iug) ${kai} ${last})  
   koniug_et = |call|koniug|et|
   special = miniverb et |apply|koniug_et|
   special = miniverb et @koniug_et
   Il et bon qu'il y ait des Français $(et|blancs|noirs|jaunes), mais sous condition quils restent minoritaires.
   C'est ce que $(au|${url}|dit) Charles de Gaulle.
   )
   (
   olst = <ol>$(map|item)</ol>
   item = <li>${1}</li>
   special = miniverb ol olst
   Maximae res publicae mundi sunt $(ol|RU|CA|US|CN).
   )
   (
   dlst = <dl>$(mapply|dtup)</dl>
   dtup = <dt>${1}</dt><dd>${2}</dd>
   special = miniverb dl dlst
   Maximae nationes mundi sunt $(dl|1=RU|2=CA|3=US|4=CN).
   )
   (
   tlst = <table>$(mapply|ttup)</table>
   tfld = <td>${1}</td>
   ttup = <tr><th>$(shift)</th>$(map|tfld)</tr>
   special = miniverb tl tlst
   Maximae nationes mundi sunt $(tl|1=RU=19k|2=CA=10k|3=US=9.9k|4=CN=9.8k).
   )
=cut
sub render_apply {
    my ($m,$varsr,@rest) = shift->vpuc(@_);
    $m->render_rest($varsr,1,@_);
}
sub render_let { # still same as progn, no special use for assigning argument 
    return shift->render_progn(@_);
}
sub render_map { # arguments are strings
}
sub render_mapply { # arguments are lists of strings
}

=head4 Function render_aresto

Application example:

    dl_item = {item id="${1}":${2}}
    desclist_join = $(aresto|${BRA}list type=dl:|1|join|call+dl_item|2|${ARB})

Syntax extension function: allow an infinite number of trailing arguments to be passed to the mother function, starting from its argument number $arg1.  Group them in groups of size $arg3, map these groups with mapping function chain $arg4, join the results together with function chain $arg2.  The function chains are scalar variables with an internal separator that is recognised by the textsep2listref subroutine

=cut

sub render_aresto {
    my ($m,$varsr,$var,@rest) = shift->vpuc(@_);
    my $ante = $m->render($varsr,$var);
    $var = pop @rest; # [ 'concat', '}' ]
    my $post = $m->render($varsr,$var);
    my $body = $m->render_rest($varsr,@rest);
    return $body unless $body;
    $ante . $body . $post;
};

sub render_foreach { 
    my ($m,$varsr,$var,$list,$body,$sep) = @_;my %vars = %$varsr;$var = $m->render($varsr,$var);$list = $m->render($varsr,$list);$sep = $m->render($varsr,$sep) || ' ';
    my @vals = split $sep, $list;
    join $sep, grep { $_ } map { $vars{$var} = $_;$m->render(\%vars,$body) } @vals;
}

=head4 Function render_if

true if the body expands to something

use as in 

	$(if ${2},${2},${1})

=cut

sub render_if { 
    my ($m,$varsr,$body,$kthen,$kelse) = shift->vpuc(@_);
    $body = $m->render($varsr, $body);
    $body ? $m->render($varsr,$kthen) : $m->render($varsr,$kelse);
}

=head4 Function render_ifn

if not, inversion of render-if

=cut

sub render_ifn { 
    my ($m,$varsr,$body,$kthen,$kelse) = shift->vpuc(@_);
    $body = $m->render($varsr,$body);
    $body ? $m->render($varsr,$kelse) : $m->render($varsr,$kthen);
}

=head4 Function render_defp

Whether the argument is defined as a variable.  Used in

    $(let $(ifn $(defp titol),$(set titol,$(shift))), ${titol} ${ARB}list type=ul:$(apply,join+${LF},call+ul_item)${BRA})

=cut

sub render_defp { 
    my ($m,$varsr,$nom) = shift->vpuc(@_);
    $m->getvar($varsr,$nom) ? 1 : 0 }

=head4 Function render_or

=cut

sub render_or { 
    my ($m,$varsr,@args) = shift->vpuc(@_);local $_;
    foreach my $arg (@args) {
	$_ = $m->render($varsr,$arg);
	last if $_;
    };
    $_ }

=head4 Function render_and

=cut

sub render_and {
    my ($m,$varsr,@args) = shift->vpuc(@_);local $_;
    foreach my $arg (@args) {
	$_ = $m->render($varsr,$arg);
	last if !$_;
    };
    $_ }

=head4 Comparison Functions

=cut

sub render_ifcmp { # if compares ok using operator from $m->{opfuns}
    my ($m,$varsr,$body,$op,$comp,$kthen,$kelse) = shift->vpuc(@_);
    $body = $m->render($varsr,$body);
    $comp = $m->render($varsr,$comp);
    $op = $m->{opfuns}->{$op} || $m->ustop('undefined comparison operator %o', o => $op) until 'CODE' eq ref $op;
    &$op($body,$comp) ? $m->render($varsr,$kthen) : $m->render($varsr,$kelse) }
sub make_ifcmp_fun {
    my ($m,$op) = @_;
    sub {
	my ($m,$varsr,$body,$comp,$kthen,$kelse) = shift->vpuc(@_);
	$m->render_ifcmp($varsr,$body,$op,$comp,$kthen,$kelse) } };
# sub render_ifeq { my ($m,$varsr,$body,$comp,$kthen,$kelse) = shift->vpuc(@_);$m->render_ifcmp($varsr,$body,'eq',$comp,$kthen,$kelse) }
# sub render_ifneq { my ($m,$varsr,$body,$comp,$kthen,$kelse) = shift->vpuc(@_);$m->render_ifcmp($varsr,$body,'eq',$comp,$kthen,$kelse) }
# sub render_ifgt { my ($m,$varsr,$body,$comp,$kthen,$kelse) = shift->vpuc(@_);$m->render_ifcmp($vars,$body,'gt',$comp,$kthen,$kelse) }
# sub render_ifge { my ($m,$varsr,$body,$comp,$kthen,$kelse) = shift->vpuc(@_);$m->render_ifcmp($vars,$body,'ge',$comp,$kthen,$kelse) }

=head4 Function render_progn

Like C<progn> of Lisp.

Allow a simple sequence of commands to be executed and return the value of the last of these, as in

    $(progn $(ifndef titol,$(set titol,$(shift))),${titol} ${BRA}list type=dl:$(apply join+${LF},call+dl_item,2)${ARB})

This requires some useful functions that operate on local variables to be implemented, e.g. set, shift.

Variables are local to this block, i.e. settings of $vars are shared inside but not passed back.

The kludgy restcall constructions can be rewritten more readably with this, and the rest verb becomes apply which has one argument less.
See examples in /adv/tmpl/a2e/lang.txt

=cut

sub render_progn {
    my ($m,$varsr,@kmds) = shift->vpuc(@_);local $_;
    foreach my $kmd (@kmds) { $_ = $m->render($varsr, $kmd) };
    $_;
}

=head4 Function render_catch

Like C<catch> of Lisp.

=cut

sub render_catch {
    my ($m,$varsr,$blok,@kmds) = shift->vpuc(@_);$blok = $m->render($varsr,$blok);local $_;
    foreach my $kmd (@kmds) { 
	$_ = $m->render($varsr,$kmd);
	next if !$varsr->{__kaptblok__};
	next if $varsr->{__kaptblok__} ne $blok;
	delete $varsr->{__kaptblok__};
	last;
    }; 
    $_;
}

=head4 Function render_throw

Like C<throw> of Lisp.

=cut

sub render_throw {
    my $m = shift;my $varsr = shift;local $_;
    $varsr->{__lastblok__} = 1;
  ARGS: {
      my $blok = shift;
      last unless $blok;
      $blok = $m->render($varsr,$blok);
      $varsr->{__kaptblok__} = $blok;
      $_ = shift;
      last unless $_;
      $_ = $m->render($varsr,$_);
    }
    $_;
};

=head4 Function render_next

Like C<next> of Perl, return success only.

=cut

sub render_next {
    my $m = shift;my $varsr = shift;
    $varsr->{__nextblok__} = 1;
  ARGS: {
      my $blok = shift;
      last unless $blok;
      $blok = $m->render($varsr,$blok);
      $varsr->{__kaptblok__} = $blok;
    }
    1;
}

=head4 Function render_while

=cut

sub render_while {
    my ($m,$varsr,$kond,@kmds) = shift->vpuc(@_);local $_;
  LOOP: while ($m->render($varsr,$kond)) { 
      foreach my $kmd (@kmds) { 
	  $_ = $m->render($varsr,$kmd);
	  do last LOOP if $varsr->{__kaptblok__};
	  do { delete $varsr->{__nextblok__};next LOOP } if $varsr->{__nextblok__};
	  do { delete $varsr->{__lastblok__};last LOOP } if $varsr->{__lastblok__};
      } };
    $_ };

=head4 Function render_until

=cut

sub render_until {
    my ($m,$varsr,$kond,@kmds) = shift->vpuc(@_);local $_;
  LOOP: until ($m->render($varsr,$kond)) {
      foreach my $kmd (@kmds) { 
	  $_ = $m->render($varsr,$kmd);
	  last LOOP if $varsr->{__kaptblok__};
	  do { delete $varsr->{__nextblok__};next LOOP } if $varsr->{__nextblok__};
	  do { delete $varsr->{__lastblok__};last LOOP } if $varsr->{__lastblok__};
      } };
    $_ }

=head4 Function render_set

Set a local variable.
Modify the local variables that are passed to functions in the first argument, therefore do not dereference this argument.

=cut

sub render_set {
    my ($m,$varsr,$nom,$val) = shift->vpuc(@_);
    $varsr->{$nom} = $val;
    $val;
}

=head4 Function render_shift

Like 'shift' of Perl, Bash et al: return the first argument and make the second the first.
Modify the local variables that are passed to functions in the first argument $varsr, therefore do not dereference this argument.

=cut

sub render_shift {
    my ($m,$varsr) = shift->vpuc(@_);
    my $body = '';
    my $i = 1;
  VARS: {
      last unless exists $varsr->{$i};
      $body = $varsr->{$i};
      while (1) {
	  my $j = $i + 1;
	  if (exists $varsr->{$j}) {
	      $varsr->{$i} = $varsr->{$j};
	      $i = $j;
	  } else {
	      delete $varsr->{$i};
	      last;
	  };
      };
    };
    $body }

=head4 Function render_pop

Like 'pop' of Perl, Bash et al: return+remove the last argument
Modify the local variables that are passed to functions in the first argument, therefore do not dereference this argument.

=cut

sub render_pop {
    my ($m,$varsr) = shift->vpuc(@_);
    my $body = '';
    my $i = 0;
    my $j = 1;
    while (1) {
	last unless exists $varsr->{$j};
	$i = $j;
	$j = $i + 1;
    };
    if ($i > 0) {
	$body = $varsr->{$i};
	delete $varsr->{$i};
    };
    $body }


=head4 Function render_value

=cut

sub render_value { 
    my ($m,$varsr,$body) = shift->vpuc(@_);
    $body = $m->render($varsr, $body);
    $m->ustop('attempting to evaluate bad variable name %e', e => $body) if ! $body =~ m(\A\w+\Z);
    $m->getvar($varsr,$body);
}

sub render_sprintf { 
    my ($m,$varsr,$fmt,@vals) = shift->vpuc(@_);$fmt = $m->render($varsr, $fmt) . '';
    @vals = map { $m->render($varsr, $_) } @vals;
    sprintf $fmt, @vals;
}

=head4 Function render_substr

    $(substr abc,0,3) => 'abc'
    $(substr abc,0,2) => 'ab'
    $(substr abc) => 'abc'

=cut

sub render_substr {
    my ($m,$varsr,$str,$pos,$len) = shift->vrpuc(@_);
    $len ||= length($str) - $pos;
    substr $str, $pos, $len;
}

sub render_lrtrim {
    my ($m,$varsr,$s) = shift->vrpuc(@_);
    my ($body) = m(\A\s*(.*\S)\s*\Z)s;
    $body;
}

sub render_strlen {
    my ($m,$varsr,$body) = shift->vrpuc(@_);
    length $body;
}

sub render_ucfirst {
    my ($m,$varsr,$body) = shift->vrpuc(@_);
    ucfirst $body;
}
sub render_lcfirst {
    my ($m,$varsr,$body) = shift->vrpuc(@_);
    lcfirst $body;
}
sub render_uc {
    my ($m,$varsr,$body) = shift->vrpuc(@_);
    uc $body;
}
sub render_lc {
    my ($m,$varsr,$body) = shift->vrpuc(@_);
    lc $body;
}
sub render_subst {
    my ($m,$varsr,$from,$to,$s) = shift->vrpuc(@_);local $_ = $s;
    s($from)($to)g;
    $_;
}

sub render_mcall {
    my ($m,$varsr,$verb,$regexp,$argstr) = shift->vrpuc(@_);
    my @args = $argstr =~ m($regexp);
    $m->nom_args_call($verb,$varsr,@args);
}

=head3 Function revargs

Invoke a function with reverse sequence of arguments. E.g.

    $(revargs|2|call|ol|${log1}|${log2}|${log3})

should result in the same as

    $(call|ol|${log3}|${log2}|${log1})

The first numeric argument says how many of the following items are not arguments and thus not reversed.

=cut

sub render_revargs {
    my ($m,$varsr,$num,@args) = shift->vrpuc(@_);
    $m->pstop('revargs needs an integer as its first argument, not %a', a => $num) unless $num =~ m(\A\d+\Z);
    my @verbs = ();
    do { my $vrb = shift @args;push @verbs, $vrb } for 1..$num;
    @args = reverse @args;
    my $verb = shift @verbs;
    my ($fun,@extra) = $m->getfuns($varsr,$verb);
    &$fun($m,$varsr,@extra,@verbs,@args);
}

=head4 numpad

Internal function underlying render_numpad

=cut
sub numpad {
    my ($m,$len,$num) = @_;
    my @nums = ();
    my $i = 0;
    my $mod = 0;
    while ($num > 0) {
	$mod = $num % 10;
	unshift @nums, $mod;
	$i++;
	$num = int $num / 10;
    };
    unshift @nums, '0' while $i++ < $len;
    join '', @nums;
}

=head4 render_numpad

Pad with leading zeroes up to a given length.

    $(numpad 3,1)
    => 001

=cut
sub render_numpad {
    my ($m,$varsr,$len,$num) = shift->vrpuc(@_);
    $m->numpad($len,$num);
}

=head4 render_dnumpad

Like render_numpad, but with double width numbers.

=cut

sub render_dnumpad {
    my ($m,$varsr,$len,$num) = shift->vrpuc(@_);
    my $i = 0;
    join '', map { $i = $_ - 48;$double_width_numbers[$i] } unpack 'U*', $m->numpad($len,$num);
};

=head4 render_dwchars

Turn argument string into double width character version, e.g.

    dok = oas_pub1704
    dwdok := $(dwchars ${dok})
    => ｏａｓ＿ｐｕｂ１７０４

=cut

sub render_dwchars {
    my ($m,$varsr,$body) = shift->vrpuc(@_);
    @{$m->{chars}} = unpack 'U*', $body;
    my $p = 0;
    my @strs = ();
    while ($p <= $#{$m->{chars}}) {
	my $num = $m->{chars}->[$p];
	$m->dpuc('p', $p, $num);
      TRY: {
	  $strs[$p] = $m->{konfig}->dwchars->{$num};
	  last if $strs[$p];
	  $num += $m->{sddelta} if $num < 128;
	  $strs[$p] = chr $num;
	};
	$m->dpop('p', $num, $strs[$p]);
    } continue { $p++ };
    @{$m->{chars}} = ();
    join '', @strs;
}
=head modnum, render_modnum

Usage Examples file:/phm/19/06/09/sig/_log09adv.phm_pub_sig190609.txt::339342

	$(modnum|15,000.00|,|.|3) => 15.000,00
        $(modnum|15,000.00|.|,|4) => 1,5000.00
        $(modnum|15,000.000|.|,|4|3) => 1,5000.000
	$(modnum|150000) => 150,000
	$(modnum|15,0000|,) => 150.000

Normally we assume two digit cent amounts after the final integer-part-concluding dot/comma, but this can be altered by another argument $4, here '4':
	$(modnum|15,000.0099|,|.|3|4) => 15.000,0099
This argument could be any expression fitting in \d{$4}, e.g. '2,4'
	$(modnum|15,000.0099|,|.|3|2,4) => 15.000,0099

=cut

sub modnum {
    my ($m,$body,$cntsep,$modsep,$gwd,$cwd) = @_;
    local $_ = $body;
    s(\s)()g;
    $cwd ||= 2;
    my ($dlr,$cnt) = m(\A(.*)[.,](\d{$cwd})\Z);
    $_ = $dlr if $dlr;
    s([,.])()g;
    do { $m->uwarn('cant modnum nonnumeric body %b', b => $body);return $body } unless m(\A\d+\Z);
    $dlr = $_;
    $cntsep ||= '.';
    $modsep ||= $cntsep eq '.' ? ',' : '.';
    $gwd ||= 3;
    my $modulo = 10 ** $gwd;
    my @mods = ();
  MODS: {
      my $mod = $dlr % $modulo;
      $dlr = int $dlr / $modulo;
    PAD: {
	last unless $dlr;
	$mod += $modulo;
	$mod = substr $mod, 1;
      };
      unshift @mods, $mod;
      redo MODS if $dlr;
    };
    local $_ = join $modsep, @mods;
    $_ = join $cntsep, $_, $cnt if $cnt;
    $_;
};    

sub render_modnum {
    my ($m,$varsr,@args) = shift->vrpuc(@_);
    $m->modnum(@args);
};

=head4 render_case

Example

    sems2subtmpl ?= $(case ${1},pub_aammdd,/^[wW]/,pub_aaww,sig,sig_ont_aamm)

This means that argument 1 is compared to regular expressions of arguments 3, 5, 7 etc and if matching 4, 6, 8 respectively are returned, otherwise the default value 2.

=cut

sub render_case { 
    my ($m,$varsr,$klav,$body,@tups) = shift->vpuc(@_);
    $klav = $m->render($varsr,$klav);
    while (@tups) {
	my $kmp = shift @tups;
	$kmp = $m->render($varsr,$kmp);
	my $rgx = qr(\A$kmp\Z);
	my $val = shift @tups;
	next unless $klav =~ m($rgx);
	$body = $val;
    };
    $m->render($varsr,$body);
}

sub render_wildcard { 
    my ($m,$varsr,$body,$sep) = shift->vrpuc(@_);
    join $sep, glob $body;
}

sub render_shell { 
    my ($m,$varsr,$body) = shift->vrpuc(@_);
    $m->ustop('Refusing for security reasons to execute command body %b in shell', b => $body) unless $m->superuserp;
    $body = `$body`;
    chomp $body;
    $body;
}

=head3 Function f2s

like the shell utility 'cat', but additionally filter like grep when further arguments are given.
The arguments are those of A2E::Prog::fayl_get_str (except coderef variant of $join).  

=cut

sub render_f2s {
    my ($m,$varsr,$fyl,@args) = shift->vrpuc(@_);
    $m->fayl_get_str($fyl,@args);
}
=head4 basename, notdir

like in GNU Make, taking output of glob filelist as argument

=cut
sub render_basename {
    my ($m,$varsr,$fyls) = shift->vrpuc(@_);
    join ' ', map { s(\.\w+\Z)() } split m(\s), $fyls;
}
sub render_notdir { 
    my ($m,$varsr,$fyls) = shift->vrpuc(@_);
    join ' ', map { s(\A.*([^\/]+)\Z)($1) } split m(\s), $fyls;
}
=head4 fromto
=cut
sub render_fromto {
    my ($m,$varsr,$from,$to,$sep) = shift->vrpuc(@_);$sep ||= ' '; 
    my @elms = ($from .. $to);
    join $sep, @elms;
}

=head3 ddmax

return last day of given month for arguments $aaaa and $mm.
Alternatively, $aaaa and $mm can be given as one concatenated argument:

    $(ddmax,2016,12)
    $(ddmax,201612)
    $(ddmax,1612)

all work.

=cut

sub render_ddmax {
    my ($m,$varsr,$aaaa,$mm,$ddmin) = shift->vrpuc(@_);
    unless ($mm) { $mm = substr $aaaa, -2;$aaaa = substr $aaaa, 0, -2;$aaaa += 2000 if $aaaa < 100 };
    $m->ustop(chkaaaa => a => $aaaa, p => 'ddmax') unless $aaaa =~ m(\A(?:19|20)\d{2}\Z);
    $m->ustop(chkmm => m => $mm, p => 'ddmax') unless $mm =~ m(\A(?:0[1-9]|1[0-2])\Z);
    $ddmin = $m->renchk_dd(ddmax => $ddmin, '01');
    my $ddmax = '30';
    ($ddmin, $ddmax) = $m->mmddminmax($ddmin, $mm, $aaaa);
    $ddmax;
}

=head3 mondags

return days of a given year/month in the form of '01 02 03 ... 30'.
take aaaa, mm and optional startday argument

=cut

sub renchk_aaaa {
    my ($m, $proc, $aaaa, $dft) = @_;
    return $dft unless $aaaa =~ m(\A\d+\Z);
    $aaaa = 2000 + $aaaa if $aaaa =~ m(\A\d{2}\Z);
    $m->ustop(chkaaaa => p => $proc, a => $aaaa) unless $aaaa =~ m(\A\d{4}\Z);
    $aaaa;
};
sub renchk_mm {
   my ($m, $proc, $mm, $dft) = @_; 
   return $dft unless $mm;
   $m->ustop(chkmm => p => $proc, m => $mm) unless $mm =~ m(\A(?:0[1-9]|1[0-2])\Z);
   $mm;
};
sub renchk_dd {
    my ($m, $proc, $dd, $dft) = @_;
    return $dft unless $dd;
    $m->ustop(chkdd => p => $proc, d => $dd) unless $dd =~ m(\A\d{2}\Z) and $dd > 0 and $dd <= 31;
    $dd;
}
sub renchk_ww {
    my ($m, $proc, $ww, $dft) = @_;
    return $dft unless $ww;
    $m->ustop(chkww => p => $proc, w => $ww) unless $ww =~ m(\A\d{2}\Z) and $ww >= 0 and $ww <= 53;
    $ww;
}
sub renchk_www {
    my ($m, $proc, $www, $dft) = @_;
    return $dft unless $www;
    my ($w, $ww) = $www =~ m(\A([wW])([0-5][0-9])\Z);
    $m->ustop('%p argument %w is not a week identifier', w => $www) unless $w;
    ($www, $w, $ww);
};
sub render_mondags {
    my ($m,$varsr,$aaaa,$mm,$ddmin,$sep) = shift->vrpuc(@_);
    $aaaa = $m->renchk_aaaa(mondags => $aaaa);
    $mm = $m->renchk_mm(mondags => $mm);
    $ddmin = $m->renchk_dd(mondags => $ddmin, '01');
    my $ddmax = '30';
    ($ddmin, $ddmax) = $m->mmddminmax($ddmin, $mm, $aaaa);
    my @elms = ($ddmin .. $ddmax);
    $sep ||= ' ';
    join $sep, map { sprintf '%02d', $_ } @elms;
}

=head3 wokdags

invocation $(wokdags|2015|07|28) should return '12 13 14 15 16 17 18'.
return days of a given year/month/week separated by blanks.
Used in /adv/tmpl/a2e/pub_aaww/lang_pre.txt to derive subdocuments from week

    sems2subsems ?= sig $(wokdags|${aaaa}|${mm}|${ww})

=cut

sub render_wokdags {
    my ($m,$varsr,$aaaa,$mm,$ww,$ddmin) = shift->vrpuc(@_);
    $aaaa = $m->renchk_aaaa(wokdags => $aaaa);
    $mm = $m->renchk_mm(wokdags => $mm);
    $ww = $m->renchk_ww(wokdags => $ww);
    $ddmin = $m->renchk_dd(wokdags => $ddmin, '01');
    my @wokdags = $m->wokdags($aaaa,$mm,$ww,$ddmin);
    join ' ', map { sprintf '%02d', $_ } @wokdags;
};

=head3 render_monwoks

Take a year and month and optionally monthday from which to begin, as renderable structures.
    
Return the weeks of a month as a two-layered structure including a symbol based on the calender week number and the initial monday and final sunday thereof.

For example

    $(monwoks 2012,07)

shall render

    |W26,01|w27,02,08|w28,09,15|w29,16,22|w30,23,29|w31,30,31|

To be used by

    /adv/tmpl/a2e/pub_ont_aamm/lang_pre.txt

Capitalisation in 'W26' means that week 26 is carried over from last month and 'w26' was therefore already used in naming.
This allows us to create document symbols like adv12w26, adv12W26, adv12w27, adv12w28, adv12w29, adv12w30, adv12w31 in parallel to adv1206, adv1207.

Last work reports
 /phm/12/07/04/sig
 /phm/12/07/09/sig
 /phm/12/09/07/sig

=cut

sub render_monwoks {
    my ($m,$varsr,$aaaa,$mm,$ddmin) = shift->vrpuc(@_);local $_;
    $aaaa = $m->renchk_aaaa(monwoks => $aaaa);
    $mm = $m->renchk_mm(monwoks => $mm);
    $ddmin = $m->renchk_dd(monwoks => $ddmin,'01');
    my $ddmax = '30';
    my $monwoks = $m->monwoks($aaaa,$mm,$ddmin);
    $_ = join '|', map { my @rek = @$_;splice @rek, 1, 1;join ',', @rek } @$monwoks;
    '|'.$_.'|';
}
=head4 Function wokmon

Derive month from week

    $(wokmon 2015,w13) => '03'
    $(wokmon|18|W48) => '12'

=cut

sub render_wokmon {
    my ($m,$varsr,$aaaa,$www) = shift->vrpuc(@_);
    $aaaa = $m->renchk_aaaa(wokmon => $aaaa);
    ($www, my $w, my $ww) = $m->renchk_www(wokmon => $www);
    $m->wokmon($aaaa,$www,$w,$ww);
};

=head4 Functions plus,minus,mult,div,intdiv,mod

    $(plus 1,4)

renders

    5

=cut

sub render_plus {
    my ($m,$varsr,@nums) = shift->vrpuc(@_);
    my $res = 0;
    foreach my $num (@nums) { $res += $num };
    $res;
}

sub render_minus {
    my ($m,$varsr,@nums) = shift->vrpuc(@_);
    my $res = shift @nums;
    foreach my $num (@nums) { $res -= $num };    
    $res;
}

sub render_mult {
    my ($m,$varsr,@nums) = shift->vrpuc(@_);
    my $res = 1;
    foreach my $num (@nums) { $res *= $num };    
    $res;
}
sub render_div {
    my ($m,$varsr,@nums) = shift->vrpuc(@_);
    my $res = shift @nums;
    foreach my $num (@nums) { $res /= $num };
    $res;
}
sub render_intdiv {
    my ($m,$varsr,@nums) = shift->vrpuc(@_);
    my $res = shift @nums;
    foreach my $num (@nums) { $res = int $res / $num };
    $res;
}
sub render_mod {
    my ($m,$varsr,@nums) = shift->vrpuc(@_);
    my $res = shift @nums;
    foreach my $num (@nums) { $res %= $num };    
    $res;
}
=head3 Functions minim, maxim

Give the smallest/biggest of the numeric arguments

=cut
sub render_minim {
    my ($m,$varsr,@nums) = shift->vrpuc(@_);
    my $i = shift @nums;
    foreach my $j (@nums) {
	$i = ($j < $i ? $j : $i);
    };
    $i;
}
sub render_maxim {
    my ($m,$varsr,@nums) = shift->vrpuc(@_);
    my $i = shift @nums;
    foreach my $j (@nums) {
	$i = ($j > $i ? $j : $i);
    };
    $i;
}

=head3 Function strfamd

As in

	$(strfamd %w,2010,4,22) => 4

invoke POSIX::strftime but with arguments as normally used in dates and in topdown order:

    args := fmt, aaaa, mm, dd, (hor, (min, sek?)?)?
    aaaa :: year
    mm :: month
    dd :: day
    hor :: hour
    min :: minute
    sek :: second

If the year is greater than 2000 and less than 2100, it can be given as either aaaa or aa.

=cut

sub render_strfamd {
    my ($m,$varsr,@vals) = shift->vrpuc(@_);
    my %arg = ();
    my %tmp = ();
    my $setargtmp = sub {
	my $nom = '';
	foreach $nom (qw(fmt annus mens dies hor min sek)) {
	    my $val = $vals[0];
	    last unless $val;
	    $arg{$nom} = $val;
	    shift @vals;
	};
	my @temp = localtime;
	foreach $nom (qw(sek min hor dies mens annus)) {
	    $tmp{$nom} = shift @temp;
	};
    };
    &$setargtmp;
    my @amd = ();  
    my $set_amd = sub {
	if ($arg{dies}) {
	    $arg{annus} += 2000 if $arg{annus} < 100;
	    @amd = map { $arg{$_} } qw(annus mens dies);
	  INVALID: {
	      last if $arg{dies} and $arg{annus} =~ m(\A\d{4}\Z) and $arg{mens} =~ m(\A\d{2}\Z) and $arg{dies} =~ m(\A\d{2}\Z) and $arg{annus} > 1970 and $arg{mens} <= 12 and $arg{dies} <= 31;
	      $m->ustop('invalid date/time arguments in strfamd: %A', A => \@amd);
	    };
	    $arg{annus} -= 1900;
	    $arg{mens} -= 1;
	};
	$arg{$_} ||= $tmp{$_} for qw(sek min hor dies mens annus);
	@amd = map { $arg{$_} } qw(fmt sek min hor dies mens annus);
    };
    &$set_amd;
    strftime(@amd);
}


=head3 Function folnum

Help generate pathnames like

    01about
    02intro
    03tutorial

Return a numeric prefix to be used in a name of a file or subdirectory to be created.
Check which prefixes have already been used and count one beyond the last.
Arg1 is the length of the prefix, default is 2.

not tested as a whole yet

=cut

sub render_folnum {
    my ($m,$varsr,$len) = shift->vrpuc(@_);
    my $num = '';$num .= '0' for 1..$len;
  NUMFOLS: {
      opendir DIR, '.';
      my @numfols = grep { m(\A\d{$len}\D) } readdir DIR;
      closedir DIR;
      last unless @numfols;
      ($num) = $numfols[$#numfols] =~ m(\A(\d{$len}));
    }
    $num++; 
    $num;
}

=head4 Function render_hannum

 $(hannum 124) => 一百二十四

=cut

sub render_hannum {
    my ($m,$varsr,$num) = shift->vrpuc(@_);
    $m->hannum($num);
};

=head4 Function inaktiv

Switch off parsing, allow verbatim rendering.
The actual work is done within C<make_pars> where the verb C<inaktiv> receives special treatment.
Cf. /phm/18/01/23/sig/_lng.phm_pub_sig180123.txt:inaktiv, /sig/adv/18/01/_lng.adv1801.txt:inaktiv

=cut

sub render_inaktiv { # dummy, work is done within make_pars
    my ($m,$varsr,$body) = shift->vrpuc(@_);
    $body;
}
sub render_list {
    my ($m,$varsr,@elms) = shift->vrpuc(@_);
    [ @elms ];
};
sub render_hash {
    my ($m,$varsr,@elms) = shift->vrpuc(@_);    
    { @elms };
};

=head1 BUGS, LIMITATIONS

=head1 TODO

=head2 List functions from_to, map, get_record, get_records

Some functions should be moved to a user-specific configuration/expansion package.

WARN: the following might be partially outdated.

The following kinds of constructions should work soon:

    patent_claims = $(call|enumerate|$(map|$(value claim${1})|$(from_to 1,14)))
    order_record080808 = $(call row,$(get_record:select name, amount from orders where customer = 'kingkong' and date = '2008-08-08'))
    orders_table = $(chain ${rows},$(map:call,row:$(get_records:orders:customer=kingkong,2008-08-01<=odate<2009-08-01:name,amount)))

The above are format-independent, because we store the format dependent stuff separately, e.g. for an imaginary HTML source format application:

    ul_li = <li>${1}</li>
    enumerate = <ul>${LF}$(rest|1|${lines}|call,ul_li)${LF}</ul>
    cell = <td>${1}</td>
    row = <tr><th>${1}</th>$(rest|2|concat|cell)</tr>
    rows = ${lines}


The C<map> verb iterates over a list of lists.
Normal verbs accept the output of map as an argument list.
While text element variables are always scalar (anything else would be incompatible with the DBM format), the output of the map, from_to and get_record function is a listref, and the output of the get_records function is a listref of listrefs. 
Consequently, the output of map, like that of rest, is never directly assigned to a text element variable.

Since the get_record and get_records functions depend on extraneous libraries like A2E::Daba, they must be implemented at a higher level, e.g. in the A2E::MLHT library.

=head1 THE END

=cut

return 1;

# Local Variables:
# coding: utf-8
# srcfile: /adv/perl/A2E/SArb/Make.pm
# url: http://a2e.de/adv/perl/A2E/SArb/Make.pm
# mode: perl
# End:
