# -*- coding: utf-8; srcfile:/adv/perl/A2E/Template/Plugin/Filter/MLHT.pm; mode: perl -*-
# Time-stamp: <2009-02-24 14:11:03 phm>

=head1 NAME

A2E::Template::Plugin::MLHT;

=head1 DESCRIPTION

A filter extension to the Perl Template Toolkit for use in the MLHT (Multilingual Hypertext) production system, implementing a syntax that allows translators to focus on their competences, reducing any non-translatable elements (especially markup elements) in the text to a minimum.

This is now obsolete, use vocabulary files with A2E::Template::Plugin::Textbau instead.

=head1 SYNOPSIS

In a vocabulary file, define a string variable containing minimal MLHT markup, e.g. one named 'sarb', as follows:

    id_est: i.e. %(a)
    uc_txt: UTF-8 or UTF-16
    uc_url: http://www.unicode.org
    isuc: Is this a %(em|%(tn|%(uc) string|%(ie|multibyte array))) ?

which contains some minimal markers, used for translations according to the MLHT syntax (see below).

In your main text, define the blocks that are referenced in the C<sarb> string:

    [% BLOCK url_anchor %]{ah url="[% url %]":[% text %]}[% END -%]
    [% BLOCK theme_note %][% thema %]{fn sym=[% sym %]:[% rhema %]}[% END -%]
    [% BLOCK emphasis %]/[% text %]/[% END -%]

Somewhere in your header you will have 

    [% USE MLHT %]

initiated this module.  Now you can filter a string like 'isuc' through the module while mapping the defined blocks to the minimal markers used in the string.

    [% isuc | mlht em='emphasis text' tn='theme_note sym=$aesym thema rhema' ie='id_est' u8='url_anchor url=$uc_url text=$uc_txt' %]

This would expands to 

    Is this a ** {ah url="http://www.unicode.org":UTF-8 & UTF-16} string{fn sym=aefn1:i.e. multibyte array} ** ?

The markers are mapped not only to functions but also to their arguments.  The positional arguments of the MLHT expression are mapped onto those named arguments which do not have a preset value.  If, as in the 'u8' case above, all are preset we have a static value without arguments.

Preset values are specified literally unless preceded by a dollar prefix.  This dollar prefix notation works within single quotes, is independent of the Template Toolkit's variable interpolation.  Unless you are sure you know what you are doing, always use single quotes and evaluate template variables/blocks only by means of the dollar prefix within single quotes.

Note that we have used the C<id_est> template variable which we initially defined as a verb with an argument.  We could have achieved the same by defining a block like

    [% BLOCK idest %]i.e. [% arg %][% END -%]

and then invoking that block with

    [% isuc | mlht ie='id_est arg' ... %]

However this kind of construction is lengthier and less likely to be understood and translated correctly by our translators.  Therefore, for multilingual expandable elements we use the MLHT notation rather than TT block notation.  The parameters within such elements are purely positional and must be named by the lower-case letters of the alphabet (a, b, c ..).

=head1 DOWNLOAD SOURCE

    http://a2e.de/adv/perl/A2E/Template/Plugin/MLHT.pm

=head1 PREREQUISITES

    A2E::SArb::MLHT(3)
    Template::Plugin::Filter(3)
    Template(3)

=head1 SEE ALSO

    mlht(1)
    http://a2e.de/adv/mlht

=head1 IMPLEMENTATION

=head2 Package A2E::Template::Plugin::MLHT;

=cut

package A2E::Template::Plugin::MLHT;
use Template::Plugin::Filter;
use Template::Parser;
use Template::Directive;
$Template::Parser::DEBUG = 1;
$Template::Directive::PRETTY = 1;
use base qw(Template::Plugin::Filter);
use A2E::SArb::MLHT;
use Data::Dumper;

sub init {
    my $m = shift;
    $m->{_DYNAMIC} = 1;
    $m->install_filter('mlht');
    $m->{SArb} = new_ready A2E::SArb::MLHT debug => $ENV{DEBUG}, verbose => $ENV{VERBOSE};
    $m->{SArb}->dpuc('init', @_);
    $m->{concat} = sub { my $keyvals = shift;join '', @_ };
    $m->{paren} = sub { my $keyvals = shift;my $left = shift;my $right = shift;return $left . join('', @_) . $right };
    $m->{mlht} = sub {
	$m->{SArb}->dpuc('mlht', [ @_ ]);
	my $keyvals = shift;
	my $x_verb = shift;
	my $res = '';
      KEYVALS: {
	  $m->{SArb}->dpuc('keyvals', $x_verb, $keyvals);
	  do { $m->{SArb}->dpop;last } unless $keyvals;
	  $res = $keyvals->{$x_verb};
	  $m->{SArb}->dpop('mlht', $res);
	  return $res;
      };
	my @args = @_;
	my $verb = $m->{konf}->{$x_verb};
	my @vargs = ();
	$m->{SArb}->dgot(x_verb => $x_verb, verb => $verb);
	if ($verb) {
	    ($verb, @vargs) = split m(\s+), $verb;
	} else {
	    $verb = $x_verb;
	};
	$keyvals = {};
      ARGS: {
	  $m->{SArb}->dpuc('args', $verb, @vargs);
	  my ($varg, $vkey, $vval);
	  my $i = ord('a');
	  while (@args || @vargs) {
	      $m->{SArb}->dpuc('vargs', @vargs, @args);
	      $varg = shift @vargs;
	      if ($varg) {
		  ($vkey, $vval) = split m(\=), $varg;
		  if ($vval) {
		      $vval = $m->{_CONTEXT}->stash->get(substr($vval, 1)) if (ord($vval) == 36);
		  } else {
		      $vval = shift @args;
		  };
	      } else {
		  $vkey = chr $i++;
		  $vval = shift @args;
	      };
	      $m->{SArb}->dgot($vkey => $vval);
	      $keyvals->{$vkey} = $vval;
	      $m->{SArb}->dpop('vargs', @args, @vargs);
	  };
	  $m->{SArb}->dpop('args', [], keyvals => $keyvals);
      };
      TRY: {
	  $m->{SArb}->dpuc('try');
	BLOCK: {
	    $m->{SArb}->dpuc('block');
	    do { $m->{SArb}->dpop;last } unless $m->{_CONTEXT}{BLOCKS}{$verb};
	    $res = $m->{_CONTEXT}->process($verb, $keyvals);
	    $m->{SArb}->dpop('mlht', $res);
	    return $res;
	};
	STASH: {
	    $m->{SArb}->dpuc('stash');
	    $res = $m->{_CONTEXT}->stash->get($verb);
	  NO_KEYVALS: {
	      $m->{SArb}->dpuc('no_keyvals', $res, $keyvals);
	      do { $m->{SArb}->dpop;last } if $res and keys %$keyvals;
	      $m->{SArb}->dpop('mlht', $res);
	      return $res;
	  };
	    my $arb = $m->{SArb}->parse($res);
	    $res = $m->render($arb, $keyvals);
	    $m->{SArb}->dpop('mlht', $res);
	    return $res;
	};
      };
	die "WEIRD: I should never be here";
    };
    $m->{SArb}->dpop('init', $m);
    return $m;
};

sub filter {
    my $m = shift;
    $m->{SArb}->dpuc('filter');
    my $text = shift;
    my $args = shift;
    my $konf = shift; 
    $m->{konf} = $m->merge_config($konf);
    # my ($dok, $des) = split m(\|), $text;
    # print STDERR "dok $dok, des $des\n";
    # see /adv/tratext/CString.pm
    # see /adv/tratext/Chars.pm
    # my $context = $m->{_CONTEXT};
    # my ($p, $s) = $mlht->render;
    $m->{SArb}->dgot(konf => $m->{konf});
    my $arb = $m->{SArb}->parse($text);
    my $res = $m->render($arb);
    # return $context->process($konf->{ah}, { dok => $dok, des => $des });
    $m->{SArb}->dpop('filter', $res);
    return $res;
};

sub render {
    my $m = shift;
    $m->{SArb}->dpuc('render', [ @_ ]);
    my $arb = shift;
    my $keyvals = shift;
    my @args = @$arb;
    my $verb = shift @args;
    my $meth = $m->{$verb};
    $m->{SArb}->dgot(verb => $verb, args => [ @args ]);
    my $res = $meth ? &$meth($keyvals, map { ref $_ ? $m->render($_, $keyvals) : $_ } @args) : Dumper($arb);
    $m->{SArb}->dpop('render', $res);
    return $res;
}

return 1;

=head1 THE END

=cut

# Local Variables:
# coding: utf-8
# srcfile:/adv/perl/A2E/Template/Plugin/MLHT.pm
# mode: perl
# End:

