# -*- coding: utf-8; mode: perl; srcfile: /adv/perl/A2E/Template/Plugin/Textbau.pm -*-
# Time-stamp: <2010-04-08 09:45:41 phm>

=head1 NAME

    A2E::Template::Plugin::Textbau

=head1 SYNOPSIS

Allow use of MLHT (multilingual hypertext, maketext) syntax at the document level, from within the Template Toolkit language.

In MLHT Document:

    [% USE Textbau 'tb' %]
    [% pi = '3.14' %]
    [% 'value of pi: ${1}' | tb 'call' pi %]
    [% 'value of pi is ${pi}' | tb %]
    [% 'ahurl = {ah url="${1}":${2}}' %]
    [% ahurl | tb 'call' 'http://www.a2e.de' 'A2E' %]
    [% 'This is $(ae:our company)' | tb 'fill' 'ae=:call:ahurl' %]

You can use variables that were defined in a maketext (vocabulary building) file.

This has been largely rendered obsolete by A2E::Template::Context(3).

=cut

package A2E::Template::Plugin::Textbau;

=head1 PREREQUISITES

    Template::Plugin::Filter(3)
    A2E::SArb::Make(3)

=cut

use Template::Plugin::Filter;
use base qw(Template::Plugin::Filter);
use A2E::SArb::Make;

=head2 Other interesting tools for Template Extension

use Template::Parser;
use Template::Directive;
$Template::Parser::DEBUG = 1;
$Template::Directive::PRETTY = 1;

=head1 IMPLEMENTATION

=head2 Function init

By default, routines are invocable with tb, as in

    [% 'value of pi: ${1}' | tb 'call' '3.14' %]

or

    [% 'value of pi is ${pi}' | tb %]

the default subcommand being 'call'.

=cut

sub init {
    my $m = shift;
    $m->{ _DYNAMIC } = 1;
    $m->install_filter($m->{ _ARGS }->[0] || 'tb');
    $m->{SArb} = new_ready A2E::SArb::Make debug => $ENV{DEBUG}, verbose => $ENV{VERBOSE}, noexec => 1;
    return $m;
}

=head2 Function fill

Part of implementation of calls like

    [% 'This is $(ae:our company)' | tb 'fill' 'ae=:call:ahurl' %]    

=cut

sub fill {
    my $m = shift;
    $m->{SArb}->dpuc('fill', @_);
    my $text = shift;
    my @args = @_;
    my $arb = $m->{SArb}->parse($text);
    my $pos = 0;
    my $key = '';
    my $vars = {};
    my $val;
    my @funs = ();
    foreach my $arg (@args) {
	($pos, $key, $val) = $m->{SArb}->fillsetvar_from_tmplvar($pos, $arg);
	$vars->{$key} = $val;
    };
    $m->{SArb}->dgot(text => $text, vars => $vars, clue => 794);
    $text = $m->{SArb}->render($vars, $arb);
    $m->{SArb}->dpop('fill', $text);
    return $text;
}

=head2 Function call

Part of implementation of calls like

    [% 'value of pi: ${1}' | tb 'call' '3.14' %]

or

    [% 'value of pi is ${pi}' | tb %]

=cut

sub call {
    my $m = shift;
    $m->{SArb}->dpuc('call', @_);
    my $text = shift;
    my @args = @_;
    my $arb = $m->{SArb}->parse($text);
    my $vars = {};
    my $n = 0;
    while (my $arg = shift @args) { $vars->{++$n} = $arg };
    $m->{SArb}->dgot(arb => $arb, text => $text, args => \@args, vars => $vars);
    my $text = $m->{SArb}->render($vars, $arb);
    $m->{SArb}->dpop('call', $text);
    return $text;
}

=head3 Function filter

Obligatory method to be implemented in any Template::Plugin::Filter extension, governs dispatching to subcommands.
Special subcommands are 'fill' and 'call' (default), all other subcommands are make-text functions from A2E::SArb::Make(3) to which rendered string-arguments are fed.  If no such function is defined, an error occurs.

=cut

sub filter {
    my $m = shift;
    $m->{SArb}->dpuc('filter', @_);
    my ($text, $args, $config) = @_;
    my $context = $m->{_CONTEXT};
    $config = $m->merge_config($config);
    # $m->{SArb}->{debug} = 0;
    $m->{SArb}->{vars} = $context->{STASH};
    $m->{SArb}->{arbs} = $context->{STASH}->{__ARBS__};
    # $m->{SArb}->{debug} = 1;
    $m->{SArb}->dgot(vars => [ keys %{$m->{SArb}->{vars}} ], arbs => [ keys %{$m->{SArb}->{arbs}} ], funs => [ keys %{$m->{SArb}->{funs}} ]);
    my @arglist = @$args;
    my $verb = shift @arglist;
    $verb ||= 'call';
  TRY: {
    ARB: {
	last unless grep { $_ eq $verb } qw(call fill);
	$text = &$verb($m, $text, @arglist);
	last TRY;
      };
      my $fun = $m->{SArb}->getfun($verb) or die sprintf 'no such function %s', $verb;
      $text = &$fun($m->{SArb}, map { $m->{SArb}->transform($_) } ($text, @arglist));
    };
    $m->{SArb}->dpop('filter', $text);
    # $m->{SArb}->{debug} = 0;
    return $text;
}

1;


