package A2E::Template::Provider;
use base 'Template::Provider';

=head1 NAME

A2E::Template::Provider

=head1 SYNOPSIS

       use A2E::Template::Provider;

       my $tmplopts = { INTERPOLATE => 1, INCLUDE_PATH => \@include_list, ENCODING => 'utf8', ... }
       ...
       my $parser = new Template::Multilingual::Parser $tmplopts;
       $tmplopts->{PARSER} = $parser;
       my $provider = new A2E::Template::Provider $tmplopts; 
       $tmplopts->{LOAD_TEMPLATES} = [ $provider ];
       my $template = new Template::Multilingual $tmplopts;

=head1 SEE ALSO

       mlht(1)
    A2E::Tmplfil(3)

=head1 DESCRIPTION

A subclass of the Template::Provider class that fixes some problems, e.g. lack
of support for utf-8, and may in the future specify additional data sources,
as suggested in the Template::Provider manual.

=head1 IMPLEMENTATION

=head2 Replacements

=head3 Routine _template_content

We had to insert the :utf8 binmode into the open command arguments.

This is arguably a bug in the original.  The documentation allows specification of a 

     binmode => :utf8

option tuple as arguments to the Template::process function, but this is not applied here.

A further improvement would be to apply it and to submit a bug report and patch to http://rt.cpan.org  

=cut

sub _template_content {
    my ($self, $path) = @_;

    return (undef, "No path specified to fetch content from ") unless $path;

    my $data;
    my $mod_date;
    my $error;

    local *FH;
    if (open(FH, '<:utf8', $path)) {
        local $/;
        $data = <FH>;
        $mod_date = (stat($path))[9];
        close(FH);
    }
    else {
        $error = "$path: $!";
    }

    return wantarray
        ? ( $data, $error, $mod_date )
        : $data;
}

return 1;

=head1 THE END

=cut

# Local Variables:
# coding: utf-8
# srcfile:/adv/perl/A2E/Template/Provider.pm
# mode: perl
# End:
