#!/usr/bin/perl
# Time-stamp: <2019-06-11 07:31:41 phm>

=head1 NAME

A2E::U2tex - convert ideal LaTeX input to currently parsable LaTeX input

=head1 VERSION

This describes version B<0.0.2> of A2E::U2tex

=cut

our $VERSION = '0.0.2';

=head1 SYNOPSIS

u2tex [options] <file>

=head1 DESCRIPTION

u2tex expects a UTF-8-coded input file and rewrites this file in such a way
that LaTeX (latex, pdflatex) can deal with it more easily, given current
limitations such as

 - in the output, \"a looks better than ä, etc
   - we do what 'recode u8..tex' fails to do
 - OBOSLETE: section headings that contain multibyte UTF-8 characters can not be numbered (temporary solution: \section{漢字} --> \section*{漢字})
   - This problem is solved by using the package CJKutf8 instead of CJK
 - \usepackage[esperanto]{babel} seems to be incompatible with the CJK package
   - not addressed yet
 - possibly more


=head1 OPTIONS

=over

=item --help | -h

Display short help and exit.

=item --manpage | -m

Display full documentation and exit.

=back

=head2 General options:

=over

=item --outfile <file> | -o <file>

Output to this file, rather than to the input file. 

=item --lang <lang> | -l <lang>

Assume that the document is written in language <lang>.

=item --quiet | -q

Let me keep silent while I run.
 
=item --verbose | -v

Let me explain what I'm doing while I run.

=item --plain_headings | -p <level>

Force non-numbered headings.

level 0: never (default)
level 1: always do it when CJK characters are found
level 2: always do it

The default value used to be 1.  It is now (since 2010-03-09) 0 because we do not need this kludge anymore since we have use the CJKutf8 package in our templates.

=back

=head1 BUGS

Some accented characters are still missing.
Ask phm to complete the list.

It is written quickly and runs slowly.
The methods of dok:tratext/CString.pm should be used, rather than regex.

=head1 PREREQUISITES

    A2E::Prog(3)

=head1 SEE ALSO

    perldoc(1)

=head1 SCRIPT CATEGORIES

text processing

=head1 ENVIRONMENT

=over

=item FILES

	/u2tex.konf
	~/u2tex.konf
        ./_u2tex    

These must contain the character transformation rules in the form

	texchars ĝ = \\^g
	texchars ç = \\c{c}

etc.

=back

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

Copyright (c) 2007-2010 Hartmut Pilch (phm)

This program is free software; you can redistribute it and/or
mmodify it under the same terms as Perl itself.

=cut

package A2E::U2tex;
use base 'A2E::Prog', 'A2E::Texchars';
use File::Spec::Functions;
use utf8;
use Encode qw(_utf8_on);

sub defvars ($) {
    my $prog = 'A2E::U2tex::defvars';my $m = shift;return if $m->{ready}->{$prog};$m->dpuc($prog, @_);
    $m->SUPER::defvars;
    $m->texchars_defvars;
    $m->konfig_define('plain_headings','i') unless $novitas > 1;
    $m->konfayl('u2tex.konf');
    $m->{ready}->{$prog} = 1;$m->dpop($prog,1);return 1;
};

=head3 Function make_protect_headings_function

=cut

sub make_protect_headings_function ($) {
    my $m = shift;
    $m->dpuc('make_protect_headings_function', @_);
    my ($ok, $sub) = ();
  ALWAYS_PLAIN_HEADINGS: {
      $ok = $m->{konfig}->plain_headings > 1;
      $m->dpuc('always_plain_headings', $ok, $m->{konfig}->plain_headings);
      do { $m->dpop;last } unless $ok;
      $sub = sub {
	  local $_ = shift;
	  s(\\(part|chapter|section|subsection|subsubsection)\{)(\\$1*\{)g;
	  return $_;
      };
      $m->dpop('make_protect_headings_function', $sub);return $sub;
  };
  CJK_PLAIN_HEADINGS: {
      $ok = $m->{konfig}->plain_headings == 1;      
      $m->dpuc('cjk_plain_headings', $ok, $m->{konfig}->plain_headings);
      do { $m->dpop;last } unless $ok;
      my $sub = sub {
	  local $_ = shift;
	  s(\\(part|chapter|section|subsection|subsubsection)\{(.*[^\p{Latin}\p{ASCII}]))(\\$1*\{$2)g;
	  return $_;
      };
      $m->dpop('make_protect_headings_function', $sub);return $sub;
    };
    $sub = sub { local $_ = shift;return $_ };
    $m->dpop('make_protect_headings_function', $sub);return $sub;
};

=head3 Function reskrib_fayl

rewrite the file

=cut

sub reskrib_fayl ($$) {
    my $m = shift;
    $m->dpuc('reskrib_fayl', @_);
    my $fi = shift;
    $m->open_fayl($fi, 'i', 'r');
    my $protect_headings_function = $m->make_protect_headings_function;
    my (@nums, @strs) = ();
    my %texchars = $m->phac('texchars');
    my $chr = 'ä';
    my $num = unpack 'U', $chr;
    _utf8_on $chr;
    $num = unpack 'U', $chr;
    my @keys = keys %texchars;
    $chr = @keys[0];
    $num = unpack 'U', $chr;
    $m->pstop('got no texchars to use in replacing') unless %texchars;
    my $nl = 0;
    my $nt = 0;
    my ($num, $chr, $tex) = ();
    local $_;
    $m->dpuc('lini');
  LINI: while ($m->getlini('i')) {
      chomp;
      $nl++;
      $m->dpuc($nl, $_);
      $_ = $m->str_u2tex($_);
      $_ = &$protect_headings_function($_);
      print $_ . "\n";
      $m->dpop($nl);
  };
    $m->dpop('lini');
    $m->dgot(nt => $nt);
    $m->kloz_fayl('i');
    $m->dpop('reskrib_fayl', 1);
    return 1;
};

=head3 Function u2tex

rewrite all files.

=cut

sub u2tex ($) {
    my $m = shift;
    $m->dpuc('u2tex', @_);
    $m->dgot(infaylz => $m->{konfig}->infaylz);
    push @{$m->{konfig}->infaylz}, @_;
    $m->dgot(infaylz => $m->{konfig}->infaylz);
    $m->out2sav($m->{konfig}->infaylz->[0]);
    $m->reskrib_fayl($_) for @{$m->{konfig}->infaylz};
    $m->sav2out;
    $m->dpop('u2tex', 1);return 1;
};

=head1 THE END

=cut

return 1;

# Local Variables: ***
# coding: utf-8 ***
# srcfile: /adv/perl/A2E/U2tex.pm.tmpl ***
# mode: perl ***
# End: ***
