=head1 NAME

A2E::Valid

=head1 VERSION

This describes version B<0.0.1>

=cut

$VERSION = '0.0.1';

=head1 DESCRIPTION

Add validators on top of A2E::Tmplfil.

Allow constructions like

   trastem_prompt = zeichenstamm wie zB epzh
   trastem_valids = |lenmin:4|lenmax:4|
   trakamd_prompt = monatskuerzel wie zB N3
   trakamd_valids = |lenmin:2|lenmax:3|
   lenmin_errmsg = Die Eingabe muss mindestens ${1} Zeichen lang sein, deine umfasst ${2} Zeichen    
   special litvalid trastem
   special litvalid trakamd
   special setopt batch 0

(found in /adv/tmpl/a2e/tra/lang.txt) to be used for validating textchunks.
These are also applied automatically during document creation to the tmpltoks (document type attributes, i.e. arguments that have to be filled with individual values whenever a document of this type is created).
The usable validators (e.g. lenmin, lenmax) are defined here or, using put_validfun, in dependent libraries.
Prompts and error messages are also taken from textchunks, but invocation of all of this prepared material happens only either automatically during document creation (on tmpltoks) or in the textchunk files when a 'special litvalid' instruction is encountered.

=head1 BUGS

just assembled from other places. 

=head1 PREREQUISITES

       A2E::Tmplfil(3)

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

    Copyright (c) 2010-2019 Hartmut Pilch (phm)

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

=head2 Header

=cut

package A2E::Valid;
use base 'A2E::Tmplfil';
use Unicode::EastAsianWidth;
use Encode qw(decode _utf8_on);
use strict;

=head2 Construction and extension of underlying object

We inherit the constructor 'new' from the base package, but supply only 'valid_postkonfig' to be used within the 'setup' routine of applications in mix-in style similar to A2E::Lokvars et al.

=head3 Function defvars

=cut

sub valid_defvars {
    my ($m,$prog,$lihs) = shift->qlpuc('valid_defvars', @_) or return;
    $m->qpop1($prog,1);
};
=head3 Function valid_postkonfig

Here the work of reading makefiles into template variables is done.  Any application that inherits this library will have the template variables read in, based on configuration and commandline options, before the main program starts.

=cut

sub valid_postkonfig {
    my ($m,$prog,$lihs) = shift->qlpuc('valid_postkonfig',@_) or return;
    $m->{validfuns} = {};
    $m->put_validfun(lenmax => sub { my $lenlim = shift;my $val = shift;my $lenval = length $val;my $res = $lenlim >= $lenval;return $res, $lenval });
    $m->put_validfun(lenmin => sub { my $lenlim = shift;my $val = shift;my $lenval = length $val;my $res = $lenlim <= $lenval;return $res, $lenval });
    $m->put_validfun(match => sub { my $re = shift;my $val = shift;return $val =~ m($re) });
    $m->put_validfun(ematch => sub { my $re = shift;my $exp = shift;my @args = @{$m->{sarb}->texsep2listref(shift)};my $val = $m->{sarb}->val_args_call($exp, undef, @args);my $res = $val =~ m($re);return $res, $val });
    $m->put_validfun(noblank => sub { my $val = shift;return $val =~ m(\A\S+\Z) });
    $m->put_validfun(integer => sub { my $val = shift;return $val =~ m(\A\d+\Z) });
    $m->put_validfun(symbol => sub { my $val = shift;return $val =~ m(\A\w+\Z) });
    $m->put_validfun(cjktut => sub { my $val = shift;return $val =~ m(\A\p{InFullwidth}+\Z) });
    $m->put_validfun(cjkmin => sub { my $num = shift;my $val = shift;my $res = $val =~ m(\p{InFullwidth}{$num});return $res, $num });
    $m->qpop1($prog,1);
};

=head3 put_validfun

Register a validator function.
$arg1 is the functions name, $arg2 its body.
Some examples are found in A2E::MLHT, e.g.

    $m->put_validfun(lenmax => sub { my $lenlim = shift;my $val = shift;my $lenval = length $val;my $res = $lenlim >= $lenval;return $res, $lenval });

As seen here, the function first consumes arguments from the validator invocation and as a final argument the value to be tested; it returns first a boolean success value and then optionally arguments usable in an error message.
Used in function invalid_msg.

=cut

sub put_validfun {
    my ($m,$prog,$verb,$body) = shift->ppuc('put_validfun', @_);
    $m->pstop('put_validfun invoked without name') unless $verb;
    $m->pstop('put_validfun invoked without body') unless $body;
    $m->{validfuns}->{$verb} = $body;
    my @keys = keys %{$m->{validfuns}};
    $m->dpop($prog,@keys);
};


=head3 Function invalid_msg

Validate value $arg1 by means of validation verb $arg2 with verb arguments @args3.
If invalid, return a message, otherwise an empty string.

Used by litvalid.

=cut

sub invalid_msg {
    my ($m,$prog,$val, $verb, @args) = shift->ppuc('invalid_msg', @_);
    my $validfun = $m->{validfuns}->{$verb} || $m->ustop('unknown validator function %v', v => $verb);
    my $errmsg_verb = join '_', $verb, 'errmsg';
    my ($valid, @msgargs) = &$validfun(@args, $val);
    my $errmsg = $valid ? '' : $m->{sarb}->nom_args_call($errmsg_verb, undef, @args, @msgargs);
    $m->dpop1($prog,$errmsg);
};

=head3 ($val,$oldp) = $m->litvalid($lit,%opts{dialog,lihs})

Validate $lit $lit based on values of lits $lit.'_valids', $lit.'_prompt' etc.
Do it always interactively if $opts{dialog} is true.

The special verb 'litvalid' behaves slightly differently: it takes its $dialog value from the invocation and, if undefined there, from the cached batch control variable as its negation i.e. !$m->{cache}->{batch}, see function valid_postkonfig.

=cut

sub litvalid {
    my ($m,$prog,$lihs,$lit,%opts) = shift->opuc('litvalid',[qw(nom)],@_);
    my $val = $lihs->getstr($lit);
    my $nom = join '_', $lit, 'prompt'; 
    my $prompt = $lihs->fgetval($nom);
    my $batch = $m->pvar('batch');
    return $m->dpop1($prog,$val) if !$prompt;
    $nom = join '_', $lit, 'valids';
    $nom  = $lihs->fgetval($nom);
    my @valids = $m->textsep2list($nom);
    @valids = map { $m->textsep2listref($_) } @valids;
    my $oldval = $val;
    $lihs->defstr(automodus => 'batch mode');
    my $readline = do { if ($ENV{TERM} eq 'dumb') {
	sub {
	    printf '#:'."\t".'%s'."\n", $lihs->getval('default');
	    printf '##:'."\t".'%s'."\n", $lihs->getval('automodus');
	    printf '%s: %s'."\n".'%s: ', $lihs->getval('default'), $val, $prompt;
	    $val = readline(*STDIN);
	    chomp $val;
	    my $lastlen = length $val;
	    return 1 if $lastlen < 1;
	    my $valsub = substr $val, 0, 1;
	    return 1 unless $valsub eq '#'; # comment entry below here
	    my $lastval = $val;
	    $val = $oldval; # restore old value in case a comment was entered
	    return 1 if $lastlen < 2;
	    $valsub = substr $lastval, 0, 2;
	    return 1 unless $valsub eq '##'; # double comment means we enter batch mode
	    $batch = 1;
	    $m->setpvar(batch => $batch);
	    1;
	};
    } else {
	sub {
	    printf '#:'."\t".'%s'."\n", $lihs->getval('batch');
	    $val = decode('UTF-8', $m->{term}->readline($prompt.': ', $val));
	    # _utf8_on $val;
	    my $lastlen = length $val;
	    return 1 if $lastlen < 1;
	    my $valsub = substr $val, 0, 1;
	    return 1 unless $valsub eq '#';
	    $batch = 1;
	    $m->setpvar(batch => $batch);
	    1;
	};
    } };
  VALID: {
      last unless $prompt;
      &$readline if $opts{dialog};
    RIID: {
	foreach my $valid (@valids) {
	    my $msg = $m->invalid_msg($val, @$valid);
	    next unless $msg;
	    print STDERR $msg, "\n";  # { $m->uwarn("%l: %v\n%m", l => $lit, v => $val, m => $msg) };
	    &$readline;
	    redo RIID; # 從頭測試
	};
      };
    };
    $lihs->putstr($lit => $val, %opts);
    $m->dpop($prog, $val, $oldval);
};

=head1 END

=cut

1;

# Local Variables:
# coding: utf-8
# srcfile: /adv/perl/A2E/Valid.pm.tmpl
# mode: perl
# End:
